#include <data/CCData.h>
#include <ccbg/CCMol.h>

#include <data/element.h>
#include <data/atomtypes.h>
#include <data/bondtypes.h>
#include <data/ffvdw.h>
#include <data/bondlength.h>
#include <data/types.h>
#include <data/resdata.h>

// global
CCElementalTable gElementalTable;
CCRefBondLengthTable gRBLTable;
CCAtomTypeTable gAtomTypeTable;
CCBondTypeTable gBondTypeTable;
CCFFVDWTable gFFVDWTable;
CCIntTyper gIntTyper;
CCResidueData gResidueData;

bool CCGlobalData::Init()
{
    if ( m_init ) {
        return true;
    }
    m_init = true;

    if ( char * p = getenv( m_env.c_str() ) ) {
        m_homeDir = p;
    }

    string dataFile;
    dataFile += m_homeDir
              + FILE_SEP_CHAR
              + m_dataDir
              + FILE_SEP_CHAR
              + m_dataFile;

    ifstream ifs( dataFile.c_str(), ios::binary );
    char line[1024];
    if ( ifs ) {
        while ( ifs.getline( line, 1024 ) ) {
            ParseLine( line );
        }
    }
    else if ( m_pData ) {
    // no data file, try to read complied values from header file
        const char* p1, * p2;
        for ( p1 = p2 = m_pData; *p2 != '\0'; ++p2 ) {
            if ( *p2 == '\n' ) {
                strncpy( line, p1, p2 - p1 );
                line[p2 - p1] = '\0';
                ParseLine( line );
                p1 = ++p2;
            }
        }
    }
    else {
        string errMsg = "Unable to open data file '" + m_dataFile + "'";
        gErrorLog.Throw( __FUNCTION__, errMsg, ccFatal );
        exit( EXIT_FAILURE );
    }

    if ( ifs ) {
        ifs.close();
    }

    return true;
}

//////////////////////////////////////////////////////////////////////////
// class CCElementalTable 
//////////////////////////////////////////////////////////////////////////
CCElementalTable::CCElementalTable()
: CCGlobalData( false, "CCBGKit_HOME", CCBGKit_HOME, "data", "element.txt", ELEMENT )
{
}

CCElementalTable::~CCElementalTable()
{
	vector<CCElement*>::iterator i = m_elements.begin();
	for ( ; i != m_elements.end(); ++i ) {
		if ( *i ) {
			delete * i;
			*i = NULL;
		}
	}
	m_elements.clear();
}

void CCElementalTable::ParseLine( const char* line )
{
    if ( CommentLine(line) || BlankLine(line) ) {
        return;
    }

    int num, maxbond;
    char symbol[5], name[256];
    double rConv, rVdw, mass, elNeg, ionize, elAffin, red, green, blue;

    if ( sscanf( line,
                 "%d %4s %lf %*f %lf %d %lf %lf %lf %lf %lf %lf %lf %255s",
                 &num, symbol, &rConv, &rVdw, &maxbond, &mass, &elNeg,
                 &ionize, &elAffin, &red, &green, &blue, name ) == 13 ) // only right format
    {
        CCElement* ccElement = new CCElement( num, symbol, rConv, rVdw,
                                              maxbond, mass, elNeg, ionize,
                                              elAffin, red, green, blue, name );
        AssertMemAlloc( ccElement );
        m_elements.push_back( ccElement );
    }
}

int CCElementalTable::GetNumofElements()
{
    if ( !m_init ) {
        Init();
    }

    return m_elements.size();
}

int CCElementalTable::GetAtomicNum( const char* symbol, int* iso )
{
    if ( !m_init ) {
        Init();
    }

    for ( vector<CCElement*>::iterator i = m_elements.begin();
          i != m_elements.end();
          ++i ) {
        if ( EQi( symbol, ( *i )->GetSymbol().c_str() ) ) {
            return ( *i )->GetAtomicNum();
        }
        else if ( EQi( symbol, "D" ) ) {
			if (iso) { 
				*iso = 2;
			}
            return 1;
        }
        else if ( EQi( symbol, "T" ) ) {
			if (iso) { 
				*iso = 3;
			}
            return 1;
        }
        else {
            continue;
        }
    }

    return 0;
}

string CCElementalTable::GetSymbol( int atomicNum )
{
    if ( !m_init ) {
        Init();
    }

    if ( atomicNum <0 || atomicNum> m_elements.size() ) {
        return "";
    }

    return m_elements[atomicNum]->GetSymbol();
}

double CCElementalTable::GetCovalentRad( int atomicNum )
{
    if ( !m_init ) {
        Init();
    }

    if ( atomicNum <0 || atomicNum> m_elements.size() ) {
        return 0.0;
    }

    return m_elements[atomicNum]->GetCovalentRad();
}

double CCElementalTable::GetVdwRad( int atomicNum )
{
    if ( !m_init ) {
        Init();
    }

    if ( atomicNum <0 || atomicNum> m_elements.size() ) {
        return 0.0;
    }

    return m_elements[atomicNum]->GetVdwRad();
}

int CCElementalTable::GetMaxBond( int atomicNum )
{
    if ( !m_init ) {
        Init();
    }

    if ( atomicNum <0 || atomicNum> m_elements.size() ) {
        return 0;
    }

    return m_elements[atomicNum]->GetMaxBond();
}

double CCElementalTable::GetMass( int atomicNum )
{
    if ( !m_init ) {
        Init();
    }

    if ( atomicNum <0 || atomicNum> m_elements.size() ) {
        return 0.0;
    }

    return m_elements[atomicNum]->GetMass();
}

double CCElementalTable::GetElectroNeg( int atomicNum )
{
    if ( !m_init ) {
        Init();
    }

    if ( atomicNum <0 || atomicNum> m_elements.size() ) {
        return 0.0;
    }

    return m_elements[atomicNum]->GetElectroNeg();
}

double CCElementalTable::GetIonization( int atomicNum )
{
    if ( !m_init ) {
        Init();
    }

    if ( atomicNum <0 || atomicNum> m_elements.size() ) {
        return 0.0;
    }

    return m_elements[atomicNum]->GetIonization();
}

double CCElementalTable::GetElectroAffin( int atomicNum )
{
    if ( !m_init ) {
        Init();
    }

    if ( atomicNum <0 || atomicNum> m_elements.size() ) {
        return 0.0;
    }

    return m_elements[atomicNum]->GetElectroAffin();
}

vector<double> CCElementalTable::GetRGB( int atomicNum )
{
    if ( !m_init ) {
        Init();
    }

    vector<double> colors( 3 );

    if ( atomicNum >= 0 && atomicNum <= m_elements.size() ) {
        colors[0] = m_elements[atomicNum]->GetRed();
        colors[1] = m_elements[atomicNum]->GetGreen();
        colors[2] = m_elements[atomicNum]->GetBlue();
    }

    return colors;
}

string CCElementalTable::GetName( int atomicNum )
{
    if ( !m_init ) {
        Init();
    }

    if ( atomicNum <0 || atomicNum> m_elements.size() ) {
        return "";
    }

    return m_elements[atomicNum]->GetName();
}

//////////////////////////////////////////////////////////////////////////
// class CCRefBondLengthTable 
//////////////////////////////////////////////////////////////////////////
CCRefBondLengthTable::CCRefBondLengthTable()
    : CCGlobalData( false, "CCBGKit_HOME", CCBGKit_HOME, "data", "bondlength.txt", BONDLENGTH )
{
}

CCRefBondLengthTable::~CCRefBondLengthTable()
{
    vector<CCRefBondLength*>::iterator i = m_refbondlength.begin();
    for ( ; i != m_refbondlength.end(); ++i ) {
        if ( *i ) {
            delete * i;
            *i = NULL;
        }
    }
}

void CCRefBondLengthTable::ParseLine( const char* line )
{
    if ( CommentLine(line) || BlankLine(line) ) {
        return;
    }

    int numbgn, numend;
    double length;

    if ( sscanf( line, "%d %d %lf", &numbgn, &numend, &length ) == 3 ) {
        CCRefBondLength* ccRefBondLength = new CCRefBondLength( numbgn,
                                                                numend,
                                                                length );
        AssertMemAlloc( ccRefBondLength );
        m_refbondlength.push_back( ccRefBondLength );
    }
}

int CCRefBondLengthTable::GetSize()
{
    if ( !m_init ) {
        Init();
    }

    return m_refbondlength.size();
}

double CCRefBondLengthTable::GetRefBondLength( int numbgn, int numend )
{
    if ( !m_init ) {
        Init();
    }

    // make the smaller atomic number to be begin atom and the larger to be end atom
    if ( numbgn > numend ) {
        int tmp = numbgn;
        numbgn = numend;
        numend = tmp;
    }

    vector<CCRefBondLength*>::iterator i;
    for ( i = m_refbondlength.begin(); i != m_refbondlength.end(); ++i ) {
        if ( numbgn == ( *i )->GetBgnAtomicNum()
          && numend == ( *i )->GetEndAtomicNum() ) {
            return ( *i )->GetRefBondLength();
        }
        else {
            continue;
        }
    }

    return 0.0;
}

//////////////////////////////////////////////////////////////////////////
// class CCAtomTypeTable 
//////////////////////////////////////////////////////////////////////////
CCAtomTypeTable::CCAtomTypeTable()
    : CCGlobalData( false, "CCBGKit_HOME", CCBGKit_HOME, "data", "atomtypes.txt", ATOMTYPES ),
      m_lineCount( 0 ), m_nCols( 0 ), m_from( -1 ), m_to( -1 )
{
}

void CCAtomTypeTable::ParseLine( const char* line )
{
    if ( CommentLine(line) || BlankLine(line) ) {
        return;
    } 

    if ( m_lineCount == 0 ) {
        sscanf( line, "%d", &m_nCols );
    }
    else if ( m_lineCount == 1 ) {
        Tokenize( m_colNames, line );
    }
    else {
        vector<string> tt;
        Tokenize( tt, line );
        if ( tt.size() == m_nCols ) {
            m_table.push_back( tt );
        }
    }
    m_lineCount++;
}

bool CCAtomTypeTable::SetFromType( const char* from )
{
    if ( !m_init ) {
        Init();
    }

    for ( unsigned i = 0; i < m_colNames.size(); ++i ) {
        if ( EQi( from, m_colNames[i].c_str() ) ) {
            m_from = i;
            return true;
        }
    }
    return false;
}

bool CCAtomTypeTable::SetToType( const char* to )
{
    if ( !m_init ) {
        Init();
    }

    for ( unsigned i = 0; i < m_colNames.size(); ++i ) {
        if ( EQi( to, m_colNames[i].c_str() ) ) {
            m_to = i;
            return true;
        }
    }
    return false;
}

string CCAtomTypeTable::GetFromType()
{
    if ( !m_init ) {
        Init();
    }

    if ( m_from > 0 && m_from < m_colNames.size() ) {
        return m_colNames[m_from];
    }
    else // reture INT by default
    {
        return m_colNames[0];
    }
}

string CCAtomTypeTable::GetToType()
{
    if ( !m_init ) {
        Init();
    }

    //if (m_to>0 && m_to < m_table.size())
    if ( m_to > 0 && m_to < m_colNames.size() ) {
        return m_colNames[m_to];
    }
    else // reture INT by default
    {
        return m_colNames[0];
    }
}

bool CCAtomTypeTable::Translate( const string from, string& to )
{
    if ( !m_init ) {
        Init();
    }

    if ( from.empty() ) {
        return false;
    }

	to.clear();
    if ( m_from >= 0
      && m_from < m_table.size()
      && m_to >= 0
      && m_to < m_table.size() ) {
        vector<vector<string> >::iterator i;
        for ( i = m_table.begin(); i != m_table.end(); ++i ) {
            if ( m_from < ( *i ).size() && EQi( ( *i )[m_from].c_str(),
                                                from.c_str() ) ) {
                to = ( *i )[m_to];
                return true;
            }
        }
    }

    return false;
}

void CCAtomTypeTable::ShowAtomTypes()
{
    if ( !m_init ) {
        Init();
    }

    for ( vector<vector<string> >::iterator it = m_table.begin();
          it != m_table.end();
          ++it ) {
        for ( int j = 0; j < m_nCols; ++j ) {
            cout << ( *it )[j] << " ";
        }
        cout << endl;
    }
}

//////////////////////////////////////////////////////////////////////////
// class CCBondTypeTable 
//////////////////////////////////////////////////////////////////////////
CCBondTypeTable::CCBondTypeTable()
    : CCGlobalData( false, "CCBGKit_HOME", CCBGKit_HOME, "data", "bondtypes.txt", BONDTYPES ),
      m_lineCount( 0 ), m_nCols( 0 ), m_from( -1 ), m_to( -1 )
{
}

void CCBondTypeTable::ParseLine( const char* line )
{
    if ( CommentLine(line) || BlankLine( line ) ) {
        return;
    } 

    if ( m_lineCount == 0 ) {
        sscanf( line, "%d", &m_nCols );
    }
    else if ( m_lineCount == 1 ) {
        Tokenize( m_colNames, line );
    }
    else {
        vector<string> tt;
        Tokenize( tt, line );
        if ( tt.size() == m_nCols ) {
            m_table.push_back( tt );
        }
    }
    m_lineCount++;
}

bool CCBondTypeTable::SetFromType( const char* from )
{
    if ( !m_init ) {
        Init();
    }

    for ( unsigned i = 0; i < m_colNames.size(); ++i ) {
        if ( EQi( from, m_colNames[i].c_str() ) ) {
            m_from = i;
            return true;
        }
    }
    return false;
}

bool CCBondTypeTable::SetToType( const char* to )
{
    if ( !m_init ) {
        Init();
    }

    for ( unsigned i = 0; i < m_colNames.size(); ++i ) {
        if ( EQi( to, m_colNames[i].c_str() ) ) {
            m_to = i;
            return true;
        }
    }
    return false;
}

string CCBondTypeTable::GetFromType()
{
    if ( !m_init ) {
        Init();
    }

    //if (m_from>0 && m_from < m_table.size())
    if ( m_from > 0 && m_from < m_colNames.size() ) {
        return m_colNames[m_from];
    }
    else // reture INT by default
    {
        return m_colNames[0];
    }
}

string CCBondTypeTable::GetToType()
{
    if ( !m_init ) {
        Init();
    }

    //if (m_to>0 && m_to < m_table.size())
    if ( m_to > 0 && m_to < m_colNames.size() ) {
        return m_colNames[m_to];
    }
    else // reture INT by default
    {
        return m_colNames[0];
    }
}

bool CCBondTypeTable::Translate( const string from, string& to )
{
    if ( !m_init ) {
        Init();
    }

    if ( from.empty() ) {
        return false;
    }

	to.clear();
    if ( m_from >= 0
      && m_from < m_table.size()
      && m_to >= 0
      && m_to < m_table.size() ) {
        vector<vector<string> >::iterator i;
        for ( i = m_table.begin(); i != m_table.end(); ++i ) {
            //if (m_from < (*i).size() && (*i)[m_from] == from)
            if ( m_from < ( *i ).size() && EQi( ( *i )[m_from].c_str(),
                                                from.c_str() ) ) {
                to = ( *i )[m_to];
                return true;
            }
        }
    }

    return false;
}

void CCBondTypeTable::ShowBondTypes()
{
    if ( !m_init ) {
        Init();
    }

    for ( vector<vector<string> >::iterator it = m_table.begin();
          it != m_table.end();
          ++it ) {
        for ( int j = 0; j < m_nCols; ++j ) {
            cout << ( *it )[j] << " ";
        }
        cout << endl;
    }
}

//////////////////////////////////////////////////////////////////////////
// class CCFFVDWTable 
//////////////////////////////////////////////////////////////////////////
CCFFVDWTable::CCFFVDWTable()
: CCGlobalData( false, "CCBGKit_HOME", CCBGKit_HOME, "data", "ffvdw.txt", FFVDW )
{
}

CCFFVDWTable::~CCFFVDWTable()
{
	map<string, CCFFVDW*>::iterator i = m_vdwelements.begin();
	for ( ; i != m_vdwelements.end(); ++i ) {
		if ( i->second ) {
			delete i->second;
			i->second = NULL;
		}
	}
	m_vdwelements.clear();
}

void CCFFVDWTable::ParseLine( const char* line )
{
	if ( CommentLine(line) || BlankLine(line) ) {
		return;
	}

	string inttype;
	double radius, epsilon;
	char ctype[10];

	if ( sscanf( line, "%s %*s %lf %lf", ctype, &radius, &epsilon ) == 3 ) {
		CCFFVDW* ccVDWElement = new CCFFVDW( radius, epsilon );
		AssertMemAlloc( ccVDWElement );
		inttype = ctype;
		m_vdwelements[inttype] = ccVDWElement;
	}
}

double CCFFVDWTable::GetVDWRadius( const string inttype ) {
	if ( m_vdwelements.find( inttype ) != m_vdwelements.end() ) {
		return m_vdwelements[inttype]->GetVDWRadius();
	}
	// return default vdw radius 1.7
	else {
		return 1.7;
	}
}
double CCFFVDWTable::GetVDWEpsilon( const string inttype ) {
	if ( m_vdwelements.find( inttype ) != m_vdwelements.end() ) {
		return m_vdwelements[inttype]->GetVDWEpsilon();
	}
	// return default vdw epsilon 0.107
	else {
		return 0.107;
	}
}

//////////////////////////////////////////////////////////////////////////
// class CCIntTyper 
//////////////////////////////////////////////////////////////////////////
CCIntTyper::CCIntTyper()
: CCGlobalData( false, "CCBGKit_HOME", CCBGKit_HOME, "data", "types.txt", TYPES ),
m_sp( NULL ), m_st( NULL )
{
}

CCIntTyper::~CCIntTyper()
{
	vector<pair<CCSmartsPattern*, vector<string>*> >::iterator it = m_fg.begin();
	for ( ; it != m_fg.end(); ++it ) {
		if ( it->first ) {
			delete it->first;
			it->first = NULL;
		}
		if ( it->second ) {
			it->second->clear(); 
			delete it->second; 
			it->second = NULL;
		}
	}
}

void CCIntTyper::ParseLine( const char* line )
{
	if ( CommentLine(line) || BlankLine( line ) ) {
		return;
	}

	vector<string> vs;

	if ( Tokenize( vs, line ) ) {
		if ( !m_sp ) {
			m_sp = new CCSmartsPattern; AssertMemAlloc( m_sp );
		}
		if ( !m_st ) {
			m_st = new vector<string>; AssertMemAlloc( m_st );
		}

		if ( vs[0] == "FGROUP" && vs.size() == 3 ) // functional group
		{
			m_sp->Init( vs[2] );
		}
		else if ( vs[0] == "CENTER" && vs.size() == 3 ) // center atom of functional group
		{
			std::copy( vs.begin() + 1, vs.end(), back_inserter( *m_st ) );
		}
		else if ( vs[0] == "BNDATM" && vs.size() == 5 ) // neighbors of center atom
		{
			std::copy( vs.begin() + 1, vs.end(), back_inserter( *m_st ) );
		}
		else if ( vs[0] == "END" ) {
			m_fg.push_back( make_pair( m_sp, m_st ) );
			m_sp = NULL;
			m_st = NULL;
		}
		else if ( vs.size() >= 2 ) // simple atom
		{
			m_sp->Init( vs[0] );
			std::copy( vs.begin() + 1, vs.end(), back_inserter( *m_st ) );
			m_fg.push_back( make_pair( m_sp, m_st ) );
			m_sp = NULL;
			m_st = NULL;
		}
		else {
			string errMsg = "Unable to parse fgroup.txt, LINE:\n";
			gErrorLog.Throw( __FUNCTION__, errMsg + line, ccWarning );
			return;
		}
	}
}

static bool sortByBondLength( pair<CCBond*, CCAtom*> p1,
							  pair<CCBond*, CCAtom*> p2 )
{
	return ( p1.first->GetLength() < p2.first->GetLength() );
}

void CCIntTyper::AssignIntType( CCMol& mol )
{
	if ( !m_init ) {
		Init();
	}

	vector<vector<int> >::iterator m;
	vector<pair<CCSmartsPattern*, vector<string>*> >::iterator it;

	for ( it = m_fg.begin(); it != m_fg.end(); ++it ) {
		CCSmartsPattern* &sp = it->first;
		vector<string>& st = *( it->second );

		int sz_of_st = st.size();

		if ( !sp || !sz_of_st ) {
			continue;
		}

		if ( sp->Match( mol ) ) {
			vector<vector<int> >& mlist = sp->GetUMapList();
			for ( m = mlist.begin(); m != mlist.end(); ++m ) {
				CCAtom* center = mol.GetAtom( ( *m )[0] );
				center->SetType( st[0] ); // internal type of central atom
				if ( sz_of_st % 4 == 2 && st[1] != "-" ) {
					center->SetFormalCharge( atoi( st[1].c_str() ) );
				}  // formal charge of central atom

				if ( sz_of_st <= 2 ) // simple atom
				{
					continue;
				}

				// sort equivalent neighbor from shortest bond length to highest
				// this needs very careful consideration
				// if 3 atoms matched, but we need only set 2 of them
				// just sort and assign first two of them
				// note: sz_of_st-2/4 is the number of neighbors to be assigned
				// which equals to vba.size()
				vector<pair<CCBond*, CCAtom*> > vba;
				for ( int i = 0, j = 1;
					i < ( sz_of_st - 2 ) / 4 && j < m->size();
					++i, ++j ) {
						pair<CCBond*, CCAtom*> ba = make_pair( mol.GetBond( ( *m )[j],
							( *m )[0] ),
							mol.GetAtom( ( *m )[j] ) );
						vba.push_back( ba );
				}
				// sort neighbor from shortest bond length to largest
				int sz_vba = vba.size();
				if ( sz_vba > 1 ) {
					bool bEqual( true ); // all nbr must be equal
					for ( int i = 0; i < sz_vba; ++i ) {
						if ( ( vba[i] ).second->GetAtomicNum()
							!= ( vba[( i + 1 ) % sz_vba] ).second->GetAtomicNum() ) {
								bEqual = false; break;
						}
					}
					if ( bEqual ) {
						std::sort( vba.begin(), vba.end(), sortByBondLength );
					}
				}

				for ( int k = 0; k < sz_vba; ++k ) {
					CCBond* nbrbond = vba[k].first;
					CCAtom* nbratom = vba[k].second;
					nbrbond->SetType( st[k * 4 + 2] ); // bond type
					nbrbond->SetOrder( atoi( st[k * 4 + 3].c_str() ) ); // bond order
					nbratom->SetType( st[k * 4 + 4] ); // atom type
					if ( st[k * 4 + 5] != "-" ) // '-' will retain original formal charge
					{
						nbratom->SetFormalCharge( atoi( st[k * 4 + 5].c_str() ) );
					} // atom formal charge
				}
			}
		}
	}
}

//////////////////////////////////////////////////////////////////////////
// class CCResidueData 
//////////////////////////////////////////////////////////////////////////
CCResidueData::RES::RES( const std::vector<string>& vs )
{
	//m_resname = vs[1];
	m_oneLetter = vs[2][0];
	m_charge = toDBL( vs[3] );
	m_logD = toDBL( vs[4] );
}

CCResidueData::ATM::ATM( const vector<string>& vs )
{
	//m_atmname = vs[1];
	m_intAtomType = vs[2];
	m_hyb = toINT( vs[3] );
	m_fcharge = toINT( vs[4] );
	m_ring = toINT( vs[5] );
	m_pcharge = toDBL( vs[6] );
	m_hbcharacter = vs[7];
	m_xtype = vs[8];
	m_logP = toDBL( vs[9] );
}

CCResidueData::BND::BND( const vector<string>& vs )
{
	m_atomName_1 = vs[1];
	m_atomName_2 = vs[2];
	m_bondOrder = toINT( vs[3] );
	m_intBondType = vs[4];
	m_ring = toINT( vs[5] );
}

CCResidueData::CCResidueData()
: CCGlobalData( false, "CCBGKit_HOME", CCBGKit_HOME, "data", "resdata.txt", RESDATA ),
m_pres( NULL ), m_patm( NULL ), m_pbnd( NULL ), m_currRes( "" )
{
}

CCResidueData::~CCResidueData()
{
	map<string, RES*>::iterator it;
	for ( it = m_res.begin(); it != m_res.end(); ++it ) {
		if ( it->second ) {
			delete it->second; it->second = NULL;
		}
	}
	m_res.clear();

	map<string, ATM*>::iterator jt;
	for ( jt = m_atm.begin(); jt != m_atm.end(); ++jt ) {
		if ( jt->second ) {
			delete jt->second; jt->second = NULL;
		}
	}
	m_atm.clear();

	map<string, BND*>::iterator kt;
	for ( kt = m_bnd.begin(); kt != m_bnd.end(); ++kt ) {
		if ( kt->second ) {
			delete kt->second; kt->second = NULL;
		}
	}
	m_bnd.clear();
}

void CCResidueData::ParseLine( const char* line )
{
	if ( CommentLine(line) || BlankLine( line ) ) {
		return;
	}

	vector<string> vs;

	if ( Tokenize( vs, line ) ) {
		if ( vs[0] == "RES" && vs.size() == 5 ) {
			m_currRes = vs[1];
			m_pres = new RES( vs ); AssertMemAlloc( m_pres );
			m_res.insert( make_pair( m_currRes, m_pres ) );
		}
		else if ( vs[0] == "ATOM" && vs.size() == 10 ) {
			m_patm = new ATM( vs ); AssertMemAlloc( m_patm );
			m_atm.insert( make_pair( m_currRes + " " + vs[1], m_patm ) );
		}
		else if ( vs[0] == "BOND" && vs.size() == 6 ) {
			m_pbnd = new BND( vs ); AssertMemAlloc( m_pbnd );
			string s = ( vs[1] < vs[2] )
				? vs[1]
			+ " " + vs[2]
			: vs[2]
			+ " " + vs[1];
			m_bnd.insert( make_pair( m_currRes + " " + s, m_pbnd ) );
		}
		else {
			string errMsg = "Unable to parse resdata.txt, LINE:\n";
			gErrorLog.Throw( __FUNCTION__, errMsg + line );
			return;
		}
	}
}

bool CCResidueData::AssignBond( CCMol& mol )
{
	if ( !m_init ) {
		Init();
	}

	// if mol is only consist of amino acid, ret = true
	// if mol contains non amino acid atoms, ret = false
	bool ret = true;
	CCAtomIter i, j;
	CCAtom* a1 = NULL, * a2 = NULL;
	void* p( NULL ); ATM* ap( NULL ); BND* bp( NULL );
	CCBond* bond( NULL );

	FOR_RESIDUES_OF_MOL( res, mol ) {
		vector<CCAtom*>& resatoms = res->GetAtoms();
		for ( i = resatoms.begin(); i != resatoms.end(); ++i ) {
			a1 = *i;
			// lookup bond
			for ( j = i + 1; j != resatoms.end(); ++j ) {
				a2 = *j;
				if ( !( p = LookupBond( a1, a2 ) ) ) {
					continue;
				}

				bp = static_cast<BND*>( p );

				if ( bond = mol.AddBond( a1->GetIdx(), a2->GetIdx(),
					bp->m_bondOrder ) ) {
						bond->SetType( bp->m_intBondType );
						if ( bp->m_ring > 0 ) {
							bond->SetInRing();
							if ( bp->m_ring == 2 ) {
								bond->SetAromatic();
							}
						}
				}
			}

			// lookup atom
			if ( a1->IsHydrogen() ) {
				// hydrogen
				a1->SetType( "H" );
				a1->SetHyb( 3 );
				continue;
			}

			if ( a1->IsOxygen() && a1->GetValence() == 0 ) {
				// WAT
				a1->SetType( "O.3" );
				a1->SetHyb( 3 );
				continue;
			}

			if ( !( p = LookupAtom( *a1 ) ) ) {
				clog << "ATOM " << res->GetSerNum( a1 )
					<< " of MOL " << a1->GetMol()->GetName()
					<< " with NAME: " << a1->GetName()
					<< " not defined in resdata.txt"  << endl;
				ret = false;
				continue;
			}

			ap = static_cast<ATM*>( p );

			a1->SetType( ap->m_intAtomType );
			a1->SetHyb( ap->m_hyb );
			a1->SetFormalCharge( ap->m_fcharge );
			a1->SetPartialCharge( ap->m_pcharge );
			// set ring info.
			if ( ap->m_ring > 0 ) {
				a1->SetInRing();
				if ( ap->m_ring == 2 ) {
					a1->SetAromatic();
				}
			}

			// set hb info.
			if ( ap->m_hbcharacter == "D" ) {
				a1->SetFlag( ATOM_IS_HBDONOR );
			}
			else if ( ap->m_hbcharacter == "A" ) {
				a1->SetFlag( ATOM_IS_HBACCEPTOR );
			}
			else if ( ap->m_hbcharacter == "DA" ) {
				a1->SetFlag( ATOM_IS_HBDONOR | ATOM_IS_HBACCEPTOR );
			}
		}
	}
	return ret;
}

void* CCResidueData::LookRes( CCResidue& res )
{
	string key = res.GetResName();

	if ( m_res.find( key ) != m_res.end() ) {
		return m_res[key];
	}

	return NULL; // RES*
}

void* CCResidueData::LookupAtom( CCAtom& atom )
{
	if ( !atom.GetResidue() ) {
		return NULL;
	}

	string str = atom.GetResidue()->GetResName() + " " + atom.GetName();    

	if ( m_atm.find( str ) != m_atm.end() ) {
		return m_atm[str];
	}
	// terminal
	string ter( "TER" ); ter += " " + atom.GetName();
	if ( m_atm.find( ter ) != m_atm.end() ) {
		return m_atm[ter];
	}
	// metal
	string mtl( "MTL" ); mtl += " " + atom.GetName();
	if ( m_atm.find( mtl ) != m_atm.end() ) {
		return m_atm[mtl];
	}

	return NULL; // ATM*
}

void* CCResidueData::LookupBond( CCAtom* a1, CCAtom* a2 )
{
	if ( !IsAcceptableBond( a1, a2 ) ) {
		return NULL;
	}

	CCResidue* res1 = a1-> GetResidue(), * res2 = a2->GetResidue();
	if ( !res1 || !res2 || res1 != res2 ) {
		return NULL;
	}

	string resname = res1->GetResName();
	string str1 = a1-> GetName(), str2 = a2->GetName();
	if ( a1->IsHydrogen() ) {
		Trim( str1, "0123456789" );
	}
	if ( a2->IsHydrogen() ) {
		Trim( str2, "0123456789" );
	}
	string a1a2 = str1 < str2 ? str1 + " " + str2 : str2 + " " + str1;

	resname += " " + a1a2;
	if ( m_bnd.find( resname ) != m_bnd.end() ) {
		return m_bnd[resname];
	}

	// terminal
	string ter = "TER";
	ter += " " + a1a2;
	if ( m_bnd.find( ter ) != m_bnd.end() ) {
		return m_bnd[ter];
	}

	return NULL; // BND*
} 

double CCResidueData::HpIndex( const string& resname )
{
	if ( !m_init ) {
		Init();
	}

	if ( m_res.find( resname ) != m_res.end() ) {
		return m_res[resname]->m_logD;
	}

	return 0.0;
}

char CCResidueData::OneLetter( const string& three )
{
	if ( !m_init ) {
		Init();
	}

	if ( three.size() == 3 && m_res.find( three ) != m_res.end() ) {
		return m_res[three]->m_oneLetter;
	}

	return 'X';
}

double CCResidueData::GetLogP( CCAtom& atom )
{
	if ( !m_init ) {
		Init();
	}

	ATM* atm = static_cast<ATM*>( LookupAtom( atom ) );
	if ( atm ) {
		return atm->m_logP;
	}

	return 0.0;
}

string CCResidueData::GetXPMFtype( CCAtom& atom )
{
	if ( !m_init ) {
		Init();
	}

	ATM* atm = static_cast<ATM*>( LookupAtom( atom ) );
	if ( atm ) {
		return atm->m_xtype;
	}

	return "";
}

string CCResidueData::HbCharacter( CCAtom& atom )
{
	if ( !m_init ) {
		Init();
	}

	ATM* atm = static_cast<ATM*>( LookupAtom( atom ) );
	if ( atm ) {
		return atm->m_hbcharacter;
	}

	return "";
}

//////////////////////////////////////////////////////////////////////////
// class CCGenericData 
//////////////////////////////////////////////////////////////////////////
CCGenericData::CCGenericData( const CCGenericData& src )
    : m_attr( src.m_attr ), m_type( src.m_type ), m_source( src.m_source )
{
}

CCGenericData& CCGenericData::operator =( const CCGenericData& src )
{
    if ( this != &src ) {
        m_attr = src.m_attr;
        m_type = src.m_type;
        m_source = src.m_source;
    }
    return *this;
}

//
//member functions for CCGMChargeData class
//
CCGMChargeData& CCGMChargeData::operator=( const CCGMChargeData& src ) {
	if ( this != &src ) {
		m_a = src.m_a;
		m_b = src.m_b;
		m_c = src.m_c;
		m_attr = src.m_attr;
		m_type = src.m_type;
		m_source = src.m_source;
	}
	return *this;
}

//
//member functions for CCExternalBondData class
//
//CCExternalBondData::CCExternalBondData()
//    : CCGenericData( "ExternalBondData", 
//	                 CCGenericDataType::ExternalBondData,
//					 perceived )
//{
//}

CCExternalBondData& CCExternalBondData::operator=( const CCExternalBondData& src )
{
	if ( this != &src ) {
		m_vexbnd = src.m_vexbnd;
		m_attr = src.m_attr;
		m_type = src.m_type;
		m_source = src.m_source;
	}
	return *this;
}

void CCExternalBondData::SetData( CCAtom* atom, CCBond* bond, int idx )
{
    CCExternalBond xb( atom, bond, idx );
    m_vexbnd.push_back( xb );
}

//
// Member functions for CCChiralDarta
//
// Chiral data is a perceived data type. We might read in some chiral info
// but this class derives and converts from whatever is read
CCChiralData& CCChiralData::operator=( const CCChiralData& src )
{
	if ( this != &src ) {
		m_atom4refs = src.m_atom4refs;
		m_atom4refo = src.m_atom4refo;
		m_atom4refc = src.m_atom4refc;
		m_parity = src.m_parity;
		m_attr = src.m_attr;
		m_type = src.m_type;
		m_source = src.m_source;
	}
	return *this;
}

void CCChiralData::Clear()
{
	m_atom4refs.clear();
	m_atom4refo.clear();
	m_atom4refc.clear();
	m_parity = 0;
}

bool CCChiralData::SetAtom4Refs( vector<int> atom4refs, atomreftype t )
{
    if ( atom4refs.size() > 4 ) {
        gErrorLog.Throw( __FUNCTION__, "More than 4 atoms in atom4refs",
                         ccDebug );
        return false;
    }
    switch ( t ) {
    case input:
        m_atom4refs = atom4refs;break;
    case output:
        m_atom4refo = atom4refs;break;
    case calcvolume:
        m_atom4refc = atom4refs;break;
    default:
        gErrorLog.Throw( __FUNCTION__, "AtomRefType called is invalid",
                         ccDebug );
        return false;
    }
    return true;
}

int CCChiralData::AddAtomRef( int atomref, atomreftype t )
{
    switch ( t ) {
    case input:
        m_atom4refs.push_back( atomref );break;
    case output:
        m_atom4refo.push_back( atomref );break;
    case calcvolume:
        m_atom4refc.push_back( atomref );break;
    default:
        gErrorLog.Throw( __FUNCTION__, "AtomRefType called is invalid",
                         ccDebug );
        return false;
    }

    return ( m_atom4refs.size() );
}

int CCChiralData::GetAtomRef( int a, atomreftype t )
{
    switch ( t ) {
    case input:
        return( m_atom4refs[a] );break;
    case output:
        return( m_atom4refo[a] );break;
    case calcvolume:
        return( m_atom4refc[a] );break;
    default:
        gErrorLog.Throw( __FUNCTION__, "AtomRefType called is invalid",
                         ccDebug );
        return false;
    }
}
vector<int> CCChiralData::GetAtom4Refs( atomreftype t ) const
{
    switch ( t ) {
    case output:
        return( m_atom4refo );
        break;
    case input:
        return( m_atom4refs );
        break;
    case calcvolume:
        return( m_atom4refc );
        break;
    default:
        gErrorLog.Throw( __FUNCTION__, "AtomRefType called is invalid",
                         ccDebug );
        return( m_atom4refo );
    }
}

int CCChiralData::GetSize( atomreftype t ) const
{
    switch ( t ) {
    case output:
        return( int ) m_atom4refo.size();
        break;
    case input:
        return( int ) m_atom4refs.size();
        break;
    case calcvolume:
        return( int ) m_atom4refc.size();
    default:
        gErrorLog.Throw( __FUNCTION__, "AtomRefType called is invalid",
                         ccDebug );
        return( 0 );
    }
}

//
// Member functions for CCFFTypeData
//
CCFFTypeData& CCFFTypeData::operator=( const CCFFTypeData& src )
{
	if ( this != &src ) {
		m_fftypes = src.m_fftypes;
		m_attr = src.m_attr;
		m_type = src.m_type;
		m_source = src.m_source;
	}
	return *this;
}

//
// Member functions for CCMultiAtomsData
//
CCMultiAtomsData& CCMultiAtomsData::operator=( const CCMultiAtomsData& src ) {
	if ( this != &src ) {
		m_multiatoms = src.m_multiatoms;
		m_attr = src.m_attr;
		m_type = src.m_type;
		m_source = src.m_source;
	}
	return *this;
}

CCMultiAtomsData::~CCMultiAtomsData() {
	vector<vector<CCAtom*> >::iterator vvi;
	vector<CCAtom*>::iterator vi;
	for ( vvi = m_multiatoms.begin(); vvi != m_multiatoms.end(); ++vvi ) {
		for ( vi = (*vvi).begin(); vi != (*vvi).end(); ++vi ) {
			if ( *vi ) {
				*vi = NULL;
			}
		}
	}
}

