#ifndef CCDATA_H
#define CCDATA_H

#include <CCConfig.h>
#include <ccbg/CCBitVec.h>
#include <ccbg/CCUtil.h>

class CCAtom;
class CCBond;
class CCResidue;
class CCMol;
class CCSmartsPattern;

// base class for global database (parameter)
class CCAPI CCGlobalData
{
public:
    CCGlobalData( bool init = false, string env = "", string homeDir = "",
                  string dataDir = "", string dataFile = "",
                  const char* pData = NULL )
        : m_init( init ), m_env( env ), m_homeDir( homeDir ),
          m_dataDir( dataDir ), m_dataFile( dataFile ), m_pData( pData )
    {
    }
    virtual ~CCGlobalData()
    {
    }

    bool Init();
    virtual void ParseLine( const char* line ) = 0;
protected:
    bool m_init;         ///< whether initilized
    string m_env;        ///< envionment variables
    string m_homeDir;    ///< home directory
    string m_dataDir;    ///< data directory
    string m_dataFile;   ///< data file
    const char* m_pData; ///< binary data of data file
};

class CCAPI CCElementalTable : public CCGlobalData
{
    class CCElement
    {
    public:
        CCElement( int num = 0, string symbol = "", double rConv = 0.0,
            double rVdw = 0.0, int maxbond = 0, double mass = 0.0,
            double elNeg = 0.0, double ionize = 0.0, double elAffin = 0.0,
            double red = 0.0, double green = 0.0, double blue = 0.0,
            string name = "" )
            : m_num( num ), m_symbol( symbol ), m_rConv( rConv ), m_rVdw( rVdw ),
            m_maxbond( maxbond ), m_mass( mass ), m_elNeg( elNeg ),
            m_ionize( ionize ), m_elAffin( elAffin ), m_red( red ),
            m_green( green ), m_blue( blue ), m_name( name )
        {
        }
        virtual ~CCElement()
        {
        }

        int GetAtomicNum()
        {
            return m_num;
        }
        string GetSymbol()
        {
            return m_symbol;
        }
        double GetCovalentRad()
        {
            return m_rConv;
        }
        double GetVdwRad()
        {
            return m_rVdw;
        }
        int GetMaxBond()
        {
            return m_maxbond;
        }
        double GetMass()
        {
            return m_mass;
        }
        double GetElectroNeg()
        {
            return m_elNeg;
        }
        double GetIonization()
        {
            return m_ionize;
        }
        double GetElectroAffin()
        {
            return m_elAffin;
        }
        double GetRed()
        {
            return m_red;
        }
        double GetGreen()
        {
            return m_green;
        }
        double GetBlue()
        {
            return m_blue;
        }
        string GetName()
        {
            return m_name;
        }
        void ShowElement();
    protected:
        int m_num, m_maxbond;
        string m_symbol, m_name;
        double m_rConv, m_rVdw, m_mass, m_elNeg, m_ionize, m_elAffin;
        double m_red, m_green, m_blue;
    };

public:
    CCElementalTable();
    virtual ~CCElementalTable();

    void ParseLine( const char* line );
    int GetNumofElements();
    int GetSize()
    {
        return GetNumofElements();
    };
    int GetAtomicNum( const char* symbol, int* iso=NULL );
    string GetSymbol( int atomicNum );
    double GetCovalentRad( int atomicNum );
    double GetVdwRad( int atomicNum );
    int GetMaxBond( int atomicNum );
    double GetMass( int atomicNum );
    double GetElectroNeg( int atomicNum );
    double GetIonization( int atomicNum );
    double GetElectroAffin( int atomicNum );
    vector<double> GetRGB( int atomicNum );
    string GetName( int atomicNum );
protected:
    vector<CCElement*> m_elements;
};
EXTERN CCElementalTable gElementalTable;

class CCAPI CCRefBondLengthTable : public CCGlobalData
{
    class CCRefBondLength
    {
    public:
        CCRefBondLength( int numbgn = 0, int numend = 0, double length = 0.0 )
            : m_numbgn( numbgn ), m_numend( numend ), m_length( length )
        {
        }
        virtual ~CCRefBondLength()
        {
        }

        int GetBgnAtomicNum()
        {
            return m_numbgn;
        }
        int GetEndAtomicNum()
        {
            return m_numend;
        }
        double GetRefBondLength()
        {
            return m_length;
        }
    protected:
        int m_numbgn;           // Atomic number of begin atom
        int m_numend;           // Atomic number of end atom
        double m_length;        // Reference single bond length
    };
public:
    CCRefBondLengthTable();
    virtual ~CCRefBondLengthTable();

    void ParseLine( const char* line );
    int GetSize();
    double GetRefBondLength( int numbgn, int numend );

protected:
    vector<CCRefBondLength*> m_refbondlength;
};
EXTERN CCRefBondLengthTable gRBLTable;

class CCAtomTypeTable : public CCGlobalData
{
protected:
    int m_lineCount;
    int m_nCols;
    int m_from, m_to;
    vector<string> m_colNames;
    vector<vector<string> > m_table;

public:
    CCAtomTypeTable();
    virtual ~CCAtomTypeTable()
    {
    }
    void ParseLine( const char* line );
    bool SetFromType( const char* from );
    bool SetToType( const char* to );
    string GetFromType();
    string GetToType();
    bool Translate( const string from, string& to );
    void ShowAtomTypes();
};
EXTERN CCAtomTypeTable gAtomTypeTable;

class CCBondTypeTable : public CCGlobalData
{
protected:
    int m_lineCount;
    int m_nCols;
    int m_from, m_to;
    vector<string> m_colNames;
    vector<vector<string> > m_table;

public:
    CCBondTypeTable();
    virtual ~CCBondTypeTable()
    {
    }
    void ParseLine( const char* line );
    bool SetFromType( const char* from );
    bool SetToType( const char* to );
    string GetFromType();
    string GetToType();
    bool Translate( const string from, string& to );
    void ShowBondTypes();
};
EXTERN CCBondTypeTable gBondTypeTable;

class CCFFVDWTable : public CCGlobalData
{
	class CCFFVDW
	{
	public:
		CCFFVDW( double radius = 0.0, double epsilon = 0.0 )
			: m_radius( radius ), m_epsilon( epsilon ) {
		}
		virtual ~CCFFVDW() {
		}

		double GetVDWRadius() {
			return m_radius;
		}
		double GetVDWEpsilon() {
			return m_epsilon;
		}
	protected:
		double m_radius, m_epsilon;
	};

public:
	CCFFVDWTable();
	virtual ~CCFFVDWTable();

	void ParseLine( const char* line );
	double GetVDWRadius( const string inttype );
	double GetVDWEpsilon( const string inttype );
protected:
	map<string, CCFFVDW*> m_vdwelements;
};
EXTERN CCFFVDWTable gFFVDWTable;

//////////////////////////////////////////////////////////////////////////
class CCAPI CCIntTyper : public CCGlobalData
{
public:
	CCIntTyper();
	virtual ~CCIntTyper();

	void ParseLine( const char* line );
	void AssignIntType( CCMol& mol );

protected:
	vector<pair<CCSmartsPattern*, vector<string>*> > m_fg;
	//variables used only temporarily for parsing fgroups.txt
	CCSmartsPattern* m_sp;
	vector<string>* m_st;
};
EXTERN CCIntTyper gIntTyper;

//////////////////////////////////////////////////////////////////////////
class CCAPI CCResidueData : public CCGlobalData
{
public:
	CCResidueData();
	virtual ~CCResidueData();

	void ParseLine( const char* line );
	bool AssignBond( CCMol& mol );
	void* LookRes( CCResidue& res );
	void* LookupAtom( CCAtom& atom );
	void* LookupBond( CCAtom* a1, CCAtom* a2 );
	double HpIndex( const string& resname );
	string HbCharacter( CCAtom& atom );
	char OneLetter( const string& three );
	double GetLogP( CCAtom& atom );
	string GetXPMFtype( CCAtom& atom );

protected:
	struct RES
	{
		RES( const vector<string>& vs );
		//string m_resname;
		char m_oneLetter;
		double m_charge;
		double m_logD;
	}* m_pres;
	struct ATM
	{
		ATM( const vector<string>& vs );
		//string m_atmname;
		string m_intAtomType;
		int m_hyb;
		int m_fcharge;
		int m_ring;
		double m_pcharge;
		string m_hbcharacter;
		string m_xtype;
		double m_logP;
	}* m_patm;
	struct BND
	{
		BND( const vector<string>& vs );
		string m_atomName_1;
		string m_atomName_2;
		int m_bondOrder;
		string m_intBondType;
		int m_ring;
	}* m_pbnd;

	map<string, RES*> m_res; // GLY => G ...
	map<string, ATM*> m_atm; // GLY-CA => C.3 ...
	map<string, BND*> m_bnd; // GLY-CA-N => SB ...

	// variables used only temporarily for parsing resdata.txt
	string m_currRes;
};
EXTERN CCResidueData gResidueData;
//////////////////////////////////////////////////////////////////////////

//////////////////////////////////////////////////////////////////////////
//! \brief Classification of data stored via CCGenericData class and subclasses.
//!
//! CCGenericDataType can be used as a faster, direct access to a particular category
//! instead of the slower access via GetData(string), which must loop
//! through all data to find a match with the supplied key. It is implemented
//! as a set of unsigned integer constants for maximum flexibility and future
//! expansion.
//! 
//! CustomData0 through CustomData15 are data slots that are not used in 
//! OpenBabel directly and are meant for use in derivative programs.
//! Macro definitions can be used to define what each data slot is used in your code.
namespace CCGenericDataType
{
//! Unknown data type (default)
static const int UndefinedData = 0;

//! Arbitrary key/value data, i.e., CCPairData
static const int PairData = 1;

//! Energetics data (e.g., total energy, heat of formation, etc.)
static const int EnergyData = 2;

//! Storing text comments (one per molecule, atom, bond, etc.) (for other data, e.g., author, keyword, ... use CCPairData)
static const int CommentData = 3;

//! Arbitrary information about conformers, i.e., CCConformerData
static const int ConformerData = 4;

//! Bond data external to OpenBabel, i.e., CCExternalBond, CCExternalBondData
static const int ExternalBondData = 5;

//! Information for generating & manipulating rotamers, i.e. CCRotamerList
static const int RotamerList = 6;

//! Info. for storing bonds to atoms yet to be added, i.e. CCVirtualBond
static const int VirtualBondData = 7;

//! Information on rings in a molecule, i.e., CCRingData
static const int RingData = 8;

//! Information about torsion/dihedral angles, i.e., CCTorsionData and CCTorsion
static const int TorsionData = 9;

//! Bond angles in a molecule, i.e., CCAngle, CCAngleData
static const int AngleData = 10;

//! Residue serial numbers
static const int SerialNums = 11;

//! Crystallographic unit cell data, i.e., CCUnitCell
static const int UnitCell = 12;

//! Spin data, including NMR, atomic and molecular spin, etc.
static const int SpinData = 13;

//! Arbitrary partial and total charges, dipole moments, etc.
static const int ChargeData = 14;

//! Symmetry data -- point and space groups, transforms, etc. i.e., CCSymmetryData
static const int SymmetryData = 15;

//! Arbitrary chiral information (atom, bond, molecule, etc.) i.e., CCChiralData
static const int ChiralData = 16;

//! Atomic and molecular occupation data
static const int OccupationData = 17;

//! Density (cube) data and surfaces
static const int DensityData = 18;

//! Electronic levels, redox states, orbitals, etc.
static const int ElectronicData = 19;

//! Vibrational modes, frequencies, etc.
static const int VibrationData = 20;

//! Rotational energy information
static const int RotationData = 21;

//! Nuclear transitions (e.g., decay, fission, fusion)
static const int NuclearData = 22;

//! Gasteiger & Marsili charge parameters
static const int GMChargeData = 23;

//! Force Field atom types
static const int FFTypeData = 24;

//! Angle, Torsion, Pair Atoms
static const int MultiAtomData = 25;
static const int AngleAtomData = 26;
static const int TorsionAtomData = 27;
static const int PairAtomData = 28;
// space for up to 2^14 more entries...

//! Custom (user-defined data)
static const int CustomData0 = 16384;
static const int CustomData1 = 16385;
static const int CustomData2 = 16386;
static const int CustomData3 = 16387;
static const int CustomData4 = 16388;
static const int CustomData5 = 16389;
static const int CustomData6 = 16390;
static const int CustomData7 = 16391;
static const int CustomData8 = 16392;
static const int CustomData9 = 16393;
static const int CustomData10 = 16394;
static const int CustomData11 = 16395;
static const int CustomData12 = 16396;
static const int CustomData13 = 16397;
static const int CustomData14 = 16398;
static const int CustomData15 = 16399;
} // end namespace

enum DataOrigin
{
    any,                 //!< Undefined or unspecified (default) 
    fileformatInput,     //!< Read from an input file
    userInput,           //!< Added by the user
    perceived,           //!< Perceived by Open Babel library methods
    external             //!< Added by an external program
};

class CCAPI CCGenericData
{
protected:
    string m_attr; //!< attribute tag (e.g., "UnitCell", "Comment" or "Author")
    int m_type; //!< attribute type -- declared for each subclass
    DataOrigin m_source;//!< source of data for accounting
public:
    CCGenericData( const string attr = "undefined",
                   const int type = CCGenericDataType::UndefinedData,
                   const DataOrigin  source = any )
        : m_attr( attr ), m_type( type ), m_source( source ) {
    }
    CCGenericData( const CCGenericData& );
	CCGenericData& operator=( const CCGenericData& src );
	virtual ~CCGenericData() {};
    virtual CCGenericData* Clone() = 0;

    void SetAttribute( const string& attr )
    {
        m_attr = attr;
    }
    virtual const string& GetAttribute() const
    {
        return m_attr ;
    }
    int GetDataType() const
    {
        return m_type;
    }
    //! Set the origin of this data, which can be used to filter the data
    void SetOrigin( const DataOrigin s )
    {
        m_source = s;
    }
    virtual const DataOrigin GetOrigin() const
    {
        return m_source;
    }
};
typedef std::vector<CCGenericData*>::iterator CCGenericDataIter;

//! \brief Used to store arbitrary attribute/value relationships of any type
template <class T>
class CCAPI CCPairTemplate : public CCGenericData
{
protected:
    T m_value; // the value of key/value pair
public:
    CCPairTemplate( const string attr = "PairData", const T value = T(),
                    const int type = CCGenericDataType::PairData,
                    const DataOrigin source = any )
        : CCGenericData( attr, type, source ), m_value( value ) {
    }
	CCPairTemplate( const CCPairTemplate &src ) 
		: m_value( src.m_value ),
		  CCGenericData( src.m_attr, src.m_type, src.m_source ) {
	}
	CCPairTemplate<T>& operator=( const CCPairTemplate<T>& src ) {
		if ( this != &src ) {
			m_value = src.m_value;
			CCGenericData( src.m_attr, src.m_type, src.m_source );
		}
		return *this;
	}
	~CCPairTemplate() {}
	CCPairTemplate* Clone()
	{
		return new CCPairTemplate( *this );
	}

    void SetValue( const T t )
    {
        m_value = t;
    }
    virtual const T& GetValue() const
    {
        return m_value;
    }
};
typedef CCPairTemplate<string> CCPairString;
typedef CCPairTemplate<int> CCPairInt;
typedef CCPairTemplate<double> CCPairDouble;

class CCAPI CCGMChargeData : public CCGenericData
{
protected:
    double m_a;          // anion value of a give atom
    double m_b;          // neutral value of a give atom
    double m_c;          // cation value of a give atom
public:
    CCGMChargeData( const string attr = "GMChargeData", double a = 0.0,
                    double b = 0.0, double c = 0.0,
                    const int type = CCGenericDataType::GMChargeData,
                    const DataOrigin source = any )
        : m_a( a ), m_b( b ), m_c( c ),
		  CCGenericData( attr, type, source ) {
    }
	CCGMChargeData( const CCGMChargeData &src ) 
		: m_a( src.m_a ), m_b( src.m_b ), m_c( src.m_c ),
		  CCGenericData( src.m_attr, src.m_type, src.m_source )
	{
	}
	CCGMChargeData& operator=( const CCGMChargeData& src );
	~CCGMChargeData() {}
	CCGMChargeData* Clone()
	{
		return new CCGMChargeData( *this );
	}

    void SetValues( const double a, const double b, const double c )
    {
        m_a = a;
        m_b = b;
        m_c = c;
    }
    double GetAnionValue() const
    {
        return m_a;
    }
    double GetNeutralValue() const
    {
        return m_b;
    }
    double GetCationValue() const
    {
        return m_c;
    }
};

//! \brief Used to store information on external bonds (e.g., in SMILES fragments)
class CCAPI CCExternalBondData : public CCGenericData
{
    //! \brief Used to store information on an external bond 
    //! (e.g., SMILES fragments)
    class CCExternalBond
    {
        int m_idx;
        CCAtom* m_atom;
        CCBond* m_bond;
    public:
        CCExternalBond( CCAtom* atom = NULL, 
			            CCBond* bond = NULL,
						int idx = 0 )
            : m_idx( idx ), m_atom( atom ), m_bond( bond ) {
        }
        CCExternalBond( const CCExternalBond& src )
            : m_idx( src.m_idx ), m_atom( src.m_atom ), m_bond( src.m_bond )
        {
        }
		CCExternalBond& operator=( const CCExternalBond& src )
		{
			if ( this != &src ) {
				m_idx = src.m_idx;
				m_atom = src.m_atom;
				m_bond = src.m_bond;
			}
			return *this;
		}
		~CCExternalBond() {
			m_idx = 0;
			m_atom = NULL;
			m_bond = NULL;
		}


        int GetIdx()  const
        {
            return( m_idx );
        }
        CCAtom* GetAtom() const
        {
            return( m_atom );
        }
        CCBond* GetBond() const
        {
            return( m_bond );
        }
        void SetIdx( int idx )
        {
            m_idx = idx;
        }
        void SetAtom( CCAtom* atom )
        {
            m_atom = atom;
        }
        void SetBond( CCBond* bond )
        {
            m_bond = bond;
        }
    };
protected:
    vector<CCExternalBond> m_vexbnd;
public:
    CCExternalBondData()
		: CCGenericData( "ExternalBondData", 
		                 CCGenericDataType::ExternalBondData,
						 perceived )
	{
		m_vexbnd.clear();
	}
	CCExternalBondData( const CCExternalBondData& src )
		: m_vexbnd( src.m_vexbnd ),
		  CCGenericData( src.m_attr, src.m_type, src.m_source )
	{
	}
	CCExternalBondData& operator=( const CCExternalBondData& src );
	~CCExternalBondData() {}
	CCExternalBondData* Clone()
	{
		return new CCExternalBondData( *this );
	}
    void SetData( CCAtom*, CCBond*, int );
    vector<CCExternalBond>* GetData()
    {
        return( &m_vexbnd );
    }
};

enum atomreftype
{
    output,
    input,
    calcvolume
}; // sets which atom4ref is accessed
//! \brief Used to hold chiral inforamtion about the atom as CCGenericData
class CCAPI CCChiralData : public CCGenericData
{
    friend class CCMol;
    friend class CCAtom;

protected:
    vector<int> m_atom4refs;
    vector<int> m_atom4refo;
    vector<int> m_atom4refc;
    int m_parity;

public:
    vector<int> GetAtom4Refs( atomreftype t ) const;
    int GetAtomRef( int a, atomreftype t );

    CCChiralData() 
		: CCGenericData( "ChiralData", 
		                 CCGenericDataType::ChiralData,
						 perceived ) 
	{
		m_atom4refs.clear();
		m_atom4refo.clear();
		m_atom4refc.clear();
		m_parity = 0;
	}
    CCChiralData( const CCChiralData& src )
		: m_atom4refs( src.m_atom4refs ),
		  m_atom4refo( src.m_atom4refo ),
		  m_atom4refc( src.m_atom4refc ),
		  m_parity( src.m_parity ),
		  CCGenericData( src.m_attr, src.m_type, src.m_source )
	{
	}
    CCChiralData& operator=( const CCChiralData& src );
    ~CCChiralData() {}
	CCChiralData* Clone()
	{
		return new CCChiralData( *this );
	}

    void Clear();
    bool SetAtom4Refs( vector<int> atom4refs, atomreftype t );
    int AddAtomRef( int atomref, atomreftype t );
    int GetSize( atomreftype t ) const;
};

//////////////////////////////////////////////////////////////////////////
// this data can be added to mol to indicate atom types in certain      //
// force field.                                                         //
//////////////////////////////////////////////////////////////////////////
class CCAPI CCFFTypeData : public CCGenericData
{
protected:
    vector<string> m_fftypes;   // force field atom types
public:
	CCFFTypeData()
		: CCGenericData( "FFTypeData",
		                 CCGenericDataType::FFTypeData,
						 perceived )
	{
	}
    CCFFTypeData( vector<string> fftypes, const string attr = "FFTypeData",
                  const int type = CCGenericDataType::FFTypeData,
                  const DataOrigin source = perceived )
        : m_fftypes( fftypes ), CCGenericData( attr, type, source )
    {
    }
	CCFFTypeData( const CCFFTypeData& src ) 
		: m_fftypes( src.m_fftypes ),
		  CCGenericData( src.m_attr, src.m_type, src.m_source )
	{
	}
	CCFFTypeData& operator=( const CCFFTypeData& src );
	~CCFFTypeData()
	{
	}
    CCFFTypeData* Clone()
    {
        return new CCFFTypeData( *this );
    }

    void SetFFTypes( vector<string>& fftypes )
    {
        m_fftypes = fftypes;
    }
    vector<string> GetFFTypes() const
    {
        return m_fftypes;
    }
};

///////////////////// added by Steven 08/05/15  //////////////////////////
// this data can be added to mol to indicate angle atoms, torsion atoms //
// and pair atoms in the molecule.                                      //
//////////////////////////////////////////////////////////////////////////
class CCAPI CCMultiAtomsData : public CCGenericData
{
protected:
    vector<vector<CCAtom*> > m_multiatoms;   // force field atom types
public:
	CCMultiAtomsData()
		: CCGenericData( "MultiAtoms",
		                 CCGenericDataType::MultiAtomData,
						 perceived )
	{
	}
    CCMultiAtomsData( vector<vector<CCAtom*> > multiatoms,
                      const string attr = "MultiAtoms",
                      const int type = CCGenericDataType::MultiAtomData,
                      const DataOrigin source = perceived )
        : m_multiatoms( multiatoms ), CCGenericData( attr, type, source )
    {
    }
	CCMultiAtomsData( const CCMultiAtomsData& src ) 
		: m_multiatoms( src.m_multiatoms ),
		  CCGenericData( src.m_attr, src.m_type, src.m_source )
	{
	}
	CCMultiAtomsData& operator=( const CCMultiAtomsData& src );
	~CCMultiAtomsData();
    CCMultiAtomsData* Clone()
    {
        return new CCMultiAtomsData( *this );
    }

    void SetMultiAtoms( vector<vector<CCAtom*> >& multiatoms )
    {
        m_multiatoms = multiatoms;
    }
    vector< vector<CCAtom*> > GetMultiAtoms() const
    {
        return m_multiatoms;
    }
};

#endif

