/** @file 
* @brief interface for handling chemical formats
* @author Yuan ZHAO, Tiejun CHENG
* @date 2008/08/23
* @version 1.0.1 
*
*/

#include <formats/CCFormat.h>
#include <ccbg/CCMol.h>

/**
* @brief new and return a map of id => format
* @remark static FMapType* ensure only one instance
*/
FMapType& FormatsMap()
{
    static FMapType* fm = NULL;
    if ( !fm ) {
        fm = new FMapType;
        AssertMemAlloc( fm );
    }

    return *fm;
}

/**
* @brief register format with id
* @param[in] id format id, e.g. mol2
* @param[in] pFormat format to be registered
* @return size of formats registered successfully
* @remark register pFormat in FormatsMap, any instance of derived class
*         of %CCFormat will register its format automatically, that's 
*         what we've done, e.g. theCCMDLFormat, theCCMol2Format
*/
int RegisterFormat( const string& id, CCFormat* pFormat )
{
    FormatsMap()[id] = pFormat;
    return FormatsMap().size();
}

/**
* @brief get format from file extension
* @param[in] fileName filename with extension
* @return found format, otherwise NULL
* @remark search the registered formats
*/
CCFormat* FormatFromExt( const string& fileName )
{
    string::size_type extPos = fileName.rfind( "." );
    CCFormat* fmt( NULL );

    if ( extPos != string::npos ) {
        fmt = GetFormat( fileName.substr( extPos + 1 ).c_str() );
    }

    if ( fmt == NULL ) {
        gErrorLog.Throw( __FUNCTION__, "Unsupported file format.", ccFatal );
        exit( EXIT_FAILURE );
    }

    return fmt;
}

/**
* @brief get format with id
* @param[in] id id to be searched (case-insensitive)
* @return found format, otherwise NULL
* @remark search the registered formats
*/
CCFormat* GetFormat( const string& id )
{
    //case insensitive
    CCFormat* fmt( NULL );
    if ( FormatsMap().find( id ) != FormatsMap().end() ) {
        fmt = FormatsMap()[id];
    }

    if ( fmt == NULL ) {
        gErrorLog.Throw( __FUNCTION__, "Unsupported file format.", ccFatal );
        exit( EXIT_FAILURE );
    }

    return fmt;
}

/** 
* @brief get option with key 
* @param[in] key key to be searched
* @return value if key exists,
*         "" otherwise
*/
string CCFormat::GetOption( const string& key )
{
    return m_popt.find( key ) == m_popt.end() ? "" : m_popt[key];
}

/** 
* @brief set key-value pair
* @param[in] key the key
* @param[in] value the value
* @note if key already exists, the former one will be overwritten
*/
void CCFormat::SetOption( const string& key, const string& value )
{
    m_popt[key] = value;
}

/** @brief has option of bopt ? */
bool CCFormat::HasOption( unsigned long bopt )
{
    return ( m_bopt & bopt ) ? true : false;
}

/** @brief set option of bopt */
void CCFormat::SetOption( unsigned long bopt )
{
    m_bopt |= bopt;
}

/** @brief unset option of bopt */
void CCFormat::UnsetOption( unsigned long bopt )
{
    m_bopt &= ~bopt;
}
