/** 
* @file 
* @brief interface for handling chemical formats
* @author Yuan ZHAO, Tiejun CHENG
* @date 2008/08/23
* @version 1.0.1 
*
* the following codes read a mol2 file (single or multi-mol2) 
* and print the molecular names
* @code
* CCFormat* fmt = FormatFromExt("test.mol2");
* // or fmt = GetFormat("mol2");
* ifstream ifs("test.mol2");
* CCMol mol;
* while (!ifs.eof()) {
*     fmt->ReadMolecule(mol, ifs);
*     cout << "read a molecule: " << mol->GetName() << endl;
* }
* ifs.close();
* @endcode
*
* Copyright (C) 2007-2008 www.sioc-ccbg.ac.cn
* 
* This file is part of the CCBGKit project.
*/

#ifndef CCFORMAT_H
#define CCFORMAT_H

#include <CCConfig.h>
#include <ccbg/CCUtil.h>

class CCMol;
class CCFormat;

typedef map<const string, CCFormat*, CharPtrLess> FMapType;

/** @brief new and return a map of id => format */
static FMapType& FormatsMap();
/** @brief get format with id */
CCFormat* GetFormat( const string& id );
/** @brief get format from file extension */
CCAPI CCFormat* FormatFromExt( const string& fileName );
/** @brief register format with id */
int RegisterFormat( const string& id, CCFormat* pFormat );

enum
{
    FMT_IS_VALID            = 1 << 0, ///< format is valid ?
    MDL_KEKULE_OUTPUT       = 1 << 1, ///< kekulize before write a MDL file
	MDL_KEEP_PROPERTY       = 1 << 2, ///< keep properties in original MDL file 
    MOL2_KEEP_COMMENT       = 1 << 3, ///< keep comments in original MOL2 file
    PDB_SKIP_WATER          = 1 << 4, ///< skip water when reading a PDB file
    PDB_SKIP_H              = 1 << 5, ///< skip hydrogen when reading a PDB file
	PDB_BREAK_METAL_BOND    = 1 << 6, ///< ignore metal bonds in CONECT section
};


/**
* @class   CCFormat CCFormat.h "formats/CCFormat.h"
* @brief   base class for handling chemical formats
*/
class CCAPI CCFormat
{
public:
	/** @brief constructor */
    CCFormat()
        : m_bopt( 0 )
    {
    }
	/** @brief destructor */
    virtual ~CCFormat()
    {
    }
	/**
	* @brief read a molecule
	* @note pure virtual function forces re-implementation in derived class
	*/
	virtual bool ReadMolecule(CCMol& mol, istream& is)=0;
	/**
	* @brief only read atoms in a molecule
	* @note pure virtual function requires re-implementation in derived class
	*/
	virtual bool ReadAtomsOnly(CCMol& mol, istream& is)=0;
	/**
	* @brief only read HETATM atoms ( exclude HOH ) in PDB file.(only for PDB)
	* @note pure virtual function requires re-implementation in derived class
	*/
	virtual bool ReadHetatmLig( CCMol& mol, istream& is ) {
		return false;
	}
	/**
	* @brief write a molecule
	* @note pure virtual function forces re-implementation in derived class
	*/
	virtual bool WriteMolecule(CCMol& mol, ostream& os)=0;
	/** @brief description of format */
    virtual const char* Description()
    {
        return "The base format\n";
    };

    /** @brief get option with key */
    string GetOption( const string& key );
	/** @brief set key-value pair */
    void SetOption( const string& key, const string& value );

    /** @brief has option of bopt ? */
    bool HasOption( unsigned long bopt );
	/** @brief set option of bopt */
    void SetOption( unsigned long bopt );
	/** @brief unset option of bopt */
    void UnsetOption( unsigned long bopt );

protected:
    map<string, string> m_popt; ///< key-value options
    unsigned long m_bopt;       ///< bitwise option
};

#endif
