/** 
* @file 
* @brief interface for MDL format
* @author Yuan ZHAO, Tiejun CHENG
* @date 2008/08/23
* @version 1.0.1 
*
*/

#include <formats/CCFormat.h>
#include <ccbg/CCMol.h>

/**
* @class   CCMDLFormat CCMDLFormat.cpp "formats/CCMDLFormat.cpp"
* @brief   MDL format interface
*
* provide read and write function for MDL format (sd, rd, sdf, 
* rdf, mol, mdl)
*/
class CCMDLFormat : public CCFormat
{
public:
	/** 
	* @brief constructor 
	* @note by default, properties will be saved and
	*       kekulize before writing a MDL file
	*/
    CCMDLFormat()
    {
        RegisterFormat( "sd", this );
        RegisterFormat( "rd", this );
        RegisterFormat( "sdf", this );
        RegisterFormat( "rdf", this );
        RegisterFormat( "mol", this );
        RegisterFormat( "mdl", this );
        SetOption( MDL_KEKULE_OUTPUT );
		SetOption( MDL_KEEP_PROPERTY );
    }
	/** @brief destructor */
    virtual ~CCMDLFormat()
    {
    }

	/** @brief read a molecule */
	bool ReadMolecule( CCMol& mol, istream& is );
	/** @brief only read atoms of a molecule */
	bool ReadAtomsOnly( CCMol& mol, istream& is );
	/** @brief write a molecule */
	bool WriteMolecule( CCMol& mol, ostream& os );

	/** @brief skip n molecules */
    bool SkipMolecule( istream& ifs, int& linenum, int n=1 );
	/** @brief read sdf v3000 */
    bool ReadV3000Block( istream& ifs, CCMol& mol );
    bool ReadV3000Line( istream& ifs, vector<string>& vs );
    bool ParseAtomBlock( istream& ifs, CCMol& mol );
    bool ParseBondBlock( istream& ifs, CCMol& mol );
	/** @brief write sdf v3000 */
    bool WriteV3000( ostream& ofs, CCMol& mol );
};
/** @brief make an instance and register %CCMDLFormat */
CCMDLFormat theCCMDLFormat;

/** 
* @brief generate time string
* @param[out] td resulting string
*/
char* GetTimeDate( char* td )
{
	//returns MMDDYYHHmm
	struct tm* ts;
	time_t long_time;
	time( &long_time );
	ts = localtime( &long_time ); //sizeof(td)
	snprintf( td, 11, "%02d%02d%02d%02d%02d\0", ts->tm_mon + 1, ts->tm_mday,
		( ( ts->tm_year >= 100 ) ? ts->tm_year - 100 : ts->tm_year ),
		ts->tm_hour, ts->tm_min );
	return td;
}

/**
* @brief skip n mol
* @param[in] ifs input file stream
* @param[out] linenum line number
* @param[in] n number of mol to be skipped
*
* skip certain mol when error encountered
*/
bool CCMDLFormat::SkipMolecule( istream& ifs, int& linenum, int n )
{
    if ( n == 0 ) {
        n++;
    }
    char line[BUFF_SIZE];
	while( ifs.getline( line, BUFF_SIZE ) ) {
		linenum++;
		if (EQn( line, "$$$$", 4 ) && (--n)==0 ) {
			break;
		}
	}
    return ifs.good();
}

/**
* @brief read a molecule
* @param[in] mol reference of molecule
* @param[in] is input file stream
*
* Return false when error encountered, else return true.
*/
bool CCMDLFormat::ReadMolecule( CCMol& mol, istream& is )
{
    // clean molecule
    mol.Clear();

    // record current line number
    static int LINENUM;
    // for buffer
    char buff[BUFF_SIZE];
    // for kekulize
    bool bKekule( false );

    // line 1 of header block -- molecular name
    if ( !is.getline( buff, BUFF_SIZE ) ) {
        ++LINENUM;
        SkipMolecule( is, LINENUM );
        return false;
    }
    string name = buff;
    Trim( name );
    mol.SetName( name );

    // line 2 of header block -- creator
    if ( !is.getline( buff, BUFF_SIZE ) ) {
        ++LINENUM;
        SkipMolecule( is, LINENUM );
        return false;
    }
    if ( strlen( buff ) > 20 && buff[20] != ' ' ) {
        mol.SetDimension( toINT( buff[20] ) );
    }

    // line 3 of header block -- comment
    if ( !is.getline( buff, BUFF_SIZE ) ) {
        ++LINENUM;
        SkipMolecule( is, LINENUM );
        return false;
    }

	// line 4 -- number of atoms and bonds
    if ( !is.getline( buff, BUFF_SIZE ) ) {
        ++LINENUM;
        SkipMolecule( is, LINENUM );
        return false;
    }

    int i = 0, nAtoms = 0, nBonds = 0;
    string line = buff;

    nAtoms = atoi( line.substr( 0, 3 ).c_str() );
	// skip empty molecule
	if ( nAtoms == 0 ) {
		SkipMolecule( is, LINENUM );
		return false;
	}
    nBonds = atoi( line.substr( 3, 3 ).c_str() );

    if ( line.find( "V3000" ) != string::npos ) {
        if ( !ReadV3000Block( is, mol ) ) {
            return false;
        }
    }
    else {
        double x( 0.0 ), y( 0.0 ), z( 0.0 );
        string symbol;
        int iso( 0 ), fcharge( 0 ), stereo( 0 );

        mol.ResizeAtoms( nAtoms );
        for ( i = 0; i < nAtoms; ++i ) {
            if ( !is.getline( buff, BUFF_SIZE ) ) {
                ++LINENUM;
                SkipMolecule( is, LINENUM );
                return false;
            }

            CCAtom atom; // a clean atom

            if ( strlen( buff ) < 34 ) {
                SkipMolecule( is, LINENUM );
                return false;
            }
            line = buff;

            x = atof( line.substr( 0, 10 ).c_str() );
            y = atof( line.substr( 10, 10 ).c_str() );
            z = atof( line.substr( 20, 10 ).c_str() );
            atom.SetXYZ( x, y, z );

            symbol = line.substr( 31, 3 );
            Trim( symbol );
            atom.SetName( symbol );
            atom.SetAtomicNum( gElementalTable.GetAtomicNum( symbol.c_str() ) );

			if ( strlen( buff ) > 33 ) {
				iso = atoi( line.substr( 33, 3 ).c_str() );
				atom.SetIsotope( iso );
			}
            if ( strlen( buff ) > 36 ) {
                fcharge = atoi( line.substr( 36, 3 ).c_str() );
                switch ( fcharge ) {
                case 0:
                    break;
                case 3:
                    atom.SetFormalCharge( 1 ); break;
                case 2:
                    atom.SetFormalCharge( 2 ); break;
                case 1:
                    atom.SetFormalCharge( 3 ); break;
                case 5:
                    atom.SetFormalCharge( -1 ); break;
                case 6:
                    atom.SetFormalCharge( -2 ); break;
                case 7:
                    atom.SetFormalCharge( -3 ); break;
                }
            }
            if ( strlen( buff ) > 39 ) // set a stereo mark
            {
                stereo = atoi( line.substr( 39, 3 ).c_str() );
                // Stereo configuration:
                // 0 none; 1 odd parity; 2 even parity; 3 unspecified)
                if ( stereo != 0 ) {
                    atom.SetChiral();
                }
            }

            if ( !mol.AddAtom( atom ) )  // add atom ERROR
            {
                return false;
            }
        }

        int start, end, order;
        mol.ResizeBonds( nBonds );
        for ( i = 0; i < nBonds; ++i ) {
            if ( !is.getline( buff, BUFF_SIZE ) ) {
                ++LINENUM;
                SkipMolecule( is, LINENUM );
                return false;
            }

            if ( strlen( buff ) < 9 ) {
                return false;
            }
            line = buff;

            start = atoi( line.substr( 0, 3 ).c_str() );
            end = atoi( line.substr( 3, 3 ).c_str() );
            order = atoi( line.substr( 6, 3 ).c_str() );
            //////////////////////////////////
            if ( order == 4 ) {
                order = 5;
                bKekule = true;
            }
            //////////////////////////////////
            if ( start == 0
              || end == 0
              || order == 0
              || start > mol.GetNumAtom()
              || end > mol.GetNumAtom() ) {
                return false;
            }

            if ( !mol.AddBond( start, end, order ) ) // add bond ERROR
            {
                ++LINENUM;
                SkipMolecule( is, LINENUM );
                return false;
            }
        }
    }

    // molecule has connection table from MDL format
    mol.SetFlag( MOL_CONNECTIVITY_PERCEIVED );

    // set dimension
    mol.SetDimensionFromXYZ();

    // kekulize if necessary
    if ( bKekule ) {
        mol.Kekulize();
        // clear flag for output
        mol.UnsetFlag( MOL_KEKULE_PERCEIVED );
    }

	if ( HasOption( MDL_KEEP_PROPERTY ) ) {
		// for generic data
		string properties;
		properties.clear();
		//Get property lines
		while ( is.getline( buff, BUFF_SIZE ) ) {
			++LINENUM;
			if ( EQn( buff, "$$$$", 4 ) || EQn( buff, "$MOL", 4 ) ||
				 EQn( buff, "M  END", 6 ) ) {
					break;
			}
			line = buff;
			properties += line + '\n';
		}

		if ( EQn( buff, "M  END", 6 ) ) {
			if ( !BlankLine( properties ) ) {
				CCGenericData* gd = new CCPairString( "#PREMENDPROP", properties );
				AssertMemAlloc( gd );
				mol.SetData( gd );
			}

			properties.clear();
			//Get property lines
			while ( is.getline( buff, BUFF_SIZE ) ) {
				++LINENUM;
				if ( EQn( buff, "$$$$", 4 ) || EQn( buff, "$MOL", 4 ) ) {
					break;
				}
				line = buff;
				properties += line + '\n';
			}

			if ( !BlankLine( properties ) ) {
				CCGenericData* gd = new CCPairString( "#POSTMENDPROP", properties );
				AssertMemAlloc( gd );
				mol.SetData( gd );
			}
		}
		else if ( !BlankLine( properties ) ) {
			CCGenericData* gd = new CCPairString( "#MDLPROPERTY", properties );
			AssertMemAlloc( gd );
			mol.SetData( gd );
		}
	}
	else {
		// ignore properties
		while ( is.getline( buff, BUFF_SIZE ) ) {
			++LINENUM;
			if ( EQn( buff, "$$$$", 4 ) || EQn( buff, "$MOL", 4 ) ) {
				break;
			}
		}
	}

	streampos oldPos = is.tellg();
	// eat up remaining blank lines if no more molecule exist
	while ( is.peek() != EOF && is.good() && 
		   ( is.peek() == '\n' || is.peek() == '\r' ) ) {
		is.getline( buff, BUFF_SIZE );
	}
	// more molecule exist, reset ifs.
	if ( !is.eof() ) {
		is.seekg( oldPos );
	}

    return true;
}

/**
* @brief only read atoms of a molecule
* @param[in] mol reference of molecule
* @param[in] is input file stream
*
* It only read atoms information in MDL file. Bond information
* will simply discarded. Return false when error encountered, 
* else return true.
*/
bool CCMDLFormat::ReadAtomsOnly( CCMol& mol, istream& is )
{
	// clean molecule
	mol.Clear();

	// record current line number
	static int LINENUM;
	// for buffer
	char buff[BUFF_SIZE];

	// line 1 of header block -- molecular name
	if ( !is.getline( buff, BUFF_SIZE ) ) {
		++LINENUM;
		SkipMolecule( is, LINENUM );
		return false;
	}
	string name = buff;
	Trim( name );
	mol.SetName( name );

	// line 2 of header block -- creator
	if ( !is.getline( buff, BUFF_SIZE ) ) {
		++LINENUM;
		SkipMolecule( is, LINENUM );
		return false;
	}
	if ( strlen( buff ) > 20 && buff[20] != ' ' ) {
		mol.SetDimension( toINT( buff[20] ) );
	}

	// line 3 of header block -- comment
	if ( !is.getline( buff, BUFF_SIZE ) ) {
		++LINENUM;
		SkipMolecule( is, LINENUM );
		return false;
	}

	// line 4 -- number of atoms and bonds
	if ( !is.getline( buff, BUFF_SIZE ) ) {
		++LINENUM;
		SkipMolecule( is, LINENUM );
		return false;
	}

	int i = 0, nAtoms = 0, nBonds = 0;
	string line = buff;

	nAtoms = atoi( line.substr( 0, 3 ).c_str() );
	// skip empty molecule
	if ( nAtoms == 0 ) {
		SkipMolecule( is, LINENUM );
		return false;
	}
	nBonds = atoi( line.substr( 3, 3 ).c_str() );

	if ( line.find( "V3000" ) != string::npos ) {
		if ( !ReadV3000Block( is, mol ) ) {
			return false;
		}
	}
	else {
		double x( 0.0 ), y( 0.0 ), z( 0.0 );
		string symbol;
		int iso( 0 ), fcharge( 0 ), stereo( 0 );

		mol.ResizeAtoms( nAtoms );
		for ( i = 0; i < nAtoms; ++i ) {
			if ( !is.getline( buff, BUFF_SIZE ) ) {
				++LINENUM;
				SkipMolecule( is, LINENUM );
				return false;
			}

			CCAtom atom; // a clean atom

			if ( strlen( buff ) < 34 ) {
				SkipMolecule( is, LINENUM );
				return false;
			}
			line = buff;

			x = atof( line.substr( 0, 10 ).c_str() );
			y = atof( line.substr( 10, 10 ).c_str() );
			z = atof( line.substr( 20, 10 ).c_str() );
			atom.SetXYZ( x, y, z );

			symbol = line.substr( 31, 3 );
			Trim( symbol );
			atom.SetName( symbol );
			atom.SetAtomicNum( gElementalTable.GetAtomicNum( symbol.c_str() ) );

			if ( strlen( buff ) > 33 ) {
				iso = atoi( line.substr( 33, 3 ).c_str() );
				atom.SetIsotope( iso );
			}
			if ( strlen( buff ) > 36 ) {
				fcharge = atoi( line.substr( 36, 3 ).c_str() );
				switch ( fcharge ) {
				case 0:
					break;
				case 3:
					atom.SetFormalCharge( 1 ); break;
				case 2:
					atom.SetFormalCharge( 2 ); break;
				case 1:
					atom.SetFormalCharge( 3 ); break;
				case 5:
					atom.SetFormalCharge( -1 ); break;
				case 6:
					atom.SetFormalCharge( -2 ); break;
				case 7:
					atom.SetFormalCharge( -3 ); break;
				}
			}
			if ( strlen( buff ) > 39 ) // set a stereo mark
			{
				stereo = atoi( line.substr( 39, 3 ).c_str() );
				// Stereo configuration:
				// 0 none; 1 odd parity; 2 even parity; 3 unspecified)
				if ( stereo != 0 ) {
					atom.SetChiral();
				}
			}

			if ( !mol.AddAtom( atom ) )  // add atom ERROR
			{
				return false;
			}
		}

		// read and skip bond information
		for ( i = 0; i < nBonds; ++i ) {
			if ( !is.getline( buff, BUFF_SIZE ) ) {
				++LINENUM;
				SkipMolecule( is, LINENUM );
				return false;
			}
		}
	}

	// set dimension
	mol.SetDimensionFromXYZ();

	if ( HasOption( MDL_KEEP_PROPERTY ) ) {
		// for generic data
		string properties;
		properties.clear();
		//Get property lines
		while ( is.getline( buff, BUFF_SIZE ) ) {
			++LINENUM;
			if ( EQn( buff, "$$$$", 4 ) || EQn( buff, "$MOL", 4 ) ||
				EQn( buff, "M  END", 6 ) ) {
					break;
			}
			line = buff;
			properties += line + '\n';
		}

		if ( EQn( buff, "M  END", 6 ) ) {
			if ( !BlankLine( properties ) ) {
				CCGenericData* gd = new CCPairString( "#PREMENDPROP", properties );
				AssertMemAlloc( gd );
				mol.SetData( gd );
			}

			properties.clear();
			//Get property lines
			while ( is.getline( buff, BUFF_SIZE ) ) {
				++LINENUM;
				if ( EQn( buff, "$$$$", 4 ) || EQn( buff, "$MOL", 4 ) ) {
					break;
				}
				line = buff;
				properties += line + '\n';
			}

			if ( !BlankLine( properties ) ) {
				CCGenericData* gd = new CCPairString( "#POSTMENDPROP", properties );
				AssertMemAlloc( gd );
				mol.SetData( gd );
			}
		}
		else if ( !BlankLine( properties ) ) {
			CCGenericData* gd = new CCPairString( "#MDLPROPERTY", properties );
			AssertMemAlloc( gd );
			mol.SetData( gd );
		}
	}
	else {
		// ignore properties
		while ( is.getline( buff, BUFF_SIZE ) ) {
			++LINENUM;
			if ( EQn( buff, "$$$$", 4 ) || EQn( buff, "$MOL", 4 ) ) {
				break;
			}
		}
	}

	streampos oldPos = is.tellg();
	// eat up remaining blank lines if no more molecule exist
	while ( is.peek() != EOF && is.good() && 
		( is.peek() == '\n' || is.peek() == '\r' ) ) {
			is.getline( buff, BUFF_SIZE );
	}
	// more molecule exist, reset ifs.
	if ( !is.eof() ) {
		is.seekg( oldPos );
	}

	return true;
}


/**
* @brief write a molecule
* @param[in] mol reference of molecule
* @param[out] os output file stream
*
* interface for writing MDL file, one molecule a time
*/
bool CCMDLFormat::WriteMolecule( CCMol& mol, ostream& os )
{
    if ( mol.IsEmpty() ) {
        gErrorLog.Throw( __FUNCTION__, "Molecule is empty." );
        return false;
    }

    char line[BUFF_SIZE];

    // line 1
    string molName = mol.GetName();
    if ( molName.empty() ) {
        //gErrorLog.Throw( __FUNCTION__,
        //                 "NULL molecular name! **** by default.", ccWarning );
        os << "****\n";
    }
    else {
		// set max length of molecular name as 50 characters
		os << molName.substr( 0, 50 ) << "\n";
    }

    char td[11] = "0000000000";
    os << "   CCBGKit" << GetTimeDate( td ) << mol.GetDimension() << "D\n"; //line 2
    os << "\n";  // line 3

    if ( mol.GetNumAtom() > 999 || mol.GetNumBond() > 999 ) {
        if ( !WriteV3000( os, mol ) ) {
            return false;
        }
    }
    else {
        snprintf( line, BUFF_SIZE, "%3d%3d%3d%3d%3d%3d%3d%3d%3d%3d%3d V2000",
                  mol.GetNumAtom(), mol.GetNumBond(), 0, 0, 
				  mol.GetNumChiralAtom(), 0, 0, 0, 0, 0,
                  999 );
        os << line << "\n";

        mol.SetAutoFormalCharge( true );

        int charge;
        FOR_ATOMS_OF_MOL( atom, mol ) {
            switch ( atom->GetFormalCharge() ) {
            case  1:
                charge = 3; break;
            case  2:
                charge = 2; break;
            case  3:
                charge = 1; break;
            case -1:
                charge = 5; break;
            case -2:
                charge = 6; break;
            case -3:
                charge = 7; break;
            default:
                charge = 0; break;
            }

			int chiral( 0 );
			if ( atom->HasFlag( ATOM_IS_CHIRAL ) ) {
				if ( atom->HasFlag( ATOM_IS_RCHIRAL ) ) {
					chiral = 1;
				}
				else if ( atom->HasFlag( ATOM_IS_SCHIRAL ) ) {
					chiral = 2;
				}
				else {
					chiral = 3;
				}
			}
            snprintf( line, BUFF_SIZE,
                      "%10.4f%10.4f%10.4f %-3s%2d%3d%3d%3d%3d", atom->GetX(),
                      atom->GetY(), atom->GetZ(), atom->GetSymbol().c_str(),
                      atom->GetIsotope() , charge, chiral, 0, 0 );
            os << line << "\n";
        }

        gBondTypeTable.SetFromType( "INT" );
        gBondTypeTable.SetToType( "MDL" );
        string fromType, toType;


        FOR_BONDS_OF_MOL( bond, mol ) {
            fromType = bond->GetType();
            gBondTypeTable.Translate( fromType, toType );

            if ( HasOption( MDL_KEKULE_OUTPUT ) ) {
                // kekulize, aromatic bond in such carboxylate will not be affected
                if ( bond->IsKTriple() ) {
                    toType = "3";
                }
                else if ( bond->IsKDouble() ) {
                    toType = "2";
                }
                else if ( bond->IsKSingle() ) {
                    toType = "1";
                }
            }

            snprintf( line, BUFF_SIZE, "%3d%3d%3d%3d%3d%3d",
                      bond->GetBeginAtomIdx(), bond->GetEndAtomIdx(),
                      atoi( toType.c_str() ), 0, 0, 0 );
            os << line << "\n";
        }
    }

	if ( HasOption( MDL_KEEP_PROPERTY ) ) {
		// ****** #PREMENDPROP/#POSTMENDPROP/#MDLPROPERTY ******
		// for generic data
		CCGenericData* gd( NULL );
		if ( gd = mol.GetData( "#PREMENDPROP" ) ) {
			os << dynamic_cast<CCPairString*>( gd )->GetValue();
			os << "M  END\n";
			if ( gd = mol.GetData( "#POSTMENDPROP" ) ) {
				os << dynamic_cast<CCPairString*>( gd )->GetValue();
				os << "$$$$" << endl;
			}
		}
		else if ( gd = mol.GetData( "#POSTMENDPROP" ) ) {
			os << "M  END\n";
			os << dynamic_cast<CCPairString*>( gd )->GetValue();
			os << "$$$$" << endl;
		}
		else if ( gd = mol.GetData( "#MDLPROPERTY" ) ) {
			os << dynamic_cast<CCPairString*>( gd )->GetValue();
			os << "$$$$" << endl;
		}
		else {
			os << "M  END\n";
			os << "$$$$" << endl;
		}
	}
	else {
		os << "M  END\n";
		os << "$$$$" << endl;
	}

    return true;
}

/** 
* @brief read sdf v3000
* @note undone
*/
bool CCMDLFormat::ReadV3000Block( istream& ifs, CCMol& mol )
{
    cout << "Read MDL V3000 Format...\n";
    return false;
}

/** 
* @brief write sdf v3000
* @note undone
*/
bool CCMDLFormat::WriteV3000( ostream& ofs, CCMol& mol )
{
    cout << "Write MDL V3000 Format...\n";
    return false;
}
