/** 
* @file 
* @brief interface for Tripos Mol2 format
* @author Yuan ZHAO, Tiejun CHENG
* @date 2008/08/23
* @version 1.0.1 
*
*/

#include <formats/CCFormat.h>
#include <ccbg/CCMol.h>

/**
* @class   CCMol2Format CCMol2Format.cpp "CCMol2Format.cpp"
* @brief   Tripos Mol2 format interface
*
* provide read and write function for Mol2 format
*/
class CCMol2Format : public CCFormat
{
public:
	/** 
	* @brief constructor 
	* @note by default, comment line will be kept
	*/
    CCMol2Format()
    {
        RegisterFormat( "mol2", this );
        SetOption( MOL2_KEEP_COMMENT );
    }
	/** @brief destructor */
    virtual ~CCMol2Format()
    {
    }

	/** @brief read a molecule */
    bool ReadMolecule( CCMol& mol, istream& is );
	/** @brief only read atoms of a molecule */
	bool ReadAtomsOnly( CCMol& mol, istream& is );
	/** @brief write a molecule */
    bool WriteMolecule( CCMol& mol, ostream& os );

	/** @brief description of Tripos Mol2 format */
    virtual const char* Description()
    {
        return
        "Sybyl Mol2 format\n" "http://www.tripos.com\n";
    };
};
/** @brief make an instance and register %CCMol2Format */
CCMol2Format theCCMol2Format;  // make an instant and register format

/**
* @brief read a molecule
* @param[out] mol mol reference of molecule
* @param[in] is input file stream
*
* interface for reading Tripos Mol2 file, one molecule a time
*/
bool CCMol2Format::ReadMolecule( CCMol& mol, istream& is )
{
    // seek beginning of next molecule if possible or error
    // clean molecule
    mol.Clear();

    // for error
    bool markError( false );
    // for buffer
    string line;
    // for Molecule
    int nAtoms(0), nBonds(0), nSubs(0);
    bool foundMoleLine( false ), foundAtomLine( false ),
         foundBondLine( false );
    // for ATOM
    int subst_id( -1 );
    char atom_name[BUFF_SIZE], atom_type[BUFF_SIZE], subst_type[BUFF_SIZE];
    double x, y, z, charge;
    string fromType, toType;
    // for BOND
    int start, end, order;
    char bond_type[BUFF_SIZE];
    // for back
    streampos oldPos;
    // for kekulize
    bool bKekule( false );
    // for generic data
    string comment;
    CCGenericData* gd( NULL );

    // skip to @<TRIPOS>MOLECULE, and record #COMMENT if any
    for ( ; ; ) {
        if ( !getline( is, line ) ) {
            return false;
        }

        if ( EQn( line.c_str(), "@<TRIPOS>MOLECULE", 17 ) ) {
            foundMoleLine = true;
            break;
        }
        else if ( EQn( line.c_str(), "@<TRIPOS>ATOM", 13 )
               || EQn( line.c_str(), "@<TRIPOS>BOND", 13 ) ) {
            markError = true;
            goto BeginOfNextMolecule;
        }
        else {
            // all others were regarded as heading comments
            comment += line + '\n';
        }
    }

    if ( !comment.empty() && HasOption( MOL2_KEEP_COMMENT ) ) {
        gd = new CCPairString( "#MOL2PRECOMMENT", comment );
        AssertMemAlloc( gd );
        mol.SetData( gd );
		comment.clear();
    }

    // ************** @<TRIPOS>MOLECULE **************
    for ( int lcount = 0; foundMoleLine && lcount < 6; ++lcount ) {
        if ( !getline( is, line ) ) {
            return false;
        }

        // some nonstandard mol2 such as FlexX output may not have all 6 lines
        // let's break out
        if ( EQn( line.c_str(), "@<TRIPOS>ATOM", 13 ) ) {
            foundAtomLine = true;
            break;
        }

        if ( lcount == 0 ) {
            // Molecular Name
            if ( BlankLine( line ) ) {
                gErrorLog.Throw( __FUNCTION__,
                                 "NULL molecular name! **** by default." );
                mol.SetName( "****" );
            }
            else {
                Trim( line );
                mol.SetName( line );
            }
        }
        else if ( lcount == 1 ) {
            // Number of atoms, bonds and substructures
            if ( sscanf( line.c_str(), "%d %d %d", 
				         &nAtoms, &nBonds, &nSubs ) < 2 ) {
                markError = true;
                goto BeginOfNextMolecule;
            }
        }
        else {
            // 2 molecular type
			// 3 charge type
			// 4 status bits
			// 5 molecular comment
			comment += line + '\n';
            
        }
    }

	if ( !comment.empty() && HasOption( MOL2_KEEP_COMMENT ) ) {
		gd = new CCPairString( "#MOL2HEADCOMMENT", comment );
		AssertMemAlloc( gd );
		mol.SetData( gd );
		comment.clear();
	}

    // skip to @TRIPOS<ATOM>
    for ( ; foundMoleLine && !foundAtomLine ; ) {
        oldPos = is.tellg();

        if ( !getline( is, line ) ) {
            return false;
        }

        if ( EQn( line.c_str(), "@<TRIPOS>ATOM", 13 ) ) {
            foundAtomLine = true;
            break;
        }
        else if ( EQn( line.c_str(), "@<TRIPOS>MOLECULE", 17 ) ) {
            is.seekg( oldPos );
            return false;
        }
        else if ( EQn( line.c_str(), "@<TRIPOS>BOND", 13 ) ) {
            markError = true;
            goto BeginOfNextMolecule;
        }
    }

    // ************** @<TRIPOS>ATOM **************
    gAtomTypeTable.SetFromType( "SYB" );
    mol.ResizeAtoms( nAtoms );
    for ( int i = 0; foundAtomLine && i < nAtoms; ++i ) {
        if ( !getline( is, line ) ) {
            return false;
        }

        CCAtom atom; // a clean atom

        int rtn = sscanf( line.c_str(), "%*s %s %lf %lf %lf %s %d %s %lf",
                          atom_name, &x, &y, &z, atom_type, &subst_id,
                          subst_type, &charge );

        if ( rtn <5 || rtn> 8 ) {
            // format must be ERROR
            gErrorLog.Throw( __FUNCTION__,
                             "Wrong mol2 format! try to skip current molecule.",
                             ccFatal );
            markError = true;
            goto BeginOfNextMolecule;
        }
        else {
            // set atomic number
            fromType = atom_type;
            // handle CL and BR atom types
            if ( fromType.size() > 1 && isupper( fromType[1] ) ) {
                fromType[1] = tolower( fromType[1] );
            }

            gAtomTypeTable.SetToType( "ATN" );
			if ( !gAtomTypeTable.Translate( fromType, toType ) ) {
				// guess atom type from atom_name
				fromType = atom_name;
				Trim( fromType, "0123456789" );
				gAtomTypeTable.Translate( fromType, toType );
			}
            // set sybyl atom type in m_type
            atom.SetType( atom_type );
            atom.SetName( atom_name );
            atom.SetAtomicNum( atoi( toType.c_str() ) );
            atom.SetXYZ( x, y, z );

            // subst_id
            gd = new CCPairInt( "SUBST_ID", subst_id );
            AssertMemAlloc( gd );
            atom.SetData( gd );

            if ( rtn > 7 ) {
                gd = new CCPairString( "SUBST_TYPE", subst_type );
                AssertMemAlloc( gd );
                atom.SetData( gd );
                atom.SetPartialCharge( charge );
            }
        }

        if ( !mol.AddAtom( atom ) ) {
            markError = true;
            goto BeginOfNextMolecule;
        }

        // Add residue information if it exists
        if ( subst_id != -1 && subst_id != 0 
            && strlen( subst_type ) > 3 && isdigit(subst_type[3]) ) { 
            // ALA34, avoid sth. like PHENYL
            string restype = subst_type;
            string resname = restype.substr( 0, 3 );

            if ( !IsNaturalAA( resname ) ) {
                // only handle standard residue
                continue;
            }

            int resnum = ( restype.size() > 3 )
                       ? toINT( restype.substr( 3 ) )
                       : subst_id;

            CCResidue* res = mol.GetResidue( mol.GetNumRes() );
            if ( res == NULL
              || res->GetResName() != resname
              || res->GetResNum() != resnum ) {
                CCResidueIter ri;
                for ( res = mol.BeginResidue( ri ) ;
                      res ;
                      res = mol.NextResidue( ri ) ) {
                    if ( res->GetResName() == resname
                      && res->GetResNum() == resnum ) {
                        break;
                    }
                }

                if ( res == NULL ) {
                    res = mol.NewResidue();
                    res->SetResName( resname );
                    res->SetResNum( resnum );

                    // set altloc
                }
            }

            CCAtom* ap = mol.GetAtom( mol.GetNumAtom() );
            res->AddAtom( ap );
            res->SetResAtomName( ap, atom_name );
            res->SetSerNum( ap, subst_id );
        }
    }

    // seek to @<TRIPOS>BOND
	// single-atom molecule has no bond
    for ( ; foundMoleLine && foundAtomLine && nBonds; ) {
        oldPos = is.tellg();

        if ( !getline( is, line ) ) {
            return false;
        }

        if ( EQn( line.c_str(), "@<TRIPOS>BOND", 13 ) ) {
            foundBondLine = true;
            break;
        }
        else if ( EQn( line.c_str(), "@<TRIPOS>MOLECULE", 17 ) ) {
            gErrorLog.Throw( __FUNCTION__,
                             "Wrong mol2 format! try to skip current molecule." );
            is.seekg( oldPos );
            return false;
        }
        else if ( EQn( line.c_str(), "@<TRIPOS>ATOM", 13 ) ) {
            markError = true;
            goto BeginOfNextMolecule;
        }
    }

    // ************** @<TRIPOS>BOND **************
    mol.ResizeBonds( nBonds );
	// single-atom molecule has no bond
    for ( int j = 0; foundBondLine && j < nBonds; ++j ) {
        if ( !getline( is, line ) ) {
            return false;
        }

        if ( sscanf( line.c_str(), "%*s %d %d %s", &start, &end, bond_type )
          != 3 ) {
            markError = true;
            goto BeginOfNextMolecule;
        }

        order = 1;
        if ( EQi( bond_type, "AR" ) ) {
            order = 5;
            bKekule = true;
        }
        else if ( EQi( bond_type, "AM" ) ) {
            order = 1;
        }
		else if ( EQi( bond_type, "DU" ) ) {
			order = 1;
		}
        else {
            order = atoi( bond_type );
        }

        if ( !mol.AddBond( start, end, order ) ) // add bond ERROR
        {
            markError = true;
            goto BeginOfNextMolecule;
        }
    }

    // molecule has connection table from SYBYL mol2 format
    mol.SetFlag( MOL_CONNECTIVITY_PERCEIVED );

    // set dimension
    mol.SetDimensionFromXYZ();

    // kekulize if necessary
    if ( bKekule ) {
        mol.Kekulize();
        // clear flag for output
        mol.UnsetFlag( MOL_KEKULE_PERCEIVED );
    }

    // go to begin of next molecule or begin of next molecule's comment
    BeginOfNextMolecule:
    if ( markError ) {
		string errMsg = "Read molecule ";
		errMsg += mol.GetName() + " failed, try to skip it.";
        gErrorLog.Throw( __FUNCTION__, errMsg, ccFatal );
    }

    for ( ; ; ) {
        oldPos = is.tellg();

        if ( !getline( is, line ) ) {
            break; // for last molecule
        }
		// # indicate the comment of the next molecule
        if ( line[0] == '#' ||
			 EQn( line.c_str(), "@<TRIPOS>MOLECULE", 17 ) ) {
            is.seekg( oldPos );
            break;
        }
		// record tailing comment line
		if ( HasOption( MOL2_KEEP_COMMENT ) ) {
			comment += line + '\n';
		}
    }

	if ( !comment.empty() && HasOption( MOL2_KEEP_COMMENT ) ) {
		gd = new CCPairString( "#MOL2POSTCOMMENT", comment );
		AssertMemAlloc( gd );
		mol.SetData( gd );
		comment.clear();
	}

    return ( markError ) ? false : true;
}

/**
* @brief only read atoms of a molecule
* @param[in] mol mol reference of molecule
* @param[in] is input file stream
*
* It only read atoms information in mol2 file. Bond information
* will simply discarded. Return false when error encountered, 
* else return true.
*/
bool CCMol2Format::ReadAtomsOnly( CCMol& mol, istream& is ) {
    // seek beginning of next molecule if possible or error
    // clean molecule
    mol.Clear();

    // for error
    bool markError( false );
    // for buffer
    string line;
    // for Molecule
    int nAtoms(0), nBonds(0), nSubs(0);
    bool foundMoleLine( false ), foundAtomLine( false );
    // for ATOM
    int subst_id( -1 );
    char atom_name[BUFF_SIZE], atom_type[BUFF_SIZE], subst_type[BUFF_SIZE];
    double x, y, z, charge;
    string fromType, toType;

    // for back
    streampos oldPos;
    // for generic data
    string comment;
    CCGenericData* gd( NULL );

    // skip to @<TRIPOS>MOLECULE, and record #COMMENT if any
    for ( ; ; ) {
        if ( !getline( is, line ) ) {
            return false;
        }

        if ( EQn( line.c_str(), "@<TRIPOS>MOLECULE", 17 ) ) {
            foundMoleLine = true;
            break;
        }
        else if ( EQn( line.c_str(), "@<TRIPOS>ATOM", 13 )
               || EQn( line.c_str(), "@<TRIPOS>BOND", 13 ) ) {
            markError = true;
            goto BeginOfNextMolecule;
        }
        else {
            // all others were regarded as comment line
            comment += line + '\n';
        }
    }

    if ( !comment.empty() && HasOption( MOL2_KEEP_COMMENT ) ) {
        gd = new CCPairString( "#MOL2PRECOMMENT", comment );
        AssertMemAlloc( gd );
        mol.SetData( gd );
    }

    // ************** @<TRIPOS>MOLECULE **************
	for ( int lcount = 0; foundMoleLine && lcount < 6; ++lcount ) {
		if ( !getline( is, line ) ) {
			return false;
		}

		// some nonstandard mol2 such as FlexX output may not have all 6 lines
		// let's break out
		if ( EQn( line.c_str(), "@<TRIPOS>ATOM", 13 ) ) {
			foundAtomLine = true;
			break;
		}

		if ( lcount == 0 ) {
			// Molecular Name
			if ( BlankLine( line ) ) {
				gErrorLog.Throw( __FUNCTION__,
					"NULL molecular name! **** by default." );
				mol.SetName( "****" );
			}
			else {
				Trim( line );
				mol.SetName( line );
			}
		}
		else if ( lcount == 1 ) {
			// Number of atoms, bonds and substructures
			if ( sscanf( line.c_str(), "%d %d %d", 
				&nAtoms, &nBonds, &nSubs ) < 2 ) {
					markError = true;
					goto BeginOfNextMolecule;
			}
		}
		else {
			// 2 molecular type
			// 3 charge type
			// 4 status bits
			// 5 molecular comment
			comment += line + '\n';

		}
	}

    // skip to @TRIPOS<ATOM>
    for ( ; foundMoleLine && !foundAtomLine ; ) {
        oldPos = is.tellg();

        if ( !getline( is, line ) ) {
            return false;
        }

        if ( EQn( line.c_str(), "@<TRIPOS>ATOM", 13 ) ) {
            foundAtomLine = true;
            break;
        }
        else if ( EQn( line.c_str(), "@<TRIPOS>MOLECULE", 17 ) ) {
            is.seekg( oldPos );
            return false;
        }
        else if ( EQn( line.c_str(), "@<TRIPOS>BOND", 13 ) ) {
            markError = true;
            goto BeginOfNextMolecule;
        }
    }

    // ************** @<TRIPOS>ATOM **************
    gAtomTypeTable.SetFromType( "SYB" );
    mol.ResizeAtoms( nAtoms );
    for ( int i = 0; foundAtomLine && i < nAtoms; ++i ) {
        if ( !getline( is, line ) ) {
            return false;
        }

        CCAtom atom; // a clean atom

        int rtn = sscanf( line.c_str(), "%*s %s %lf %lf %lf %s %d %s %lf",
                          atom_name, &x, &y, &z, atom_type, &subst_id,
                          subst_type, &charge );

        if ( rtn <5 || rtn> 8 ) {
            // format must be ERROR
            gErrorLog.Throw( __FUNCTION__,
                             "Wrong mol2 format! try to skip current molecule.",
                             ccFatal );
            markError = true;
            goto BeginOfNextMolecule;
        }
        else {
            // set atomic number
            fromType = atom_type;
            // handle CL and BR atom types
            if ( fromType.size() > 1 && isupper( fromType[1] ) ) {
                fromType[1] = tolower( fromType[1] );
            }

			gAtomTypeTable.SetToType( "ATN" );
			if ( !gAtomTypeTable.Translate( fromType, toType ) ) {
				// guess atom type from atom_name
				fromType = atom_name;
				Trim( fromType, "0123456789" );
				gAtomTypeTable.Translate( fromType, toType );
			}
            // set sybyl atom type in m_type
            atom.SetType( atom_type );
            atom.SetName( atom_name );
            atom.SetAtomicNum( atoi( toType.c_str() ) );
            atom.SetXYZ( x, y, z );

			// subst_id
			gd = new CCPairInt( "SUBST_ID", subst_id );
			AssertMemAlloc( gd );
			atom.SetData( gd );

			if ( rtn > 7 ) {
				gd = new CCPairString( "SUBST_TYPE", subst_type );
				AssertMemAlloc( gd );
				atom.SetData( gd );
				atom.SetPartialCharge( charge );
			}
        }

        if ( !mol.AddAtom( atom ) ) {
            markError = true;
            goto BeginOfNextMolecule;
        }

        // Add residue information if it exists
        if ( subst_id != -1 && subst_id != 0 
            && strlen( subst_type ) > 3 && isdigit(subst_type[3]) ) { 
            // ALA34, avoid sth. like PHENYL
            string restype = subst_type;
            string resname = restype.substr( 0, 3 );

            if ( !IsNaturalAA( resname ) ) {
                // only handle standard residue
                continue;
            }

            int resnum = ( restype.size() > 3 )
                       ? toINT( restype.substr( 3 ) )
                       : subst_id;

            CCResidue* res = mol.GetResidue( mol.GetNumRes() );
            if ( res == NULL
              || res->GetResName() != resname
              || res->GetResNum() != resnum ) {
                CCResidueIter ri;
                for ( res = mol.BeginResidue( ri ) ;
                      res ;
                      res = mol.NextResidue( ri ) ) {
                    if ( res->GetResName() == resname
                      && res->GetResNum() == resnum ) {
                        break;
                    }
                }

                if ( res == NULL ) {
                    res = mol.NewResidue();
                    res->SetResName( resname );
                    res->SetResNum( resnum );
                }
            }

            CCAtom* ap = mol.GetAtom( mol.GetNumAtom() );
            res->AddAtom( ap );
            res->SetResAtomName( ap, atom_name );
            res->SetSerNum( ap, subst_id );
        }
    }

    // set dimension
    mol.SetDimensionFromXYZ();

    // go to begin of next molecule or begin of next molecule's comment
    BeginOfNextMolecule:
    if ( markError ) {
        gErrorLog.Throw( __FUNCTION__,
                         "Read molecule failed, try to skip it.", ccFatal );
    }

	comment.clear();
    for ( ; ; ) {
        oldPos = is.tellg();

        if ( !getline( is, line ) ) {
            break; // for last molecule
        }
        if ( line[0] == '#' ||
			 EQn( line.c_str(), "@<TRIPOS>MOLECULE", 17 ) ) {
            is.seekg( oldPos );
            break;
        }
		// record tail comment line
		if ( EQn( line.c_str(), "@<TRIPOS>COMMENT", 16 ) ) {
			comment += line + '\n';
			for ( ; ; ) {
				if ( !getline( is, line ) ) {
					break; // for last molecule
				}
				// # indicate the comment of the next molecule
				if ( line[0] == '#' ||
					EQn( line.c_str(), "@<TRIPOS>MOLECULE", 17 ) ) {
						is.seekg( oldPos );
						break;
				}
				else {
					comment += line + '\n';
				}
			}
			break;
		}
	}

	if ( !comment.empty() && HasOption( MOL2_KEEP_COMMENT ) ) {
		gd = new CCPairString( "#MOL2POSTCOMMENT", comment );
		AssertMemAlloc( gd );
		mol.SetData( gd );
	}

    return ( markError ) ? false : true;
}

/**
* @brief write a molecule
* @param[in] mol reference of molecule
* @param[out] os output file stream
*
* interface for writing Tripos Mol2 file, one molecule a time
*/
bool CCMol2Format::WriteMolecule( CCMol& mol, ostream& os )
{
    if ( mol.IsEmpty() ) {
        gErrorLog.Throw( __FUNCTION__, "Molecule is empty." );
        return false;
    }

    // for generic data
    CCGenericData* gd( NULL );

    char line[BUFF_SIZE];

    // ************** #COMMENT if any **************
    if ( ( gd = mol.GetData( "#MOL2PRECOMMENT" ) ) &&
		   HasOption( MOL2_KEEP_COMMENT ) ) {
        os << dynamic_cast<CCPairString*>( gd )->GetValue();
    }

    // ************** @<TRIPOS>MOLECULE **************
    os << "@<TRIPOS>MOLECULE\n";
    string molName = mol.GetName();
    if ( molName.empty() ) {
        gErrorLog.Throw( __FUNCTION__,
                         "NULL molecular name! **** by default.", ccWarning );
        os << "****\n";
    }
    else {
		// set max length of molecular name as 50 characters
        os << molName.substr( 0, 50 ) << "\n";
    }

    snprintf( line, BUFF_SIZE, "%5d %5d %5d %5d %5d", mol.GetNumAtom(),
              mol.GetNumBond(), mol.GetNumSubs(), 0, 0 );
    os << line << "\n";

	// ************** #COMMENT if any **************
	if ( ( gd = mol.GetData( "#MOL2HEADCOMMENT" ) ) &&
		HasOption( MOL2_KEEP_COMMENT ) ) {
		os << dynamic_cast<CCPairString*>( gd )->GetValue();
	}
	else {
		os << "SMALL\n"; // MOL_TYPE
		os << "USER_CHARGES\n"; // CHARGE_TYPE
		os << "\n"; // STATUS_BITS
		os << "\n"; // MOL_COMMENT
	}
    
    // ************** @<TRIPOS>ATOM **************
    gAtomTypeTable.SetFromType( "INT" );
    gAtomTypeTable.SetToType( "SYB" );

    os << "@<TRIPOS>ATOM\n";
    string fromType, toType, subst_type;
    int subst_id;
    char label[BUFF_SIZE], rlabel[BUFF_SIZE];
    vector<int> labelcount( gElementalTable.GetNumofElements() + 1 );
    CCResidue* res( NULL );

    FOR_ATOMS_OF_MOL( atom, mol ) {
        res = atom->GetResidue();
        if ( res ) {
            // Use original atom_name if residue exists
            snprintf( label, BUFF_SIZE, "%s",
                      ( res->GetResAtomName( &*atom ) ).c_str() );
        }
        else {
            // Use sequentially numbered atom names if no residue
            snprintf( label, BUFF_SIZE, "%s%d", atom->GetSymbol().c_str(),
                      ++labelcount[atom->GetAtomicNum()] );
        }
        subst_id = ( res ) ? res->GetResNum() : atom->GetPart();
        subst_type = ( res )
                   ? res->GetResName()
                   + toSTR( res->GetResNum() )
                   : ( gd = atom->GetData( "SUBST_TYPE" ) )
                   ? ( dynamic_cast<CCPairString*>( gd ) )->GetValue().c_str()
                   : "SUB1";

        snprintf( rlabel, BUFF_SIZE, "%s", subst_type.c_str() );

        fromType = atom->GetType();
        gAtomTypeTable.Translate( fromType, toType );

        snprintf( line, BUFF_SIZE,
                  "%7d %-6s  %10.4f%10.4f%10.4f %-5s %5d %-8s%10.4f",
                  atom->GetIdx(), label, atom->GetX(), atom->GetY(),
                  atom->GetZ(), toType.c_str(), subst_id, rlabel,
                  atom->GetPartialCharge() );
        os << line << "\n";
    }

    // ************** @<TRIPOS>BOND **************
	// single-atom molecule has no bond, e.g. He
	if ( mol.GetNumBond() > 0 ) {
		gBondTypeTable.SetFromType( "INT" );
		gBondTypeTable.SetToType( "SYB" );
		string fromBondType, toBondType;

		os << "@<TRIPOS>BOND\n";
		FOR_BONDS_OF_MOL( bond, mol ) {
			fromBondType = bond->GetType();
			gBondTypeTable.Translate( fromBondType, toBondType );

			snprintf( line, BUFF_SIZE, "%6d%5d%5d %-2s", bond->GetIdx(),
				bond->GetBeginAtomIdx(), bond->GetEndAtomIdx(),
				toBondType.c_str() );
			os << line << "\n";
		}
	}    

    // ************** #COMMENT if any **************
	if ( ( gd = mol.GetData( "#MOL2POSTCOMMENT" ) ) &&
		   HasOption( MOL2_KEEP_COMMENT ) ) {
		os << dynamic_cast<CCPairString*>( gd )->GetValue();
	}
	else { //@<TRIPOS>SUBSTRUCTURE
		os << "@<TRIPOS>SUBSTRUCTURE\n";
		set<string> seenSUB;
		int cntSUB( 0 );
		FOR_ATOMS_OF_MOL( atom, mol ) {
			res = atom->GetResidue();
			subst_type = ( res )
				? res->GetResName()
				+ toSTR( res->GetResNum() )
				: "SUB";

			if ( seenSUB.find( subst_type ) != seenSUB.end() ) {
				continue;
			}
			else {
				seenSUB.insert( subst_type );
			}
			snprintf( rlabel, BUFF_SIZE, "%s", subst_type.c_str() );

			snprintf( line, BUFF_SIZE, "%6d %-8s%5d %-6s", ++cntSUB, rlabel,
				atom->GetIdx(),
				res
				? "RESIDUE"
				: ( gd = atom->GetData( "SUBST_TYPE" ) )
				? ( static_cast<CCPairString*>( gd ) )->GetValue().c_str()
				: "GROUP" );
			os << line << "\n";
		}
	}
    os << endl;

    return true;
}
