#include <formats/CCFormat.h>
#include <ccbg/CCMol.h>

class CCPDBFormat : public CCFormat
{
public:
    CCPDBFormat()
    {
        RegisterFormat( "pdb", this );
        RegisterFormat( "ent", this );
        SetOption( PDB_SKIP_WATER );
		SetOption( PDB_BREAK_METAL_BOND );
    }
    virtual ~CCPDBFormat()
    {
    }

    bool ReadMolecule( CCMol& mol, istream& is );
	bool ReadAtomsOnly( CCMol& mol, istream& is );
	bool ReadHetatmLig( CCMol& mol, istream& is );
    bool WriteMolecule( CCMol& mol, ostream& os );
};
CCPDBFormat theCCPDBFormat;  // make an instant and register format

//////////////////////////////////////////////////////////////////////////
/** @brief parse atom record in pdb file. */
static bool ParseATOM( char*, CCMol&, char&, bool& );
static bool ParseCONECT( char*, CCMol& );
static bool ParseSSBOND( char*, CCMol&, map<string, string>& );
//
void SetHelixSheet( CCMol&, map<char, vector<pair<int,int> > >&, string );
void MakeSSBOND( CCMol&, map<string, string>& );
void ConnectResidues( CCMol& );
void HandleHet( CCMol& mol );
//////////////////////////////////////////////////////////////////////////
bool CCPDBFormat::ReadMolecule( CCMol& mol, istream& is )
{
    // seek beginning of next molecule if possible or error
    // for buffer
    char line[BUFF_SIZE];

    // indicate a new chain
    char currChainID = '-';

	bool atomline( false ), hetatmline( false );

    // for generic data
    CCGenericData* gd( NULL );
	// for HELIX & SHEET
	char chainID;
	int initResNum, endResNum;
	map<char, vector<pair<int,int> > > helix, sheet;
	// for SSBOND
    map<string, string> ssbond;
	// whether it is amber format.
	bool amber(false);

    mol.Clear();

    while ( is.getline( line, BUFF_SIZE ) && !EQn( line, "END", 3 ) ) {
        if ( EQn( line, "ATOM", 4 ) || EQn( line, "HETATM", 6 ) ) {
			if ( !atomline && EQn( line, "ATOM", 4 ) ) {
				atomline = true;
			}
			if ( !hetatmline && EQn( line, "HETATM", 6 ) ) {
				hetatmline = true;
			}
            if ( !ParseATOM( line, mol, currChainID, amber ) ) {
				// try amber pdb format
				if ( !ParseATOM( line, mol, currChainID, amber ) ) {
					string errMsg = "Parse atom error in LINE\n";
					gErrorLog.Throw( __FUNCTION__, errMsg + line );
					// just skip this line
					continue;
				}
            }
        }
        else if ( EQn( line, "TER", 3 ) ) {
            currChainID = '-'; // expect an new chain
        }
        else if ( EQn( line, "CONECT", 6 ) ) {
            ParseCONECT( line, mol );
        }
		//COMPND does not describe a complex name
        /*else if ( EQn( line, "COMPND", 6 ) && ( mol.GetName() ).empty() ) {
            string name = line;
            if ( name.size() > 10 ) {
                name = name.substr( 10, BUFF_SIZE );
                Trim( name );
                mol.SetName( name );
            }
        }*/
        else if ( EQn( line, "HEADER", 6 ) && ( mol.GetName() ).empty() ) {
            string header = line;
			if ( header.size()>63 && header[62]!=' ') {
				// four character unique PDB ID
				header = header.substr(62,4);
				Trim(header);
				mol.SetName( header );
				if (!header.empty()) {
					gd = new CCPairString( "PDBID", header );
					AssertMemAlloc( gd );
					mol.SetData( gd );
				}
			}
			else {
				header = header.substr(6);
				Trim(header);
				mol.SetName( header );
			}
        }
		else if ( EQn( line, "HELIX", 5 ) ) {
			chainID = line[19];
			initResNum = toINT(string(&line[21],4));
			endResNum = toINT( string(&line[33],4));
			helix[chainID].push_back(make_pair(initResNum,endResNum));
		}
		else if ( EQn( line, "SHEET", 5 ) ) {
			chainID = line[21];
			initResNum = toINT(string(&line[22],4));
			endResNum = toINT( string(&line[33],4));
			sheet[chainID].push_back(make_pair(initResNum,endResNum));
		}
        /*
        else if ( EQn( line, "SEQRES", 6 ) ) {
                    // has sequence info.
                    strSEQRES = strSEQRES + line + '\n';
        }*/
        else if ( EQn( line, "SSBOND", 6 ) ) {
            // disulfide bond
            ParseSSBOND( line, mol, ssbond );
        }
    }

	// set HELIX & SHEET info.
	SetHelixSheet(mol, helix, "H");
	SetHelixSheet(mol, sheet, "S");

	// make SSBond if it was missed in CONECT
	MakeSSBOND(mol, ssbond);
    
    // assign bond according to template
	if ( atomline ) {
		if ( !hetatmline ) {
			hetatmline = !gResidueData.AssignBond( mol );
		}
		else {
			gResidueData.AssignBond( mol );
		}
	}

    // connect adjacent residues
    ConnectResidues( mol );

	// if has HETAM, perceive structure using I-interpret
    if ( hetatmline ) {
      HandleHet( mol );
    }

	// sort bond by atom index
    mol.SortBonds();

    // set flag to avoid repeated perception
    mol.SetFlag( MOL_HYBRID_PERCEIVED
	           | MOL_HBDONORACCEPTOR_PERCEIVED
			   | MOL_RING_PERCEIVED
			   | MOL_AROMATICITY_PERCEIVED
			   | MOL_AUTOFCHARGE_PERCEIVED
			   | MOL_KEKULE_PERCEIVED
			   | MOL_BONDORDER_PERCEIVED
			   | MOL_INT_TYPE_PERCEIVED );

    // remove NULL pointer at the end of m_vatom and m_vbond
    mol.ResizeAtoms( mol.GetNumAtom() );
    mol.ResizeBonds( mol.GetNumBond() );

    // eat up remaining blank lines
    while ( is.peek() != EOF
         && is.good()
         && ( is.peek() == '\n' || is.peek() == '\r' ) ) {
        is.getline( line, BUFF_SIZE );
    }

    return !mol.IsEmpty();
}

void HandleHet( CCMol& mol )
{
    map<CCAtom*, CCAtom*> amap;
    CCAtom* atm;
    CCMol lig;
    FOR_ATOMS_OF_MOL( a, mol ) {
        if ( !a->HasFlag( ATOM_IS_IN_NATAA ) ) {
            atm = lig.AddAtom( *a );
            if ( atm ) {
                atm->ClearBond();
                amap[atm] = &*a;
            }
        }
    }

    // using XYZ
    lig.HandlePDB();
	lig.Kekulize();
    map<CCBond*, CCBond*> bmap;
    FOR_BONDS_OF_MOL( b, lig ) {
        CCBond* bnd;
        bnd = mol.AddBond( amap[b->GetBeginAtom()]->GetIdx(),
                           amap[b->GetEndAtom()]->GetIdx(), b->GetOrder(),
                           b->GetFlag() );
        if ( bnd == NULL ) {
            bnd = mol.GetBond( amap[b->GetBeginAtom()]->GetIdx(),
                               amap[b->GetEndAtom()]->GetIdx() );
            if ( bnd == NULL ) {
                continue;
            }
        }
        bnd->SetType( b->GetType() );
        bmap[&*b] = bnd;
    }
    FOR_ATOMS_OF_MOL( a, lig ) {
        atm = amap[&*a];
        atm->SetType( a->GetType() );
        atm->SetImpVal( a->GetImpVal() );
        atm->SetHyb( a->GetHyb() );
        atm->SetFlag( a->GetFlag() );
        FOR_BONDS_OF_ATOM( b, &*a ) {
            atm->AddBond( bmap[&*b] );
        }
    }
}

/**
* @brief parse atom record in pdb file.
* @param[in] line atom record line for parsing.
* @param[out] mol molecule for holding the parsed atom.
* @param[out] currChainID current chain index for chain information.
* @param[out] whether the atom record line is in amber pdb format.
*
* //////////////////////////////////////////////////////////////////  <br>
* ATOMFORMAT "(a6,i5,1x,a4,a1,a3,1x,a1,i4,a1,3x,3f8.3,2f6.2,1x,i3)"   <br>
*                                                                     <br>
* ATOM/HETATM Record Fromat                                           <br>
* http://www.wwpdb.org/documentation/format23/sect9.html#ATOM         <br>
* COLUMNS    DATA TYPE       FIELD         DEFINITION                 <br>
* --------------------------------------------------------------      <br>
* 1  - 6   Record name    "ATOM  "(or"HETATM)                         <br>
* 7  - 11   Integer       serial      Atom serial number              <br>
* 13 - 16   Atom          name        Atom name                       <br>
* 17        Character     altLoc      Alternate location indicator    <br>
* 18 - 20   Residue name  resName     Residue name                    <br>
* 22        Character     chainID     Chain identifier                <br>
* 23 - 26   Integer       resSeq      Residue sequence number         <br>
* 27        AChar         iCode       Code for insertion of residues  <br>
* 31 - 38   Real(8.3)     x           coordinates for X in Angstroms  <br>
* 39 - 46   Real(8.3)     y           coordinates for Y in Angstroms  <br>
* 47 - 54   Real(8.3)     z           coordinates for Z in Angstroms  <br>
* 55 - 60   Real(6.2)     occupancy   Occupancy                       <br>
* 61 - 66   Real(6.2)     tempFactor  Temperature factor              <br>
* 77 - 78   LString(2)    element     Element symbol, right-justified <br>
* 79 - 80   LString(2)    charge      Charge on the atom              <br>
*/
static bool ParseATOM( char* line, CCMol& mol,
							 char& currChainID, bool& amber )
{
    string sbuf = line;
    if ( sbuf.size() < 54 ) {
        return false;
    }

    /* "ATOM  "(or"HETATM) 1-6 */
    bool hetatm = ( EQn( line, "HETATM", 6 ) ) ? true : false;

    /* Atom serial number 7-11 */
    int serial = toINT(sbuf.substr( 6, 5 ));

    /* Atom name 13-16 */
    string name = sbuf.substr( 12, 4 );
    if ( name[0] == 'L' && name[1] == 'P' ) {
        return true; // skip lone pair
    }
	// amber pdb format
	if ( amber && name[3] == ' ' ) {
		name[3] = name[2];
		name[2] = name[1];
		name[1] = name[0];
		name[0] = ' ';
	}

    /* Alternate location indicator 17 */
    char altLoc = sbuf[16];

    /* Residue name 18-20 */
    string resName = sbuf.substr( 17, 3 );
    if ( resName == "   " ) {
        resName = "Xaa";
    }
    else if ( ( resName == "HOH" || resName == "WAT" )
           && theCCPDBFormat.HasOption( PDB_SKIP_WATER ) ) {
        return true; // skip water
    }
    Trim( resName );

    /* Chain identifier 22 */
    char chainID = sbuf[21];

    /* Residue sequence number 23-26 */
    int resSeq = toINT( sbuf.substr( 22, 4 ) );

	/* Code for insertion of residues */
	char iCode = sbuf[26];

    /* X 31-38, Y 39-46, Z 47-54 */
    double x = toDBL( sbuf.substr( 30, 8 ) );
    double y = toDBL( sbuf.substr( 38, 8 ) );
    double z = toDBL( sbuf.substr( 46, 8 ) );

    /* Occupancy 55-60 */
	double occupancy(0.0);
	if ( sbuf.size() > 55 ) {
		occupancy = toDBL( sbuf.substr( 54, 6 ) );
	}

    /* Temperature factor 61-66 */
	double tempFactor(0.0);
	if ( sbuf.size() > 61 ) {
		tempFactor = toDBL( sbuf.substr( 60, 6 ) );
	}

    /* element 77-78 */
    string element("  ");
    if ( sbuf.size() > 77 ) {
        element = sbuf.substr( 76, 2 );
    }

    /* Charge on the atom 79-80 */
	int charge(0);
	if ( sbuf.size() > 79 ) {
		charge = toINT( sbuf.substr( 78, 2 ) );
	}

    /* Six characters (columns) are reserved for atom names, assigned as follows.
        COLUMN     VALUE
        -----------------------------------------------------------------------
        13 - 14    Chemical symbol - right justified, except for hydrogen atoms
        15         Remoteness indicator (alphabetic)
        16         Branch designator (numeric)
        77 - 78    Element symbol, right-justified
    */
    string ele;

    // parse from element symbol, if exists.
    if ( isalpha( element[1] ) ) {
        if ( isalpha( element[0] ) ) {
            ele = element.substr( 0, 2 );
        }
        else {
            ele = element.substr( 1, 1 );
        }
    }
    else {
        // parse from atom name
        if ( EQn( line, "ATOM", 4 ) ) {
            // ATOM
            if ( isdigit( name[0] ) ) {
                if ( isdigit( name[1] ) ) {
                    // sometimes non-standard files have, e.g 11HG (ILE, 1HG1)
                    ele = name.substr( 2, 1 );
                    // revise the atmname
                }
                else {
                    // 1HG
                    ele = name.substr( 1, 1 );
                }
            }
            else if ( name[0] == ' ' ) {
                // one-char element
                ele = name.substr( 1, 1 );
            }
            else if ( name[0] == 'H' ) {
                // eg. HNCA, HB1, HG23
                ele = "H";
            }
            else {
                // fail
				amber = !amber;
                return false;
            }
        }
        else {
            //must be hetatm record
            if ( name[0] == ' ' || isdigit( name[0] ) ) {
                // one-char element
                ele = name.substr( 1, 1 );
            }
            else if ( isalpha( name[0] ) ) {
                ele = name.substr( 0, 2 );
            }
            else {
                // fail
				amber = !amber;
                return false;
            }

            // double check
            if ( name == resName ) {
                ele = name;
            }
            else {
                // trim digit
                Trim( ele, "0123456789" );
            }
        }
    }

    // skip hydrogen if required
    if ( theCCPDBFormat.HasOption( PDB_SKIP_H ) && ele == "H" ) {
        return true; // skip hydrogen
    }

    CCAtom* atom = mol.NewAtom();
    Trim( name ); atom->SetName( name );
    atom->SetXYZ( x, y, z );
    atom->SetAtomicNum( gElementalTable.GetAtomicNum( ele.c_str() ) );
    atom->SetType( ele );
	atom->SetFormalCharge(charge);
    if ( IsNaturalAA( resName ) ) {
        atom->SetFlag( ATOM_IS_IN_NATAA );
    }

    // Get the latest residue and last chain
    CCResidue* residue = mol.GetResidue( mol.GetNumRes() );
    CCChain* chain = mol.GetChain( mol.GetNumChain() );
    if ( /*bNaturalAA && // consider only natural amino acid */
        residue == NULL ||     // first residue
        residue->GetChainID() != currChainID ||     // residue in new chain
        residue->GetResName() != resName ||     // residue with new resname
        residue->GetResNum() != resSeq ||        // residue with new resnum
        residue->GetAltLoc() != altLoc || // residue with new altLoc
		residue->GetiCode() != iCode) { // reside with new iCode
        // check whether it is a new residue
        residue = NULL; chain = NULL;
        // find the last chain of currChainID for the residue
        for ( int i = mol.GetNumChain(); i > 0; --i ) {
            CCChain* chn = mol.GetChain( i );
            if ( chn && chn->GetChainID() == currChainID ) {
                chain = chn;
                break;
            }
        }

        if ( chain == NULL) {
            // a new chain
            chain = mol.NewChain();
            if ( currChainID == '-' ) {
                // first time comes to this chain
                currChainID = chainID;
            }
            chain->SetChainID( currChainID );
        }
        else {
            // an existing chain
            // sometimes, atoms in an existing residue may be disordered,
            // e.g. later added hydrogens come at the end of all residues.
            // so let's check whether it is an existing residue in this chain,
            // to save time, check only for hydrogen atoms because generally
            // we won't come there.
            if ( atom->IsHydrogen() ) {
                for ( int j = chain->GetNumRes(); j > 0; --j ) {
                    CCResidue* res = chain->GetResidue( j );
                    if ( res
                      && res->GetChainID() == currChainID
                      && res->GetResName() == resName
                      && res->GetResNum() == resSeq
                      && res->GetAltLoc() == altLoc
					  && res->GetiCode() == iCode) {
                        residue = res;
                        break; // an existing residue
                    }
                }
            }
        }

        if ( residue == NULL ) {
            // a new residue
            residue = mol.NewResidue();
            residue->SetResNum( resSeq );
            residue->SetResName( resName );
            // set altLoc
            if ( !( altLoc == ' ' || altLoc == 'A' ) ) {
                residue->SetAltLoc( altLoc );
            }
			residue->SetiCode(iCode);
            // add residue to chain if it is a new residue
            chain->AddResidue( residue );
        }
    }

    // add to residue
    if ( residue ) {
        residue->AddAtom( atom );
        residue->SetHetAtom( atom, hetatm );
        residue->SetResAtomName( atom, name );
        residue->SetSerNum( atom, serial );
		residue->SetOccupancy( atom, occupancy );
		residue->SetTempFactor( atom, tempFactor );
    }

    return true;
}

//////////////////////////////////////////////////////////////////////////
/* Connect Record Format
   http://www.wwpdb.org/documentation/format23/sect10.html#CONECT
COLUMNS       DATA TYPE       FIELD         DEFINITION
-------------------------------------------------------
 1 -  6       Record name     "CONECT"
 7 - 11        Integer         serial        Atom serial number
12 - 16        Integer         serial        Serial number of bonded atom
17 - 21        Integer         serial        Serial number of bonded atom
22 - 26        Integer         serial        Serial number of bonded atom
27 - 31        Integer         serial        Serial number of bonded atom
*/
static bool ParseCONECT( char* line, CCMol& mol )
{
    string sbuf = line;
    sbuf = sbuf.substr( 0, 31 );
    RTrim( sbuf );

    if ( sbuf.size() < 7 ) {
        return false;
    }

    // Serial number of the first atom, read from column 7-11 of the
    // connect record, to which the other atoms connect to.
    long int rootAtomSN = atol( sbuf.substr( 6, 5 ).c_str() );

    CCAtom* rootAtom = NULL;

    FOR_ATOMS_OF_MOL( atom, mol ) {
        if ( atom->GetResidue() // atom may have no residue
      && static_cast<long int>( atom->GetResidue()->GetSerNum( &*atom ) )
          == rootAtomSN ) {
            rootAtom = &*atom;
            break;
        }
    }

    if ( rootAtom == NULL ) {
        gErrorLog.Throw( __FUNCTION__,
                         "  Problems reading a CONECT record.\n");
        return false;
    }

	if (rootAtom->IsMetal()
		&& theCCPDBFormat.HasOption(PDB_BREAK_METAL_BOND)) {
		return true; // break metal bonds
	}

    long int boundedAtomsSNs[4] = {
        0, 0, 0, 0
    };
    int numNbr = 0;
    if ( sbuf.size() > 11 ) {
        numNbr++;
        boundedAtomsSNs[0] = atol( sbuf.substr( 11, 5 ).c_str() );
    }
    if ( sbuf.size() > 16 ) {
        numNbr++;
        boundedAtomsSNs[1] = atol( sbuf.substr( 16, 5 ).c_str() );
    }
    if ( sbuf.size() > 21 ) {
        numNbr++;
        boundedAtomsSNs[2] = atol( sbuf.substr( 21, 5 ).c_str() );
    }
    if ( sbuf.size() > 26 ) {
        numNbr++;
        boundedAtomsSNs[3] = atol( sbuf.substr( 26, 5 ).c_str() );
    }

    // Now iterate over the VALID boundedAtomsSNs and connect
    // the atoms.
    for ( int k = 0; k < numNbr; k++ ) {
        // Find atom that is connected to, write an error message
        CCAtom* connectedAtom = NULL;
        FOR_ATOMS_OF_MOL( atom, mol ) {
            if ( atom->GetResidue() && // atom may have no residue
                static_cast<long int>( atom->GetResidue()->GetSerNum( &*atom ) )
              == boundedAtomsSNs[k] ) {
                connectedAtom = &*atom;
                break;
            }
        }

        if ( connectedAtom == NULL ) {
            gErrorLog.Throw( __FUNCTION__,
                             "Problems reading a CONECT record.\n", ccWarning );
            return false;
        }

        // generate bond
        CCBond* lastbond = mol.AddBond( rootAtom->GetIdx(),
                                        connectedAtom->GetIdx(), 1 );
        if ( lastbond ) {
            lastbond->SetFlag( BOND_IS_INPUT );
            lastbond->SetType( "SB" );
        }
    }
    return true;
}

//////////////////////////////////////////////////////////////////////////
/*
COLUMNS        DATA TYPE       FIELD         DEFINITION
-------------------------------------------------------------------
1 -  6        Record name     "SSBOND"
8 - 10        Integer         serNum       Serial number.
12 - 14        LString(3)      "CYS"        Residue name.
16             Character       chainID1     Chain identifier.
18 - 21        Integer         seqNum1      Residue sequence number.
22             AChar           icode1       Insertion code.
26 - 28        LString(3)      "CYS"        Residue name.
30             Character       chainID2     Chain identifier.
32 - 35        Integer         seqNum2      Residue sequence number.
36             AChar           icode2       Insertion code.
60 - 65        SymOP           sym1         Symmetry oper for 1st resid
67 - 72        SymOP           sym2         Symmetry oper for 2nd resid
*/
static bool ParseSSBOND( char* line, CCMol& mol,
                               map<string, string>& ssbond )
{
    string ssline = line;
    if ( ssline.size() < 35 ) {
        gErrorLog.Throw( __FUNCTION__,
                         "  Problems reading a SSBOND record.\n");
        return false;
    }

    // Residue name1
    string resname1 = ssline.substr( 11, 3 );
    if ( resname1 != "CYS" ) {
        return false;
    }

    // ChainID1
    char chainID1 = ssline[15];
    /*if (chainID1 == ' ') {
        chainID1 = 'A';
    }*/

    // Residue sequence number
    int serNum1 = toINT( ssline.substr( 17, 4 ) );

    // Residue name2
    string resname2 = ssline.substr( 25, 3 );
    if ( resname2 != "CYS" ) {
        return false;
    }

    // ChainID2
    char chainID2 = ssline[29];
    /*if (chainID2 == ' ') {
        chainID2 = 'A';
    }*/

    // Residue sequence number2
    int serNum2 = toINT( ssline.substr( 31, 4 ) );

    stringstream s1, s2;
    s1 << resname1 << ' ' << chainID1 << ' ' << serNum1;
    s2 << resname2 << ' ' << chainID2 << ' ' << serNum2;

    // record twice for convenience
    ssbond.insert( make_pair( s1.str(), s2.str() ) );
    ssbond.insert( make_pair( s2.str(), s1.str() ) );

    return true;
}

void SetHelixSheet( CCMol& mol, map<char, vector<pair<int,int> > >& hs, string t )
{
	if (hs.empty()) {
		return;
	}

	FOR_CHAINS_OF_MOL(chn, mol) {
		if (hs.find(chn->GetChainID()) == hs.end()) {
			continue;
		}
		FOR_RESIDUES_OF_CHAIN(res, *chn) {
			vector<pair<int, int> >& vp = hs[chn->GetChainID()];
			for (int i=0; i<vp.size(); ++i) {
				if (vp[i].first<=res->GetResNum() &&
					vp[i].second>=res->GetResNum()) {
					res->SetSecStruct(t);
				}
			}
		}
	}
}

// connect CYS with SSBOND
// generally this should have a record in CONECT
void MakeSSBOND( CCMol& mol, map<string, string>& ssbond )
{
	if (ssbond.empty()) {
		return;
	}

    CCMolResidueIter r1( mol), r2( mol );
    for ( ; r1; ++r1 ) {
        stringstream s1;
        s1 << r1->GetResName() << ' ' << r1->GetChainID() << ' '
           << r1->GetResNum();
        if ( ssbond.find( s1.str() ) == ssbond.end() ) {
            continue;
        }

        CCAtom* sa = r1->GetAtomOfName( "SG" );
        if ( sa == NULL ) {
            continue;
        }

        for ( r2 = r1, ++r2; r2; ++r2 ) {
            stringstream s2;
            s2 << r2->GetResName() << ' ' << r2->GetChainID() << ' '
               << r2->GetResNum();
            if ( ssbond.find( s2.str() ) == ssbond.end() ) {
                continue;
            }

            CCAtom* sb = r2->GetAtomOfName( "SG" );
            if ( sb == NULL ) {
                continue;
            }

            // standard S-S 2.05A
            if ( IsAcceptableBond( sa, sb ) ) {
                mol.AddBond( sa->GetIdx(), sb->GetIdx(), 1, BOND_IS_SSBOND );
                break;
            }
        }
    }
}

// Connect all C and N on backbone
void ConnectResidues( CCMol& mol )
{
    FOR_CHAINS_OF_MOL( chain, mol ) {
        CCChainResIter rn;
        CCAtom* a1, * a2;
        CCBond* b;
        FOR_RESIDUES_OF_CHAIN( rc, &*chain ) {
            if ( !( a1 = rc->GetAtomOfName( "C" ) ) ) {
                continue;
            }

            for ( rn = rc, ++rn; rn; ++rn ) {
                if ( !( a2 = rn->GetAtomOfName( "N" ) )
                  || !IsAcceptableBond( &*a1,
                                        &*a2 ) ) {
                    continue;
                }

                if ( b = mol.AddBond( a1->GetIdx(), a2->GetIdx(), 1 ) ) {
                    b->SetType( "AM" );
                    b->SetFlag( BOND_IS_PEPTIDE );
                    break;
                }
            }
        }
    }
}

////////////////////////////////////////////////////////////////////////
void WriteSEQRES( CCMol& mol, ostream& ofs )
{
    char line[BUFF_SIZE];
    CCResidue* res( NULL );
    FOR_CHAINS_OF_MOL( chain, mol ) {
        int i( 0 ), s( 0 );
		bool bAA(true);
        for ( i = 0; i < chain->GetNumRes(); ) {
            res = chain->GetResidue( i + 1 );
            if ( !res || !res->IsNaturalAA() ) {
				++i;
				bAA = false;
                continue; // empty res or not a natural amino acid
            }

			if ( i % 13 == 0 ) {
				snprintf( line, BUFF_SIZE, "SEQRES  %2d %c %4d  ", ++s,
					chain->GetChainID(), chain->GetNumRes() );
				ofs << line;
			}

			//write residue name
			ofs << res->GetResName() << ' ';

            if ( ++i % 13 == 0 ) {
                ofs << "\n";
            }
        }

        if ( i % 13 != 0 && bAA ) {
            ofs << "\n";
        }
    }
}

void WriteSSBOND( CCMol& mol, ostream& ofs )
{
    char line[BUFF_SIZE];
    int ss( 0 );
    vector<CCResidue*> vCYS = mol.GetResOfName( "CYS" );
    for ( int i = 0; i < (int)vCYS.size() - 1; ++i ) {
        CCAtom* sa = vCYS[i]->GetAtomOfName( "SG" );
        if ( !sa ) {
            continue;
        }
        for ( int j = i + 1; j < (int)vCYS.size(); ++j ) {
            CCAtom* sb = vCYS[j]->GetAtomOfName( "SG" );
            if ( !sb ) {
                continue;
            }
            if ( IsAcceptableBond( sa, sb ) ) {
                // SSBOND, 1-72
                snprintf( line, BUFF_SIZE,
                          "%6s %3d %3s %c %4d%c   %3s %c %4d%c%23s%6s %6s",
                          "SSBOND",                      // Record Name
                  ++ss,                          // Serial number
                  vCYS[i]->GetResName().c_str(), // Residue name1
                  vCYS[i]->GetChainID(),         // Chain identifier1
                  vCYS[i]->GetResNum(),          // Residue sequence number1
                  ' ',                           // Insertion code1
                  vCYS[j]->GetResName().c_str(), // Residue name2
                  vCYS[j]->GetChainID(),         // Chain identifier2
                  vCYS[j]->GetResNum(),          // Residue sequence number2
                  ' ',                           // Insertion code1
                  " ",                           // 22 blanks
                  " ",                           // Symmetry oper1
                  " " );                          // Symmetry oper2
                ofs << line;
                ofs << "\n";
                // set flag for CONECT
                sa->SetFlag( ATOM_IS_IN_SSBOND );
                sb->SetFlag( ATOM_IS_IN_SSBOND );
            }
        }
    }
}

bool CCPDBFormat::ReadAtomsOnly(CCMol& mol, istream& is)
{
	// seek beginning of next molecule if possible or error
	// for buffer
	char line[BUFF_SIZE];

	// indicate a new chain
	char currChainID = '-';

	bool atomline( false ), hetatmline( false );

	// for generic data
	CCGenericData* gd( NULL );
	// for SEQRES
	string strSEQRES( "" );
	// for SSBOND
	string strSSBOND( "" );
	map<string, string> ssbond;
	// whether it is amber format.
	bool amber(false);

	mol.Clear();

	while ( is.getline( line, BUFF_SIZE ) && !EQn( line, "END", 3 ) ) {
		if ( EQn( line, "ATOM", 4 ) || EQn( line, "HETATM", 6 ) ) {
			if ( !atomline && EQn( line, "ATOM", 4 ) ) {
				atomline = true;
			}
			if ( !hetatmline && EQn( line, "HETATM", 6 ) ) {
				hetatmline = true;
			}
			if ( !ParseATOM( line, mol, currChainID, amber ) ) {
				// try amber pdb format
				if ( !ParseATOM( line, mol, currChainID, amber ) ) {
					string errMsg = "Parse atom error in LINE\n";
					gErrorLog.Throw( __FUNCTION__, errMsg + line );
					// just skip this line
					continue;
				}
			}
		}
		else if ( EQn( line, "TER", 3 ) ) {
			currChainID = '-'; // expect an new chain
		}
		else if ( EQn( line, "CONECT", 6 ) ) {
			continue;
		}
		else if ( EQn( line, "COMPND", 6 ) && ( mol.GetName() ).empty() ) {
			string name = line;
			if ( name.size() > 10 ) {
				name = name.substr( 10, BUFF_SIZE );
				Trim( name );
				mol.SetName( name );
			}
		}
		else if ( EQn( line, "HEADER", 6 ) && ( mol.GetName() ).empty() ) {
			string name = &line[6];
			Trim( name );
			mol.SetName( name );

			if ( strlen( line ) > 63 ) {
				// four character unique PDB ID
				string pdbID = line;
				pdbID = pdbID.substr( 62, 4 );
				Trim( pdbID );
				if ( !pdbID.empty() ) {
					gd = new CCPairString( "PDBID", pdbID );
					AssertMemAlloc( gd );
					mol.SetData( gd );
				}
			}
		}
		else if ( EQn( line, "SEQRES", 6 ) ) {
			// has sequence info.
			strSEQRES = strSEQRES + line + '\n';
		}
		else if ( EQn( line, "SSBOND", 6 ) ) {
			// disulfide bond
			strSSBOND = strSSBOND + line + '\n';
			ParseSSBOND( line, mol, ssbond );
		}
	}

	/*
	// store SEQRES if any
	if ( ! strSEQRES.empty() ) {
	gd = new CCPairString( "SEQRES", strSEQRES );
	AssertMemAlloc( gd );
	mol.SetData(gd);
	} */
	// make SSBond if it was missed in CONECT
	if ( !strSSBOND.empty() && !ssbond.empty() ) {
		/*
		gd = new CCPairString( "SSBOND", strSSBOND );
		AssertMemAlloc( gd );
		mol.SetData( gd ); */
		MakeSSBOND( mol, ssbond );
	}

	// assign bond according to template
	if ( atomline ) {
		if ( !hetatmline ) {
			hetatmline = !gResidueData.AssignBond( mol );
		}
		else {
			gResidueData.AssignBond( mol );
		}
	}

	// connect adjacent residues
	ConnectResidues( mol );

	if ( hetatmline ) {
		HandleHet( mol );
	}
	mol.SortBonds();

	// set flag to avoid repeated perception
	mol.SetFlag( MOL_HYBRID_PERCEIVED );
	mol.SetFlag( MOL_HBDONORACCEPTOR_PERCEIVED );
	mol.SetFlag( MOL_RING_PERCEIVED );
	mol.SetFlag( MOL_AROMATICITY_PERCEIVED );
	mol.SetFlag( MOL_AUTOFCHARGE_PERCEIVED );
	mol.SetFlag( MOL_KEKULE_PERCEIVED );
	mol.SetFlag( MOL_BONDORDER_PERCEIVED );
	mol.SetFlag( MOL_INT_TYPE_PERCEIVED );

	// remove NULL pointer at the end of m_vatom and m_vbond
	mol.ResizeAtoms( mol.GetNumAtom() );
	mol.ResizeBonds( mol.GetNumBond() );

	// eat up remaining blank lines
	while ( is.peek() != EOF
		&& is.good()
		&& ( is.peek() == '\n' || is.peek() == '\r' ) ) {
			is.getline( line, BUFF_SIZE );
	}

	return !mol.IsEmpty();
}

/** 
* @brief Read HETATM parts ( exclude HOH ) from PDB file.
* @param[out] mol molecule to save the HETATM(ligand) structure
* @param[in] is file point to read HETATM part
*/
bool CCPDBFormat::ReadHetatmLig( CCMol& mol, istream& is ) {
    // seek beginning of next molecule if possible or error
    // for buffer
    char line[BUFF_SIZE];

    // indicate a new chain
    char currChainID = '-';

    // for generic data
    CCGenericData* gd( NULL );

	// whether it is amber format.
	bool amber(false);

    mol.Clear();

    while ( is.getline( line, BUFF_SIZE ) && !EQn( line, "END", 3 ) ) {
        if ( EQn( line, "HETATM", 6 ) ) {
            if ( !ParseATOM( line, mol, currChainID, amber ) ) {
				// try amber pdb format
				if ( !ParseATOM( line, mol, currChainID, amber ) ) {
					string errMsg = "Parse atom error in LINE\n";
					gErrorLog.Throw( __FUNCTION__, errMsg + line );
					// just skip this line
					continue;
				}
            }
        }
        else if ( EQn( line, "TER", 3 ) ) {
            currChainID = '-'; // expect an new chain
        }
        //else if ( EQn( line, "CONECT", 6 ) ) {
        //    ParseCONECT( line, mol );
        //}
        else if ( EQn( line, "COMPND", 6 ) && ( mol.GetName() ).empty() ) {
            string name = line;
            if ( name.size() > 10 ) {
                name = name.substr( 10, BUFF_SIZE );
                Trim( name );
                mol.SetName( name );
            }
        }
        else if ( EQn( line, "HEADER", 6 ) && ( mol.GetName() ).empty() ) {
            string name = &line[6];
            Trim( name );
            mol.SetName( name );

            if ( strlen( line ) > 63 ) {
                // four character unique PDB ID
                string pdbID = line;
                pdbID = pdbID.substr( 62, 4 );
                Trim( pdbID );
                if ( !pdbID.empty() ) {
                    gd = new CCPairString( "PDBID", pdbID );
                    AssertMemAlloc( gd );
                    mol.SetData( gd );
					mol.SetName( pdbID + "_ligand" );
                }
            }
        }
    }

    HandleHet( mol );
    mol.SortBonds();

    // set flag to avoid repeated perception
    mol.SetFlag( MOL_HYBRID_PERCEIVED
	           | MOL_HBDONORACCEPTOR_PERCEIVED
			   | MOL_RING_PERCEIVED
			   | MOL_AROMATICITY_PERCEIVED
			   | MOL_AUTOFCHARGE_PERCEIVED
			   | MOL_KEKULE_PERCEIVED
			   | MOL_BONDORDER_PERCEIVED
			   | MOL_INT_TYPE_PERCEIVED );

    // remove NULL pointer at the end of m_vatom and m_vbond
    mol.ResizeAtoms( mol.GetNumAtom() );
    mol.ResizeBonds( mol.GetNumBond() );

    // eat up remaining blank lines
    while ( is.peek() != EOF
         && is.good()
         && ( is.peek() == '\n' || is.peek() == '\r' ) ) {
        is.getline( line, BUFF_SIZE );
    }

    return !mol.IsEmpty();
}
//////////////////////////////////////////////////////////////////////////
bool CCPDBFormat::WriteMolecule( CCMol& mol, ostream& os )
{
    if ( mol.IsEmpty() ) {
        gErrorLog.Throw( __FUNCTION__, "Molecule is empty." );
        return false;
    }

    string molName = mol.GetName();
    if ( molName.empty() ) {
        gErrorLog.Throw( __FUNCTION__,
                         "NULL molecular name! UNNAMED by default." );
        os << "HEADER    UNNAMED\n";
        os << "COMPND    UNNAMED\n";
    }
    else {
		// set max length of molecular name as 50 characters
        os << "HEADER    " << molName.substr( 0, 50 ) << "\n";
        os << "COMPND    " << molName.substr( 0, 50 ) << "\n";
    }

    os << "REMARK    GENERATED BY CCBGKit " 
		<< CCBGKit_VERSION << "           " << GetTime() << "\n";

    //CCGenericData* gd( NULL );
    // SEQRES
    WriteSEQRES( mol, os );

    // SSBOND
    WriteSSBOND( mol, os );

    CCResidue* res( NULL );
    CCChain* lastChain( NULL );
    string atomName;
    string element;
    int resNum( 0 );
    string resName;
    bool het( false );
    int l( 0 );
	char line[BUFF_SIZE];

    FOR_ATOMS_OF_MOL( atom, mol ) {
        element = atom->GetSymbol();
        //two char elements are on position 13 and 14 one char start at 14
        if ( element.size() > 1 ) {
            element[1] = toupper( element[1] );
        }

        res = atom->GetResidue();
        if ( res ) {
            het = res->IsHetAtom( &*atom );
            resNum = res->GetResNum();
            resName = res->GetResName();
            atomName = res->GetResAtomName( &*atom );

            char tmp[5] = "    ";

            // format atomName
            if ( atomName.empty() ) {
                atomName = element + "  ";
            }
            else if ( atomName.size() == 1 ) {
                snprintf( tmp, 5, " %1s  ", atomName.c_str() );
                atomName = tmp;
            }
            else if ( atomName.size() == 2 ) {
                if ( element.size() > 1 || isdigit(atomName[0]) ) { //e.g. 1H__
                    snprintf( tmp, 5, "%2s  ", atomName.c_str() );
                    atomName = tmp;
                }
                else {
                    snprintf( tmp, 5, " %2s ", atomName.c_str() );
                    atomName = tmp;
                }
            }
            else if ( atomName.size() == 3 ) {
                if ( element.size() > 1 || isdigit(atomName[0])) { //e.g. 1HB_
                    snprintf( tmp, 5, "%3s ", atomName.c_str() );
                    atomName = tmp;
                }
                else {
                    snprintf( tmp, 5, " %3s", atomName.c_str() );
                    atomName = tmp;
                }
            }
        }
        else {
            // atom may have no residue
            het = true;
            resNum = 1;
            resName = element;
            atomName = element + "  ";
        }

		// insertion code if any
		char iCode = ( res ) ? res->GetiCode() : ' ';
        // altLoc (if any)
        char altLoc = ( res ) ? res->GetAltLoc() : ' ';
        // chainID (if any)
        char chainID = ( res ) ? res->GetChainID() : ' ';
        // occupancy (if any)
		double occupancy = ( res ) ? res->GetOccupancy(&*atom) : 0.0;
        // tempFactor (if any)
		double tempFactor = (res) ? res->GetTempFactor(&*atom) : 0.0;
		// charge (if not NULL)
		int charge = atom->GetFormalCharge();

        if ( lastChain &&
			( ( res && res->GetChain() != lastChain ) || !res ) ) {
		// begin of new chain
        // end of last chain
        // write a TER before ATOM section
            os << "TER\n";
        }
        // update lastChain for next atom
        lastChain = ( res ) ? res->GetChain() : NULL;

        snprintf( line, BUFF_SIZE,
                  "%6s%5d %4s%c%3s %c%4d%c   %8.3f%8.3f%8.3f%6.2f%6.2f          %2s",
                  het
                ? "HETATM"
                : "ATOM  ", ++l, atomName.c_str(), altLoc,
                  resName.c_str(), chainID, resNum, iCode, atom->GetX(),
                  atom->GetY(), atom->GetZ(), occupancy, tempFactor,
                  element.c_str() );
		os << line;
		if (charge) {
			os << abs(charge) << (charge>0 ? "+" : "-");
		}
		else {
			os << "  ";
		}
		os << "\n";
    }

    // CONECT
	FOR_ATOMS_OF_MOL( atom, mol ) {
        if ( ( atom->GetValence() == 0 )
			|| ( atom->GetValence() > atom->GetMaxBond()) ) {
			continue;
		}

		if ( !(res = atom->GetResidue()) )
		{
			// write CONECT only for nonstandard group and SSBOND
			continue;
		}

		// write CONECT only for nonstandard group and SSBOND
		line[0] = '\0';
        FOR_NBORS_OF_ATOM( nbr, &*atom ) {
			if (nbr->GetIdx() < atom->GetIdx() ||
				(res->IsNaturalAA() && nbr->GetResidue()->IsNaturalAA()
				&& !(atom->IsSulfur()&&nbr->IsSulfur())) )
			{
				continue;
			}
            snprintf( line, BUFF_SIZE, "%5d", nbr->GetIdx() );
        }
		if (line[0]) {
			os << "CONECT" << setw(5) << atom->GetIdx() << line << "\n";
		}
    }

    snprintf( line, BUFF_SIZE,
              "MASTER        0    0    0    0    0    0    0    0 %4d    0 %4d    0",
              mol.GetNumAtom(), mol.GetNumBond() );
    os << line << "\n";
    os << "END" << endl;

    return true;
}
