/**********************************************************************
Copyright (C) 1998-2001 by OpenEye Scientific Software, Inc.
Some portions Copyright (C) 2001-2006 by Geoffrey R. Hutchison
Some portions Copyright (C) 2004 by Chris Morley

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation version 2 of the License.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
***********************************************************************/
//Contains CCSMIFormat
// TODO: Rewrite. Use std::string in place of char * to avoid buffer overflow
//  use std::string::reserve (or different allocator) to avoid resize slowdown

#include <formats/CCFormat.h>
#include <ccbg/CCMol.h>



class CCSMIFormat : public CCFormat
{
public:
    //Register this format type ID
    CCSMIFormat()
    {
        RegisterFormat( "smi", this );
    }
    ~CCSMIFormat()
    {
    }

    ////////////////////////////////////////////////////
	bool ReadMolecule( CCMol& mol, istream& is );
	bool ReadAtomsOnly( CCMol& mol, istream& is );
	bool WriteMolecule( CCMol& mol, ostream& os );
    ///////////////////////////////////////////////////////

    virtual const char* Description()
    {
        return
        "SMILES format\n"
        "A linear text format which can describe the connectivity\n"
        "and chirality of a molecule\n";
    }
};
CCSMIFormat theCCSMIFormat; // make an instant and register format

//////////////////////////////////////////////////////////////////
class CCSmiNode
{
    CCAtom* _atom, * _parent;
    std::vector<CCSmiNode*> _nextnode;
    std::vector<CCBond*> _nextbond;
public:
    CCSmiNode( CCAtom* atom );
    ~CCSmiNode();
    int Size()
    {
        return _nextnode.empty() ? 0 : _nextnode.size();
    }
    void SetParent( CCAtom* a )
    {
        _parent = a;
    }
    void SetNextNode( CCSmiNode*, CCBond* );
    CCAtom* GetAtom()
    {
        return _atom;
    }
    CCAtom* GetParent()
    {
        return _parent;
    }
    CCAtom* GetNextAtom( int i )
    {
        return _nextnode[i]->GetAtom();
    }
    CCBond* GetNextBond( int i )
    {
        return _nextbond[i];
    }
    CCSmiNode* GetNextNode( int i )
    {
        return _nextnode[i];
    }
};

class CCMol2Smi
{
    std::vector<int> _atmorder;
    std::vector<int> _storder;
    std::vector<bool> _aromNH;
    CCBitVec _uatoms, _ubonds;
    std::vector<CCBond*> _vclose;
    std::vector<std::pair<CCAtom*, std::pair<int, int> > > _vopen;
public:
    CCMol2Smi()
    {
        _vclose.clear();
    }
    ~CCMol2Smi()
    {
    }
    int GetUnusedIndex();
    void Init();
    void CreateSmiString( CCMol&, char* );
    void GetClosureAtoms( CCAtom*, std::vector<CCAtom*>& );
    void FindClosureBonds( CCMol& );
    void ToSmilesString( CCSmiNode* node, char* buffer );
    void RemoveUsedClosures();
    void AssignCisTrans( CCSmiNode* );
    char GetCisTransBondSymbol( CCBond*, CCSmiNode* );
    bool BuildTree( CCSmiNode* );
    bool GetSmilesElement( CCSmiNode*, char* );
    bool GetChiralStereo( CCSmiNode*, char* );
    void CorrectAromaticAmineCharge( CCMol& );
    std::vector<std::pair<int, CCBond*> > GetClosureDigits( CCAtom* );
    std::vector<int>& GetOutputOrder()
    {
        return( _atmorder );
    }
};

bool WriteTheSmiles( CCMol& mol, char* out );

/////////////////////////////////////////////////////////////////
class CCSmilesParser
{
    int _bondflags;
    int _order;
    int _prev;
    char* _ptr;
    vector<int> _vprev;
    vector<vector<int> > _rclose;
    vector<vector<int> > _extbond;
    vector<int> _path;
    vector<bool> _avisit;
    vector<bool> _bvisit;
    char _buffer[BUFF_SIZE];
    vector<int> PosDouble; //for extension: lc atoms as conjugated double bonds
    bool chiralWatch; // set when a chiral atom is read
    map<CCAtom*, CCChiralData*> _mapcd; // map of ChiralAtoms and their data
public:

    CCSmilesParser()
    {
    }
    ~CCSmilesParser()
    {
    }

    bool SmiToMol( CCMol&, string& );
    bool ParseSmiles( CCMol& );
    bool ParseSimple( CCMol& );
    bool ParseComplex( CCMol& );
    bool ParseRingBond( CCMol& );
    bool ParseExternalBond( CCMol& );
    bool CapExternalBonds( CCMol& mol );
    void FindAromaticBonds( CCMol& mol, CCAtom*, int );
    void FindAromaticBonds( CCMol& );
    void FindOrphanAromaticAtoms( CCMol& mol ); //CM 18 Sept 2003
    void FixCisTransBonds( CCMol& );
    void CorrectUpDownMarks( CCBond*, CCAtom* );
};

/////////////////////////////////////////////////////////////////
bool CCSMIFormat::ReadMolecule( CCMol& mol, istream& is )
{
    // clean molecule
    mol.Clear();
    // molecule has connection table from smi format
    mol.SetFlag( MOL_CONNECTIVITY_PERCEIVED );

    //Taken unchanged from ReadSmiles
    char buffer[BUFF_SIZE];

    if ( !is.getline( buffer, BUFF_SIZE ) ) {
        return false;
    }
    vector<string> vs;
    Tokenize( vs, buffer );

    // Essentially everything after the first space on a SMILES file line
    // is treated as the name.
    if ( vs.size() > 2 ) {
        for ( int i = 2; i < vs.size(); i++ ) {
            vs[1] = vs[1] + " " + vs[i];
        }
    }

    if ( vs.empty() ) {
        return false;
    }
    mol.SetDimension( 0 );

    if ( vs.size() >= 2 ) {
        mol.SetName( vs[1].c_str() );
    }
    else {
        gErrorLog.Throw( __FUNCTION__,
                         "NULL molecular name! **** by default.", ccWarning );
        mol.SetName( "****" );
    }

    CCSmilesParser sp;
    return sp.SmiToMol( mol, vs[0] );
}

bool CCSMIFormat::ReadAtomsOnly( CCMol& mol, istream& is )
{
	return ReadMolecule( mol, is );
}
//////////////////////////////////////////////////
bool CCSMIFormat::WriteMolecule( CCMol& mol, ostream& os )
{
    if ( mol.IsEmpty() ) {
        gErrorLog.Throw( __FUNCTION__, "Molecule is empty." );
        return false;
    }

    char buffer[BUFF_SIZE] = ""; //empty buffer

    // This is a hack to prevent recursion problems.
    // we still need to fix the underlying problem (mainly chiral centers) -GRH
    if ( mol.GetNumAtom() > 1000 ) {
        stringstream errorMsg;
        errorMsg
     << "SMILES Conversion failed: Molecule is too large to convert, limited to 1000 atoms.\n";
        errorMsg << "Molecule size: " << mol.GetNumAtom() << " atoms.";
        gErrorLog.Throw( __FUNCTION__, errorMsg.str() );
        return false;
    }

    if ( mol.GetNumAtom() != 0 ) {
        CCMol2Smi m2s;
        m2s.Init();
        m2s.CorrectAromaticAmineCharge( mol );
        m2s.CreateSmiString( mol, buffer );
    }

	string molName = mol.GetName();
	// set max length of molecular name as 50 characters
    os << buffer << '\t' << molName.substr( 0, 50 );
    os << "\n";

    return true;
}

//////////////////////////////////////////////

bool CCSmilesParser::SmiToMol( CCMol& mol, string& s )
{
    strncpy( _buffer, s.c_str(), BUFF_SIZE );
    _buffer[BUFF_SIZE - 1] = '\0';

    _vprev.clear();
    _rclose.clear();
    _prev = 0;
    chiralWatch = false;

    if ( !ParseSmiles( mol ) ) {
        mol.Clear();
        return false;
    }

    mol.SetAutoFormalCharge( false );

    return true;
}

bool CCSmilesParser::ParseSmiles( CCMol& mol )
{
    for ( _ptr = _buffer; *_ptr; _ptr++ ) {
        if ( isspace( *_ptr ) ) {
            continue;
        }
        else if ( isdigit( *_ptr ) || *_ptr == '%' ) //ring open/close
        {
            if ( !ParseRingBond( mol ) ) {
                return false;
            };
            continue;
        }
        else if ( *_ptr == '&' ) //external bond
        {
            ParseExternalBond( mol );
            continue;
        }
        else {
            switch ( *_ptr ) {
            case '.':
                _prev = 0;
                break;
            case '(':
                _vprev.push_back( _prev );
                break;
            case ')':
                if ( _vprev.empty() ) //CM
                {
                    return false;
                }
                _prev = _vprev.back();
                _vprev.pop_back();
                break;
            case '[':
                if ( !ParseComplex( mol ) ) {
                    mol.Clear();
                    return false;
                }
                break;
            case '-':
                _order = 1;
                break;
            case '=':
                _order = 2;
                break;
            case '#':
                _order = 3;
                break;
            case ':':
                _order = 5;
                break;
            case '/':
                _bondflags |= BOND_IS_TORDOWN;   // initial mark, see FixCisTransBonds() below 
                break;
            case '\\':
                _bondflags |= BOND_IS_TORUP;     // initial mark, see FixCisTransBonds() below 
                break;
            default:
                if ( !ParseSimple( mol ) ) {
                    mol.Clear();
                    return false;
                }
            }
        } // end switch
    } // end for _ptr

    // place dummy atoms for each unfilled external bond
    if ( !_extbond.empty() ) {
        CapExternalBonds( mol );
    }

    //set aromatic bond orders
    mol.SetFlag( MOL_AROMATICITY_PERCEIVED );
    FindAromaticBonds( mol );
    FindOrphanAromaticAtoms( mol );// CM 18 Sept 2003
    mol.AssignSpin();
    mol.UnsetFlag( MOL_AROMATICITY_PERCEIVED );

    FixCisTransBonds( mol );

    //Extension which interprets cccc with conjugated double bonds if niether
    //of its atoms is aromatic.
    vector<int>::iterator itr;
    for ( itr = PosDouble.begin(); itr != PosDouble.end(); ++itr ) {
        CCBond* bond = mol.GetBond( *itr );
        if ( !bond->GetBeginAtom()->IsAromatic()
          && !bond->GetEndAtom()->IsAromatic() ) {
            bond->SetOrder( 2 );
            mol.UnsetFlag( MOL_IMPVALENCE_PERCEIVED );
        }
    }

    //NE add the CCChiralData stored inside the _mapcd to the atoms now after end
    // modify so they don't get lost.
    if ( _mapcd.size() > 0 ) {
        CCAtom* atom;
        CCChiralData* cd;
        map<CCAtom*, CCChiralData*>::iterator ChiralSearch;
        for ( ChiralSearch = _mapcd.begin();
              ChiralSearch != _mapcd.end();
              ChiralSearch++ ) {
            atom = ChiralSearch->first;
            cd = ChiralSearch->second;
            atom->SetData( cd );
        }
    }

    return true;
}

void CCSmilesParser::FixCisTransBonds( CCMol& mol )
{
    // OpenBabel's internal model for cis/trans uses an imaginary drawing,
    // in which the double bond is horizontal on the page, and an "up" bond
    // means, "above the double-bond on the page", and "down" means, "below
    // the double bond on the page".  Thus, a cis configuration can be
    // represented as either "up/up" or "down/down", and a trans
    // configuration can be either "up/down" or "down/up".
    //
    // When parsing a SMILES, '/' and '\' bonds are initially marked as
    // "down" and "up", respectively, but they don't mean "down" and "up"
    // yet -- they only mean "forward slash" and "backslash".  These bond
    // types have to be converted to an actual cis/trans specification.
    //
    // For example, consider the following trans ethene:
    //
    //   SMILES     Configuration  Initial parsing    Final designation
    //   ------     -------------  ---------------    -----------------
    //   C/C=C/C        trans        down/down           down/up
    //   C(\C)=C/C      trans        up/down             down/up
    //   C/C=C(/C)      trans        down/down           down/up
    //   C(=C/C)\C      trans        down/up             up/down
    //
    // The SMILES parsing rule is:
    //   Before double-bonded atom: '/' means down, '\' means up
    //   After  double-bonded atom: '/' means up,   '\' means down
    //
    // Note: This must be called *after* aromaticity detection.

    FOR_BONDS_OF_MOL( dbi, mol ) {
        CCBond* dbl_bond = &( *dbi );

        // Not a double bond?
        if ( !dbl_bond->IsDouble() || dbl_bond->IsAromatic() ) {
            continue;
        }

        // Find the single bonds around the atoms connected by the double bond.
        // While we're at it, note whether the pair of atoms on either end are
        // identical, in which case it's not cis/trans.

        CCAtom* a1 = dbl_bond->GetBeginAtom();
        CCAtom* a2 = dbl_bond->GetEndAtom();

        // Check that both atoms on the double bond have at least one
        // other neighbor, but not more than two other neighbors;
        int v1 = a1->GetValence();
        int v2 = a2->GetValence();
        if ( v1 <2 || v1> 3 || v2 <2 || v2> 3 ) {
            continue;
        }

        // Get the bonds of neighbors of atom1 and atom2
        CCBond* a1_b1 = NULL, * a1_b2 = NULL, * a2_b1 = NULL, * a2_b2 = NULL;

        FOR_BONDS_OF_ATOM( bi, a1 ) {
            CCBond* b = &( *bi );
            if ( ( b ) == ( dbl_bond ) ) {
                continue;
            }  // skip the double bond we're working on
            if ( NULL == a1_b1 ) {
                a1_b1 = b;
            }    // remember 1st bond of Atom1
            else {
                a1_b2 = b;
            }    // remember 2nd bond of Atom1
        }

        FOR_BONDS_OF_ATOM( bi, a2 ) {
            CCBond* b = &( *bi );
            if ( b == dbl_bond ) {
                continue;
            }
            if ( NULL == a2_b1 ) {
                a2_b1 = b;
            }    // remember 1st bond of Atom2
            else {
                a2_b2 = b;
            }    // remember 2nd bond of Atom2
        }

        // Now check that at least two are marked up/down.
        int count = 0;
        if ( a1_b1 && ( a1_b1->IsUp() || a1_b1->IsDown() ) ) {
            count++;
        }
        if ( a1_b2 && ( a1_b2->IsUp() || a1_b2->IsDown() ) ) {
            count++;
        }
        if ( a2_b1 && ( a2_b1->IsUp() || a2_b1->IsDown() ) ) {
            count++;
        }
        if ( a2_b2 && ( a2_b2->IsUp() || a2_b2->IsDown() ) ) {
            count++;
        }
        if ( count < 2 ) {
            continue;
        }

        // OK, now do what we're here for.  We have two, three or four
        // bonds, marked "up" or "down", but at this point it just
        // means '/' or '\', respectively.  In order to decide whether
        // a bond is "up" or "down", we examine the order of atoms in
        // the molecule.  See the comments at the top of this function.

        CorrectUpDownMarks( a1_b1, a1 );
        CorrectUpDownMarks( a1_b2, a1 );
        CorrectUpDownMarks( a2_b1, a2 );
        CorrectUpDownMarks( a2_b2, a2 );
    }
}

void CCSmilesParser::CorrectUpDownMarks( CCBond* b, CCAtom* a )
{
    // This is an adjunct to FixCisTransBonds(), above.  See the comments
    // there.  In this function, atom a is one of the double-bonded atoms,
    // and bond b is a bond from atom a to one of the substituent atoms.

    if ( !b || !a || !( b->IsUp() || b->IsDown() ) ) {
        return;
    }

    CCAtom* ax = b->GetNbrAtom( a );

    // If the substituent comes before the double-bonded atom, then
    // the initial marks from the SMILES are correct.
    if ( ax->GetIdx() < a->GetIdx() ) {
        return;
    }

    // The substituent comes after the double-bonded atom, so
    // the bond '/' means "up", and '\' means "down".
    if ( b->IsUp() ) {
        b->SetFlag( BOND_IS_TORDOWN );
    }
    else {
        b->SetFlag( BOND_IS_TORUP );
    }
}


void CCSmilesParser::FindOrphanAromaticAtoms( CCMol& mol )
{
    //Facilitates the use lower case shorthand for radical entry
    //Atoms which are marked as aromatic but have no aromatic bonds
    //are taken to be radical centres
    CCAtom* atom;
    vector<CCAtom*>::iterator j;

    for ( atom = mol.BeginAtom( j ); atom; atom = mol.NextAtom( j ) ) {
        if ( atom->IsAromatic() ) {
            if ( atom->CountBondWithOrder( 5 ) < 2 ) //bonds order 5 set in FindAromaticBonds()
                //not proper aromatic atoms - could be conjugated chain or radical centre
            {
                atom->UnsetAromatic();
            }
            else {
                //recognized as aromatic, so are not radicals
                atom->SetSpin( 0 );
            }
        }
    }
}

void CCSmilesParser::FindAromaticBonds( CCMol& mol )
{
    _path.clear();
    _avisit.clear();
    _bvisit.clear();
    _avisit.resize( mol.GetNumAtom() );
    _bvisit.resize( mol.GetNumBond() );
    _path.resize( mol.GetNumAtom() );

    CCBond* bond;
    vector<CCBond*>::iterator i;
    for ( bond = mol.BeginBond( i ); bond; bond = mol.NextBond( i ) ) {
        if ( !bond->GetBeginAtom()->IsAromatic()
          || !bond->GetEndAtom()->IsAromatic() ) {
            _bvisit[bond->GetIdx()] = true;
        }
    }

    CCAtom* atom;
    vector<CCAtom*>::iterator j;

    for ( atom = mol.BeginAtom( j ); atom; atom = mol.NextAtom( j ) ) {
        if ( !_avisit[atom->GetIdx()] && atom->IsAromatic() ) {
            FindAromaticBonds( mol, atom, 0 );
        }
    }
}

void CCSmilesParser::FindAromaticBonds( CCMol& mol, CCAtom* atom, int depth )
{
    CCBond* bond;
    vector<CCBond*>::iterator k;

    if ( _avisit[atom->GetIdx()] ) {
        int j = depth - 1;
        bond = mol.GetBond( _path[j--] );
        bond->SetOrder( 5 );
        while ( j >= 0 ) {
            bond = mol.GetBond( _path[j--] );
            bond->SetOrder( 5 );
            if ( bond->GetBeginAtom() == atom || bond->GetEndAtom() == atom ) {
                break;
            }
        }
    }
    else {
        _avisit[atom->GetIdx()] = true;
        for ( bond = atom->BeginNbrBond( k );
              bond;
              bond = atom->NextNbrBond( k ) ) {
            if ( !_bvisit[bond->GetIdx()] ) {
                _path[depth] = bond->GetIdx();
                _bvisit[bond->GetIdx()] = true;
                FindAromaticBonds( mol, bond->GetNbrAtom( atom ), depth + 1 );
            }
        }
    }
}


bool CCSmilesParser::ParseSimple( CCMol& mol )
{
    char symbol[3];
    int element;
    bool arom = false;
    memset( symbol, '\0', sizeof( char ) * 3 );

    if ( isupper( *_ptr ) ) {
        switch ( *_ptr ) {
        case 'C':
            _ptr++;
            if ( *_ptr == 'l' ) {
                strcpy( symbol, "Cl" );
                element = 17;
            }
            else {
                symbol[0] = 'C';
                element = 6;
                _ptr--;
            }
            break;

        case 'N':
            element = 7;
            symbol[0] = 'N';
            break;
        case 'O':
            element = 8;
            symbol[0] = 'O';
            break;
        case 'S':
            element = 16;
            symbol[0] = 'S';
            break;
        case 'P':
            element = 15;
            symbol[0] = 'P';
            break;
        case 'F':
            element = 9;
            symbol[0] = 'F';
            break;
        case 'I':
            element = 53;
            symbol[0] = 'I';
            break;

        case 'B':
            _ptr++;
            if ( *_ptr == 'r' ) {
                element = 35;
                strcpy( symbol, "Br" );
            }
            else {
                element = 5;
                symbol[0] = 'B';
                _ptr--;
            }
            break;
        default:
            return false;
        }
    }
    else {
        arom = true;
        switch ( *_ptr ) {
        case 'c':
            element = 6;
            symbol[0] = 'C';
            break;
        case 'n':
            element = 7;
            symbol[0] = 'N';
            break;
        case 'o':
            element = 8;
            symbol[0] = 'O';
            break;
        case 'p':
            element = 15;
            symbol[0] = 'P';
            break;
        case 's':
            element = 16;
            symbol[0] = 'S';
            break;
        case '*':
            element = 0;
            strcpy( symbol, "Du" );
            arom = false;
            break;
        case 'b':
            gErrorLog.Throw( __FUNCTION__, "Illegal aromatic element b",
                             ccWarning );
            element = 5;
            strcpy( symbol, "B" );
            break;
        default:
            return false;
        }
    }

    CCAtom* atom = mol.NewAtom();
    atom->SetAtomicNum( element );
    atom->SetType( symbol );
    if ( arom ) {
        atom->SetAromatic();
        atom->SetSpin( 2 ); // CM 18 Sept 2003
    }

    // Untrue, but necessary to avoid perception being called in CCAtom::IsAromatic()
    // on incomplete molecule. Undone at end of function. 
    mol.SetFlag( MOL_AROMATICITY_PERCEIVED );

    if ( _prev ) //need to add bond
    {
        /* CM 18 Sept 2003
            An extension to the SMILES format has been added so that lower case c,n,o can 
            represent a radical centre: CcC is isopropyl radical;
            and cccc... a carbon chain bonded by conjugated double bonds.
            Fails sometimes when using c as both aromatic and as the extened form.
            For benzyl radical C6H5CH2. c1ccccc1c is ok; c1cc(c)ccc1 fails.
            Radical centres should not be involved in ring closure:
            for cyclohexyl radical C1cCCCC1 is ok, c1CCCCC1 is not.  
            Implementation
            Atoms c,n,o, etc initially added as a radical centre
            unless _prev is a radical centre when both are made a normal atoms
            connected by a double bond. But making this bond double is deferred until
            the molecule has been constructed, because it is not appropriate if
            either of the atoms is really part of an aromatic ring.
            Since they are still marked as aromatic, FindAromaticBonds() will
            replace the bonds by aromatic bonds if they are in a ring.
            FindOrphanAromand removes the aromatic tag from the atoms not found in this way
            and removes stray radical centres.
            */
        CCAtom* prevatom = mol.GetAtom( _prev );
        if ( arom && prevatom->IsAromatic() ) {
            _order = 5; //Potential aromatic bond

            if ( prevatom->GetSpin() ) {
                //Previous atom had been marked, so bond is potentially a double bond
                //if it is not part of an aromatic ring. This will be decided when all
                //molecule has been constructed.
                PosDouble.push_back( mol.GetNumBond() ); //saves index of bond about to be added
                prevatom->SetSpin( 0 );
                atom->SetSpin( 0 );
            }
        }

        mol.AddBond( _prev, mol.GetNumAtom(), _order, _bondflags );

        //NE iterate through and see if atom is bonded to chiral atom
        map<CCAtom*, CCChiralData*>::iterator ChiralSearch;
        ChiralSearch = _mapcd.find( mol.GetAtom( _prev ) );
        if ( ChiralSearch != _mapcd.end() && ChiralSearch->second != NULL ) {
            ( ChiralSearch->second )->AddAtomRef( mol.GetNumAtom(), input );
            // cout << "Line 650: Adding "<<mol.GetNumAtom()<<" to "<<ChiralSearch->second<<endl;
        }
    }

    //set values
    _prev = mol.GetNumAtom();
    _order = 1;
    _bondflags = 0;

    mol.UnsetFlag( MOL_AROMATICITY_PERCEIVED ); //undo 
    return true;
}

bool CCSmilesParser::ParseComplex( CCMol& mol )
{
    char symbol[7];
    int element = 0;
    int isotope = 0;
    int isoPtr = 0;
    bool arom = false;
    memset( symbol, '\0', sizeof( char ) * 7 );

    _ptr++;

    //grab isotope information
    for ( ; *_ptr && isdigit( *_ptr ) && ( isoPtr <= 6 ); _ptr++ ) {
        symbol[isoPtr] = *_ptr;
        isoPtr++;
    }
    if ( isoPtr >= 6 ) {
        return false;
    }
    isotope = atoi( symbol );

    //parse element data
    if ( isupper( *_ptr ) ) {
        switch ( *_ptr ) {
        case 'C':
            _ptr++;
            switch ( *_ptr ) {
            case 'a':
                element = 20;
                strcpy( symbol, "Ca" );
                break;
            case 'd':
                element = 48;
                strcpy( symbol, "Cd" );
                break;
            case 'e':
                element = 58;
                strcpy( symbol, "Ce" );
                break;
            case 'f':
                element = 98;
                strcpy( symbol, "Cf" );
                break;
            case 'l':
                element = 17;
                strcpy( symbol, "Cl" );
                break;
            case 'm':
                element = 96;
                strcpy( symbol, "Cm" );
                break;
            case 'o':
                element = 27;
                strcpy( symbol, "Co" );
                break;
            case 'r':
                element = 24;
                strcpy( symbol, "Cr" );
                break;
            case 's':
                element = 55;
                strcpy( symbol, "Cs" );
                break;
            case 'u':
                element = 29;
                strcpy( symbol, "Cu" );
                break;
            default:
                element = 6;
                symbol[0] = 'C';
                _ptr--;
            }
            break;

        case 'N':
            _ptr++;
            switch ( *_ptr ) {
            case 'a':
                element = 11;
                strcpy( symbol, "Na" );
                break;
            case 'b':
                element = 41;
                strcpy( symbol, "Nb" );
                break;
            case 'd':
                element = 60;
                strcpy( symbol, "Nd" );
                break;
            case 'e':
                element = 10;
                strcpy( symbol, "Ne" );
                break;
            case 'i':
                element = 28;
                strcpy( symbol, "Ni" );
                break;
            case 'o':
                element = 102;
                strcpy( symbol, "No" );
                break;
            case 'p':
                element = 93;
                strcpy( symbol, "Np" );
                break;
            default:
                element = 7;
                symbol[0] = 'N';
                _ptr--;
            }
            break;

        case('O'):
            _ptr++;
            if ( *_ptr == 's' ) {
                element = 76;
                strcpy( symbol, "Os" );
            }
            else {
                element = 8;
                symbol[0] = 'O';
                _ptr--;
            }
            break;

        case 'P':
            _ptr++;
            switch ( *_ptr ) {
            case 'a':
                element = 91;
                strcpy( symbol, "Pa" );
                break;
            case 'b':
                element = 82;
                strcpy( symbol, "Pb" );
                break;
            case 'd':
                element = 46;
                strcpy( symbol, "Pd" );
                break;
            case 'm':
                element = 61;
                strcpy( symbol, "Pm" );
                break;
            case 'o':
                element = 84;
                strcpy( symbol, "Po" );
                break;
            case 'r':
                element = 59;
                strcpy( symbol, "Pr" );
                break;
            case 't':
                element = 78;
                strcpy( symbol, "Pt" );
                break;
            case 'u':
                element = 94;
                strcpy( symbol, "Pu" );
                break;
            default:
                element = 15;
                symbol[0] = 'P';
                _ptr--;
            }
            break;

        case('S'):
            _ptr++;
            switch ( *_ptr ) {
            case 'b':
                element = 51;
                strcpy( symbol, "Sb" );
                break;
            case 'c':
                element = 21;
                strcpy( symbol, "Sc" );
                break;
            case 'e':
                element = 34;
                strcpy( symbol, "Se" );
                break;
            case 'i':
                element = 14;
                strcpy( symbol, "Si" );
                break;
            case 'm':
                element = 62;
                strcpy( symbol, "Sm" );
                break;
            case 'n':
                element = 50;
                strcpy( symbol, "Sn" );
                break;
            case 'r':
                element = 38;
                strcpy( symbol, "Sr" );
                break;
            default:
                element = 16;
                symbol[0] = 'S';
                _ptr--;
            }
            break;

        case 'B':
            _ptr++;
            switch ( *_ptr ) {
            case 'a':
                element = 56;
                strcpy( symbol, "Ba" );
                break;
            case 'e':
                element = 4;
                strcpy( symbol, "Be" );
                break;
            case 'i':
                element = 83;
                strcpy( symbol, "Bi" );
                break;
            case 'k':
                element = 97;
                strcpy( symbol, "Bk" );
                break;
            case 'r':
                element = 35;
                strcpy( symbol, "Br" );
                break;
            default:
                element = 5;
                symbol[0] = 'B';
                _ptr--;
            }
            break;

        case 'F':
            _ptr++;
            switch ( *_ptr ) {
            case 'e':
                element = 26;
                strcpy( symbol, "Fe" );
                break;
            case 'm':
                element = 100;
                strcpy( symbol, "Fm" );
                break;
            case 'r':
                element = 87;
                strcpy( symbol, "Fr" );
                break;
            default:
                element = 9;
                symbol[0] = 'F';
                _ptr--;
            }
            break;

        case 'I':
            _ptr++;
            switch ( *_ptr ) {
            case 'n':
                element = 49;
                strcpy( symbol, "In" );
                break;
            case 'r':
                element = 77;
                strcpy( symbol, "Ir" );
                break;
            default:
                element = 53;
                symbol[0] = 'I';
                _ptr--;
            }
            break;

        case 'A':
            _ptr++;
            switch ( *_ptr ) {
            case 'c':
                element = 89;
                strcpy( symbol, "Ac" );
                break;
            case 'g':
                element = 47;
                strcpy( symbol, "Ag" );
                break;
            case 'l':
                element = 13;
                strcpy( symbol, "Al" );
                break;
            case 'm':
                element = 95;
                strcpy( symbol, "Am" );
                break;
            case 'r':
                element = 18;
                strcpy( symbol, "Ar" );
                break;
            case 's':
                element = 33;
                strcpy( symbol, "As" );
                break;
            case 't':
                element = 85;
                strcpy( symbol, "At" );
                break;
            case 'u':
                element = 79;
                strcpy( symbol, "Au" );
                break;
            default:
                _ptr--;
                return false;
            }
            break;

        case 'D':
            _ptr++;
            if ( *_ptr == 'y' ) {
                element = 66;
                strcpy( symbol, "Dy" );
            }
            else {
                _ptr--;
                return false;
            }
            break;

        case 'E':
            _ptr++;
            switch ( *_ptr ) {
            case 'r':
                element = 68;
                strcpy( symbol, "Er" );
                break;
            case 's':
                element = 99;
                strcpy( symbol, "Es" );
                break;
            case 'u':
                element = 63;
                strcpy( symbol, "Eu" );
                break;
            default:
                _ptr--;
                return false;
            }
            break;

        case 'G':
            _ptr++;
            switch ( *_ptr ) {
            case 'a':
                element = 31;
                strcpy( symbol, "Ga" );
                break;
            case 'd':
                element = 64;
                strcpy( symbol, "Gd" );
                break;
            case 'e':
                element = 32;
                strcpy( symbol, "Ge" );
                break;
            default:
                _ptr--;
                return false;
            }
            break;

        case 'H':
            _ptr++;
            switch ( *_ptr ) {
            case 'e':
                element = 2;
                strcpy( symbol, "He" );
                break;
            case 'f':
                element = 72;
                strcpy( symbol, "Hf" );
                break;
            case 'g':
                element = 80;
                strcpy( symbol, "Hg" );
                break;
            case 'o':
                element = 67;
                strcpy( symbol, "Ho" );
                break;
            default:
                element = 1;
                symbol[0] = 'H';
                _ptr--;
            }
            break;

        case 'K':
            _ptr++;
            if ( *_ptr == 'r' ) {
                element = 36;
                strcpy( symbol, "Kr" );
            }
            else {
                element = 19;
                symbol[0] = 'K';
                _ptr--;
            }
            break;

        case 'L':
            _ptr++;
            switch ( *_ptr ) {
            case 'a':
                element = 57;
                strcpy( symbol, "La" );
                break;
            case 'i':
                element = 3;
                strcpy( symbol, "Li" );
                break;
            case 'r':
                element = 103;
                strcpy( symbol, "Lr" );
                break;
            case 'u':
                element = 71;
                strcpy( symbol, "Lu" );
                break;
            default:
                _ptr--;
                return false;
            }
            break;

        case 'M':
            _ptr++;
            switch ( *_ptr ) {
            case 'd':
                element = 101;
                strcpy( symbol, "Md" );
                break;
            case 'g':
                element = 12;
                strcpy( symbol, "Mg" );
                break;
            case 'n':
                element = 25;
                strcpy( symbol, "Mn" );
                break;
            case 'o':
                element = 42;
                strcpy( symbol, "Mo" );
                break;
            default:
                _ptr--;
                return false;
            }
            break;

        case 'R':
            _ptr++;
            switch ( *_ptr ) {
            case 'a':
                element = 88;
                strcpy( symbol, "Ra" );
                break;
            case 'b':
                element = 37;
                strcpy( symbol, "Rb" );
                break;
            case 'e':
                element = 75;
                strcpy( symbol, "Re" );
                break;
            case 'h':
                element = 45;
                strcpy( symbol, "Rh" );
                break;
            case 'n':
                element = 86;
                strcpy( symbol, "Rn" );
                break;
            case 'u':
                element = 44;
                strcpy( symbol, "Ru" );
                break;
            default:
                _ptr--;
                return false;
            }
            break;

        case 'T':
            _ptr++;
            switch ( *_ptr ) {
            case 'a':
                element = 73;
                strcpy( symbol, "Ta" );
                break;
            case 'b':
                element = 65;
                strcpy( symbol, "Tb" );
                break;
            case 'c':
                element = 43;
                strcpy( symbol, "Tc" );
                break;
            case 'e':
                element = 52;
                strcpy( symbol, "Te" );
                break;
            case 'h':
                element = 90;
                strcpy( symbol, "Th" );
                break;
            case 'i':
                element = 22;
                strcpy( symbol, "Ti" );
                break;
            case 'l':
                element = 81;
                strcpy( symbol, "Tl" );
                break;
            case 'm':
                element = 69;
                strcpy( symbol, "Tm" );
                break;
            default:
                _ptr--;
                return false;
            }
            break;

        case('U'):
            element = 92;
            symbol[0] = 'U';
            break;
        case('V'):
            element = 23;
            symbol[0] = 'V';
            break;
        case('W'):
            element = 74;
            symbol[0] = 'W';
            break;

        case('X'):
            _ptr++;
            if ( *_ptr == 'e' ) {
                element = 54;
                strcpy( symbol, "Xe" );
            }
            else {
                _ptr--;
                return false;
            }
            break;

        case('Y'):
            _ptr++;
            if ( *_ptr == 'b' ) {
                element = 70;
                strcpy( symbol, "Yb" );
            }
            else {
                element = 39;
                symbol[0] = 'Y';
                _ptr--;
            }
            break;

        case('Z'):
            _ptr++;
            switch ( *_ptr ) {
            case 'n':
                element = 30;
                strcpy( symbol, "Zn" );
                break;
            case 'r':
                element = 40;
                strcpy( symbol, "Zr" );
                break;
            default:
                _ptr--;
                return false;
            }
            break;
        }
    }
    else {
        arom = true;
        switch ( *_ptr ) {
        case 'c':
            element = 6;
            symbol[0] = 'C';
            break;
        case 'n':
            element = 7;
            symbol[0] = 'N';
            break;
        case 'o':
            element = 8;
            symbol[0] = 'O';
            break;
        case 'p':
            element = 15;
            symbol[0] = 'P';
            break;
        case 'a':
            _ptr++;
            if ( *_ptr == 's' ) {
                element = 33;
                strcpy( symbol, "As" );
            }
            else {
                return false;
            }
            break;
        case '*':
            element = 0;
            strcpy( symbol, "Du" );
            arom = false;
            break;
        case 's':
            //note fall through
            _ptr++;
            if ( *_ptr == 'e' ) {
                element = 34;
                strcpy( symbol, "Se" );
                break;
            }
            else if ( *_ptr == 'i' || *_ptr == 'n' || *_ptr == 'b' ) {
                _ptr--;
            }
            else {
                element = 16;
                symbol[0] = 'S';
                _ptr--;
                break;
            }
            //fall through
        default:
            strncpy( symbol, _ptr, 2 );
            string symb( symbol );
            symbol[0] = toupper( symbol[0] );
            gErrorLog.Throw( __FUNCTION__, "Illegal aromatic element " + symb,
                             ccWarning );
            //But convert anyway
            ++_ptr;
            if ( symb == "si" ) {
                element = 14;
                break;
            }
            else if ( symb == "ge" ) {
                element = 32;
                break;
            }
            else if ( symb == "sb" ) {
                element = 51;
                break;
            }
            else if ( symb == "bi" ) {
                element = 83;
                break;
            }
            else if ( symb == "te" ) {
                element = 52;
                break;
            }
            else if ( symb == "sn" ) {
                element = 50;
                break;
            }
            else {
                return false;
            }
        }
    }

    //handle hydrogen count, stereochemistry, and charge

    CCAtom* atom = mol.NewAtom();
    int hcount = 0;
    int charge = 0;
    int rad = 0;
    char tmpc[2];
    tmpc[1] = '\0';
    for ( _ptr++; *_ptr && *_ptr != ']'; _ptr++ ) {
        switch ( *_ptr ) {
        case '@':
            _ptr++;
            chiralWatch = true;
            _mapcd[atom] = new CCChiralData;
            if ( *_ptr == '@' ) {
                atom->SetFlag( ATOM_IS_RCHIRAL | ATOM_IS_CHIRAL );
            }
            else {
                atom->SetFlag( ATOM_IS_SCHIRAL | ATOM_IS_CHIRAL );
                _ptr--;
            }
            break;
        case '-':
            _ptr++;
            if ( isdigit( *_ptr ) ) {
                tmpc[0] = *_ptr;
                charge = -atoi( tmpc );
            }
            else {
                charge--;
                _ptr--;
            }
            break;
        case '+':
            _ptr++;
            if ( isdigit( *_ptr ) ) {
                tmpc[0] = *_ptr;
                charge = atoi( tmpc );
            }
            else {
                charge++;
                _ptr--;
            }
            break;

        case 'H':
            _ptr++;
            if ( isdigit( *_ptr ) ) {
                tmpc[0] = *_ptr;
                hcount = atoi( tmpc );
            }
            else {
                hcount = 1;
                _ptr--;
            }
            break;
        case '.':
            //CM Feb05
            rad = 2;
            if ( *( ++_ptr ) == '.' ) {
                rad = 3;
            }
            else {
                _ptr--;
            }
            break;

        default:
            return false;
        }
    }

    if ( !*_ptr || *_ptr != ']' ) {
        return false;
    } // we should have a trailing ']' now

    if ( charge ) {
        atom->SetFormalCharge( charge );
    }
    if ( rad ) {
        atom->SetSpin( rad );
    }
    atom->SetAtomicNum( element );
    atom->SetIsotope( isotope );
    atom->SetType( symbol );
    if ( arom ) {
        atom->SetAromatic();
    }

    if ( _prev ) //need to add bond
    {
        mol.AddBond( _prev, mol.GetNumAtom(), _order, _bondflags );
        if ( chiralWatch ) // if chiral atom need to add its previous into atom4ref
        {
            if ( _mapcd[atom] == NULL ) {
                _mapcd[atom] = new CCChiralData;
            }

            ( _mapcd[atom] )->AddAtomRef( ( int ) _prev, input );
            // cout <<"line 1405: Added atom ref "<<_prev<<" to "<<_mapcd[atom]<<endl;
        }
        map<CCAtom*, CCChiralData*>::iterator ChiralSearch;
        ChiralSearch = _mapcd.find( mol.GetAtom( _prev ) );
        if ( ChiralSearch != _mapcd.end() && ChiralSearch->second != NULL ) {
            ( ChiralSearch->second )->AddAtomRef( mol.GetNumAtom(), input );
            // cout <<"line 1431: Added atom ref "<<mol.GetNumAtom()<<" to "<<ChiralSearch->second<<endl;
        }
    }          

    //set values
    _prev = mol.GetNumAtom();
    _order = 1;
    _bondflags = 0;

    //now add hydrogens
    if ( hcount == 0 ) {
        atom->ForceNoH();
    }//ensure AssignMultiplicity regards [C] as C atom

    for ( int i = 0; i < hcount; i++ ) {
        atom = mol.NewAtom();
        atom->SetAtomicNum( 1 );
        atom->SetType( "H" );
        mol.AddBond( _prev, mol.GetNumAtom(), 1 );
        if ( chiralWatch ) {
            if ( _mapcd[mol.GetAtom( _prev )] != NULL ) {
                ( _mapcd[mol.GetAtom( _prev )] )->AddAtomRef( mol.GetNumAtom(),
                                                              input );
            }
            // cout << "line 1434: Added atom ref "<<mol.GetNumAtom()<<" to "<<_mapcd[mol.GetAtom(_prev)]<<endl;
        }
    }
    chiralWatch = false;
    return true;
}

bool CCSmilesParser::CapExternalBonds( CCMol& mol )
{
    if ( _extbond.empty() ) {
        return true;
    }

    CCAtom* atom;
    vector<vector<int> >::iterator bond;

    for ( bond = _extbond.begin(); bond != _extbond.end(); bond++ ) {
        // create new dummy atom
        atom = mol.NewAtom();
        atom->SetAtomicNum( 0 );
        atom->SetType( "*" );

        // bond dummy atom to mol via external bond
        mol.AddBond( ( *bond )[1], atom->GetIdx(), ( *bond )[2], ( *bond )[3] );
        CCBond* refbond = atom->GetBond( mol.GetAtom( ( *bond )[1] ) );

        //record external bond information
        CCExternalBondData* xbd;
        if ( mol.HasData( CCGenericDataType::ExternalBondData ) ) {
            xbd = ( CCExternalBondData * )
                  mol.GetData( CCGenericDataType::ExternalBondData );
        }
        else {
            xbd = new CCExternalBondData;
            xbd->SetOrigin( fileformatInput );
            mol.SetData( xbd );
        }
        xbd->SetData( atom, refbond, ( *bond )[0] );
        //this data gets cleaned up in mol.Clear.
    }

    return true;
}

bool CCSmilesParser::ParseExternalBond( CCMol& mol )
{
    int digit;
    char str[10];

    //*_ptr should == '&'
    _ptr++;

    switch ( *_ptr ) // check for bond order indicators CC&=1.C&1
    {
    case '-':
        _order = 1;
        _ptr++;
        break;
    case '=':
        _order = 2;
        _ptr++;
        break;
    case '#':
        _order = 3;
        _ptr++;
        break;
    case ';':
        _order = 5;
        _ptr++;
        break;
    case '/':
        //chiral, but _order still == 1
        _bondflags |= BOND_IS_TORDOWN;
        _ptr++;
        break;
        _ptr++;
    case '\\':
        // chiral, but _order still == 1
        _bondflags |= BOND_IS_TORUP;
        _ptr++;
        break;
    default:
        // no bond indicator just leave order = 1
        break;
    }

    if ( *_ptr == '%' ) // external bond indicator > 10
    {
        _ptr++;
        str[0] = *_ptr;
        _ptr++;
        str[1] = *_ptr;
        str[2] = '\0';
    }
    else // simple single digit external bond indicator
    {
        str[0] = *_ptr;
        str[1] = '\0';
    }
    digit = atoi( str );  // convert indicator to digit

    //check for dot disconnect closures
    vector<vector<int> >::iterator j;
    int bondFlags, bondOrder;
    for ( j = _extbond.begin(); j != _extbond.end(); j++ ) {
        if ( ( *j )[0] == digit ) {
            bondFlags = ( _bondflags > ( *j )[3] ) ? _bondflags : ( *j )[3];
            bondOrder = ( _order > ( *j )[2] ) ? _order : ( *j )[2];
            mol.AddBond( ( *j )[1], _prev, bondOrder, bondFlags );

            // after adding a bond to atom "_prev"
            // search to see if atom is bonded to a chiral atom
            map<CCAtom*, CCChiralData*>::iterator ChiralSearch;
            ChiralSearch = _mapcd.find( mol.GetAtom( _prev ) );
            if ( ChiralSearch != _mapcd.end() && ChiralSearch->second != NULL ) {
                ( ChiralSearch->second )->AddAtomRef( ( *j )[1], input );
                // cout << "Added external "<<(*j)[1]<<" to "<<ChiralSearch->second<<endl;
            }

            _extbond.erase( j );
            _bondflags = 0;
            _order = 0;
            return true;
        }
    }

    //since no closures save another ext bond
    vector<int> vtmp( 4 );
    vtmp[0] = digit;
    vtmp[1] = _prev;
    vtmp[2] = _order;
    vtmp[3] = _bondflags;

    _extbond.push_back( vtmp );
    _order = 1;
    _bondflags = 0;

    return true;
}

bool CCSmilesParser::ParseRingBond( CCMol& mol )
{
    int digit;
    char str[10];

    if ( *_ptr == '%' ) {
        _ptr++;
        str[0] = *_ptr;
        _ptr++;
        str[1] = *_ptr;
        str[2] = '\0';
    }
    else {
        str[0] = *_ptr;
        str[1] = '\0';
    }
    digit = atoi( str );

    int bf, ord;
    vector<vector<int> >::iterator j;
    for ( j = _rclose.begin(); j != _rclose.end(); j++ ) {
        if ( ( *j )[0] == digit ) {
            bf = ( _bondflags > ( *j )[3] ) ? _bondflags : ( *j )[3];
            ord = ( _order > ( *j )[2] ) ? _order : ( *j )[2];
            mol.AddBond( ( *j )[1], _prev, ord, bf, ( *j )[4] );

            // after adding a bond to atom "_prev"
            // search to see if atom is bonded to a chiral atom
            // need to check both _prev and (*j)[1] as closure is direction independant
            map<CCAtom*, CCChiralData*>::iterator ChiralSearch, cs2;
            ChiralSearch = _mapcd.find( mol.GetAtom( _prev ) );
            cs2 = _mapcd.find( mol.GetAtom( ( *j )[1] ) );
            if ( ChiralSearch != _mapcd.end() && ChiralSearch->second != NULL ) {
                ( ChiralSearch->second )->AddAtomRef( ( *j )[1], input );
                //cout << "Added ring closure "<<(*j)[1]<<" to "<<ChiralSearch->second<<endl;
            }
            if ( cs2 != _mapcd.end() && cs2->second != NULL ) {
                ( cs2->second )->AddAtomRef( _prev, input );
                // cout <<"Added ring opening "<<_prev<<" to "<<cs2->second<<endl;
            }

            //CM ensure neither atoms in ring closure is a radical centre
            CCAtom* patom = mol.GetAtom( _prev );
            patom->SetSpin( 0 );
            patom = mol.GetAtom( ( *j )[1] );
            patom->SetSpin( 0 );
            //CM end
            _rclose.erase( j );
            _bondflags = 0;
            _order = 1;
            return true;
        }
    }

    vector<int> vtmp( 5 );
    vtmp[0] = digit;
    vtmp[1] = _prev;
    vtmp[2] = _order;
    vtmp[3] = _bondflags;
    CCAtom* atom = mol.GetAtom( _prev );
    if ( !atom ) {
        gErrorLog.Throw( __FUNCTION__,
                         "Number not parsed correctly as a ring bond",
                         ccWarning );
        return false;
    }

    vtmp[4] = atom->GetValence(); //store position to insert closure bond
    for ( j = _rclose.begin();
          j != _rclose.end();
          j++ ) //correct for multiple closure bonds to a single atom
    {
        if ( ( *j )[1] == _prev ) {
            vtmp[4]++;
        }
    }

    _rclose.push_back( vtmp );
    _order = 1;
    _bondflags = 0;

    return true;
}

static bool IsCisOrTransH( CCAtom* atom )
{
    if ( !atom->IsHydrogen() ) {
        return false;
    }
    else {
        FOR_BONDS_OF_ATOM( bond, atom ) {
            if ( bond->IsUp() || bond->IsDown() ) {
                return true;
            }
        }
    } 
    return false;
}

void CCMol2Smi::CreateSmiString( CCMol& mol, char* buffer )
{
    CCAtom* atom;
    CCSmiNode* root = NULL;
    buffer[0] = '\0';
    vector<CCAtom*>::iterator i;
	_uatoms.Resize( mol.GetNumAtom() + 1 );
	_ubonds.Resize( mol.GetNumBond() + 1 );

    for ( atom = mol.BeginAtom( i ); atom; atom = mol.NextAtom( i ) )
        // don't use a hydrogen as the root node unless it's not bonded
        // or it's involved in a cis/trans '/' or '\' specification
    {
        if ( ( !atom->IsHydrogen()
            || atom->GetValence() == 0
            || IsCisOrTransH( atom ) )
          && !_uatoms[atom->GetIdx()] ) {
            if ( !atom->IsChiral() || !mol.HasNonZeroCoords() )
                // don't use chiral root atoms unless this is from a SMILES
            {
                //clear out closures in case structure is dot disconnected
                _vclose.clear();
                _atmorder.clear();
                _storder.clear();
                _vopen.clear();
                //dot disconnected structure
                if ( strlen( buffer ) > 0 ) {
                    strcat( buffer, "." );
                }
                root = new CCSmiNode( atom );
                BuildTree( root );
                FindClosureBonds( mol );
                if ( mol.Has2D() ) {
                    AssignCisTrans( root );
                }
                ToSmilesString( root, buffer );
                delete root;
            }
        }
    }

    //If no starting node found e.g. [H][H] CM 21Mar05
    if ( root == NULL && !mol.IsEmpty() ) {
        root = new CCSmiNode( mol.GetAtom( 1 ) ); // get first atom
        BuildTree( root );
        ToSmilesString( root, buffer );
        delete root;
    }
}

bool CCMol2Smi::BuildTree( CCSmiNode* node )
{
    vector<CCBond*>::iterator i;
    CCAtom* nbr, * atom = node->GetAtom();

    _uatoms.Set( atom->GetIdx() ); //mark the atom as visited
    _atmorder.push_back( atom->GetIdx() ); //store the atom ordering
    _storder.push_back( atom->GetIdx() ); //store the atom ordering for stereo

    for ( nbr = atom->BeginNbrAtom( i ); nbr; nbr = atom->NextNbrAtom( i ) ) {
        // Normally skip hydrogens
        // but D,T is explicit and so is H2
        // or an H atom involved in a cis/trans '/' or '\' bond spec
        // or an H atom bonded to a chiral atom (see ToSmilesString below)
        if ( ( !nbr->IsHydrogen()
            || nbr->GetIsotope()
            || atom->IsHydrogen()
            || ( atom->HasChiralitySpecified() )
            || ( ( ( CCBond * ) *i )->IsUp() || ( ( CCBond * ) *i )->IsDown() ) )
          && !_uatoms[nbr->GetIdx()] ) {
            _ubonds.Set( ( *i )->GetIdx() );
            CCSmiNode* next = new CCSmiNode( nbr );
            next->SetParent( atom );
            node->SetNextNode( next, ( CCBond * ) *i );
            BuildTree( next );
        }
    }

    return true;
}

void CCMol2Smi::ToSmilesString( CCSmiNode* node, char* buffer )
{
    char tmpbuf[16];
    CCAtom* atom = node->GetAtom();

    //write the current atom to the string
    GetSmilesElement( node, tmpbuf );
    strcat( buffer, tmpbuf );

    //handle ring closures here
    vector<pair<int, CCBond*> > vc = GetClosureDigits( atom );
    if ( !vc.empty() ) {
        vector<pair<int, CCBond*> >::iterator bpi;
        for ( bpi = vc.begin(); bpi != vc.end(); bpi++ ) {
            if ( bpi->second ) {
                char bs[2];
                bs[0] = GetCisTransBondSymbol( bpi->second, node );
                bs[1] = '\0';
                if ( bs[0] ) {
                    strcat( buffer, bs );   // append "/" or "\"
                }
                else {
#ifndef KEKULE
                    if ( bpi->second->GetOrder() == 2
                      && !bpi->second->IsAromatic() ) {
                        strcat( buffer, "=" );
                    }
#else
                    if ( bpi->second->GetOrder() == 2 ) {
                        strcat( buffer, "=" );
                    }
#endif
                    if ( bpi->second->GetOrder() == 3 ) {
                        strcat( buffer, "#" );
                    }
                }
            }

            if ( bpi->first > 9 ) {
                strcat( buffer, "%" );
            }
            snprintf( tmpbuf, sizeof( tmpbuf ), "%d", bpi->first );
            strcat( buffer, tmpbuf );
        }
    }

    // Follow path to child atoms.
    //
    // Note: Cis/trans bonds are tricky, for example, C/C=C/C is trans,
    // but C(/C)=C/C is cis.  See the comments in FixCisTransBonds(), above.

    // Also make sure we don't output any explicit "H" atoms used for 
    // chiral specifications

    CCBond* bond;
    CCAtom* next;
    for ( int i = 0; i < node->Size(); i++ ) {
        bond = node->GetNextBond( i );
        next = node->GetNextNode( i )->GetAtom();

        // this was a "H" included by BuildTree we should omit
        // it's only included to get the current atom's chirality correct
        // (e.g., for a chiral first atom)
        if ( atom->HasChiralitySpecified() && next->IsHydrogen() ) {
            continue;
        }

        if ( i + 1 < node->Size() ) {
            strcat( buffer, "(" );
        }
        if ( bond->IsUp() || bond->IsDown() ) {
            char cc[2];
            cc[0] = GetCisTransBondSymbol( bond, node );
            cc[1] = '\0';
            strcat( buffer, cc );
        }
#ifndef KEKULE
        if ( bond->GetOrder() == 2 && !bond->IsAromatic() ) {
            strcat( buffer, "=" );
        }
#else
        if ( bond->GetOrder() == 2 ) {
            strcat( buffer, "=" );
        }
#endif
        if ( bond->GetOrder() == 3 ) {
            strcat( buffer, "#" );
        }

        ToSmilesString( node->GetNextNode( i ), buffer );
        if ( i + 1 < node->Size() ) {
            strcat( buffer, ")" );
        }
    }
}

void CCMol2Smi::GetClosureAtoms( CCAtom* atom, vector<CCAtom*>& va )
{
    //look through closure list for start atom
    vector<CCBond*>::iterator i;
    for ( i = _vclose.begin(); i != _vclose.end(); i++ ) {
        if ( *i ) {
            if ( ( ( CCBond * ) *i )->GetBeginAtom() == atom ) {
                va.push_back( ( ( CCBond * ) *i )->GetEndAtom() );
            }
            if ( ( ( CCBond * ) *i )->GetEndAtom() == atom ) {
                va.push_back( ( ( CCBond * ) *i )->GetBeginAtom() );
            }
        }
    }

    CCAtom* nbr;
    vector<pair<CCAtom*, pair<int, int> > >::iterator j;
    for ( j = _vopen.begin(); j != _vopen.end(); j++ ) {
        for ( nbr = atom->BeginNbrAtom( i );
              nbr;
              nbr = atom->NextNbrAtom( i ) ) {
            if ( nbr == j->first ) {
                va.push_back( nbr );
            }
        }
    }
}

vector<pair<int, CCBond*> > CCMol2Smi::GetClosureDigits( CCAtom* atom )
{
    vector<pair<int, CCBond*> > vc;
    vc.clear();

    //look through closure list for start atom
    int idx, bo;
    CCBond* bond;
    vector<CCBond*>::iterator i;
    for ( i = _vclose.begin(); i != _vclose.end(); i++ ) {
        if ( ( bond = ( CCBond * ) *i ) ) {
            if ( bond->GetBeginAtom() == atom || bond->GetEndAtom() == atom ) {
                idx = GetUnusedIndex();
                vc.push_back( pair<int, CCBond*>( idx, bond ) );
                bo = ( bond->IsAromatic() ) ? 1 : bond->GetOrder();
                _vopen.push_back( pair<CCAtom*, pair<int, int> >( bond->GetNbrAtom( atom ),
                                                                  pair<int, int>( idx,
                                                                                  bo ) ) );
                *i = NULL;//remove bond from closure list
            }
        }
    }

    //try to complete closures
    if ( !_vopen.empty() ) {
        vector<pair<CCAtom*, pair<int, int> > >::iterator j;
        for ( j = _vopen.begin(); j != _vopen.end(); ) {
            if ( j->first == atom ) {
                vc.push_back( pair<int, CCBond*>( j->second.first,
                                                  NULL ) );
                _vopen.erase( j );
                j = _vopen.begin();
            }
            else {
                j++;
            }
        }
    }

    return( vc );
}

void CCMol2Smi::FindClosureBonds( CCMol& mol )
{
    //find closure bonds
    CCAtom* a1, * a2;
    CCBond* bond;
    vector<CCBond*>::iterator i;
    CCBitVec bv;
    bv.FromTrueIdx( _storder );

    for ( bond = mol.BeginBond( i ); bond; bond = mol.NextBond( i ) ) {
        if ( !_ubonds[bond->GetIdx()] && bv[bond->GetBeginAtomIdx()] ) {
            a1 = bond->GetBeginAtom();
            a2 = bond->GetEndAtom();
            if ( !a1->IsHydrogen() && !a2->IsHydrogen() ) {
                _vclose.push_back( bond );
            }
        }
    }

    vector<CCBond*>::reverse_iterator j;
    vector<int>::iterator k;

    //modify _order to reflect ring closures
    for ( j = _vclose.rbegin(); j != _vclose.rend(); j++ ) {
        bond = ( CCBond * ) *j;
        a1 = a2 = NULL;

        for ( k = _storder.begin(); k != _storder.end(); k++ ) {
            if ( bond->GetBeginAtomIdx() == *k || bond->GetEndAtomIdx() == *k ) {
                if ( !a1 ) {
                    a1 = mol.GetAtom( *k );
                }
                else if ( !a2 ) {
                    a2 = mol.GetAtom( *k )
                    ;
                    _storder.erase( k );
                    break;
                }
            }
        }

        for ( k = _storder.begin(); k != _storder.end(); k++ ) {
            if ( a1->GetIdx() == *k ) {
                k++;
                if ( k != _storder.end() ) {
                    _storder.insert( k, a2->GetIdx() );
                }
                else {
                    _storder.push_back( a2->GetIdx() );
                }
                break;
            }
        }
    }
}

int CCMol2Smi::GetUnusedIndex()
{
    int idx = 1;

    vector<pair<CCAtom*, pair<int, int> > >::iterator j;
    for ( j = _vopen.begin(); j != _vopen.end(); ) {
        if ( j->second.first == idx ) {
            idx++; //increment idx and start over if digit is already used
            j = _vopen.begin();
        }
        else {
            j++;
        }
    }

    return( idx );
}

void CCMol2Smi::CorrectAromaticAmineCharge( CCMol& mol )
{
    CCAtom* atom;
    vector<CCAtom*>::iterator i;

    _aromNH.clear();
    _aromNH.resize( mol.GetNumAtom() );

    for ( atom = mol.BeginAtom( i ); atom; atom = mol.NextAtom( i ) ) {
        if ( atom->IsNitrogen() && atom->IsAromatic() ) {
            if ( atom->GetHvyVal() == 2 ) {
                if ( atom->GetValence() == 3 || atom->GetImpVal() == 3 ) {
                    _aromNH[atom->GetIdx()] = true;
                }
            }
        }
    }
}

void CCMol2Smi::AssignCisTrans( CCSmiNode* node )
{
    //traverse the tree searching for acyclic olefins - if it
    //has at least one heavy atom attachment on each end assign stereochem

    CCBond* bond;
    for ( int i = 0; i < node->Size(); i++ ) {
        bond = node->GetNextBond( i );
        if ( bond->GetOrder() == 2 && !bond->IsInRing() ) {
            CCAtom* b = node->GetAtom();
            CCAtom* c = bond->GetNbrAtom( b );

            //skip allenes
            if ( b->GetHyb() == 1 || c->GetHyb() == 1 ) {
                continue;
            }

            if ( b->GetHvyVal() > 1 && c->GetHvyVal() > 1 ) {
                CCAtom* a, * d;
                vector<CCBond*>::iterator j, k;

                //look for bond with assigned stereo as in poly-ene
                for ( a = b->BeginNbrAtom( j ); a; a = b->NextNbrAtom( j ) ) {
                    if ( ( ( CCBond * ) *j )->IsUp()
                      || ( ( CCBond * ) *j )->IsDown() ) {
                        break;
                    }
                }

                if ( !a ) {
                    for ( a = b->BeginNbrAtom( j );
                          a;
                          a = b->NextNbrAtom( j ) ) {
                        if ( a != c && !a->IsHydrogen() ) {
                            break;
                        }
                    }
                }
                for ( d = c->BeginNbrAtom( k ); d; d = c->NextNbrAtom( k ) ) {
                    if ( d != b && !d->IsHydrogen() ) {
                        break;
                    }
                }
                //                obAssert(a);
                //                obAssert(d);

                // Calculate the torsion angle between the "substituent" atoms.  This is an
                // odd use of the Torsion() function.  It measures how much you'd
                // have to twist around the double bond to bring both substituents to the
                // same side.  Cis bonds are already on the same side, so they'l have a
                // torsion angle of zero.  Trans bonds are opposite, so you'd have to twist
                // around the double bond by 180 degrees.  So small (near zero) means cis,
                // and large (near 180) means trans.  This is cool because it also works in
                // any 3D orientation.
                double angle = fabs( Torsion( a->GetXYZ(), b->GetXYZ(),
                                              c->GetXYZ(), d->GetXYZ() ) );

                if ( ( *j )->IsUp() || ( *j )->IsDown() ) //stereo already assigned
                {
                    if ( angle > 10.0 ) {
                        // 180 degrees == trans configuration
                        if ( ( *j )->IsUp() ) {
                            ( *k )->SetFlag( BOND_IS_TORDOWN );
                        }   // set bonds "opposite sides" (up/down)
                        else {
                            ( *k )->SetFlag( BOND_IS_TORUP );
                        }       // set bonds "same side" (both up)
                    }
                    else {
                        // small angle == cis configuration
                        if ( ( *j )->IsUp() ) {
                            ( *k )->SetFlag( BOND_IS_TORUP );
                        }
                        else {
                            ( *k )->SetFlag( BOND_IS_TORDOWN );
                        }
                    }
                }
                else //assign stereo to both ends
                {
                    ( *j )->SetFlag( BOND_IS_TORUP );
                    // See comments above re: angle between substituents
                    if ( angle > 10.0 ) {
                        ( *k )->SetFlag( BOND_IS_TORDOWN ); // trans configuration, set bonds "opposite sides" (up/down)
                    }
                    else {
                        ( *k )->SetFlag( BOND_IS_TORUP );   // cis configuration, set bonds "same side" (both up)
                    }
                }
            }
        }
        AssignCisTrans( node->GetNextNode( i ) );
    }
}

char CCMol2Smi::GetCisTransBondSymbol( CCBond* bond, CCSmiNode* node )
{
    // This is the converse of CorrectUpDownMarks() above, for writing
    // the SMILES back out.  Given a cis/trans bond and the node in the
    // SMILES tree, figures out whether to write a '/' or '\' symbol.
    // See the comments in FixCisTransBonds(), above.
    // 
    // The CCSmiNode is the most-recently-written atom in the SMILES string
    // we're creating.  If it is the double-bonded atom, then the substituent
    // follows, so that "up" means '/' and "down" means '\'.  If the CCSmiNode
    // atom is the single-bonded atom then the double-bonded atom comes next,
    // in which case "up" means '\' and "down" means '/'.
    //
    // Note that there's an ambiguity: What if both ends of the bond are
    // double-bonded atoms?  In this case, the first one takes precedence.

    if ( !bond || ( !bond->IsUp() && !bond->IsDown() ) ) {
        return '\0';
    }

    CCAtom* atom = node->GetAtom();
    if ( atom->HasBondWithOrder( 2 ) ) {
        // double-bonded atom is first in the SMILES?
        if ( bond->IsUp() ) {
            return '/';
        }
        else {
            return '\\';
        }
    }
    else {
        // double-bonded atom is second in the SMILES
        if ( bond->IsUp() ) {
            return '\\';
        }
        else {
            return '/';
        }
    }
}


void CCMol2Smi::Init()
{
    _vclose.clear();
    _atmorder.clear();
    _storder.clear();
    _aromNH.clear();
    _uatoms.Clear();
    _ubonds.Clear();
    _vopen.clear();
}

bool CCMol2Smi::GetSmilesElement( CCSmiNode* node, char* element )
{
    //***handle reference atom stuff here and return***
    char symbol[16];
    bool bracketElement = false;
    bool normalValence = true;

    CCAtom* atom = node->GetAtom();

    //int bosum = atom->KBOSum();
    //atom->SumBO(); //CM temp
    int bosum = atom->SumBO();
    switch ( atom->GetAtomicNum() ) {
    case 0:
        break;
    case 5:
        /*bracketElement = !(normalValence = (bosum == 3)); break;*/
        break;
    case 6:
        break;
    case 7:
        if ( atom->IsAromatic()
          && atom->GetHvyVal() == 2
          && atom->GetImpVal() == 3 ) {
            bracketElement = !( normalValence = false );
            break;
        }
        else {
            bracketElement = !( normalValence = ( bosum == 3 || bosum == 5 ) );
        }
        break;
    case 8:
        break;
    case 9:
        break;
    case 15:
        break;
    case 16:
        bracketElement = !( normalValence = ( bosum == 2
                                           || bosum == 4
                                           || bosum == 6 ) );
        break;
    case 17:
        break;
    case 35:
        break;
    case 53:
        break;

    default:
        bracketElement = true;
    }

    if ( atom->GetHvyVal() > 2 && atom->IsChiral() ) {
        if ( atom->GetMol()->HasNonZeroCoords()
          || atom->HasChiralitySpecified() ) {
            bracketElement = true;
        }
    }

    if ( atom->GetFormalCharge() != 0 ) //bracket charged elements
    {
        bracketElement = true;
    }

    if ( atom->GetIsotope() ) //CM 19Mar05
    {
        bracketElement = true;
    }

    //CM begin 18 Sept 2003

    //This outputs form [CH3][CH3] rather than CC if -h option has been specified
    if ( atom->GetMol()->HasFlag( MOL_HYDROGEN_ADDED ) ) {
        bracketElement = true;
        normalValence = false;
    }
    else {
        if ( atom->GetSpin() ) {
            //For radicals output bracket form anyway
            bracketElement = true;
            normalValence = false;
        }
    }
    //CM end

    if ( !bracketElement ) {
        if ( !atom->GetAtomicNum() ) {
            bool external = false;
            vector<pair<int, pair<CCAtom *, CCBond*> > >* externalBonds = ( vector<pair<int, pair<CCAtom *, CCBond*> > >* )
                                                                          ( ( CCMol* )
                                                                            atom->GetMol() )->GetData( CCGenericDataType::ExternalBondData );
            vector<pair<int, pair<CCAtom *, CCBond*> > >::iterator externalBond;

            if ( externalBonds ) {
                for ( externalBond = externalBonds->begin();
                      externalBond != externalBonds->end();
                      externalBond++ ) {
                    if ( externalBond->second.first == atom ) {
                        external = true;
                        strcpy( symbol, "&" );
                        CCBond* bond = externalBond->second.second;
                        if ( bond->IsUp() ) {
                            if ( ( bond->GetBeginAtom() )->HasBondWithOrder( 2 )
                              || ( bond->GetEndAtom() )->HasBondWithOrder( 2 ) ) {
                                strcat( symbol, "/" );
                            }
                        }
                        if ( bond->IsDown() ) {
                            if ( ( bond->GetBeginAtom() )->HasBondWithOrder( 2 )
                              || ( bond->GetEndAtom() )->HasBondWithOrder( 2 ) ) {
                                strcat( symbol, "\\" );
                            }
                        }
#ifndef KEKULE
                        if ( bond->GetOrder() == 2 && !bond->IsAromatic() ) {
                            strcat( symbol, "=" );
                        }
                        if ( bond->GetOrder() == 2 && bond->IsAromatic() ) {
                            strcat( symbol, ":" );
                        }
#else
                        if ( bond->GetOrder() == 2 ) {
                            strcat( symbol, "=" );
                        }
#endif

                        if ( bond->GetOrder() == 3 ) {
                            strcat( symbol, "#" );
                        }
                        sprintf( symbol, "%s%d", symbol, externalBond->first );
                        break;
                    }
                }
            }

            if ( !external ) {
                strcpy( symbol, "*" );
            }
        }
        else {
            //strcpy(symbol,etab.GetSymbol(atom->GetAtomicNum()));
            strcpy( symbol, atom->GetSymbol().c_str() );
#ifndef KEKULE
            if ( atom->IsAromatic() ) {
                symbol[0] = tolower( symbol[0] );
            }
#endif

            //Radical centres lc if r option set
            if ( atom->GetSpin() ) {
                symbol[0] = tolower( symbol[0] );
            }
        }
        strcpy( element, symbol );

        return true;
    }

    strcpy( element, "[" );
    if ( atom->GetIsotope() ) //CM 19Mar05
    {
        char iso[4];
        sprintf( iso, "%d", atom->GetIsotope() );
        strcat( element, iso );
    }
    if ( !atom->GetAtomicNum() ) {
        strcpy( symbol, "*" );
    }
    else {
        //strcpy(symbol,etab.GetSymbol(atom->GetAtomicNum()));
        strcpy( symbol, atom->GetSymbol().c_str() );
#ifndef KEKULE
        if ( atom->IsAromatic() ) {
            symbol[0] = tolower( symbol[0] );
        }
#endif
    }
    strcat( element, symbol );

    //if (atom->IsCarbon() && atom->GetHvyVal() > 2 && atom->IsChiral())
    if ( atom->GetHvyVal() > 2 && atom->IsChiral() ) {
        char stereo[5];
        if ( GetChiralStereo( node, stereo ) ) {
            strcat( element, stereo );
        }
    }

    //add extra hydrogens
    //  if (!normalValence && atom->ImpHydrogenCount())
    //Reduce implicit H count by number of D,T (and explicit 1H) because they will be
    //output explicitly
    //          int nHisotopes = atom->ExpHydrogenCount() - atom->ExpHydrogenCount(true);
    //    int nH = atom->ImpHydrogenCount() - nHisotopes;
    int nH = atom->ImpHydrogenCount() + atom->ExpHydrogenCount( true );//excludes H isotopes
    if ( nH && !atom->IsHydrogen() ) {
        strcat( element, "H" );
        if ( nH > 1 ) {
            char tcount[10];
            sprintf( tcount, "%d", nH );
            strcat( element, tcount );
        }
    }

    //cat charge on the end
    if ( atom->GetFormalCharge() != 0 ) {
        if ( atom->GetFormalCharge() > 0 ) {
            strcat( element, "+" );
        }
        else {
            strcat( element, "-" );
        }

        if ( abs( atom->GetFormalCharge() ) > 1 ) {
            char tcharge[10];
            sprintf( tcharge, "%d", abs( atom->GetFormalCharge() ) );
            strcat( element, tcharge );
        }
    }

    strcat( element, "]" );

    return true;
}

bool CCMol2Smi::GetChiralStereo( CCSmiNode* node, char* stereo )
{
    bool is2D = false;
    double torsion;
    CCAtom* a, * b, * c, * d, * atom, hydrogen;

    b = node->GetAtom();
    CCMol* mol = b->GetMol();

    if ( !mol->HasNonZeroCoords() ) //must have come in from smiles string, but not neccessarily the same string! must check order.
        //This section re-written to use the ChiralData object but the non 0D co-ord routines have been left alone. Might use 
        //different names for things. In this section a,b,c,d are the neighbours of the chrial atom, atom.
    {
        if ( !b->HasChiralitySpecified() ) {
            return false;
        }

        atom = node->GetAtom();  
        b = c = d = NULL;
        a = node->GetParent(); // a is parent, atom is chiral atom b/c/d are ones off it.
        CCChiralData* cd = ( CCChiralData* )
                           atom->GetData( CCGenericDataType::ChiralData );

        if ( !cd ) {
            //if no Chiral Data Set, need to make one!
            cd = new CCChiralData;
            atom->SetData( cd );
        }

        //get connected atoms in order
        //   CCAtom *nbr;
        vector<int>::iterator j;
        vector<int> vnbor, vsmiles;
        FOR_NBORS_OF_ATOM( nbr, atom ) {
            vnbor.push_back( nbr->GetIdx() );
        }

        for ( j = _storder.begin(); j != _storder.end(); j++ ) {
            for ( int x = 0; x < vnbor.size(); x++ ) {
                // this cast is not ideal but what else?
                if ( static_cast<unsigned>( *j ) == vnbor[x] ) {
                    vsmiles.push_back( vnbor[x] );
                }
            }
        }
        if ( vsmiles.size() == 3 ) {
            vsmiles.insert( ++( vsmiles.begin() ), atom->GetIdx() );
        }

        cd->SetAtom4Refs( vsmiles, output );   // This saves the output atom4refs calculated above
        CorrectChirality( atom );


        if ( atom->IsRChiral() ) {
            strcpy( stereo, "@@" );
        }
        else if ( atom->IsSChiral() ) {
            strcpy( stereo, "@" );
        }
        else {
            return false;
        }

        return true;
    }

    //give peudo Z coords if mol is 2D
    if ( !mol->Has3D() ) {
        CCVector v, vz( 0.0, 0.0, 1.0 );
        is2D = true;
        CCAtom* nbr;
        CCBond* bond;
        vector<CCBond*>::iterator i;
        for ( bond = b->BeginNbrBond( i ); bond; bond = b->NextNbrBond( i ) ) {
            nbr = bond->GetEndAtom();
            if ( nbr != b ) {
                v = nbr->GetXYZ();
                if ( bond->IsWedge() ) {
                    v += vz;
                }
                else if ( bond->IsHash() ) {
                    v -= vz;
                }

                nbr->SetXYZ( v );
            }
            else {
                nbr = bond->GetBeginAtom();
                v = nbr->GetXYZ();
                if ( bond->IsWedge() ) {
                    v -= vz;
                }
                else if ( bond->IsHash() ) {
                    v += vz;
                }

                nbr->SetXYZ( v );
            }
        }
    }

    c = d = NULL;
    a = node->GetParent();
    //    obAssert(a); //chiral atom can't be used as root node - must have parent

    if ( b->GetHvyVal() == 3 ) //must have attached hydrogen
    {
        if ( b->GetValence() == 4 )//has explicit hydrogen
        {
            vector<CCBond*>::iterator i;
            for ( c = b->BeginNbrAtom( i ); c; c = b->NextNbrAtom( i ) ) {
                if ( c->IsHydrogen() ) {
                    break;
                }
            }
            //            obAssert(c);
        }
        else  //implicit hydrogen
        {
            CCVector v;
            b->GetNewBondVector( v, 1.0 );
            hydrogen.SetXYZ( v );
            c = &hydrogen;
        }
    }

    //get connected atoms in order
    CCAtom* nbr;
    vector<int>::iterator j;

    //try to get neighbors that are closure atoms in the order they appear in the string
    vector<CCAtom*> va;
    GetClosureAtoms( b, va );
    if ( !va.empty() ) {
        CCAtomIter k;
        for ( k = va.begin(); k != va.end(); k++ ) {
            if ( *k != a ) {
                if ( !c ) {
                    c = ( CCAtom * ) *k;
                }
                else if ( !d ) {
                    d = ( CCAtom * ) *k;
                }
            }
        }
    }

    for ( j = _storder.begin(); j != _storder.end(); j++ ) {
        nbr = mol->GetAtom( *j );
        if ( !b->IsConnected( nbr ) ) {
            continue;
        }
        if ( nbr == a || nbr == b || nbr == c ) {
            continue;
        }
        if ( !c ) {
            c = nbr;
        }
        else if ( !d ) {
            d = nbr;
        }
    }

    // ************* our torsion is opposite to openbabel's *************
    torsion = -Torsion( a->GetXYZ(), b->GetXYZ(), c->GetXYZ(), d->GetXYZ() );
    // if torsion == 0 or torsion == 180, there's no stereo
    // otherwise, copy the stereo flag

    if ( !IsNegligible( torsion, 1.0, 1.0e-3 ) && !IsApprox( torsion, 180.0,
                                                             1.0e-3 ) ) {
        strcpy( stereo, ( torsion < 0.0 ) ? "@" : "@@" );
    }
    else {
        stereo[0] = '\0';
    }

    //re-zero psuedo-coords
    if ( is2D ) {
        CCVector v;
        CCAtom* atom;
        CCAtomIter k;
        for ( atom = mol->BeginAtom( k ); atom; atom = mol->NextAtom( k ) ) {
            v = atom->GetXYZ();
            v.SetZ( 0.0 );
            atom->SetXYZ( v );
        }
    }

    return true;
}

CCSmiNode::CCSmiNode( CCAtom* atom )
{
    _atom = atom;
    _parent = NULL;
    _nextnode.clear();
    _nextbond.clear();
}

void CCSmiNode::SetNextNode( CCSmiNode* node, CCBond* bond )
{
    _nextnode.push_back( node );
    _nextbond.push_back( bond );
}

CCSmiNode::~CCSmiNode()
{
    vector<CCSmiNode*>::iterator i;
    for ( i = _nextnode.begin(); i != _nextnode.end(); i++ ) {
        delete ( *i );
    }
}


bool WriteTheSmiles( CCMol& mol, char* out )
{
    char buffer[2 * BUFF_SIZE];

    CCMol2Smi m2s;

    m2s.Init();
    m2s.CorrectAromaticAmineCharge( mol );
    m2s.CreateSmiString( mol, buffer );

    strcpy( out, buffer );
    return true;
}

