/**********************************************************************
XYZ cartesian coordinates
***********************************************************************/
#include <formats/CCFormat.h>
#include <ccbg/CCMol.h>

class XYZFormat : public CCFormat
{
public:
    //Register this format type ID
    XYZFormat()
    {
        RegisterFormat( "xyz", this );
    }

    virtual const char* Description()
    {
        return
        "XYZ cartesian coordinates format\n";
    };

	bool ReadMolecule( CCMol& mol, istream& is );
	bool ReadAtomsOnly( CCMol& mol, istream& is );
	bool WriteMolecule( CCMol& mol, ostream& os );
};
//Make an instance of the format class
XYZFormat theXYZFormat;

/////////////////////////////////////////////////////////////////
bool XYZFormat::ReadMolecule( CCMol& mol, istream& is )
{
    char buffer[BUFF_SIZE];
    stringstream errMsg;

    if ( !is.getline( buffer, BUFF_SIZE ) ) {
        gErrorLog.Throw( __FUNCTION__,
                         "Problems reading an XYZ file: Cannot read the first line." );
        return false;
    }

    int natoms;
    if ( sscanf( buffer, "%d", &natoms ) == 0 || !natoms ) {
        gErrorLog.Throw( __FUNCTION__,
                         "Problems reading an XYZ file: The first line must contain the number of atoms." );
        return false;
    }

    mol.ResizeAtoms( natoms );

    // The next line contains a title string for the molecule. Use this
    // as the title for the molecule if the line is not
    // empty. Otherwise, use the title given by the calling function.
    if ( !is.getline( buffer, BUFF_SIZE ) ) {
        gErrorLog.Throw( __FUNCTION__,
                         "Problems reading an XYZ file: Cannot read molecular title line." );
        return false;
    }
    Trim( buffer );
    mol.SetName( buffer );

    // The next lines contain four items each, separated by white
    // spaces: the atom type, and the coordinates of the atom
    vector<string> vs;
    for ( int i = 1; i <= natoms; i ++ ) {
        if ( !is.getline( buffer, BUFF_SIZE ) ) {
            gErrorLog.Throw( __FUNCTION__,
                             "Problems reading an XYZ file: Cannot read coordinate line." );
            return false;
        }

        Tokenize( vs, buffer );
        if ( vs.size() != 4 ) {
            errMsg
         << "LINE: "
         << buffer
         << "\n"
         << "should contain exactly 4 entries, separated by white space.\n"
         << "However, there are "
         << vs.size()
         << " items."
         << endl;

            gErrorLog.Throw( __FUNCTION__, errMsg.str() );
            return false;
        }

        // read elemental type
        CCAtom* atom = mol.NewAtom();
        int atomicNum = gElementalTable.GetAtomicNum( vs[0].c_str() );
        atom->SetAtomicNum( atomicNum );

        // Read the atom coordinates
        double x( 0.0 ), y( 0.0 ), z( 0.0 );
        x = atof( vs[1].c_str() );
        y = atof( vs[2].c_str() );
        z = atof( vs[3].c_str() );
        atom->SetXYZ( x, y, z ); //set coordinates
    }

    // clean out any remaining blank lines
    while ( is.peek() != EOF
         && is.good()
         && ( is.peek() == '\n' || is.peek() == '\r' ) ) {
        is.getline( buffer, BUFF_SIZE );
    }

    return true;
}

bool XYZFormat::ReadAtomsOnly( CCMol& mol, istream& is )
{
	return ReadMolecule( mol, is );
}

bool XYZFormat::WriteMolecule( CCMol& mol, ostream& os )
{
	if ( mol.IsEmpty() ) {
		gErrorLog.Throw( __FUNCTION__, "Molecule is empty." );
		return false;
	}

    char buffer[BUFF_SIZE];

    snprintf( buffer, BUFF_SIZE, "%d\n", mol.GetNumAtom() );
    os << buffer;
	string molName = mol.GetName();
	// set max length of molecular name as 50 characters
    os << molName.substr( 0, 50 ) << "\n";

    FOR_ATOMS_OF_MOL( atom, mol ) {
        snprintf( buffer, BUFF_SIZE, "%-3s%15.5f%15.5f%15.5f\n", "K",//atom->GetSymbol().c_str(),
                  atom->GetX(), atom->GetY(), atom->GetZ() );
        os << buffer;
    }

    return true;
}

