/** @file 
* @brief set aromatic flag on atoms and bonds without changing bond order. 
* @author Yuan ZHAO, Tiejun CHENG
* @date 2008/08/22
* @version 1.0.1 
*
*	This file is part of the CCBGKit project.
*/ 

#ifndef CCAROMATIC_H
#define CCAROMATIC_H

#include <CCConfig.h>
#include <ccbg/CCBitVec.h>

class CCAtom;
class CCBond;
class CCMol;

//////////////////////////////////////////////////////////////////////////
//                       CCAromaticTyper Class                          //
//////////////////////////////////////////////////////////////////////////
/** 
* @class CCAromaticTyper CCAromatic.h "include/ccbg/CCAromatic.h"
* @brief CCAromaticTyper class. 
* 
* set aromatic flag but don't change bond order.
*/ 
class CCAPI CCAromaticTyper
{
public:
	/** @brief clear all members to default */
    void Clear();
	/** @brief constructor */
    CCAromaticTyper();
	/** @brief destructor */
    virtual ~CCAromaticTyper()
    {
    }

	/** @brief assign aromaticity on molecule */
    void AssignAromaticity( CCMol& mol );
	/** @brief exclude atoms that can't be aromatic */
    bool ExcludeNonAromaticAtom( CCMol& mol );
	/** @brief propagate potential aromatic atoms in groups */
    bool PropagateAromaticAtom( CCMol& mol );
	/** @brief select a root atom in each ring */
    bool SelectRootAtom( CCMol& mol, bool avoidInnerAtom = true );
	/** @brief check the aromaticity starting from atom */
    void CheckAromaticity( CCAtom* atom, int depth );
	/** @brief depth-first-search traverse */
    bool TraverseCycle( CCAtom* root, CCAtom* atom, CCBond* prev, int depth );
	/** @brief center atom has an external double bond ? */
    bool CheckConsecutiveSingleBond( CCBond* prev, CCAtom* center, CCBond* curr );
	/** @brief perceive odd ring aromaticity ? */
    void OddRingAromaticity( bool odd )
    {
        m_odd = odd;
    }

protected:
    CCBitVec m_exAtoms;      ///< excluded atoms index
    CCBitVec m_rootAtoms;    ///< root atoms index
    CCBitVec m_visit;        ///< visited atom
    CCBitVec m_marked;       ///< perceived atom
    pair<int, int> m_piElec; ///< current min and max pi electrons
    int m_maxDepth;          ///< max depth for search ring
    bool m_odd;              ///< perceive odd rings aromaticity ?
    CCBond* m_bgnBond;       ///< starting bond for a potential aromatic ring
};
EXTERN CCAromaticTyper gAromaticTyper;

#endif
