/** @file 
* @brief CCAtom class and relevant functions.
* @author Yuan ZHAO, Tiejun CHENG
* @date 2008/08/23
* @version 1.0.1 
* 
* CCAtom class was a data container. It contains many of the
* atom-specific routines that facilitates extraction of data
* perceived from both the atom and the molecule.
* 
* The following code demonstrates how to print out the atom numbers,
* element numbers, and coordinates of a molecule:
* @code
* CCMol mol;
* 
* FOR_ATOMS_IN_MOL(atom, mol) {
* 	cout << atom->GetIdx() << " ";
* 	cout << atom->GetAtomicNum() << " ";
* 	cout << atom->GetX() << " " 
* 	     << atom->GetY() << " "
* 	     << atom->GetZ() << " " << endl;
* }
* @endcode
*
* Copyright (C) 2007-2008 www.sioc-ccbg.ac.cn
* 
* This file is part of the CCBGKit project.
*/

#ifndef CCATOM_H
#define CCATOM_H

#include <ccbg/CCBase.h>
#include <ccbg/CCVector.h>


class CCAtom;
class CCBond;
class CCMol;
class CCRing;
class CCResidue;

/** @brief abbreviation for std::vector<CCAtom*>::iterator */
typedef std::vector<CCAtom*>::iterator CCAtomIter;

//////////////////////////////////////////////////////////////////////////
//                           CCAtom class                               //
//////////////////////////////////////////////////////////////////////////
/** @brief A enum of atom flags, with inline docs */ 
enum
{
    ATOM_IS_VALID           = 1 << 0,  ///< atom is valid (used in interpret pdb)
    ATOM_IS_IN_RING         = 1 << 1,  ///< atom is in ring
    ATOM_IS_AROMATIC        = 1 << 2,  ///< atom is aromatic
    ATOM_IS_CHIRAL          = 1 << 3,  ///< atom is chiral
    ATOM_IS_RCHIRAL         = 1 << 4,  ///< atom is R-chiral
    ATOM_IS_SCHIRAL         = 1 << 5,  ///< atom is S-chiral
	ATOM_IS_STEREO          = 1 << 6,  ///< atom is stereo
	ATOM_IS_ESTEREO         = 1 << 7,  ///< atom is E-stereo
	ATOM_IS_ZSTEREO         = 1 << 8,  ///< atom is Z-stereo
    ATOM_HAS_NO_H           = 1 << 9,  ///< atom has no hydrogen neighbor
    ATOM_IS_IN_FGROUP       = 1 << 10, ///< atom is in functional groups
    ATOM_IS_HBDONOR         = 1 << 11, ///< atom is hydrogen bond dono
    ATOM_IS_HBACCEPTOR      = 1 << 12, ///< atom is hydrogen bond acceptor
    ATOM_IS_IN_SSBOND       = 1 << 13, ///< atom is in S-S bond
    ATOM_IS_IN_NATAA        = 1 << 14, ///< atom is in protein 
    ATOM_IS_CONSTRAIN       = 1 << 15, ///< atom is constrained when doing minimization
	ATOM_IS_PROBLEMATIC     = 1 << 16, ///< report error message only once
};

/** 
* @class CCAtom CCAtom.h "include/ccbg/CCAtom.h"
* @brief CCAtom class. 
* 
* a data container of atoms properties and relevant functions.
*/ 
class CCAPI CCAtom : public CCBase
{
protected:
    CCMol* m_mol;      ///< parent molecule
    int m_idx;         ///< atomic index
    int m_ele;         ///< atomic number
    string m_name;     ///< atomic name
    string m_type;     ///< internal atomic type
    int m_impval;      ///< implicit valence
    CCVector m_coor;   ///< atomic coordinates
    int m_hyb;         ///< hybridization
    int m_part;        ///< component indicator
    int m_fcharge;     ///< formal charge
    double m_pcharge;  ///< partial charge
    vector<CCBond*> m_vbond; ///< neighboring bonds(coordinate with m_vatom);
    CCResidue* m_residue;    ///< parent residue
    //////////////////////////////////////////////////////////////////////////
    int m_spin;        ///< atomic spin, e.g., 2 for radical  1 or 3 for carbene
    int m_isotope;     ///< isotope
    //////////////////////////////////////////////////////////////////////////
	string m_xtype;    ///< XLOGP3 type

public:
	/** @brief CCAtom constructor */
    CCAtom();
	/** @brief CCAtom copy constructor */
    CCAtom( const CCAtom& src );
	/** @brief CCAtom copy assignment operator */
    CCAtom& operator =( const CCAtom& src );
	/** @brief CCAtom destructor */
	virtual ~CCAtom();
	/** @brief clear members of CCAtom */
	void Clear();

	/** @brief get atomic parent molecule */
    CCMol* GetMol() const
    {
        return m_mol;
    }
	/** @brief get atomic index */
    int GetIdx() const
    {
        return m_idx;
    }
	/** @brief get element numbers */
    int GetAtomicNum() const
    {
        return m_ele;
    }
	/** @brief get atomic name */
    string GetName() const
    {
        return m_name;
    }
	/** @brief get element symbol */
    string GetSymbol() const;
	/** @brief get atomic weight */
    double GetMass() const;
	/** @brief get covalent radius */
    double GetCovRad() const;
	/** @brief get van de waals radius */
    double GetVdwRad() const;
	/** @brief get the number of max bonded atoms */
    int GetMaxBond() const;
	/** @brief get the electro negativity */
	double GetEleNeg();

	/** @brief get internal atom type */
    string GetType() const;
	/** @brief get implicit valence (implicit bonded atoms) */
    int GetImpVal() const;
	/** @brief get explicit valence (explicit bonded atoms) */
    int GetValence() const
    {
        return m_vbond.size();
    }
	/** @brief get atomic coordinates */
    const CCVector& GetXYZ() const
    {
        return m_coor;
    }
	/** @brief get x coordinate */
    double GetX() const
    {
        return m_coor.GetX();
    }
	/** @brief get y coordinates */
    double GetY() const
    {
        return m_coor.GetY();
    }
	/** @brief get z coordinates */
    double GetZ() const
    {
        return m_coor.GetZ();
    }
	/** @brief get hybridization state */
    int GetHyb() const;
	/** @brief get the number of fragment part which the atom belonged to */
    int GetPart() const;
	/** @brief get formal charge */
    int GetFormalCharge();
	/** @brief get partial charge */
    double GetPartialCharge() const
    {
        return m_pcharge;
    }
	/** @brief get atomic parent residue */
    CCResidue* GetResidue() const
    {
        return m_residue;
    }

	/** @brief get atomic parent residue name */
	string GetResName() const;

	/** @brief set atomic parent molecule */
    void SetMol( CCMol* mol )
    {
        m_mol = mol;
    }
	/** @brief set atomic index */
    void SetIdx( int idx )
    {
        m_idx = idx;
    }
	/** @brief set internal atom type */
    void SetType( string type )
    {
        m_type = type;
    }
	/** @brief set internal atom type as atomic symbol */
    void SetTypeFromSymbol();
	/** @brief set internal atom type as "atomic symbol.1" */
    void SetSp1Type();
	/** @brief set internal atom type as "atomic symbol.2" */
    void SetSp2Type();
	/** @brief set internal atom type as "atomic symbol.3" */
    void SetSp3Type();
	/** @brief set element numbers */
    void SetAtomicNum( int ele )
    {
        m_ele = ele;
    }
	/** @brief set atomic name */
    void SetName( string name )
    {
        m_name = name;
    }
	/** @brief set implicit valence */
    void SetImpVal( int impval )
    {
        m_impval = impval;
    }
	/** @brief increse implicit valence by 1 */
    void IncreseImpVal()
    {
        m_impval++;
    }
	/** @brief decrese implicit valence by 1 */
    void DecreseImpVal()
    {
        m_impval > 0 ? m_impval-- : 0;
    }
	/** @brief set atomic coordinates from a CCVector */
    void SetXYZ( const CCVector& coor )
    {
        m_coor = coor;
    }
	/** @brief set atomic coordinates from three double value */
    void SetXYZ( double x, double y, double z )
    {
        m_coor.SetXYZ( x, y, z );
    }
	/** @brief set atomic coordinates from double array */
    void SetXYZ( double* c )
    {
        m_coor.SetXYZ( c );
    }
	/** @brief set hybridization state */
    void SetHyb( int hyb )
    {
        m_hyb = hyb;
    }
	/** @brief set the number of fragment part which the atom belonged to */
    void SetPart( int part )
    {
        m_part = part;
    }
	/** @brief set formal charge */
    void SetFormalCharge( short fcharge )
    {
        m_fcharge = fcharge;
    }
	/** @brief set partial charge */
    void SetPartialCharge( double pcharge )
    {
        m_pcharge = pcharge;
    }
	/** @brief set atomic parent residue */
    void SetResidue( CCResidue* residue )
    {
        m_residue = residue;
    }
	/** @brief check whether the atom had ATOM_IS_VALID flag */
    bool IsValid() const
    {
        return HasFlag( ATOM_IS_VALID );
    }
	/** @brief assign ATOM_IS_VALID flag to the atom */
    void SetValid()
    {
        SetFlag( ATOM_IS_VALID );
    }
	/** @brief check whether the atom had ATOM_IS_IN_FGROUP flag */
    bool IsInFGroup()
    {
        return HasFlag( ATOM_IS_IN_FGROUP );
    }
	/** @brief assign ATOM_IS_IN_FGROUP flag to the atom */
    void SetFGroup()
    {
        SetFlag( ATOM_IS_IN_FGROUP );
    }
	/** @brief assign ATOM_IS_VALID and ATOM_IS_IN_FGROUP flag to the atom */
    void SetValidFGroup()
    {
        SetFlag( ATOM_IS_IN_FGROUP ); 
        SetFlag( ATOM_IS_VALID );
    }

	/** @brief get atomic spin state */
    int GetSpin() const
    {
        return m_spin;
    }
	/** @brief set atomic spin state */
    void SetSpin( int spin )
    {
        m_spin = spin;
    }
	/** @brief get atomic isotope state */
    int GetIsotope() const
    {
        return m_isotope;
    }
	/** @brief set atomic isotope state */
    void SetIsotope( int iso );

	/** @brief add a connective bond to the atom */
    bool AddBond( CCBond* bond );
	/** @brief remove a bond from atom's neighboring bonds list */
    bool RemoveBond( CCBond* bond );
	/** @brief clear atom's neighboring bonds list */
    void ClearBond()
    {
        m_vbond.clear();
    }
	/** @brief get the bond connects this and na */
	CCBond* GetBond(CCAtom* na) const;

	/** @brief get first neighboring atom */
    CCAtom* BeginNbrAtom( vector<CCBond*>::iterator& i );
	/** @brief get next neighboring atom */
    CCAtom* NextNbrAtom( vector<CCBond*>::iterator& i );
	/** @brief get first neighboring bond */
    CCBond* BeginNbrBond( vector<CCBond*>::iterator& i );
	/** @brief get next neighboring bond */
    CCBond* NextNbrBond( vector<CCBond*>::iterator& i );

	/** @brief get nth neighboring atom */
    CCAtom* GetNbrAtom( int n );
	/** @brief get nth neighboring heavy atom */
    CCAtom* GetNbrHvyAtom( int n );
	/** @brief get nth neighboring bond */
    CCBond* GetNbrBond( int n );

	/** @brief get the smallest bond angle of center atom */
    double SmallestBondAngle();
	/** @brief get the average bond angle of center atom */
    double AverageBondAngle();

	/** @brief assign ATOM_IS_AROMATIC flag to the atom */
    void SetAromatic()
    {
        SetFlag( ATOM_IS_AROMATIC );
    }
	/** @brief clear ATOM_IS_AROMATIC flag in the atom */
    void UnsetAromatic()
    {
        UnsetFlag( ATOM_IS_AROMATIC );
    }
	/** @brief assign ATOM_IS_IN_RING flag to the atom */
    void SetInRing()
    {
        SetFlag( ATOM_IS_IN_RING );
    }
	/** @brief assign ATOM_IS_CHIRAL flag to the atom */
    void SetChiral()
    {
        SetFlag( ATOM_IS_CHIRAL );
    }
	/** @brief clear chiral relevant flag to the atom */
    void UnsetChiral()
    {
        UnsetFlag( ATOM_IS_SCHIRAL );
        UnsetFlag( ATOM_IS_RCHIRAL );
        UnsetFlag( ATOM_IS_CHIRAL );
    }
	/** @brief assign ATOM_IS_STEREO flag to the atom */
	void SetStereo()
	{
		SetFlag( ATOM_IS_STEREO );
	}
	/** @brief clear stereo relevant flag to the atom */
	void UnsetStereo()
	{
		UnsetFlag( ATOM_IS_ESTEREO );
		UnsetFlag( ATOM_IS_ZSTEREO );
		UnsetFlag( ATOM_IS_STEREO );
	}

    ////IS////IS////IS////IS////IS////IS////IS////IS////IS////IS////IS////IS///
	/** @brief check whether this atom was bonded to atom */
    bool IsConnected( CCAtom* atom ) const;
	/** @brief check whether the atom is chiral atom */
	bool IsChiral();
	/** @brief check whether the atom is R-chiral atom */
    bool IsRChiral() const
    {
        return HasFlag( ATOM_IS_RCHIRAL );
    }
	/** @brief check whether the atom is S-chiral atom */
    bool IsSChiral() const
    {
        return HasFlag( ATOM_IS_SCHIRAL );
    }
	/** @brief check whether the atom is stereo atom */
	bool IsStereo();
	/** @brief check whether the atom is E-stereo atom */
	bool IsEStereo() const
	{
		return HasFlag( ATOM_IS_ESTEREO );
	}
	/** @brief check whether the atom is Z-stereo atom */
	bool IsZStereo() const
	{
		return HasFlag( ATOM_IS_ZSTEREO );
	}

	/** @brief check whether the atom is aromatic atom */
    bool IsAromatic();
	/** @brief check whether the atom is in ring */
    bool IsInRing();
	/** @brief check whether the atom is in certain membered ring */
    bool IsInRingSize( int size );
	/** @brief get the smallest ring size this atom belonged to */
    int InSmallestRing();
	/** 
	* @brief check whether an atom is in the eleList ( comma-separated elements ).
	* eg. atom->IsOneOf("N,O,S")
	*/
    bool IsOneOf( const string& eleList );
	/** @brief check whether the atom is metal atom */
    bool IsMetal()        const;
	/** @brief check whether the atom is terminal atom */
    bool IsTerAtom()      const
    {
        return GetImpVal() == 1;
    }
	/** @brief check whether the atom is heavy atom */
    bool IsHvyAtom()      const
    {
        return m_ele != 1;
    }
	/** @brief check whether the atom is hetero atom ( N, O, S, P ) */
    bool IsHetAtom()      const
    {
        return ( m_ele == 7 || m_ele == 8 || m_ele == 15 || m_ele == 16 );
    }
	/** @brief check whether the atom is halogen atom ( F, Cl, Br, I ) */
    bool IsHalAtom()      const
    {
        return ( m_ele == 9 || m_ele == 17 || m_ele == 35 || m_ele == 53 );
    }
	/** @brief check whether the atom is hydrogen */
    bool IsHydrogen()     const
    {
        return m_ele == 1;
    }
	/** @brief check whether the atom is apolar hydrogen */
    bool IsApolarHydrogen() const;
	/** @brief check whether the atom is carbon */
    bool IsCarbon()       const
    {
        return m_ele == 6;
    }
	/** @brief check whether the atom is nitrogen */
    bool IsNitrogen()     const
    {
        return m_ele == 7;
    }
	/** @brief check whether the atom is oxygen */
    bool IsOxygen()       const
    {
        return m_ele == 8;
    }
	/** @brief check whether the atom is fluorine */
    bool IsFluorine()     const
    {
        return m_ele == 9;
    }
	/** @brief check whether the atom is chlorine */
    bool IsChlorine()     const
    {
        return m_ele == 17;
    }
	/** @brief check whether the atom is bromine */
    bool IsBromine()      const
    {
        return m_ele == 35;
    }
	/** @brief check whether the atom is indine */
    bool IsIodine()       const
    {
        return m_ele == 53;
    }
	/** @brief check whether the atom is silicon */
    bool IsSilicon()      const
    {
        return m_ele == 14;
    }
	/** @brief check whether the atom is phosphorus */
    bool IsPhosphorus()   const
    {
        return m_ele == 15;
    }
	/** @brief check whether the atom is sulfur */
    bool IsSulfur()       const
    {
        return m_ele == 16;
    }
	/** @brief check whether the atom is arsenic */
    bool IsArsenic()      const
    {
        return m_ele == 33;
    }
	/** @brief check whether the atom is selenium */
    bool IsSelenium()     const
    {
        return m_ele == 34;
    }
    /*bool IsAntimony()     const
    {
        return m_ele == 51;
    }*/
	/** @brief check whether the atom is terminal carbon */
    bool IsTerCarbon()    const
    {
        return ( m_ele == 6 && GetValence() == 1 );
    }
	/** @brief check whether the atom is sp1 carbon */
    bool IsSp1Carbon()    const
    {
        return ( m_ele == 6 && GetHyb() == 1 );
    }
	/** @brief check whether the atom is sp2 carbon */
    bool IsSp2Carbon()    const
    {
        return ( m_ele == 6 && GetHyb() == 2 );
    }
	/** @brief check whether the atom is sp3 carbon */
    bool IsSp3Carbon()    const
    {
        return ( m_ele == 6 && GetHyb() == 3 );
    }
	/** @brief check whether the atom is terminal nitrogen */
    bool IsTerNitrogen()  const
    {
        return ( m_ele == 7 && GetHvyVal() == 1 );
    }
	/** @brief check whether the atom is potential sp2 nitrogen which is
	* defined as a nitrogen atom with only one heavy atom neighbor and three
	* links. This is special for Nitrogen ( Hvy=N-H ).
	*/
    bool IsPotentialSp2N() const
    {
        return ( m_ele == 7 && GetHvyVal() == 1 && GetValence() < 3 );
    }
	/** @brief check whether the atom is sp1 nitrogen */
    bool IsSp1Nitrogen()  const
    {
        return ( m_ele == 7 && GetHyb() == 1 );
    }
	/** @brief check whether the atom is sp2 nitrogen */
    bool IsSp2Nitrogen()  const
    {
        return ( m_ele == 7 && GetHyb() == 2 );
    }
	/** @brief check whether the atom is sp3 nitrogen */
    bool IsSp3Nitrogen()  const
    {
        return ( m_ele == 7 && GetHyb() == 3 );
    }
	/** @brief check whether the atom is terminal oxygen */
    bool IsTerOxygen()    const
    {
        return ( m_ele == 8 && GetValence() == 1 );
    }
	/** @brief check whether the atom is sp2 oxygen */
    bool IsSp2Oxygen()    const
    {
        return ( m_ele == 8 && GetHyb() == 2 );
    }
	/** @brief check whether the atom is sp3 oxygen */
    bool IsSp3Oxygen()    const
    {
        return ( m_ele == 8 && GetHyb() == 3 );
    }
	/** @brief check whether the atom is terminal sulfur */
    bool IsTerSulfur()    const
    {
        return ( m_ele == 16 && GetValence() == 1 );
    }
	/** @brief check whether the atom is sp2 sulfur */
    bool IsSp2Sulfur()    const
    {
        return ( m_ele == 16 && GetHyb() == 2 );
    }
	/** @brief check whether the atom is sp3 sulfur */
    bool IsSp3Sulfur()    const
    {
        return ( m_ele == 16 && GetHyb() == 3 );
    }
	/** @brief check whether the atom is sp2 selenium */
    bool IsSp2Selenium()  const
    {
        return ( m_ele == 34 && GetHyb() == 2 );
    }
	/** @brief check whether the atom is hydrogen bond donor atom */
    bool IsHBDonor();
	/** @brief check whether the atom is hydrogen bond acceptor atom */
    bool IsHBAcceptor();
	/** @brief check whether the atom is donor hydrogen */
    bool IsDornorH();
	/** @brief check whether the atom is acceptor hydrogen */
    bool IsAcceptorH();
	/** @brief check whether the atom is amide nitrogen */
    bool IsAmideNitrogen() const;
	/** @brief check whether the atom is in a closure bond */
    bool IsInClosureBond();
	/** @brief check whether the atom is polar to hydrogen */
    bool IsPolarToHydrogen() const
    {
        return ! ( IsCarbon() || IsHydrogen() );
    }

    /** @brief get the number of bonded heavy atoms */
    int GetHvyVal()     const;
	/** @brief get the min and max number of pi electrons 
	* that contribute to aromaticity */
    pair<int, int> GetPIelectron() const;
    /** @brief get max bond order of neighboring bonds */
    int GetMaxBO();

	/** @brief get sum of neighboring bond orders */
	int SumBondOrder( bool hydrogen = true );
	/** @brief count bond with order */
    int CountBondWithOrder( int order ) const;
	/** @brief count bond with larger order */
    int CountBondWithLargerOrder( int order ) const;

    ///HAS///HAS///HAS///HAS///HAS///HAS///HAS///HAS///HAS///HAS///HAS///HAS//
	/** @brief whether the atom connected with at least one hydrogen atom */
    bool HasH( bool includeImpH = true, bool excludeIsotopes = false );
	/** @brief check whether the atom has pi electron */
    bool HasPiElectron();
	/** @brief check whether the atom bonded to a conjugate system */
    bool HasAlphaBetaUnsat( bool includePandS = true );
	/** @brief has bond with order ? */
    bool HasBondWithOrder( int order ) const;
	/** @brief has bond with larger order ? */
	bool HasBondWithLargerOrder( int order ) const;
	/** @brief check whether the atom has aromatic 
	* neighboring bond ( ring bond with order of 5 ) */
    bool HasAromaticBond() const;

	/** @brief whether chiral information is assigned */
    bool HasChiralitySpecified() const
    {
        return HasFlag( ATOM_IS_RCHIRAL | ATOM_IS_SCHIRAL );
    }

	/** @brief count the number of implicit hydrogens */
    int ImpHydrogenCount();
	/** @brief return the number of hydrogens explicitly
	* bound to this atom, optionally excluding D,T and 
	* isotope explicitly set to 1 */
    int ExpHydrogenCount( bool excludeIsotopes = false ) const;
	/** @brief mark that atom has no hydrogens attached */
    void ForceNoH()
    {
        SetFlag( ATOM_HAS_NO_H );
    }

	/** @brief count the sum of the bond orders of the bonds to 
	* the atom (i.e. double bond = 2, aromatic bond = 3...) */
    int SumBO() const;
	/** @brief used by smi format to get chiral and stereo information */
    bool GetNewBondVector( CCVector& v, double length );
	/** @brief count the number of rings that contain this atom */
    int MemberOfRingCount() const;
	/** @brief return the size of the first ring that contains this
	* atom (0 if not in a ring) */
    int MemberOfRingSize() const;
	/** @brief count the number of explicit ring connections to this atom */
    int CountRingBonds() const;
    //////////////////////////////////////////////////////////////////////////
	/** @brief check whether the atom match the smarts pattern */
    bool MatchesSMARTS( const char* pattern );
	/** @brief get the max weighted neighboring atom */
    CCAtom* GetMaxWeightedNbr( CCAtom* excludeatom = NULL ) const;
	/** @brief add hydrogen atom to this atom */
    int AddHydrogen( bool polarHydrogenOnly=false );
	/** @brief add a hydrogen atom in certain position. */
	bool AddSpecificHydrogen( const CCVector &coor, int order = 1 );
	/** @brief delete attached non-polar hydrogen/all hydrogen */
    int DeleteHydrogen( bool nonPolarHydrogenOnly = false );
	//////////////////////////////////////////////////////////////////////////
	/** @brief get XLOGP3 type */
	string GetXtype();
	/** @brief set XLOGP3 type */
	void SetXtype( const string& xtype ) { m_xtype = xtype; }
};

//////////////////////////////////////////////////////////////////////////
//                       Atom related functions                         //
//////////////////////////////////////////////////////////////////////////
// for center atom check, bit 1-> O; 2-> S; 4-> N
/** @brief check whether an atom is in carbonyl group */
CCAPI int IsCarbonyl( CCAtom* atom, int bit = 1 );
/** @brief check whether an atom is in carboxyl group */
CCAPI int IsCarboxyl( CCAtom* atom, int bit = 1 );
/** @brief check whether an atom is in carboxylate group */
CCAPI int IsCarboxylate( CCAtom* atom, int bit = 1 );
/** @brief check whether an atom is in cyano group */
CCAPI int IsCyano( CCAtom* atom );
/** @brief check whether an atom is in nitro group */
CCAPI int IsNitro( CCAtom* atom );
/** @brief check whether an atom is in nitroso group */
CCAPI int IsNitroso( CCAtom* atom );
/** @brief check whether an atom is in nitro-oxide group */
CCAPI int IsNitroOxide( CCAtom* atom );
/** @brief check whether an atom is in sulfone group */
CCAPI int IsSulfone( CCAtom* atom, int bit = 1 );
/** @brief whether an atom is in sulfonate group */
CCAPI int IsSulfonate( CCAtom* atom, int bit = 1 );
/** @brief check whether an atom is in sulfine group */
CCAPI int IsSulfine( CCAtom* atom, int bit = 1 );
/** @brief check whether an atom is in phosphone group */
CCAPI int IsPhosphone( CCAtom* atom, int bit = 1 );
/** @brief check whether an atom is in phosphate group */
CCAPI int IsPhosphate( CCAtom* atom, int bit = 1 );
/** @brief check whether an atom is in hydroxyl group */
CCAPI int IsHydroxyl( CCAtom* atom, int bit = 1 );
/** @brief check whether an atom is in azide group */
CCAPI bool IsAzide( CCAtom* atom );
/** @brief check whether an atom is in acyl group */
CCAPI bool IsAcyl( CCAtom* atom, int bit = 3 );
/** @brief check whether an atom is in electron withdraw group */
CCAPI bool IsEWG( CCAtom* atom );
/** @brief check whether an atom is a hydrogen bond donor */
CCAPI bool IsDonor( CCAtom* atom );
/** @brief check whether an atom is a hydrogen bond acceptor */
CCAPI bool IsAcceptor( CCAtom* atom );
CCAPI bool Is15CoPlanar( CCAtom* a1, CCAtom* a2 );
/** @brief check whether an atom is in pi system */
CCAPI bool IsPiSystem( CCAtom* atom );
/** @brief check whether an atom is a special nitrogen ( >N- ) */
CCAPI bool IsAliphaticNitrogen( CCAtom* atom );
/** @brief check whether an atom is in symmetric group.
* x=2,3,4 -NH2 -N[+]H3 NH3 -NO2 -CH3 CH4 -CX3 -PO3 -SO3 -CO2 */
CCAPI bool IsATx( CCAtom* atom, int preT = -1 );
/** @brief check whether an atom is in t-butyl group */
CCAPI bool IsTButyl( CCAtom* atom );
/** @brief check whether an atom is Positively Ionizable(PI) atom */
CCAPI bool IsPositivelyIonizable( CCAtom *atom );
/** @brief check whether an atom is Negatively Ionizable(NI) atom */
CCAPI bool IsNegativelyIonizable( CCAtom *atom );

// for adjacent atom check
/** @brief check whether an atom has adjacent halogen atom */
CCAPI int HasAdjacentHalogen( CCAtom* atom );
/** @brief check whether an atom has adjacent sulfone S atom */
CCAPI bool HasAdjacentSulfone( CCAtom* atom, int bit );
/** @brief check whether an atom has adjacent sulfine S atom */
CCAPI bool HasAdjacentSulfine( CCAtom* atom, int bit );
/** @brief check whether an atom has adjacent phosphone P atom */
CCAPI bool HasAdjacentPhosphone( CCAtom* atom, int bit );
/** @brief check whether an atom has adjacent acyl group */
CCAPI bool HasAdjacentAcyl( CCAtom* atom, int bit );
/** @brief check whether an atom has adjacent hydroxyl group */
CCAPI bool HasAdjacentHydroxyl( CCAtom* atom, int bit );
/** @brief check whether an atom has adjacent EWG */
CCAPI bool HasAdjacentEWG( CCAtom* atom );
/** @brief check whether an atom has adjacent aromatic neighbor */
CCAPI bool HasAdjacentAromatic( CCAtom* atom );
/** @brief check whether an atom has adjacent pi system */
CCAPI bool HasAdjacentPiSystem( CCAtom* atom );
/** @brief check whether an atom has adjacent -OH */
CCAPI bool HasAdjacentHydroxyl( CCAtom* atom, int bit = 1 ); // bit 1=OH, 2=SH
/** @brief check whether an atom has adjacent ring */
CCAPI bool HasAdjacentRing( CCAtom* atom, CCRing** ring = NULL );
/** @brief check whether two atoms have the same altloc data (see pdb format) */
CCAPI bool HasSameAltlocData( CCAtom* a1, CCAtom* a2 );
/** @brief search the shortest path length of the two atoms */
CCAPI int ShortestPathLength( CCAtom* a1, CCAtom* a2 );
/** @brief the shortest path length of the two atoms */
CCAPI int ShortestPath( CCAtom* a1, CCAtom* a2, 
                        vector<CCAtom*>* sPath = NULL,
                        bool impH = false ); // BFS
/** @brief use depth first search method to find all the atoms of nDepth */
CCAPI bool DFSnDepth( CCAtom* root, vector<vector<CCAtom*> >& paths,
					  int nDepth = 4, bool impH = false );
/** @brief compare two atoms in ascending order of their coordinates */
CCAPI bool CmpXYZ( const CCAtom* a, const CCAtom* b );

#endif


