/** @file 
* @brief Base classes to build a molecule
* @author Yuan ZHAO, Tiejun CHENG
* @date 2008/08/24
* @version 1.0.1 
*
* Basic class for holding different data.
*
* Copyright (C) 2007-2008 www.sioc-ccbg.ac.cn
* 
* This file is part of the CCBGKit project.
*/

#ifndef CCBASE_H
#define CCBASE_H

#include <CCConfig.h>
#include <data/CCData.h>

//////////////////////////////////////////////////////////////////////////
//                           CCBase class                               //
//////////////////////////////////////////////////////////////////////////
/** 
* @class CCBase CCBase.h "include/ccbg/CCBase.h"
* @brief CCBase class. 
*
* CCBase is just a placeholder class.
* The various classes ( CCAtom, CCBond, CCMol) inherit from CCBase class.
*/ 
class CCAPI CCBase
{
protected:
    vector<CCGenericData*> m_vdata;    ///< Generic data
    unsigned long m_flags;             ///< Bitwise flag data

public:
	/** @brief clear all generic data in CCBase */
    void Clear();
	/** @brief CCBase constructor */
    CCBase();
	/** @brief CCBase destructor */
    virtual ~CCBase()
    {
        DestroyGD();
    }

	/** @brief copy CCBase data */
    void Clone( const CCBase& src );

	/** @brief return the size of generic data */
    int DataSize() const
    {
        return( m_vdata.size() );
    }

	/** @brief get first generic data of certain attribute */
	CCGenericData* GetData( const string& attr ) const;
	/** @brief get first generic data of certain data type */
    CCGenericData* GetData( const int type ) const;
	/** @brief get all generic data */
    const vector<CCGenericData*>& GetData() const
    {
        return( m_vdata );
    }
	/** @brief get all generic data of certain type */
    vector<CCGenericData*> GetData( DataOrigin source ) const;
	/** @brief check whether it has certain type of attribute */
    bool HasData( const string& attr ) const;
	/** @brief check whether it has certain type of data type */
    bool HasData( const int type ) const;
	/** @brief set generic data */
    void SetData( CCGenericData* gd )
    {
        m_vdata.push_back( gd );
    }

    // flags
	/** @brief return flag data */
    unsigned long GetFlag() const
    {
        return( m_flags );
    }
	/** @brief set certain flag data */
    void SetFlag( unsigned long flag )
    {
        m_flags |= flag;
    }
	/** @brief clear certain flag data */
    void UnsetFlag( unsigned long flag )
    {
        m_flags &= ~flag;
    }
	/** @brief clear all flag data */
	void ClearFlag() {
		m_flags = 0;
	}
	/** @brief check whether has certain flag */
    bool HasFlag( unsigned long flag ) const
    {
        return ( m_flags & flag );
    }

	/** @brief destroy generic data */
    void DestroyGD();
};

#endif

