/** @file
* @brief Vector of bits.
* @author Yuan ZHAO, Tiejun CHENG
* @date 2008/08/24
* @version 1.0.1
*
* The CCBitVec class is a fast and efficient bitstring class that is
* handy to use as a truth table. Truth tables are an easy way to store
* whether a list of items has a particular propery. Instances of
* CCBitVec can by dynamically resized, and have a number of overloaded
* operators that make code simple and readable. The following examples
* demonstrate how to use the CCBitVec class:
*
* @code
*	CCBitVec bv1,bv2,bv3;
*	bv1.Set( 10 );
*	bv2.Reset( 20 );
*   bv2.Flip( 15 );
*	bv1 |= bv2;
*	bv1 = bv1 & bv2;
*   //Any() returns true if at least one bit is set on
*	if ( !bv1.Any() ) {
*		cout << "bv1 = " << bv1 << endl;
*	}
* @endcode
*
* Copyright (C) 2007-2008 www.sioc-ccbg.ac.cn
*
* This file is part of the CCBGKit project.
*/

#ifndef CCBITVEC_H
#define CCBITVEC_H

#include <CCConfig.h>

#ifndef WORDSIZE
#define WORDSIZE 32
#endif


/**
* @class CCBitVec CCBitVec.h "include/ccbg/CCBitVec.h"
* @brief CCBitVec class.
*
* Fast and efficient bitstring class.
*
* @note bit position starts form 0.
*/

class CCAPI CCBitVec
{
protected:
    vector<int> m_set; ///< data container of CCBitVec

public:
	/** @brief clear the data of m_set */
    void Clear();
	/** @brief CCBitVec constructor */
    CCBitVec();
	/** @brief initialize CCBitVec with an integer */
    CCBitVec( int bits );
	/** @brief initialize CCBitVec with a string */
	CCBitVec( const string s );
	/** @brief CCBitVec copy constructor */
    CCBitVec( const CCBitVec& bv );
	/** @brief CCBitVec destructor */
    virtual ~CCBitVec() {};

	/** @brief resize the size of m_set to bits length */
    void Resize( int bits );
	/** @brief set the pos bit on */
    void Set( int pos );
	/** @brief set all bits on */
    void Set();
	/** @brief set the pos bit off */
    void Reset( int pos );
	/** @brief set all bits off */
    void Reset();
	/** @brief flip the pos bit */
    void Flip( int pos );
	/** @brief flip all bits */
    void Flip();
	/** @brief check whether pos bit is on */
    bool operator[]( int pos ) const; // subscript starts from 0

	/** @brief initialize all bits from a string */
    void FromString( const string& s );
	/** @brief output all bits to a string */
    string ToString() const;
	/** @brief initialize from an int vector of indices where bits are set to 1 */
    void FromTrueIdx( vector<int>& v );
	/** @brief return indices where bits are set to 1 */
	vector<int> ToTureIdx() const;
	/** @brief initialize all bits from an integer vector */
	void FromVecInt( vector<int>& v );
	/** @brief output all bits an integer vector. */
	vector<int> ToVecInt() const;
	/** @brief return the size of the bit vector */
    int Size() const
    {
        return m_set.size();
    }
	/** @brief return the memory size of the bit vector */
    int Bits() const
    {
        return m_set.size() * WORDSIZE;
    }

	/** @brief return the total number of on bit */
    int Count() const;
	/** @brief check whether at least one bit is on */
    bool Any() const;
	/** @brief check whether all bits is off */
    bool None() const;
	/** @brief check whether all bits is on */
    bool All() const;  // all element is 1
	/** @brief test the state of pos bit */
	bool Test( int pos ) const;

	/** @brief CCBitVec copy assignment operator */
    CCBitVec& operator =( const CCBitVec& bv );
	/** @brief bit and assignment operator */
    CCBitVec& operator &=( const CCBitVec& bv );
	/** @brief bit or assignment operator */
    CCBitVec& operator |=( const CCBitVec& bv );
	/** @brief bit xor assignment operator */
    CCBitVec& operator ^=( const CCBitVec& bv );

	/** @brief equal operator of two CCBitVec */
    friend bool operator ==( const CCBitVec& bv1, const CCBitVec& bv2 );
	/** @brief unequal operator of two CCBitVec */
    friend bool operator !=( const CCBitVec& bv1, const CCBitVec& bv2 );

	/** @brief bit and operator of two CCBitVec */
    friend CCBitVec operator &( const CCBitVec& bv1, const CCBitVec& bv2 );
	/** @brief bit or operator of two CCBitVec */
    friend CCBitVec operator |( const CCBitVec& bv1, const CCBitVec& bv2 );
	/** @brief bit xor operator of two CCBitVec */
    friend CCBitVec operator ^( const CCBitVec& bv1, const CCBitVec& bv2 );
	/** @brief overload input operator on CCBitVec */
    friend istream& operator >>( istream& is, CCBitVec& bv );
	/** @brief overload output operator on CCBitVec */
    friend ostream& operator <<( ostream& os, const CCBitVec& bv );
};

#endif
