/** @file 
* @brief CCBond class and relevant functions.
* @author Yuan ZHAO, Tiejun CHENG
* @date 2008/08/26
* @version 1.0.1 
* 
* CCBond class was a data container. It contains many of the
* bond-specific routines that facilitates extraction of data
* perceived from both the bond and the molecule.
* 
* The following code demonstrates how to print out the bond index,
* the beginning atom index of the bond, the end atom index of the bond:
* @code
* CCMol mol;
* 
* FOR_BONDS_IN_MOL(bond, mol) {
* 	cout << bond->GetIdx() << " ";
* 	cout << bond->GetBeginAtomIdx() << " ";
* 	cout << bond->GetEndAtomIdx() << endl;
* }
* @endcode
*
* Copyright (C) 2007-2008 www.sioc-ccbg.ac.cn
* 
* This file is part of the CCBGKit project.
*/

#ifndef CCBOND_H
#define CCBOND_H

#include <ccbg/CCBase.h>

class CCAtom;
class CCBond;
class CCMol;
class CCVector;

/** @brief abbreviation for std::vector<CCBond*>::iterator. */
typedef std::vector<CCBond*>::iterator CCBondIter;

/************************************************************************/
/*                           CCBond class                               */
/************************************************************************/
/** @brief A enum of bond flags, with inline docs. */ 
enum
{
    BOND_IS_VALID           = 1 << 0,  /**< bond is valid (used in interpret pdb). */
    BOND_IS_IN_RING         = 1 << 1,  /**< bond is in ring. */
    BOND_IS_AROMATIC        = 1 << 2,  /**< bond is aromatic. */
    BOND_IS_CLOSURE         = 1 << 3,  /**< bond is closure bond (used in perceiving rings). */
    BOND_IS_TORUP           = 1 << 4,  /**< "upper" bond in a double bond cis/trans isomer (i.e., "/" in SMILES). */
    BOND_IS_TORDOWN         = 1 << 5,  /**< "down" bond in a double bond cis/trans isomer (i.e., "\" in SMILES). */
    BOND_IS_WEDGE           = 1 << 6,  /**< solid black wedge in 2D representations -- i.e., "up" from the 2D plane. */
    BOND_IS_HASH            = 1 << 7,  /**< dashed "hash" bond in 2D representations -- i.e., "down" from the 2D plane. */
    BOND_IS_IN_FGROUP       = 1 << 8,  /**< bond is in functional groups. */
    BOND_IS_ROTATABLE       = 1 << 9,  /**< bond is rotatable bond. */
	BOND_IS_TAUTOMERAR      = 1 << 10, /**< bond is tautomer aromatic bond (only in even membered ring). */
    BOND_IS_KSINGLE         = 1 << 11, /**< bond is single bond after kekulizing. */
    BOND_IS_KDOUBLE         = 1 << 12, /**< bond is double bond after kekulizing. */
    BOND_IS_KTRIPLE         = 1 << 13, /**< bond is triple bond after kekulizing. */
    BOND_IS_SSBOND          = 1 << 14, /**< bond is S-S bond. */
    BOND_IS_PEPTIDE         = 1 << 15, /**< bond is peptide bond ( only for protein ). */
    BOND_IS_INPUT           = 1 << 16, /**< bond exists in pdb "CONECT" record. */
	BOND_IS_IN_MATCH        = 1 << 17, /**< bond is included in searching matching bond. */
	BOND_IS_LEADOPT         = 1 << 18, /**< bond is transplanting bond in lead optimization. */
	BOND_IS_OUT_MATCH       = 1 << 19, /**< bond is excluded from searching matching bond. */
	BOND_IS_RECAP           = 1 << 20, /**< bond is in 11 RECAP bond-cleavage types. */
};

/** 
* @class CCBond CCBond.h "include/ccbg/CCBond.h"
* @brief CCBond class. 
* 
* a data container of bonds properties and relevant functions.
*/ 
class CCAPI CCBond : public CCBase
{
protected:
    CCMol*  m_mol; ///< parent molecule
    int     m_idx; ///< bond index
    CCAtom* m_bgn; ///< begin atom
    CCAtom* m_end; ///< end atom
    short m_order; ///< bond order
    string m_type; ///< bond type

public:
	/** @brief CCBond constructor. */
    CCBond();
	/** @brief CCBond copy constructor. */
    CCBond( const CCBond& src );
	/** @brief CCBond copy assignment operator. */
    CCBond& operator =( const CCBond& src );
	/** @brief CCBond destructor. */
	virtual ~CCBond();
	/** @brief clear members of CCBond. */
	void Clear();

	///GET///GET///GET///GET///GET///GET///GET///GET///GET///GET///GET///GET//
	/** @brief get the parent molecule of current bond. */
    CCMol* GetMol() const
    {
        return m_mol;
    }
	/** @brief get index of current bond. */
    int GetIdx() const
    {
        return m_idx;
    }
	/** @brief get beginning atom of current bond. */
    CCAtom* GetBeginAtom() const
    {
        return m_bgn;
    }
	/** @brief get end atom of current bond. */
    CCAtom* GetEndAtom() const
    {
        return m_end;
    }
	/** @brief get bond order. */
    short GetOrder() const
    {
        return m_order;
    }
	/** @brief get internal bond type. */
    string GetType() const;

	/** @brief get the atom in current bond which is 
	* different from curr atom. */
	CCAtom* GetNbrAtom( CCAtom* curr )  const;
	/** @brief get the index of atom in current bond 
	* which is different from curr atom. */
	int GetNbrAtomIdx( CCAtom* curr )   const;
	/** @brief get the index of beginning atom in current bond. */
	int GetBeginAtomIdx()               const;
	/** @brief get the index of end atom in current bond. */
	int GetEndAtomIdx()                 const;
	/** @brief get bond length. */
	double GetLength()                  const;
	/** @brief get standard bond length of current bond. */
	double GetSTLength()                const;
	/** @brief get the coordinates of the center of current bond. */
	CCVector GetCenter()                const;

	///SET///SET///SET///SET///SET///SET///SET///SET///SET///SET///SET///SET//
	/** @brief set the parent molecule of current bond. */
    void SetMol( CCMol* mol )
    {
        m_mol = mol;
    }
	/** @brief set index of current bond. */
    void SetIdx( int idx )
    {
        m_idx = idx;
    }
	/** @brief set beginning atom of current bond. */
    void SetBegin( CCAtom* bgn )
    {
        m_bgn = bgn;
    }
	/** @brief set end atom of current bond. */
    void SetEnd( CCAtom* end )
    {
        m_end = end;
    }
	/** @brief set bond order. */
    void SetOrder( short order )
    {
        m_order = order;
    }
	/** @brief set internal bond type. */
    void SetType( const string type )
    {
        m_type = type;
    }
	/** @brief set internal bond type from bond order. */
    void SetTypeFromBO();
	/** @brief set bond index, beginning atom, end atom, order and flag. */
    void Set( int idx, CCAtom* bgn, CCAtom* end, int order, unsigned long flag );
	/** @brief assign BOND_IS_VALID flag to current bond. */
	void SetValid()
	{
		SetFlag( BOND_IS_VALID );
	}
	/** @brief assign BOND_IS_IN_FGROUP flag to current bond. */
	void SetFGroup()
	{
		SetFlag( BOND_IS_IN_FGROUP );
	}
	/** @brief assign BOND_IS_IN_FGROUP and BOND_IS_VALID 
	* flag to current bond. */
	void SetValidFGroup()
	{
		SetFlag( BOND_IS_IN_FGROUP ); 
		SetFlag( BOND_IS_VALID );
	}
	/** @brief assign BOND_IS_AROMATIC flag to current bond. */
	void SetAromatic()
	{
		SetFlag( BOND_IS_AROMATIC );
	}
	/** @brief clear BOND_IS_AROMATIC flag to current bond. */
	void UnsetAromatic()
	{
		UnsetFlag( BOND_IS_AROMATIC );
	}
	/** @brief assign BOND_IS_IN_RING flag to current bond. */
	void SetInRing()
	{
		SetFlag( BOND_IS_IN_RING );
	}
	/** @brief assign BOND_IS_KSINGLE flag to current bond. */
	void SetKSingle();
	/** @brief assign BOND_IS_KDOUBLE flag to current bond. */
	void SetKDouble();
	/** @brief assign BOND_IS_KTRIPLE flag to current bond. */
	void SetKTriple();

	////IS////IS////IS////IS////IS////IS////IS////IS////IS////IS////IS////IS///
	/** @brief check whether current bond had BOND_IS_VALID flag. */
	bool IsValid() const
	{
		return HasFlag( BOND_IS_VALID );
	}
	/** @brief check whether current bond had BOND_IS_IN_FGROUP flag. */
	bool IsInFGroup()
	{
		return HasFlag( BOND_IS_IN_FGROUP );
	}
	/** @brief check whether current bond had BOND_IS_TORUP flag. */
    bool IsUp() const
    {
        return HasFlag( BOND_IS_TORUP );
    }
	/** @brief check whether current bond had BOND_IS_TORDOWN flag. */
    bool IsDown() const
    {
        return HasFlag( BOND_IS_TORDOWN );
    }
	/** @brief check whether current bond had BOND_IS_WEDGE flag. */
    bool IsWedge() const
    {
        return HasFlag( BOND_IS_WEDGE );
    }
	/** @brief check whether current bond had BOND_IS_HASH flag. */
    bool IsHash() const
    {
        return HasFlag( BOND_IS_HASH );
    }

	/** @brief check whether current bond is RECAP bond. */
	bool IsRECAP( bool strict = false );
	/** @brief check whether current bond is aromatic bond. */
    bool IsAromatic();
	/** @brief check whether current bond is in ring. */
    bool IsInRing();
	/** @brief check whether current bond is closure ring bond. */
    bool IsClosure();
	/** @brief check whether current bond is rotatable bond. */
    bool IsRotatable();
	/** @brief check whether current bond is tautomer aromatic bond. */
	bool IsTautomerAR();

	/** @brief check whether current bond is single bond. */
    bool IsSingle();
	/** @brief check whether current bond is double bond. */
    bool IsDouble();
	/** @brief check whether current bond is triple bond. */
    bool IsTriple();
	/** @brief check whether current bond is amide bond. */
    bool IsAmide();
	/** @brief check whether current bond is sulfone amide bond. */
	bool IsSulfoneAmide();
	/** @brief check whether current bond is ester bond. */
	bool IsEster();

	/** @brief check whether current bond is kekulized single bond. */
    bool IsKSingle();
	/** @brief check whether current bond is kekulized double bond. */
    bool IsKDouble();
	/** @brief check whether current bond is kekulized triple bond. */
    bool IsKTriple();

	/** @brief check whether current bond is near planar. */
    bool IsNearPlanar( double cutoff = 30.0 );
	/** @brief check whether the length of current bond is less than 
	* standard single bond length. */
    bool IsLengthInSB()
    {
        return ( GetLength() < GetSTLength() );
    }
	/** @brief check whether the length of current bond is less than 
	* double bond length. */
    bool IsLengthInDB()
    {
        return ( GetLength() < GetSTLength() - 0.11 );
    }
	/** @brief check whether the length of current bond is less than 
	* triple bond length. */
    bool IsLengthInTB()
    {
        return ( GetLength() < GetSTLength() - 0.25 );
    }
	/** @brief check whether current bond is polar bond. */
    bool IsPolar();  // electronegativities difference of two bonded atoms >0.45
};

/************************************************************************/
/* CCBond related functions                                             */
/************************************************************************/
/** @brief check whether two atoms a1 and a2 can form an acceptable bond. */
bool IsAcceptableBond( CCAtom* a1, CCAtom* a2, double cutoff = 0.4 );
/** @brief check whether the bond is in a conjugated system. */
bool HasConjugateBond( CCBond* bond );
/** @brief check whether b1 is more close to form a double bond than b2. */
bool DoubleBondDev( CCBond* b1, CCBond* b2 );

#endif

