/** @file 
* @brief Perceive R/S chirality and E/Z stereo of each atom.
* @author Yuan ZHAO, Tiejun CHENG
* @date 2008/08/26
* @version 1.0.1 
* 
* ref: 1. CahnCIngoldCPrelog priority rules<br>
* http://en.wikipedia.org/wiki/Cahn_Ingold_Prelog_priority_rules <br>
* 2. P. Labute, An Efficient Algorithm For The Determination
* Of Topological RS Chirality. 1996
* 
*	Example of R/S chirality and E/Z stereo perceiving.
* @code
*	CCMol mol;
*	...
*	cout << mol->GetName() << endl;
*	mol->PerceiveChirality();
*	mol->PerceiveStereo();
*	FOR_ATOMS_OF_MOL( a, *mol ) {
*		if ( a->IsRChiral() ) {
*			cout << a->GetIdx() << "R-";
*		}
*		else if ( a->IsSChiral() ) {
*			cout << a->GetIdx() << "S-";
*		}
*		else if ( a->IsEStereo() ) {
*			cout << a->GetIdx() << "E-";
*		}
*		else if ( a->IsZStereo() ) {
*			cout << a->GetIdx() << "Z-";
*		}
*	}
* @endcode
*
* Copyright (C) 2007-2008 www.sioc-ccbg.ac.cn
* 
* This file is part of the CCBGKit project.
*/

#ifndef CCCHIRAL_H
#define CCCHIRAL_H

#include <CCConfig.h>
#include <data/CCData.h>

class CCAtom;
class CCMol;
class CCVector;

/** @brief correct chirality of patm. */
CCAPI bool CorrectChirality( CCAtom* patm, atomreftype i = input,
                             atomreftype o = output );

/** @brief calculate the signed volume for an atom. */
CCAPI double CalcSignedVolume( CCAtom* patm, vector<int> &gidv, bool ReZeroZ = true );
/** @brief calculate a signed volume given a set of 4 coordinates. */
CCAPI double signed_volume( const CCVector& a, const CCVector& b,
                            const CCVector& c, const CCVector& d );

#endif

