#ifndef CCITER_H
#define CCITER_H

#include <CCConfig.h>
#include <ccbg/CCError.h>

class CCMol;
class CCAtom;
class CCBond;
class CCRing;
class CCResidue;
class CCChain;

template<class TParent, class TChild, typename Container>
class CCAPI ParentChildIter
{
protected:
	typename Container::iterator m_iter;
	TParent* m_parent;
	TChild* m_child;

public:
	// need to be implemented for each instance
	virtual void Init( TParent* parent );
	ParentChildIter()
		:m_parent(NULL), m_child(NULL){}
	ParentChildIter( TParent* parent ) 
	{
		if (!parent) {
			gErrorLog.Throw(__FUNCTION__, "NULL parent pointer.", ccFatal);
			exit( EXIT_FAILURE );
		}
		Init(parent);
	}
	ParentChildIter( TParent& parent )
	{
		Init(&parent);
	}
	ParentChildIter( const ParentChildIter& src )
	{
		m_iter = src.m_iter;
		m_parent = src.m_parent;
		m_child = src.m_child;
	}
	ParentChildIter& operator=( const ParentChildIter& src )
	{
		if (&src != this) {
			m_iter = src.m_iter;
			m_parent = src.m_parent;
			m_child = src.m_child;
		}

		return *this;
	}

	operator bool() const 
	{
		return m_child != NULL;
	}

	// need to be implemented for each instance
	virtual ParentChildIter operator++(); // ++i
	ParentChildIter operator++( int ) // i++
	{
		ParentChildIter tmp(*this);
		operator++();
		return tmp;
	}

	TChild* operator->() const
	{
		return m_child;
	}
	TChild& operator*() const
	{
		return *m_child;
	}
	// need to be implemented for each necessary instance
	virtual void* operator~() const;
};

//! Iterate over all atoms of CCMol
typedef ParentChildIter<CCMol, CCAtom, vector<CCAtom*> > CCMolAtomIter;
//! Iterate over all bonds of CCMol
typedef ParentChildIter<CCMol, CCBond, vector<CCBond*> > CCMolBondIter;
//! Iterate over all neighboring atoms of CCAtom
typedef ParentChildIter<CCAtom, CCAtom, vector<CCBond*> > CCAtomAtomIter;
//! Iterate over all neighboring bonds of CCAtom
typedef ParentChildIter<CCAtom, CCBond, vector<CCBond*> > CCAtomBondIter;
//! Iterate over all rings in CCMol
typedef ParentChildIter<CCMol, CCRing, vector<CCRing*> > CCMolRingIter;
//! Iterate over all chains of CCMol
typedef ParentChildIter<CCMol, CCChain, vector<CCChain*> > CCMolChainIter;
//! Iterate over all residues of CCMol
typedef ParentChildIter<CCMol, CCResidue, vector<CCResidue*> > CCMolResidueIter;
//! Iterate over all residues of CCChain
typedef ParentChildIter<CCChain, CCResidue, vector<CCResidue*> > CCChainResIter;
//! Iterate over all atoms of CCResidue
typedef ParentChildIter<CCResidue, CCAtom, vector<CCAtom*> > CCResidueAtomIter;
//! Iterate over all atoms in CCRing
typedef ParentChildIter<CCRing, CCAtom, vector<CCAtom*> > CCRingAtomIter;

//! Iterate over all angle of CCMol
class CCAPI CCMolAngleIter
{
    vector<vector<CCAtom*> >::iterator m_i;
    CCMol* m_mol;
    vector<vector<CCAtom*> > m_angles;  // three atoms of the angle

public:
    CCMolAngleIter()
        : m_mol( NULL )
    {
    }
    CCMolAngleIter( CCMol* mol );
    CCMolAngleIter( CCMol& mol );
    CCMolAngleIter( const CCMolAngleIter& ai );
    CCMolAngleIter& operator=( const CCMolAngleIter& ai );

    operator bool() const
    {
        return m_i != m_angles.end();
    }
    CCMolAngleIter& operator++();       //  ++i
    vector<CCAtom*> operator *() const
    {
        return *m_i;
    }
};

//! Iterate over all torsion of CCMol
class CCAPI CCMolTorsionIter
{
    vector<vector<CCAtom*> >::iterator m_i;
    CCMol* m_mol;
    vector<vector<CCAtom*> > m_torsions;  // four atoms of the torsion

public:
    CCMolTorsionIter()
        : m_mol( NULL )
    {
    }
    CCMolTorsionIter( CCMol* mol );
    CCMolTorsionIter( CCMol& mol );
    CCMolTorsionIter( const CCMolTorsionIter& ti );
    CCMolTorsionIter& operator=( const CCMolTorsionIter& ti );

    operator bool() const
    {
        return m_i != m_torsions.end();
    }
    CCMolTorsionIter& operator++();       //  ++i
    vector<CCAtom*> operator *() const
    {
        return *m_i;
    }
};

//! Iterate over all pairs of CCMol
class CCAPI CCMolPairIter
{
    vector<vector<CCAtom*> >::iterator m_i;
    CCMol* m_mol;
    vector<vector<CCAtom*> > m_pairs;  // two atoms of the pair

public:
    CCMolPairIter()
        : m_mol( NULL )
    {
    }
    CCMolPairIter( CCMol* mol );
    CCMolPairIter( CCMol& mol );
    CCMolPairIter( const CCMolPairIter& pi );
    CCMolPairIter& operator=( const CCMolPairIter& pi );

    operator bool() const
    {
        return m_i != m_pairs.end();
    }
    CCMolPairIter& operator++();       //  ++i
    vector<CCAtom*> operator *() const
    {
        return *m_i;
    }
};

//! Iterate over all 1-4 pairs of CCMol
class CCAPI CCMol14PairIter
{
    vector<vector<CCAtom*> >::iterator m_i;
    CCMol* m_mol;
    vector<vector<CCAtom*> > m_pairs;  // two atoms of the pair

public:
    CCMol14PairIter()
        : m_mol( NULL )
    {
    }
    CCMol14PairIter( CCMol* mol );
    CCMol14PairIter( CCMol& mol );
    CCMol14PairIter( const CCMol14PairIter& pi );
    CCMol14PairIter& operator=( const CCMol14PairIter& pi );

    operator bool() const
    {
        return m_i != m_pairs.end();
    }
    CCMol14PairIter& operator++();       //  ++i
    vector<CCAtom*> operator *() const
    {
        return *m_i;
    }
};

//! Iterate over all 1-5+ pairs of CCMol
class CCAPI CCMol1NPairIter
{
    vector<vector<CCAtom*> >::iterator m_i;
    CCMol* m_mol;
    vector<vector<CCAtom*> > m_pairs;  // two atoms of the pair

public:
    CCMol1NPairIter()
        : m_mol( NULL )
    {
    }
    CCMol1NPairIter( CCMol* mol );
    CCMol1NPairIter( CCMol& mol );
    CCMol1NPairIter( const CCMol1NPairIter& pi );
    CCMol1NPairIter& operator=( const CCMol1NPairIter& pi );

    operator bool() const
    {
        return m_i != m_pairs.end();
    }
    CCMol1NPairIter& operator++();       //  ++i
    vector<CCAtom*> operator *() const
    {
        return *m_i;
    }
};

// template for iteration over a container
template <typename T, typename Container>
class EIV
{
public :
    EIV( Container& v )
    {
        m_it = v.begin();
        m_itend = v.end();
    }
    EIV( const EIV& src )
    {
        m_it = src.m_it;
        m_itend = src.m_itend;
    }
    EIV& operator =( const EIV& src )
    {
        if ( this != &src ) {
            m_it = src.m_it;
            m_itend = src.m_itend;
        }
        return *this;
    }
    ~EIV()
    {
    };

    operator bool() const
    {
        return m_it != m_itend;
    }
    T operator ->() const
    {
        return *m_it;
    }
    T& operator *() const
    {
        return *m_it;
    }
    EIV operator++() //  ++i
    {
        ++m_it;
        return *this;
    }
    EIV operator++( int ) //  i++
    {
        EIV tmp( *this );
        operator++();
        return tmp;
    }

private:
    typename Container::iterator m_it;
    typename Container::iterator m_itend;
};

// ģػ
typedef EIV<CCAtom*, vector<CCAtom*> > AtomV;
typedef EIV<CCBond*, vector<CCBond*> > BondV;
typedef EIV<CCChain*, vector<CCChain*> > ChainV;
typedef EIV<CCResidue*, vector<CCResidue*> > ResidueV;
typedef EIV<CCMol*, vector<CCMol*> > MolV;
typedef EIV<CCRing*, vector<CCRing*> > RingV;
// Macro for convenience
#define FOR_ATOMS_IN_VECTOR(a,v)    for (AtomV a(v); a; ++a )
#define FOR_BONDS_IN_VECTOR(b,v)    for (BondV b(v); b; ++b )
#define FOR_RESIDUES_IN_VECTOR(r,v) for (ResidueV r(v); r; ++r )
#define FOR_CHAINS_IN_VECTOR(c,v)   for (ChainV c(v); c; ++c )
#define FOR_MOLS_IN_VECTOR(m,v)     for (MolV m(v); m; ++m )
#define FOR_RINGS_IN_VECTOR(r,v)    for (RingV r(v); r; ++r )

// Macro for convenience
#define FOR_ATOMS_OF_MOL(a,m)       for( CCMolAtomIter      a(m); a; ++a )
#define FOR_BONDS_OF_MOL(b,m)       for( CCMolBondIter      b(m); b; ++b )
#define FOR_RINGS_OF_MOL(r,m)       for( CCMolRingIter      r(m); r; ++r )
#define FOR_RESIDUES_OF_MOL(r,m)    for( CCMolResidueIter   r(m); r; ++r )
#define FOR_CHAINS_OF_MOL(c,m)      for( CCMolChainIter     c(m); c; ++c )
#define FOR_NBORS_OF_ATOM(n,a)      for( CCAtomAtomIter     n(a); n; ++n )
#define FOR_BONDS_OF_ATOM(b,a)      for( CCAtomBondIter     b(a); b; ++b )
#define FOR_ATOMS_OF_RING(a,r)      for( CCRingAtomIter     a(r); a; ++a )
#define FOR_ATOMS_OF_RESIDUE(a,r)   for( CCResidueAtomIter  a(r); a; ++a )
#define FOR_RESIDUES_OF_CHAIN(r,c)  for( CCChainResIter     r(c); r; ++r )
#define FOR_ANGLES_OF_MOL(a,m)      for( CCMolAngleIter     a(m); a; ++a )
#define FOR_TORSIONS_OF_MOL(t,m)    for( CCMolTorsionIter   t(m); t; ++t )
#define FOR_PAIRS_OF_MOL(p,m)       for( CCMolPairIter      p(m); p; ++p )
#define FOR_14PAIRS_OF_MOL(p,m)     for( CCMol14PairIter    p(m); p; ++p )
#define FOR_1NPAIRS_OF_MOL(p,m)     for( CCMol1NPairIter    p(m); p; ++p )


#endif // CCITER_H

