/** @file
* @brief 3x3 matrix and its operators.
* @author Yuan ZHAO, Tiejun CHENG
* @date 2008/08/30
* @version 1.0.1
*
* CCMatrix is an arbitrary 3x3 matrix. Some operations
* were define to handle it conveniently.
*
* Copyright (C) 2007-2008 www.sioc-ccbg.ac.cn
*
* This file is part of the CCBGKit project.
*/
#ifndef CCMATRIX_H
#define CCMATRIX_H

#include <CCConfig.h>

class CCVector;

/** 
* @class CCMatrix CCMatrix.h "include/ccbg/CCMatrix.h"
* @brief CCMatrix class. 
* 
* a data container of 3x3 matrix and its operators.
*/ 
class CCAPI CCMatrix
{
protected:
    double m_ele[3][3];  ///< element of the 3x3 matrix
public:
	/** @brief CCMatrix constructor. */
    CCMatrix();
	/** @brief Constructs a matrix from row vectors. */
    CCMatrix( const CCVector& row1, const CCVector& row2, const CCVector& row3 );
	/** @brief Constructs a matrix from a 3x3-array of doubles. */
    CCMatrix( const double d[3][3] );
	/** @brief CCMatrix copy constructor. */
	CCMatrix( const CCMatrix& src );
	/** @brief CCMatrix copy assignment operator. */
	CCMatrix& operator=( const CCMatrix& src );
	/** @brief CCMatrix destructor */
	~CCMatrix() {};

	/** @brief operator for get the value of certain element. */
    const double& operator()( int row, int column ) const;
	/** @brief operator for get the reference of certain element. */
    double& operator()( int row, int column );
	/** @brief calculate the determinant of current matrix. */
    double Determinant() const;
	
	/** @brief calculate the inverse of current matrix. */
    CCMatrix Inverse() const;
	/** @brief calculate the transpose of current matrix. */
    CCMatrix Transpose() const;
	/** @brief calculates a rotation matrix. */
    void RotAboutAxisByAngle( const CCVector& axis, double angle );
	/** @brief multiplies all entries of the matrix by a scalar c. */
    CCMatrix& operator *=( double c )
    {
        for ( int i = 0; i < 3; i++ ) {
            for ( int j = 0; j < 3; j++ ) {
                m_ele[i][j] *= c;
            }
        }

        return *this;
    }

	/** @brief divides all entries of the matrix by a scalar c. */
    CCMatrix& operator /=( double c )
    {
        return ( ( *this ) *= ( 1.0 / c ) );
    }

	/** @brief matrix-vector multiplication m(3*3) * v(3*1). */
    friend CCVector operator*( const CCMatrix& m, const CCVector& v );
	/** @brief vector-matrix multiplication v(1*3) * m(3*3). */
    friend CCVector operator*( const CCVector& v, const CCMatrix& m );
	/** @brief matrix-matrix multiplication. */
    friend CCMatrix operator*( const CCMatrix& A, const CCMatrix& B );
};

#endif


