/** @file 
* @brief CCMol class and relevant functions.
* @author Yuan ZHAO, Tiejun CHENG
* @date 2008/08/23
* @version 1.0.1 
* 
* CCMol class was a data container. It contains many of the
* molecular specific routines that facilitates extraction of data
* perceived from both the molecule.
* 
* The following code demonstrates how to read and write molecules,
* get an atom, get molecular formula. It also indicates how to
* traverse between each atom and bond.
*
* @code
*     CCMol mol;
*     mol.ReadMolecule( "test.sdf" );
*     mol.WriteMolecule( "result.mol2" );
*     CCAtom *patom = mol.GetAtom( 10 );
*     string formula = mol.GetFormula();
*     FOR_ATOMS_OF_MOL( atm, mol ) {
*         cout << atm->GetIdx() << endl;
*     }
*     FOR_BONDS_OF_MOL( bnd, mol ) {
*         cout << bnd->GetIdx() << endl;
*     }
* @endcode
*
* Copyright (C) 2007-2008 www.sioc-ccbg.ac.cn
* 
* This file is part of the CCBGKit project.
*/

#ifndef CCMOL_H
#define CCMOL_H

#include <CCConfig.h>

#include <ccbg/CCBase.h>
#include <ccbg/CCFileHandler.h>
#include <ccbg/CCUtil.h>
#include <ccbg/CCError.h>
#include <ccbg/CCVector.h>
#include <ccbg/CCMatrix.h>
#include <ccbg/CCBitVec.h>
#include <ccbg/CCIter.h>

#include <formats/CCFormat.h>
#include <data/CCData.h>

#include <ccbg/CCAtom.h>
#include <ccbg/CCBond.h>
#include <ccbg/CCChain.h>
#include <ccbg/CCResidue.h>
#include <ccbg/CCRing.h>
#include <ccbg/CCAromatic.h>
#include <ccbg/CCChiral.h>
#include <ccbg/parSMART.h>

typedef pair<CCAtom*,CCAtom*> CCAAPair;
typedef pair<CCAtom*,CCBond*> CCABPair;
typedef pair<CCBond*,CCBond*> CCBBPair;
typedef pair<CCBond*,CCAtom*> CCBAPair;

/************************************************************************/
/*                           CCMol class                                */
/************************************************************************/
enum
{
    MOL_IS_VALID                        = 1 << 0,
    MOL_RING_PERCEIVED                  = 1 << 1,
    MOL_AROMATICITY_PERCEIVED           = 1 << 2,
    MOL_SSSR_PERCEIVED                  = 1 << 3,
    MOL_ALL_RING_PERCEIVED              = 1 << 4,
    MOL_FRAGMENT_PERCEIVED              = 1 << 5,
    MOL_PUTATIVE_HYBRID_PERCEIVED       = 1 << 6,
    MOL_HYBRID_PERCEIVED                = 1 << 7,
    MOL_IMPVALENCE_PERCEIVED            = 1 << 8,
    MOL_INT_TYPE_PERCEIVED              = 1 << 9,
    MOL_CLOSURE_BOND_PERCEIVED          = 1 << 10,
    MOL_CHIRALITY_PERCEIVED             = 1 << 11,
	MOL_STEREO_PERCEIVED                = 1 << 12,
    MOL_FGROUP_PERCEIVED                = 1 << 13,
    MOL_CONNECTIVITY_PERCEIVED          = 1 << 14,
    MOL_BONDORDER_PERCEIVED             = 1 << 15,
    MOL_ROTATABLEBOND_PERCEIVED         = 1 << 16,
	MOL_TAUTOMERARBOND_PERCEIVED        = 1 << 17,
    MOL_HBDONORACCEPTOR_PERCEIVED       = 1 << 18,
    MOL_HYDROGEN_ADDED                  = 1 << 19,
    MOL_KEKULE_PERCEIVED                = 1 << 20,
    MOL_AUTOFCHARGE_TRUE                = 1 << 21,
    MOL_AUTOPCHARGE_TRUE                = 1 << 22,
    MOL_AUTOFCHARGE_PERCEIVED           = 1 << 23,
    MOL_AUTOPCHARGE_PERCEIVED           = 1 << 24,
	MOL_RECAPBOND_PERCEIVED             = 1 << 25,
	MOL_XTYPE_PERCEIVED                 = 1 << 26,///< XLOGP3 type perceived
	MOL_CFACOTR_PERCEIVED               = 1 << 27,///< XLOGP3 correction factor perceived
};

class CCAPI CCMol : public CCBase
{
protected:
    string m_name;   ///< Molecular name
    int m_natom;     ///< Number of atoms
    int m_nbond;     ///< Number of bonds
    int m_nsubs;     ///< Number of substructures
    vector<CCAtom*> m_vatom;     ///< Atoms
    vector<CCBond*> m_vbond;     ///< Bonds
    vector<CCRing*> m_vring;     ///< Rings
    vector<CCResidue*> m_vresidue;  ///< Residues
    vector<CCChain*> m_vchain;   ///< Chains
    int m_dimension; ///< Dimension
    //////////////////////////////////////////////////////////////////////////
	map<string, size_t> m_cfactor;   ///< XLOGP3 correction factors

public:
	/** @brief CCMol constructor */
    CCMol();
	/** @brief CCMol destructor */
    virtual ~CCMol();
	/** @brief clear members of CCMol */
    void Clear();
	/** @brief destroy members of CCMol */
	void Destroy();
	/** @brief CCMol copy constructor */
    CCMol( const CCMol& src );
	/** @brief CCMol copy assignment operator */
    CCMol& operator =( const CCMol& src );
	/** @brief Clone an instance of src */
    void Clone( const CCMol& src );

    string GetName()    const
    {
        return m_name;
    }
    int GetNumAtom() const
    {
        return m_natom;
    }
	int GetNumHvyAtom();
	int GetNumChiralAtom();
    int GetNumBond();
    int GetNumRes()  const
    {
        return( m_vresidue.size() );
    }
    int GetNumChain() const
    {
        return( m_vchain.size() );
    }
    int GetNumSubs();
    double GetMW() const;
    string GetFormula();

    void SetName( const string name )
    {
        m_name = name;
    }
    void SetNumSubs( int nsubs )
    {
        m_nsubs = nsubs;
    }

    CCAtom* NewAtom();
    CCBond* NewBond();
    CCResidue* NewResidue();
    CCChain* NewChain();

    CCAtom* AddAtom( CCAtom& atom );
    CCBond* AddBond( CCBond& bond );
    CCBond* AddBond( int first, int second, int order, unsigned long flag = 0,
                     int insertpos = -1 );
    CCResidue* AddResidue( CCResidue& residue );
    CCChain* AddChain( CCChain& chain );

    CCAtom* BeginAtom( CCAtomIter& i );
    CCAtom* NextAtom( CCAtomIter& i );
    CCBond* BeginBond( CCBondIter& i );
    CCBond* NextBond( CCBondIter& i );
    CCRing* BeginRing( CCRingIter& i );
    CCRing* NextRing( CCRingIter& i );
    CCResidue* BeginResidue( CCResidueIter& i );
    CCResidue* NextResidue( CCResidueIter& i );
    CCChain* BeginChain( CCChainIter& i );
    CCChain* NextChain( CCChainIter& i );

    void ResizeAtoms( int natm )
    {
        m_vatom.resize( natm );
    }
    void ResizeBonds( int nbnd )
    {
        m_vbond.resize( nbnd );
    }

    CCAtom* GetAtom( int idx ) const;
    CCBond* GetBond( int idx ) const;
    CCBond* GetBond( int start, int end ) const;
    CCBond* GetBond( CCAtom* bgn, CCAtom* end ) const;
    CCResidue* GetResidue( int idx ) const;
    CCChain* GetChain( int idx ) const;
    vector<CCResidue*> GetResOfName( string resname ) const;

    double GetDistance( int a1, int a2 ) const;
    double GetDistance( CCAtom* a1, CCAtom* a2 ) const;
    double GetAngle( int a1, int a2, int a3 ) const;
    double GetAngle( CCAtom* a1, CCAtom* a2, CCAtom* a3 ) const;
    double GetTorsion( int a1, int a2, int a3, int a4 ) const;
    double GetTorsion( CCAtom* a1, CCAtom* a2, CCAtom* a3, CCAtom* a4 ) const;

    bool IsEmpty()
    {
        // Is it a empty molecule
        return ( m_natom == 0 || m_vatom.empty() );
    }

    vector<CCRing*>& GetSSSR();
    vector<CCRing*>& GetAllRings();
    void BFSFindSSSR();
    void FindAllRingsfromSSSR( vector<CCRing*>& allRings, int sizeLimit = 14 );
    void DFSFindAllRings( int sizeLimit = 14 );

    void PerceiveRingAtomsAndBonds();
    void PerceiveHybridization();
    void PerceiveFragment();
    void PerceiveImplicitValence();
    void PerceiveAromaticity();
    void PerceiveClosureBond();
	/** @brief perceive RECAP bond in molecule. */
	void PerceiveRECAPBond( bool strict = false );

    void PerceiveInternalAtomBondTypes();
	/** @brief perceive each chiral atom's R/S chirality. */
    void PerceiveChirality();
	/** @brief perceive each stereo atom's E/Z stereo. */
	void PerceiveStereo();
    void PerceiveRotatableBond();
	void PerceiveTautomerARBond();
    void PerceiveHBDonorAcceptor();

    void UpdateRingBVPath();
    //////////////////////////////////////////////////////////////////////////
    // handle PDB file
    void HandlePDB();
    void PerceiveConnectivityFromXYZ();
    void PerceivePutativeHybFromXYZ();
    void PerceiveFGroupFromXYZ();
    void PerceiveAromaticityFromXYZ();
    void PerceiveBondOrderFromXYZ();
    void PerceiveAtomTypesFromXYZ();
    //////////////////////////////////////////////////////////////////////////
	bool HasExplicitH();
    bool HasNonZeroCoords();
    bool Has2D();
    bool Has3D();
    int  GetDimension();
    void SetDimension( int d )
    {
        m_dimension = d;
    }
    void SetDimensionFromXYZ();
    bool AssignSpin()
    {
        return true;
    }

    //! Set the flag for determining automatic formal charges with pH (default=true)
    void SetAutoFormalCharge( bool );
    //! Set the flag for determining partial charges automatically (default=true)
    void SetAutoPartialCharge( bool );
    // documented in CCChiral.cpp: calculate symmetry-unique identifiers
    void GetGIDVector( vector<int>& vgid );
    //////////////////////////////////////////////////////////////////////////
    void Kekulize();
    void NewPerceiveKekuleBonds();
    void start_kekulize( vector <CCAtom*>& cycle, vector<int>& electron );
    bool expand_kekulize( CCAtom* atom1, CCAtom* atom2,
                          vector<int>& currentState, vector<int>& initState,
                          vector<int>& bcurrentState, vector<int>& binitState,
                          vector<bool>& mark );
    int getorden( CCAtom* atom );
    void expandcycle( CCAtom* atom, CCBitVec& avisit );
    //////////////////////////////////////////////////////////////////////////
	template <class T> void DestroyIt( T* p ) {
		if ( p ) {
			delete p; p = NULL;
		}
	};
    //////////////////////////////////////////////////////////////////////////
    // sort atoms by index
    void SortAtoms();
	// sort atoms according to PDB rule
	void SortPDBAtoms();
    // sort bonds by begin and end atoms
    void SortBonds(bool updateRingBVPath=false);
    // center of coordinates of all atoms
    CCVector GetCentroid( bool includeHydrogen = false ) const;
    // add hydrogen (return number of hydrogen added)
    int AddHydrogen( bool polarHydrogenOnly = false );
	// add hydrogen for protein
	int AddBioHydrogen( bool polarHydrogenOnly = true );
    // delete atom
    bool DeleteBond( CCBond* bond );
    bool DeleteAtom( CCAtom* atom );
    bool DeleteAtoms( vector<CCAtom*> atoms );
    bool DeleteAtom( int idx );
    bool DeleteBonds( vector<CCBond*> bonds );  
    void DeleteHydrogen( bool apolarHydrogenOnly = false );
    //////////////////////////////////////////////////////////////////////////
    // shortcut for reading a single-entry molecule
    bool ReadMolecule( const string& inFilename, unsigned long flag = 0 );
    bool WriteMolecule( const string& outFilename );

    void ShowConnectivity();
    //////////////////////////////////////////////////////////////////////////
	/** @brief Tailor from the bond tbond to get two fragments. */
	void Tailoring( CCBond *tbond, CCMol &bgnpart, CCMol &endpart,
		            CCAtom **bgnpoint, CCAtom **endpoint );
	/** @brief add mol to current molecule and keep mol. */
    void AddMol( CCMol& mol );
	/** @brief merge mol to current molecule and then delete mol. */
	void MergeMol( CCMol& mol );
	/**
	* @brief integrate parent molecule of point1 and parent 
	*        molecule of point2.
	*/
    CCMol& Integrate( CCAtom* point1, CCAtom* point2,
		              int order = 1, int flag = 0,
					  CCAtom **newpoint1 = NULL,
					  CCAtom **newpoint2 = NULL );
	/** 
	* @brief integrate parent molecule of points1 and each parent 
	*        molecule of atoms in points2.
	*/
	bool MultiIntegrate( vector<CCAtom*> &points1,
						 vector<CCAtom*> &points2,
						 vector<int> &orders,
						 vector<int> &flags );
    //////////////////////////////////////////////////////////////////////////
    //void SortBondID();
    //////////////////////////////////////////////////////////////////////////
	/** @brief find all the atoms in angles and save them to mol data. */
    void FindAngles();  
	/** @brief find all the atoms in torsions and save them to mol data. */
    void FindTorsions();
	/** @brief find all pair of atoms( 1-N(N>4) ) and save them to mol data. */
    void FindPairs();
	////////////////////////////////////////////////////////////////////////// XLOGP3
	/** @brief XLOGP3 atom typing scheme */
	void AtomTyping();
	/** @brief get XLOGP3 correction factors */
	map<string, size_t>& GetCFactor();
};

/************************************************************************/
/*                    Molecular related functions                       */
/************************************************************************/
/** @brief check whether two molecules have same formula. */
bool SameFormula( CCMol& mol1, CCMol& mol2 );
/** @brief check whether two molecules have same formula. */
bool SameFormula( CCMol* mol1, CCMol* mol2 );
/** @brief check whether molecule has hydrophobic region. */
bool HasHydrophobicRegion( CCMol& mol );

/** @brief cutoff value for collision. */
const double COLLISION_DIST = 1.0;
const double VDW_BUMP_RANGE = 0.60;
/** @brief check whether two molecules have collision. */
bool OuterCollisionCheck( CCMol &m1, CCMol &m2 );
/** @brief check whether two molecules have collision. */
bool InnerCollisionCheck( CCMol &mol );
/** @brief delete small fragments and only keep the largest one. */
void KeepLargestFragment( CCMol &mol );

#endif
