#ifndef CCRESIDUE_H
#define CCRESIDUE_H

#include <map>
#include <ccbg/CCBase.h>

class CCAtom;
class CCBond;
class CCMol;
class CCResidue;
class CCChain;

typedef std::vector<CCResidue*>::iterator CCResidueIter;

/************************************************************************/
/*                           CCResidue class                            */
/************************************************************************/

enum
{
    RES_IS_VALID            = 1 << 0,
    RES_KEYS_PERCEIVED      = 1 << 1,
    RES_IS_TERMINAL         = 1 << 2,
};

class RESATM // PDB ATOM section
{
public:
	RESATM(bool hetatm=false,
		int serial=0,
		string name="",
		char altLoc=' ',
		string resName="",
		char chainID=' ',
		int resSeq=0,
		char iCode=' ',
		double x=0.0, double y=0.0, double z=0.0,
		double occupancy=0.0,
		double tempFactor=0.0,
		string element="",
		int charge=0
		)
		:_hetatm(hetatm),
		_serial(serial),
		_name(name),
		_altLoc(altLoc),
		_resName(resName),
		_chainID(chainID),
		_resSeq(resSeq),
		_iCode(iCode),
		_x(x), _y(y), _z(z),
		_tempFactor(tempFactor),
		_charge(charge)
	{}

private:
	bool _hetatm;       // 1- 6 ATOM or HETATM
	int _serial;        // 7-11 Atom  serial number
	string _name;       //13-16 Atom name
	char _altLoc;       //17    Alternate location indicator
	string _resName;    //18-20 Residue name
	char _chainID;      //22    Chain identifier
	int _resSeq;        //23-26 Residue sequence number
	char _iCode;        //27    Code for insertion of residues
	double _x;          //31-38 x coordinate
	double _y;          //39-46 y coordinate
	double _z;          //47-54 z coordinate
	double _occupancy;  //55-60 Occupancy
	double _tempFactor; //61-66 Temperature factor
	string _element;    //77-78 Element symbol, right-justified
	int _charge;        //79-80 Charge on the atom
};

class CCAPI CCResidue : public CCBase
{
protected:
    CCMol* m_mol;                // Parent molecule
    CCChain* m_chain;            // Parent Chain
    int m_idx;                   // Residue index
    int m_resNum;                // Residue number
    string m_resName;            // Residue text name
    //----------- ATOM ----------//
    vector<CCAtom*> m_vatom;     // List of atoms in this residue
	vector<bool> m_hetatm;       // Is a given atom a HETAM
	vector<int> m_serial;        // List of atom serial numbers
	vector<string> m_atmName;    // List of atom name
	char m_altLoc;               // Alternate location indicator
	char m_iCode;                // Code for insertion of residues
	vector<double> m_occupancy;  // List of atom occupancy
	vector<double> m_tempFactor; // List of atom temperature factor
	string m_ss;                 // Type of second structure: H=Helix S=Sheet

public:
    CCResidue();
    virtual ~CCResidue();
    void Clear();
    CCResidue( const CCResidue& src );
    CCResidue& operator =( const CCResidue& src );

    CCMol* GetMol() const
    {
        return m_mol;
    }
    int GetIdx() const
    {
        return m_idx;
    }
    CCChain* GetChain() const
    {
        return m_chain;
    }
    char GetChainID() const;
    int GetResNum() const
    {
        return m_resNum;
    }
    string GetResName() const
    {
        return m_resName;
    }
    vector<CCAtom*>& GetAtoms()
    {
        return m_vatom;
    }
    bool IsHetAtom( CCAtom* atom ) const;
    string GetResAtomName( CCAtom* atom ) const;
    CCAtom* GetAtomOfName( string atom_name );
    int GetSerNum( CCAtom* atom ) const;
    char GetAltLoc() const
    {
        return m_altLoc;
    }
	char GetiCode() const
	{
		return m_iCode;
	}
	double GetOccupancy( CCAtom* atom ) const;
	double GetTempFactor( CCAtom* atom ) const;
	string GetSecStruct() const { return m_ss; }

    void SetMol( CCMol* mol )
    {
        m_mol = mol;
    }
    void SetIdx( int idx )
    {
        m_idx = idx;
    }
    void SetChain( CCChain* chain )
    {
        m_chain = chain;
    }
    void SetResNum( int resnum )
    {
        m_resNum = resnum;
    }
    void SetResName( const string& resname )
    {
        m_resName = resname;
    }
    void SetHetAtom( CCAtom* atom, bool hetatm );
    void SetResAtomName( CCAtom* atom, const string& atmname );
    void SetSerNum( CCAtom* atom, int sernum );
    void SetAltLoc( char altloc )
    {
        m_altLoc = altloc;
    }
	void SetiCode( char icode )
	{
		m_iCode = icode;
	}
	void SetOccupancy( CCAtom* atom, double occu );
	void SetTempFactor( CCAtom* atom, double tempFactor );
	void SetSecStruct( const string& ss ) { m_ss = ss; }

    CCAtom* BeginAtom( vector<CCAtom*>::iterator& i );
    CCAtom* NextAtom( vector<CCAtom*>::iterator& i );
    void AddAtom( CCAtom* atom );
    void RemoveAtom( CCAtom* atom );

    bool IsNaturalAA()
    {
        return OneLetter() != 'X';
    }
    char OneLetter()
    {
        return gResidueData.OneLetter( m_resName );
    }
    double HpIndex() // the hydrophobic index
    {
        return gResidueData.HpIndex( m_resName );
    }
};

/************************************************************************/
/*                     Residue related functions                        */
/************************************************************************/
void SetResKeys( const string& resname, unsigned int& reskey,
                 unsigned int& aakey );
unsigned int GetResKey( const string& resname );
//# return if a given residue name is natural amino acid
bool IsNaturalAA( const string& resname );


#endif

