#ifndef CCRING_H
#define CCRING_H

#include <CCConfig.h>
#include <ccbg/CCBitVec.h>
#include <ccbg/CCBase.h>

class CCAtom;
class CCBond;
class CCMol;
class CCRing;
class CCRingNode;
class CCRingTable;

typedef std::vector<CCRing*>::iterator CCRingIter;

CCAPI bool BuildRingTables( CCMol& mol, vector<CCRingTable*>& rtables );
CCAPI bool AnExistRing( CCBitVec& newRingBV, vector<CCRing*>& curRings );
CCAPI void AssembleRings( const vector<CCRing*>& sssr,
                          vector<vector<CCRing*> >& ringAssemblys );
CCAPI void ConstructAllRingsfromAssembly( vector<CCRing*>& allring,
                                          vector<CCRing*>& ringAssembly,
                                          int sizeLimit = -1 );

//////////////////////////////////////////////////////////////////////////
enum
{
    RING_IS_VALID           = 1 << 0,
    RING_IS_NORMAL          = 1 << 1,
    RING_IS_PLANAR          = 1 << 2,
    RING_IS_AROMATIC        = 1 << 3,
};

class CCAPI CCRing : public CCBase
{
protected:
    CCMol* m_mol;      // parent molecule
    vector<CCAtom*> m_vatom;    // atoms in ring
    CCBitVec m_bvpath;   // bitwise atom path

public:
    CCRing()
        : m_mol( NULL )
    {
    }
    virtual ~CCRing()
    {
    }
    CCRing( const CCRing& src );
    CCRing& operator=( const CCRing& src );

    CCMol* GetMol()      const
    {
        return m_mol;
    }

    CCAtom* BeginAtom( vector<CCAtom*>::iterator& i );
    CCAtom* NextAtom( vector<CCAtom*>::iterator& i );

    vector<int> GetAtomPath() const;
    CCBitVec GetAVPath()   const;
    vector<int> GetBondPath() const;
    CCBitVec GetBVPath() const
    {
        return m_bvpath;
    }
    vector<CCAtom*>& GetAtoms()
    {
        return m_vatom;
    }
    vector<CCBond*> GetBonds() const;

    int GetSize() const
    {
        return m_vatom.size();
    }
    void SetMol( CCMol* mol )
    {
        m_mol = mol;
    }

    void SetAtoms( const vector<CCAtom*>& atoms )
    {
        m_vatom = atoms;
    }
    void SetAtoms( const vector<int>& apath );
    void SetBVPath( const CCBitVec& bvpath )
    {
        m_bvpath = bvpath;
    }
    void SetAtomsFromBVPath( const CCBitVec& bvpath );
    void WriteRingAtoms( ostream& os = cout ) const;
    void WriteRingBonds( ostream& os = cout ) const;

    bool IsAtomInRing( CCAtom* atom ) const;
    bool IsAtomInRing( int idx ) const;
    bool IsBondInRing( CCBond* bond ) const;
    bool IsBondInRing( int idx ) const
    {
        return m_bvpath[idx];
    }
    bool IsAromatic();
    //////////////////////////////////////////////////////////////////////////
    bool IsNearPlanar();
    // use m_vatom to update bvpath
    void UpdateBVPath();
};

class CCAPI CCRingNode
{
protected:
    CCAtom* m_atom;
    short m_degree;
    vector<CCRingNode*> m_nbrs;
public:
    CCRingNode()
        : m_atom( NULL ), m_degree( 0 )
    {
    }
    CCRingNode( const CCRingNode& rd );
    CCRingNode& operator=( const CCRingNode& rd );
    virtual ~CCRingNode()
    {
    }

    CCAtom* GetAtom() const
    {
        return m_atom;
    }
    int GetIdx() const;
    short GetDegree() const
    {
        return m_degree;
    }
    void SetAtom( CCAtom* atom )
    {
        m_atom = atom;
    }
    void SetDegree( int degree )
    {
        m_degree = degree;
    }       
    CCRingNode* BeginNbr();
    CCRingNode* NextNbr( CCRingNode* last );
    void AddNbr( CCRingNode* nbr ); // Add or restore a neighbor
    CCRingNode* RemoveNbr( CCRingNode* nbr );
};

class CCAPI CCRingTable
{
protected:
    CCMol* m_mol;
    vector<CCRingNode*> m_rt;
    vector<CCRing*> m_vring;
public:
    CCRingTable()
    {
        m_mol = NULL;
    }
    CCRingTable( const CCRingTable& rt );
    CCRingTable& operator=( const CCRingTable& rt );
    virtual ~CCRingTable();

    CCMol* GetMol() const
    {
        return m_mol;
    }
    void SetMol( CCMol* mol )
    {
        m_mol = mol;
    }
    CCRingNode* BeginNode();
    CCRingNode* NextNode( CCRingNode* last );
    CCRingNode* GetNode( CCAtom* atom ) const;
    int FindSSSRing( CCRingNode* rootnode, CCRing& sssr );
    void DFSRing( CCRingNode* rootnode, vector<CCRing*>& allRings,
                  int sizeLimit );
    void AddNode( CCRingNode* rnode )
    {
        m_rt.push_back( rnode );
    };
    CCRingNode* RemoveNode( CCRingNode* rnode );
    void TrimBranches();
    CCRingNode* ProcessNodeTwo();
    void BreakEdge( CCRing* sssr );
    bool Empty()
    {
        return m_rt.empty();
    }
    int GetSize()
    {
        return m_rt.size();
    }
    ////////////////////////////////////////////////////////////////////////// 2008/04/23
    vector<CCRing*>& GetRings()
    {
        return m_vring;
    };
    void AddUniqueSSSR( const CCRing& sssr );
    void BFSFindSSSR();
};

#endif
