/** 
* @file 
* @brief interface for utilities
* @author Yuan ZHAO, Tiejun CHENG
* @date 2008/08/23
* @version 1.0.1 
*
* collections of small useful routines
*
* Copyright (C) 2007-2008 www.sioc-ccbg.ac.cn
* 
* This file is part of the CCBGKit project.
*/

#ifndef CCUTIL_H
#define CCUTIL_H

#include <CCConfig.h>
#include<strings.h>

/* -------------- MISC -------------- */

/** @brief print the cpu time for a statement */
#define TIMER(statement) { \
	                       clock_t _t_=clock(); \
                           statement; \
                           cout << "Time elapsed: " << clock()-_t_ << "ms\n"; \
                         }
/** @brief debug to a statement */
#define DEBUGHERE(condition) { \
                               if ( condition ) { \
	                               cout << "DEBUG TO HERE" << endl; \
                                   system( "pause" ); \
                               } \
							 }

/** @brief assert whether memory allocated successfully */
CCAPI void AssertMemAlloc( void* p );
/** @brief memory allocation error */
CCAPI void MemAllocErr();
/** @brief comment line ? */
CCAPI bool CommentLine( const string& line, char* delimstr = "#" );
/** @brief blank line ? */
CCAPI bool BlankLine( const string& line, char* delimstr = " \t\r\n" );
/** @brief return current time in ASCII mode */
CCAPI char* GetTime();
/** @brief print a line */
CCAPI void PrintLine(std::ostream &os=cout, char c='-', int num=77 );

/* -------------- STRING -------------- */
// macro for c-string utility 
#ifdef WIN32
#ifndef strcasecmp
#define strcasecmp _stricmp
#endif

#ifndef strncasecmp
#define strncasecmp _strnicmp
#endif

#ifndef snprintf
#define snprintf _snprintf
#endif
#endif

#ifndef EQ
#define EQ(a, b)     (!strcmp((a), (b)))
#endif

#ifndef EQi
#define EQi(a, b)    (!strcasecmp((a), (b)))
#endif

#ifndef EQn
#define EQn(a, b, n) (!strncmp((a), (b), (n)))
#endif

#ifndef EQni
#define EQni(a, b, n) (!strncasecmp((a), (b), (n)))
#endif

//# trim string
CCAPI void Trim( string& s, char* delimstr = " \t\n\r" );
CCAPI void Trim( char* s, char* delimstr = " \t\n\r" );
CCAPI void RTrim( string& s, char* delimstr = " \t\n\r" );
CCAPI void RTrim( char* s, char* delimstr = " \t\n\r" );
CCAPI void LTrim( string& s, char* delimstr = " \t\n\r" );
CCAPI void LTrim( char* s, char* delimstr = " \t\n\r" );
//# string conversion
CCAPI void ToUpper( std::string& );
CCAPI void ToUpper( char* );
CCAPI void ToLower( std::string& );
CCAPI void ToLower( char* );
CCAPI void InvertCase( std::string&, int );
CCAPI void InvertCase( char* );
//# compare two string case-insensitively
CCAPI bool iCaseEQ( const string& s1, const string& s2 );
//# breaking a given string $s into pieces, saved in $vstr
CCAPI int Tokenize( vector<string>& vstr, const string& s,
				   const char* delimstr = " \t\n\r", int limit = -1 );
//# whether substr is one of str (items were separated by ',')
//# e.g. IsOneOf( "Ca", "Zn,Ca,Mg" )
CCAPI bool IsOneOf( const string& substr, const string& str );
//# Case insensitive string comparison, can be used in map or set
struct CharPtrLess : public binary_function<const string&, const string&, bool>
{
	bool operator()( const string& s1, const string& s2 ) const
	{
		return strcasecmp( s1.c_str(), s2.c_str() ) < 0;
	}
};

/** @brief join several c-style string with the "\\" or "/" */
// e.g. Concat("D:", "test", "a.txt", "\\") -> "D:\\test\\a.txt"
string Concat(char *, ...);

/* -------------- MATH -------------- */
// macro for math utility
#ifndef SQUARE
#define SQUARE(x) ((x)*(x))
#endif

extern const double pINF;
extern const double nINF;
extern const double EPSILON;
extern const double PI;
extern const double RAD_TO_DEG;
extern const double DEG_TO_RAD;

CCAPI inline bool IsApprox( const double& a, const double& b,
						   const double& precision = EPSILON )
{
	return( fabs( a - b ) < precision );
}

CCAPI inline bool IsZero( const double& a, 
						 const double& precision = EPSILON )
{
	return ( fabs(a) < precision );
}
CCAPI inline bool IsBetween( const double& y, const double& x, const double& z )
{
	return ( ( y > x && y < z ) || ( y > z && y < x ) );
}

CCAPI inline bool IsNegligible( const double& a, const double& b,
							   const double &precision = EPSILON )
{
	return( fabs( a ) <= precision * fabs( b ) );
}

CCAPI bool CanBeSquared( const double& a );

//! Comparison for nan (not a number)
CCAPI inline bool IsNan( const double& a )
{
	return ( ( a ) != ( a ) );
}
// get a random number that is average distributed between min and max
// the default precision is 4. [min, max]  
CCAPI double AverageRandom( double min = 0.0, double max = 1.0,
						   int precision = 4 );

// return the index of bin which 'd' belongs to,
// otherwise return 0
int BinnedDistance( double d, 
				   double flr, double cil, double spacing );

/* Bernstein's hash
djb2 is one of the best string hash functions. it has excellent
distribution and speed on many different sets of keys and table sizes.

this algorithm (k=33) was first reported by dan bernstein many years
ago in comp.lang.c. another version of this algorithm (now favored
by bernstein) uses xor: hash(i) = hash(i - 1) * 33 ^ str[i]; the
magic of number 33 (why it works better than many other constants,
prime or not) has never been adequately explained.
*/
CCAPI unsigned long DJB2Hash( string key );
CCAPI unsigned long DJB2Hash( vector<unsigned int> key );

/*
this algorithm was created for sdbm (a public-domain reimplementation of
ndbm) database library. it was found to do well in scrambling bits, causing
better distribution of the keys and fewer splits. it also happens to be a 
good general hashing function with good distribution. the actual function 
is hash(i) = hash(i - 1) * 65599 + str[i]; what is included below is the 
faster version used in gawk. [there is even a faster, duff-device version] 
the magic constant 65599 was picked out of thin air while experimenting with
different constants, and turns out to be a prime. this is one of the 
algorithms used in berkeley db (see sleepycat) and elsewhere. 
*/
CCAPI unsigned long SDBMHash( string key );
CCAPI unsigned long SDBMHash( vector<unsigned int> key );

/* -------------- TEMPLATE -------------- */

/** @brief transform between two types */
template <class T1, class T2>
inline void TypeTrans( const T1& from, T2& to )
{
    stringstream ssm;
    ssm << from;
    ssm >> to;
}

/** @brief convert to integer type */
template <class T>
inline int toINT( const T& from )
{
    int to( 0 );
    TypeTrans( from, to );
    return to;
}

/** @brief convert to double type */
template <class T>
inline double toDBL( const T& from )
{
    double to( 0.0 );
    TypeTrans( from, to );
    return to;
}

/** @brief convert to string type */
template <class T>
inline string toSTR( const T& from )
{
    string to;
    TypeTrans( from, to );
    return to;
}

/** @brief template class for compare pair data. */
template <class T1, class T2>
class lesspairfirst
{
public:
	// functor for operator <
	bool operator() ( pair<T1,T2> a, pair<T1,T2> b ) const {
		return ( a.first < b.first );
	}
};

/** @brief template class for compare pair data. */
template <class T1, class T2>
class lesspairsecond
{
public:
	// functor for operator <
	bool operator() ( pair<T1,T2> a, pair<T1,T2> b ) const {
		return ( a.second < b.second );
	}
};

/** @brief template class for compare pair data. */
template <class T1, class T2>
class greaterpairfirst
{
public:
	// functor for operator >
	bool operator() ( pair<T1,T2> a, pair<T1,T2> b ) const {
		return ( a.first > b.first );
	}
};

/** @brief template class for compare pair data. */
template <class T1, class T2>
class greaterpairsecond
{
public:
	// functor for operator >
	bool operator() ( pair<T1,T2> a, pair<T1,T2> b ) const {
		return ( a.second > b.second );
	}
};

/** @brief A 3-element template, based on the design of the STL pair<> */
template <class T1, class T2, class T3>
struct triple
{
    //type names for the values
    typedef T1 first_type;
    typedef T2 second_type;
    typedef T3 third_type;

    //member
    T1 first;
    T2 second;
    T3 third;

    /** Default constructor
    *   T1() and T2() and T3() force initialization for built in types
    **/
    triple()
        : first( T1() ), second( T2() ), third( T3() )
    {
    }

    //! Constructor for 3 values
    triple( const T1& a, const T2& b, const T3& c )
        : first( a ), second( b ), third( c )
    {
    }

    //! Copy constructor with implicit conversions
    template <class U, class V, class W>
    triple( const triple<U, V, W>& t )
        : first( t.first ), second( t.second ), third( t.third )
    {
    }
};

/** @brief A 4-element template, based on the design of the STL pair<> */
template <class T1, class T2, class T3, class T4>
struct quad
{
    //type names for the values
    typedef T1 first_type;
    typedef T2 second_type;
    typedef T3 third_type;
    typedef T4 fourth_type;

    //member
    T1 first;
    T2 second;
    T3 third;
    T4 fourth;

    /*! default constructor
    *   T1() and T2() and T3() and T4() force initialization for built in types
    */
    quad()
        : first( T1() ), second( T2() ), third( T3() ), fourth( T4() )
    {
    }

    //! constructor for 4 values
    quad( const T1& a, const T2& b, const T3& c, const T4& d )
        : first( a ), second( b ), third( c ), fourth( d )
    {
    }

    //! copy constructor with implicit conversions
    template <class U, class V, class W, class X>
    quad( const quad<U, V, W, X>& q )
        : first( q.first ), second( q.second ), third( q.third ),
          fourth( q.fourth )
    {
    }
};
#endif


