#ifndef CCVECTOR_H
#define CCVECTOR_H

#include <CCConfig.h>

extern const double EPSILON;

class CCAPI CCVector
{
private:
    double m_x;
    double m_y;
    double m_z;

public:
    CCVector( double x = 0.0, double y = 0.0, double z = 0.0 )
        : m_x( x ), m_y( y ), m_z( z )
    {
    }
    virtual ~CCVector()
    {
    }
    void Clear()
    {
        m_x = m_y = m_z = 0.0;
    }
    CCVector( const CCVector& v )
        : m_x( v.m_x ), m_y( v.m_y ), m_z( v.m_z )
    {
    }
    CCVector& operator=( const CCVector& v )
    {
        if ( this != &v ) {
            m_x = v.m_x;
            m_y = v.m_y;
            m_z = v.m_z;
        }

        return *this;
    }

    //! Access function to x: [0], y: [1], and z: [2]
    double operator[]( int i );
    const double& x() const
    {
        return m_x;
    }
    const double& y() const
    {
        return m_y;
    }
    const double& z() const
    {
        return m_z;
    }
    double& x()
    {
        return m_x;
    }
    double& y()
    {
        return m_y;
    }
    double& z()
    {
        return m_z;
    }
    double GetX() const
    {
        return m_x;
    }
    double GetY() const
    {
        return m_y;
    }
    double GetZ() const
    {
        return m_z;
    }
    void SetX( const double x )
    {
        m_x = x;
    }
    void SetY( const double y )
    {
        m_y = y;
    }
    void SetZ( const double z )
    {
        m_z = z;
    }
    void SetXYZ( double x, double y, double z )
    {
        m_x = x; m_y = y; m_z = z;
    }
    void SetXYZ( const double* c )
    {
        m_x = c[0]; m_y = c[1]; m_z = c[2];
    }
	void SetXYZ( const CCVector& src )
    {
        m_x = src.m_x; m_y = src.m_y; m_z = src.m_z;
    }

    friend bool operator==( const CCVector& v1, const CCVector& v2 )
    {
        return ( ( fabs( v1.m_x - v2.m_x ) < EPSILON )
              && ( fabs( v1.m_y - v2.m_y ) < EPSILON )
              && ( fabs( v1.m_z - v2.m_z ) < EPSILON ) );
    }

    friend bool operator!=( const CCVector& v1, const CCVector& v2 )
    {
        return !( v1 == v2 );
    }

    //# vector addition
    friend CCVector operator+( const CCVector& v1, const CCVector& v2 )
    {
        return CCVector( v1.m_x + v2.m_x, v1.m_y + v2.m_y, v1.m_z + v2.m_z );
    };
    //# vector subtraction
    friend CCVector operator-( const CCVector& v1, const CCVector& v2 )
    {
        return CCVector( v1.m_x - v2.m_x, v1.m_y - v2.m_y, v1.m_z - v2.m_z );
    };
    //# unary minus
    friend CCVector operator-( const CCVector& v )
    {
        return CCVector( -v.m_x, -v.m_y, -v.m_z );
    };
    //# multiplication with a scalar
    friend CCVector operator*( double  c, const CCVector& v )
    {
        return CCVector( c * v.m_x, c * v.m_y, c * v.m_z );
    };
    //# multiplication with a scalar
    friend CCVector operator*( const CCVector& v, double  c )
    {
        return CCVector( c * v.m_x, c * v.m_y, c * v.m_z );
    };
    //# division by a scalar
    friend CCVector operator/( const CCVector& v, double  c )
    {
        return CCVector( v.m_x / c, v.m_y / c, v.m_z / c );
    };

    //! \return The length of the vector squared
    double Length_2() const
    {
        return m_x * m_x + m_y * m_y + m_z * m_z;
    };
    //! \return The vector length
    double Length() const
    {
        return sqrt( Length_2() );
    };

    friend void GetSp3Coordinates3( const CCVector& v1, const CCVector& vref,
                                    CCVector& v2, CCVector& v3, CCVector& v4,
                                    double angle = 109.40 );

    friend void GetSp3Coordinates2( const CCVector& v1, const CCVector& v2,
                                    CCVector& v3, CCVector& v4 );

    friend void GetSp3Coordinates( const CCVector& v1, const CCVector& v2,
                                   const CCVector& v3, CCVector& v4 );

    friend void GetSp2Coordinates2( const CCVector& v1, const CCVector& vref,
                                    CCVector& v2, CCVector& v3 );

    friend void GetSp2Coordinates( const CCVector& v1, const CCVector& v2,
                                   CCVector& v3 );

    friend void GetSp1Coordinates( const CCVector& v1, CCVector& v2 );
    //////////////////////////////////////////////////////////////////////////
    // Immediate Sum, Difference, Scalar Product
    CCVector& operator+=( const CCVector& v )
    {
        m_x += v.m_x;
        m_y += v.m_y;
        m_z += v.m_z;
        return *this;
    };
    CCVector& operator-=( const CCVector& v )
    {
        m_x -= v.m_x;
        m_y -= v.m_y;
        m_z -= v.m_z;
        return *this;
    };
    // Scalar multiplication (multiply *this by @p c)
    CCVector& operator*=( double c )
    {
        m_x *= c;
        m_y *= c;
        m_z *= c;
        return *this;
    };
    // Scalar division
    CCVector& operator/=( double c )
    {
        double inv = 1.0 / c;
        return( ( *this ) *= inv );
    };
    double distSq( const CCVector& vv ) const
    {
        double dx = m_x - vv.m_x;
        double dy = m_y - vv.m_y;
        double dz = m_z - vv.m_z;
        return( dx * dx + dy * dy + dz * dz );
    }
    bool IsApprox( const CCVector& other, double precision ) const;
	bool IsZeroVec() const;
    bool CanBeNormalized() const; // return Whether a vector can be normalized
    CCVector& Normalize();
    //////////////////////////////////////////////////////////////////////
    void RandomUnitVector();
};

/************************************************************************/
/*                     CCVector related functions                       */
/************************************************************************/
//# the square of distance between p1 and p2
CCAPI double Distance2( const CCVector& p1, const CCVector& p2 );
CCAPI double Distance2( const CCVector& p1, double x, double y, double z );
//# the distance between p1 and p2
CCAPI double Distance( const CCVector& p1, const CCVector& p2 );
CCAPI double Distance( const CCVector& p1, double c[] );
CCAPI double Distance( const CCVector& p1, double x, double y, double z );
//# dot product of two vectors
CCAPI double Dot( const CCVector& v1, const CCVector& v2 );
//# cross product of two vectors
CCAPI CCVector Cross( const CCVector& v1, const CCVector& v2 );
//# returns the angle of two vectors (in degrees 0-180)
CCAPI double AngleOfTwoVectors( const CCVector& v1, const CCVector& v2 );
//# returns the angle of three points v1-v2-v3(in degrees 0-180)
CCAPI double Angle( const CCVector& p1, const CCVector& p2, const CCVector& p3 );
//# returns the torsion angle of four points p1-p2-p3-p4 
// (in degrees -180~180)
CCAPI double Torsion( const CCVector& p1, const CCVector& p2,
                      const CCVector& p3, const CCVector& p4 );
//# return whether 3 points are in a line
CCAPI bool IsInALine( const CCVector& p1, const CCVector& p2,
                      const CCVector& p3 );
//# return the acute angle (<90)
CCAPI double AcuteAngle( double angle );
CCAPI double MinTorsion( double angle );
//# return the distance from point to line
CCAPI double DistanceFromPointToLine( const CCVector& axis,
                                      const CCVector& origin,
                                      const CCVector& point );
CCAPI double DistanceOfPointToPlane( const CCVector& p1,
                                    const CCVector& p2,
                                    const CCVector& p3,
                                    const CCVector& p4 );
//# return angle of two planes
CCAPI double AngleOfTwoPlanes( const CCVector& p1, const CCVector& p2,
                               const CCVector& p3, const CCVector& q1,
                               const CCVector& q2, const CCVector& q3 );
//# return new point that rotates around an axis
CCAPI CCVector RotatePoint( const CCVector& axis, const CCVector& origin,
                            const CCVector& start, double angle );
//////////////////////////////////////////////////////////////////////////
EXTERN const CCVector VZero;
EXTERN const CCVector VX;
EXTERN const CCVector VY;
EXTERN const CCVector VZ;
/////////////////////////////////////////////////////////////////////////
// Vector Analysis                                                     //
/////////////////////////////////////////////////////////////////////////
/*! Calculate the derivative of a vector length. 
*   The vector is given by a - b, 
* the length of this vector rab = sqrt(ab.x^2 + ab.y^2 + ab.z^2).
* \param a atom a (coordinates), will be changed to -drab/da
* \param b atom b (coordinates), will be changed to  drab/db
* \return The distance between a and b 
* (bondlength for bond stretching, separation for vdw, electrostatic)
*/
double VectorLengthDerivative( CCVector& a, CCVector& b );
/*! Calculate the derivative of a angle a-b-c. The angle is 
*   given by dot(ab,cb)/rab*rcb.
* \param a atom a (coordinates), will be changed to -dtheta/da
* \param b atom b (coordinates), will be changed to -dtheta/db
* \param c atom c (coordinates), will be changed to -dtheta/dc
* \return The angle between a-b-c
*/
double VectorAngleDerivative( CCVector& a, CCVector& b, CCVector& c );
/*! Calculate the derivative of a out of plane.
* \param a atom a (coordinates), will be changed to -dr/da
* \return The distance of a to b-c-d plane
*/
double VectorOOPDerivative( CCVector& a, CCVector& b, CCVector& c, CCVector& d );
/*! Calculate the derivative of a torsion angle a-b-c-d. The torsion
*   angle is given by dot(corss(ab,bc),cross(bc,cd)/rabbc*rbccd.
* \param a atom a (coordinates), will be changed to -dtheta/da
* \param b atom b (coordinates), will be changed to -dtheta/db
* \param c atom c (coordinates), will be changed to -dtheta/dc
* \param d atom d (coordinates), will be changed to -dtheta/dd
* \return The tosion angle for a-b-c-d
*/
double VectorTorsionDerivative( CCVector& a, CCVector& b, CCVector& c,
                                CCVector& d );

CCVector RandomVector();
//////////////////////////////////////////////////////////////////////////
#endif
