/** @file
* @brief set aromatic flag on atoms and bonds without changing bond order.
* @author Yuan ZHAO, Tiejun CHENG
* @date 2008/08/22
* @version 1.0.1
*
*/

#include <ccbg/CCAromatic.h>
#include <ccbg/CCMol.h>

/** @brief global CCAromaticTyper object for interpret aromaticity */
CCAromaticTyper gAromaticTyper;

/** @brief whether consider aromaticity in odd ring */
const bool ODD_RING_AROMATICITY = true;

void CCAromaticTyper::Clear()
{
    m_exAtoms.Clear();
    m_rootAtoms.Clear();
    m_visit.Clear();
    m_marked.Clear();
    m_piElec = make_pair( 0, 0 );
    m_maxDepth = 0;
    m_bgnBond = NULL;
}

CCAromaticTyper::CCAromaticTyper()
    : m_piElec( make_pair( 0, 0 ) ), m_maxDepth( 0 ), m_bgnBond( NULL )
{
    m_odd = ODD_RING_AROMATICITY;
}

/**
* @brief assign aromaticity on molecule.
* @param[in] mol molecule for assigning aromaticity.
*
* The atoms and bonds in input molecule will be assign
* aromatic flag if they are in aromatic ring. The max
* aromatic ring size is set as 14.
*/
void CCAromaticTyper::AssignAromaticity( CCMol& mol )
{
    // restore to initial state
    // necessary, coz there is only one AromaticTyper
    Clear();

    // unset aromaticity for all atoms and bonds
    FOR_ATOMS_OF_MOL( atom, mol ) {
        atom->UnsetAromatic();
    }
    FOR_BONDS_OF_MOL( bond, mol ) {
        bond->UnsetAromatic();
    }

    if ( !ExcludeNonAromaticAtom( mol ) ) {
        return;
    }

    if ( !PropagateAromaticAtom( mol ) ) {
        return;
    }

    if ( !SelectRootAtom( mol ) ) {
        return;
    }

    int numAtoms = mol.GetNumAtom();
    m_visit.Resize( numAtoms + 1 );
    m_marked = m_exAtoms;

    if ( m_marked.Count() < numAtoms ) {
        FOR_ATOMS_OF_MOL( atom, mol ) {
            if ( m_rootAtoms[atom->GetIdx()] ) {
                // try as large as 14 atoms
                CheckAromaticity( &*atom, 14 );
                /*// allow early termination when all atoms have been marked before
                        if ( m_marked.Count() >= numAtoms ) {
                            break;
                        }*/
            }
        }
    }

    /*// cleanup -- check for ring bonds where both atoms are aromatic
    // but for some reason the bond is not (e.g., between a fused system)
    FOR_BONDS_OF_MOL(bond, mol)
    {
    if (bond->IsInRing() &&
    (bond->GetBeginAtom())->IsAromatic() &&
    (bond->GetEndAtom())->IsAromatic())
    bond->SetAromatic();
    }*/
}

/**
* @brief exclude atoms that can't be aromatic.
* @param[in] mol molecule for excluding non-aromatic atoms.
* @return true if there is potential aromatic atom.
*         false if all the atoms are excluded.
*
* The atoms in the following situation will be excluded:<br>
*   1. atoms not in ring<br>
*   2. atoms in 3 or 4 membered ring<br>
*   3. atoms whose implicit connections greater than 3<br>
*   4. sp1 atoms<br>
*   5. metal atoms<br>
* The excluded atoms index are saved in m_exAtoms.
*/
bool CCAromaticTyper::ExcludeNonAromaticAtom( CCMol& mol )
{
    // mark atoms that are definitely not aromatic
    int numAtom = mol.GetNumAtom();
    m_exAtoms.Resize( numAtom + 1 );

    FOR_ATOMS_OF_MOL( atom, mol ) {
        if ( !atom->IsInRing() ||         // exclude atom not in ring
            atom->GetImpVal() > 3 ||
            // exlcude atom with >3 implicit connections (sp3 C and sulfoneS)
            atom->GetHyb() == 1 ||         // exclude sp1 center ???
            atom->IsMetal() )          // exclude metal
        {
            m_exAtoms.Set( atom->GetIdx() );
        }
    }

    if ( m_exAtoms.Count() == numAtom ) // no aromatic atoms in this molecule
    {
        return false;
    }

    // exclude small ring
    FOR_RINGS_OF_MOL( ring, mol ) {
        if ( ring->GetSize() < 5 ) {
            vector<int>::iterator l;
            vector<int> ra = ring->GetAtomPath();
            for ( l = ra.begin(); l != ra.end(); ++l ) {
                m_exAtoms.Set( *l );
            }
        }
    }

    if ( m_exAtoms.Count() == numAtom ) // no aromatic atoms in this molecule
    {
        return false;
    }

    return true;
}

/**
* @brief propagate potential aromatic atoms in groups
* @param[in] mol molecule for checking.
* @return true if there is potential aromatic atom.
*         false otherwise.
*
* check whether every potential aromatic atom has more than two
* potential aromatic neighboring atoms. If not, the atom will
* be put into the excluded aromatic atom list. It will update
* variable m_exAtoms.
*/
bool CCAromaticTyper::PropagateAromaticAtom( CCMol& mol )
{
    // propagate aromatic atoms
    CCAtom* root = NULL;

    FOR_ATOMS_OF_MOL( atom, mol ) {
        if ( !m_exAtoms[atom->GetIdx()] ) {
            queue<CCAtom*> q;
            q.push( &*atom );

            while ( !q.empty() ) {
                root = q.front();
                q.pop();

                int count = 0;
                FOR_NBORS_OF_ATOM( nbr, root ) {
                    if ( /*nbr->IsInRing() && */!m_exAtoms[nbr->GetIdx()] ) {
                        count++;
                    }
                }

                if ( count < 2 ) {
                    m_exAtoms.Set( root->GetIdx() );
                    if ( count == 1 ) {
                        FOR_NBORS_OF_ATOM( nbr, root ) {
                            if ( /*nbr->IsInRing()&& */!m_exAtoms[nbr->GetIdx()] ) {
                                q.push( &*nbr );
                            }
                        }
                    }
                }
            }
        }
    }

    // no aromatic atoms in this molecule
    if ( m_exAtoms.Count() == mol.GetNumAtom() ) {
        return false;
    }

    return true;
}

/**
* @brief select a root atom in each ring
* @param[in] mol molecule for checking.
* @param[in] avoidInnerAtom whether avoid two root atoms in the same ring (default true).
* @return true if at least one root atom exist,
*         false otherwise.
*
* first make the root atom is the begin or end atom of the closure
* bond. If avoidInnerAtom parameter is true, try its best to make the
* root atom has only two ring bonds and avoid two root atoms in the
* same ring.
*/
bool CCAromaticTyper::SelectRootAtom( CCMol& mol,
                                      bool avoidInnerAtom/* =true */ )
{
    // select a root atom
    CCAtom* atom, * nbr, * nbr2;
    CCAtomIter i;
    CCBondIter j, nbrIter, nbr2Iter;
    m_rootAtoms.Resize( mol.GetNumAtom() + 1 );

    int rootAtom;
    vector<int> rootAtoms;
    vector<int>::iterator k;

    FOR_BONDS_OF_MOL( bond, mol ) {
        if ( bond->IsClosure() ) {
            // BASIC APPROACH
            // pick beginning atom at closure bond
            // this is really ready, isn't it ! ;-)
            if ( !m_exAtoms[bond->GetBeginAtomIdx()] ) {
                rootAtom = bond->GetBeginAtomIdx();
            }
            else if ( !m_exAtoms[bond->GetEndAtomIdx()] ) {
                rootAtom = bond->GetEndAtomIdx();
            }
            else {
                continue;
            } // go for next closure bond

            // record root atom
            m_rootAtoms.Set( rootAtom );
            rootAtoms.push_back( rootAtom );
        }
    }

    // EXTENDED APPROACH
    if ( avoidInnerAtom ) {
        for ( int i = 0; i < rootAtoms.size(); ++i ) {
            int ringNbrs = 0;
            rootAtom = rootAtoms[i];
            atom = mol.GetAtom( rootAtom );

            for ( nbr = atom->BeginNbrAtom( nbrIter );
                  nbr;
                  nbr = atom->NextNbrAtom( nbrIter ) ) {
                if ( /*nbr->IsInRing() &&*/!m_exAtoms[nbr->GetIdx()] ) {
                    ringNbrs++;
                }
            }

            // if this atom has more than 2 neighbor ring atoms
            // we could get trapped later when traversing cycles
            // which can cause aromaticity false detection
            if ( ringNbrs > 2 ) {
                int newRoot = -1;

                // try to find another root atom
                FOR_RINGS_OF_MOL( ring, mol ) {
                    if ( !ring->IsAtomInRing( rootAtom ) ) {
                        continue;
                    } // go for next ring

                    vector<int> tmpPath = ring->GetAtomPath();
                    bool checkThisRing = false;
                    int rootAtomNum = 0;

                    // avoiding two root atoms in one ring!
                    for ( k = tmpPath.begin(); k != tmpPath.end(); ++k ) {
                        if ( m_rootAtoms[*k] ) {
                            rootAtomNum++;
                        }
                        if ( rootAtomNum >= 2 ) {
                            break;
                        }
                    }

                    if ( rootAtomNum < 2 ) {
                        checkThisRing = true;
                    }

                    // check ring for getting another
                    // root atom to avoid aromaticity typer problems
                    if ( checkThisRing ) {
                        // check if we can find another root atom
                        for ( k = tmpPath.begin(); k != tmpPath.end(); ++k ) {
                            ringNbrs = 0;
                            for ( nbr2 = mol.GetAtom( *k )->BeginNbrAtom( nbr2Iter );
                                  nbr2;
                                  nbr2 = mol.GetAtom( *k )->NextNbrAtom( nbr2Iter ) ) {
                                if ( /*nbr->IsInRing() &&*/!m_exAtoms[nbr2->GetIdx()] ) {
                                    ringNbrs++;
                                }
                            }

                            // if the number of neighboured heavy atoms is also
                            // the number of neighboured ring atoms, the aromaticity
                            // typer could be stuck in a local traversing trap
                            if ( ringNbrs
                              <= 2/* && ring->IsAtomInRing((mol->GetAtom(*k)->GetIdx()))*/
                              && *k
                              != rootAtom ) {
                                newRoot = *k;
                                break;
                            }
                        }
                    }

                    // found and reset root atom
                    if ( ( newRoot != -1 )/* && (rootAtom != newRoot)*/ ) {
                        // unset root atom
                        m_rootAtoms.Reset( rootAtom );
                        // pick new root atom
                        m_rootAtoms.Set( newRoot );
                        break;
                    }
                }
            }
        }
    }

    return !m_rootAtoms.None();
}

/**
* @brief check the aromaticity starting from atom
* @param[in] atom atom for checking aromaticity.
* @param[in] depth max size of aromatic ring.
*
* all non-ring atoms should be excluded before.
* The aromatic perceiving will begin from the input atom. It
* search the rings that contain the input atom and calculate
* the pi electrons of the ring. If it satisfies 4n+2 rule, the
* input atom and corresponding bonds will be set aromatic flag.
* The max size of the ring is set as depth.
*
*/
void CCAromaticTyper::CheckAromaticity( CCAtom* atom, int depth )
{
    // init
    m_piElec = atom->GetPIelectron();
    m_maxDepth = depth;

    CCAtom* nbr;
    CCBondIter i;

    for ( nbr = atom->BeginNbrAtom( i ); nbr; nbr = atom->NextNbrAtom( i ) ) {
        if ( /*(*i)->IsInRing() &&*/ !m_exAtoms[nbr->GetIdx()] )
            // all non-ring atoms have been excluded before
        {
            m_bgnBond = *i;
            if ( TraverseCycle( atom, nbr, *i, depth - 1 ) ) {
                // captured an aromatic atom
                m_marked.Set( atom->GetIdx() );
                atom->SetAromatic();
                ( *i )->SetAromatic();
            }

            /*
            // allow early termination if all atom have been marked
            if ( m_marked.Count() >= atom->GetMol()->GetNumAtom() ) {
                break;
            }*/
        }
    }
}

/**
* @brief depth-first-search traverse
* @param[in] root root atom.
* @param[in] atom neighboring atom.
* @param[in] prev previous bond.
* @param[in] depth max size of aromatic ring.
* @return whether the traverse cycle is aromatic
*
* iteratively depth step to search a ring that contains root,
* atom, and prev bond. Then check the aromaticity of the ring.
* return true if the root, atom and prev bond are in an
* aromatic ring.
*/
bool CCAromaticTyper::TraverseCycle( CCAtom* root, CCAtom* atom, CCBond* prev,
                                     int depth )
{
    if ( atom == root ) // a cycle is found
    {
        if ( !m_odd && ( m_maxDepth - depth ) % 2 == 1 ) // a odd ring
        {
            return false;
        }

        // check this ring for two continual single bonds
        // external-ring double bond can be recognized
        if ( !CheckConsecutiveSingleBond( prev, atom, m_bgnBond ) ) {
            return false;
        }

        for ( int i = m_piElec.first; i <= m_piElec.second; ++i ) {
            if ( i % 4 == 2 && i > 2 ) {
                return true;
            }
        } // aromatic ring
        return false;
    }

    if ( depth == 0
        /*|| m_exAtoms[atom->GetIdx()] || m_visit[atom->GetIdx()]*/ ) {
        return false;
    }

    depth--;
    bool result = false;
    pair<int, int> pi = atom->GetPIelectron();

    m_piElec.first += pi.first;
    m_piElec.second += pi.second;

    m_visit.Set( atom->GetIdx() );

    CCAtom* nbr; CCBondIter i;
    for ( nbr = atom->BeginNbrAtom( i ); nbr; nbr = atom->NextNbrAtom( i ) ) {
        if ( *i != prev && !m_visit[nbr->GetIdx()] &&
            /*(*i)->IsInRing() &&*/ !m_exAtoms[nbr->GetIdx()] ) {
            if ( !CheckConsecutiveSingleBond( prev, atom, *i ) ) {
                continue;
            }

            if ( TraverseCycle( root, nbr, ( *i ), depth ) ) {
                result = true;
                ( *i )->SetAromatic();
                //(*i)->SetOrder(5);
            }
        }
    }

    m_visit.Reset( atom->GetIdx() );
    if ( result ) {
        // captured an aromatic atom
        m_marked.Set( atom->GetIdx() );
        atom->SetAromatic();
    }

    // restore pi electron
    m_piElec.first -= pi.first;
    m_piElec.second -= pi.second;

    return result;
}

/**
* @brief center atom has two consecutive sing bonds ?
* @param[in] prev previous bond of center atom.
* @param[in] center center atom.
* @param[in] next next bond of center atom.
* @return true only if the center atom is one of N, sp3 O,
*         sp3 or sp2 S, sp2 Se and both prev and next are
*         single bonds <BR>
*         false otherwise.
*
* this function can be used to exclude center atoms such as
* C atoms with two consecutive single bonds and a ext-ring
* double bonds, atoms of this kind are not considered as
* aromatic in our algorithm.
*/
bool CCAromaticTyper::CheckConsecutiveSingleBond( CCBond* prev, CCAtom* center,
                                               CCBond* next )
{
    if ( prev->GetOrder() == 1 && next->GetOrder() == 1 ) {
        if ( !center->IsNitrogen()
          && !center->IsSp3Oxygen()
          && !center->IsSp3Sulfur()
          // >S=O (may also be S=,
		  // but OK because it can't pass unaromaticity check
          && !center->IsSp2Sulfur()
          // >Se=O (may also be Se=,
		  // but OK because it can't pass unaromaticity check
          && !center->IsSp2Selenium() ) {
            return false;
        }
    }
    return true;
}


