/** @file 
* @brief CCAtom class and relevant functions.
* @author Yuan ZHAO, Tiejun CHENG
* @date 2008/08/23
* @version 1.0.1
*
*/
#include <ccbg/CCAtom.h>
#include <ccbg/CCMol.h>

/** @brief CCAtom constructor */
CCAtom::CCAtom()
    : m_mol( NULL ), m_idx( 0 ), m_ele( 0 ), //m_name,
                   //m_type,
                   m_impval( 0 ), m_coor( 0.0, 0.0, 0.0 ), m_hyb( 0 ),
      m_part( 0 ), m_fcharge( 0 ), m_pcharge( 0.0 ), //m_vbond,
                   m_residue( NULL ), m_isotope( 0 ), m_spin( 0 )
{
    CCBase::CCBase();
}

/** 
* @brief CCAtom copy constructor.
* @param[in] src atom for copy.
*/
CCAtom::CCAtom( const CCAtom& src )
    : m_mol( src.m_mol ), m_idx( src.m_idx ), m_ele( src.m_ele ),
      m_name( src.m_name ), m_type( src.m_type ), m_impval( src.m_impval ),
      m_coor( src.m_coor ), m_hyb( src.m_hyb ), m_part( src.m_part ),
      m_fcharge( src.m_fcharge ), m_pcharge( src.m_pcharge ),
      m_vbond( src.m_vbond ), m_residue( src.m_residue ),
      m_spin( src.m_spin ), m_isotope( src.m_isotope )
{
    CCBase::Clone( src );
}

/** 
* @brief CCAtom copy assignment operator.
* @param[in] src atom for copy assignment.
*/
CCAtom& CCAtom::operator =( const CCAtom& src )
{
    if ( this != &src ) {
        CCBase::Clone( src );
        m_mol = src.m_mol;
        m_idx = src.m_idx;
        m_ele = src.m_ele;
        m_name = src.m_name;
        m_type = src.m_type;
        m_impval = src.m_impval;
        m_coor = src.m_coor;
        m_hyb = src.m_hyb;
        m_part = src.m_part;
        m_fcharge = src.m_fcharge;
        m_pcharge = src.m_pcharge;
        m_vbond = src.m_vbond;
        m_residue = src.m_residue;
        m_spin = src.m_spin;
        m_isotope = src.m_isotope;
    }   
    return *this;
}

/** @brief CCAtom destructor */
CCAtom::~CCAtom()
{
	CCBase::DestroyGD();

	if ( m_residue != NULL ) {
		m_residue->RemoveAtom( this );
	}
}

/** @brief clear members of CCAtom */
void CCAtom::Clear()
{
	CCBase::Clear();
	m_mol = NULL;
	m_idx = 0;
	m_ele = 0;
	m_name.clear();
	m_type.clear();
	m_impval = 0;
	m_coor.Clear();
	m_hyb = 0;
	m_part = 0;
	m_fcharge = 0;
	m_pcharge = 0.0;
	m_vbond.clear();
	m_residue = NULL;
	m_isotope = 0;
	m_spin = 0;
}

/**
* @brief add neighboring bond
* @param[in] bond bond to be added to the atom
* @return true if bond was successfully added
*
* If atom is not in the bond or the bond has already
* been in the atom's neighboring bond list, it will
* return false. Else add the bond to atom's neighboring
* bond list and return true
*/
bool CCAtom::AddBond( CCBond* bond )
{
    if ( this != bond->GetBeginAtom() && this != bond->GetEndAtom() ) {
        return false;
    }

    if ( std::find( m_vbond.begin(), 
		            m_vbond.end(),
					bond ) == m_vbond.end() ) {
        m_vbond.push_back( bond );
        return true;
    }
    else {
        return false;
    }
}

/**
* @brief remove a bond from atom's neighboring bonds list.
* @param[in] bond bond to be removed.
* @return true if bond was successufully removed.
*
* If the bond was not in the atom's neighboring bond list, it
* will return false. Else remove the bond from atom's neighboring
* bond list and return true.
*/
bool CCAtom::RemoveBond( CCBond* bond )
{
    CCBondIter it = std::find( m_vbond.begin(), m_vbond.end(), bond );
    if ( it != m_vbond.end() ) {
        m_vbond.erase( it );
        return true;
    }
    else {
        return false;
    }
}

/**
* @brief check whether this atom was bonded to atom.
* @param[in] atom atom for checking connection.
* @return whether two atoms were bonded.
*
* Check whether atom was existing in this atom's neighboring
* atom list.
*/
bool CCAtom::IsConnected( CCAtom* atom ) const
{
    FOR_NBORS_OF_ATOM( nbr, ( CCAtom * )this ) {
        if ( &*nbr == atom ) {
            return true;
        }
    }

    return false;
}

/**
* @brief check whether the atom is chiral atom.
* @return whether the atom is chiral atom.
*
* If atom had ATOM_IS_CHIRAL flag, it will return true.
* Else it will touch off the PerceiveChirality function
* if the molecular chirality was not perceived.
*
* @see CCMol::PerceiveChirality()
*/
bool CCAtom::IsChiral()
{
	if ( HasFlag( ATOM_IS_CHIRAL ) ) {
		return true;
	}

	if ( m_mol && !m_mol->HasFlag( MOL_CHIRALITY_PERCEIVED ) ) {
		m_mol->PerceiveChirality();
	}

	return HasFlag( ATOM_IS_CHIRAL );
}

/**
* @brief check whether the atom is stereo atom.
* @return whether the atom is stereo atom.
*
* If atom had ATOM_IS_STEREO flag, it will return true.
* Else it will touch off the PerceiveStereo function
* if the molecular stereo was not perceived.
*
* @see CCMol::PerceiveStereo()
*/
bool CCAtom::IsStereo()
{
	if ( HasFlag( ATOM_IS_STEREO ) ) {
		return true;
	}

	if ( m_mol && !m_mol->HasFlag( MOL_STEREO_PERCEIVED ) ) {
		m_mol->PerceiveStereo();
	}

	return HasFlag( ATOM_IS_STEREO );
}

/**
* @brief get first neighboring atom.
* @param[out] i current iterator of neighboring bond
* @return the first neighboring atom
*
* i was first iterated to the first neighboring bond, through i
* the neighboring atom can be returned by (*i)->GetNbrAtom(this). 
* If i reaches the end of m_vbond, return NULL.
* 
* @see CCBond::GetNbrAtom(), CCAtom::NextNbrAtom()
*/
CCAtom* CCAtom::BeginNbrAtom( vector<CCBond*>::iterator& i )
{
    i = m_vbond.begin();
    return ( i == m_vbond.end()
           ? NULL
           : ( *i )->GetNbrAtom( this ) );
}

/**
* @brief get next neighboring atom.
* @param[out] i current iterator of neighboring bond
* @return the next neighboring atom.
*
* i was first iterated to the next neighboring bond, through i
* the neighboring atom ban be returned by (*i)->GetNbrAtom(this).
* If i reaches the end of m_vbond, return NULL. 
* 
* @see CCBond::GetNbrAtom(), CCAtom::BeginNbrAtom()
*/
CCAtom* CCAtom::NextNbrAtom( vector<CCBond*>::iterator& i )
{
    ++i;
    return ( i == m_vbond.end()
           ? NULL
           : ( *i )->GetNbrAtom( this ) );
}

/**
* @brief get first neighboring bond
* @param[out] i current iterator of neighboring bond
* @return the first neighboring bond
*
* i was first iterated to the first neighboring bond, return
* that neighboring bond. If i reaches the end of m_vbond, return NULL.
* 
* @see CCAtom::NextNbrBond()
*/
CCBond* CCAtom::BeginNbrBond( vector<CCBond*>::iterator& i )
{
    i = m_vbond.begin();
    return ( i == m_vbond.end() ? NULL : *i );
}

/**
* @brief get next neighboring bond
* @param[out] i current iterator of neighboring bond
* @return the next neighboring bond
*
* i was first iterated to the next neighboring bond, return
* that neighboring bond. If i reaches the end of m_vbond, return NULL.
* 
* @see CCAtom::BeginNbrBond()
*/
CCBond* CCAtom::NextNbrBond( vector<CCBond*>::iterator& i )
{
    ++i;
    return ( i == m_vbond.end() ? NULL : *i );
}

/**
* @brief get nth neighboring atom
* @param[in] n nth neighbor
* @return NULL if n<1 or n>max neighboring atom,
*         otherwise, return the nth neighboring atom
*/
CCAtom* CCAtom::GetNbrAtom( int n )
{
    if ( n > m_vbond.size() || n < 1 ) {
        return NULL;
    }
    else {
        return m_vbond[n - 1]->GetNbrAtom( this );
    }
}

/**
* @brief get nth neighboring heavy atom
* @param[in] n nth neighbor
* @return the nth neighboring heavy atom.
*
* If this atom had more than n neighboring heavy 
* atoms, it will return the nth neighboring heavy
* atom. Otherwise return NULL.
*/
CCAtom* CCAtom::GetNbrHvyAtom( int n )
{
    if ( n > m_vbond.size() || n < 1 ) {
        return NULL;
    }
    else {
        int i = 0;
        FOR_NBORS_OF_ATOM( nbr, ( CCAtom * )this ) {
            if ( nbr->IsHvyAtom() && ( ( ++i ) == n ) ) {
                return &*nbr;
            }
        }
    }
    return NULL;
}

/**
* @brief get nth neighboring bond
* @param[in] n nth neighbor
* @return NULL if n<1 or n>max neighboring bond,
*         otherwise, return the nth neighboring bond
*/
CCBond* CCAtom::GetNbrBond( int n )
{
    if ( n > m_vbond.size() || n < 1 ) {
        return NULL;
    }
    else {
        return m_vbond[n - 1];
    }
}

/**
* @brief get formal charge of the atom
* @return the formal charge of the atom.
*
* It will touch off PerceiveInternalAtomBondTypes funtion
* if the molecule had MOL_AUTOFCHARGE_TRUE flag and 
* the flag MOL_AUTOFCHARGE_PERCEIVED was not assigned.
*
* @see CCMol::PerceiveInternalAtomBondTypes()
*/
int CCAtom::GetFormalCharge()
{
    if ( m_mol
      && m_mol->HasFlag( MOL_AUTOFCHARGE_TRUE )
      && !m_mol->HasFlag( MOL_AUTOFCHARGE_PERCEIVED ) ) {
        m_mol->SetFlag( MOL_AUTOFCHARGE_PERCEIVED );
        m_mol->PerceiveInternalAtomBondTypes();
    }

    return m_fcharge;
}

/**
* @brief get element symbol
* @return element symbol of the atom.
*
* Use global variable gElementalTable to get element
* symbol from atomic number.
*
* @see CCElementalTable::GetSymbol()
*/
string CCAtom::GetSymbol() const
{
    return gElementalTable.GetSymbol( m_ele );
}

/**
* @brief get atomic weight
* @return atomic weight of the atom.
*
* Use global variable gElementalTable to get atomic
* weight from atomic number.
*
* @see CCElementalTable::GetMass()
*/
double CCAtom::GetMass() const
{
    return gElementalTable.GetMass( m_ele );
}

/**
* @brief get covalent radius
* @return covalent radius of the atom.
*
* Use global variable gElementalTable to get covalent
* radius from atomic number.
*
* @see CCElementalTable::GetCovalentRad()
*/
double CCAtom::GetCovRad() const
{
    return gElementalTable.GetCovalentRad( m_ele );
}

/**
* @brief get van de waals radius
* @return van de waals radius of the atom.
*
* Use global variable gElementalTable to get van de 
* waals radius from atomic number.
*
* @see CCElementalTable::GetVdwRad()
*/
double CCAtom::GetVdwRad() const
{
    return gElementalTable.GetVdwRad( m_ele );
}

/**
* @brief get the number of max bonded atoms
* @return the number of max bonded atoms.
*
* Use global variable gElementalTable to get the
* number of max bonded atoms from atomic number.
*
* @see CCElementalTable::GetMaxBond()
*/
int CCAtom::GetMaxBond() const
{
    return gElementalTable.GetMaxBond( m_ele );
}

/**
* @brief get the electro negativity
* @return the electro negativity.
*
* Use global variable gElementalTable to get electro
* negativity from atomic number.
*
* @see CCElementalTable::GetElectroNeg()
*/
double CCAtom::GetEleNeg()
{
	return gElementalTable.GetElectroNeg( m_ele );
}

/** 
* @brief get atomic parent residue name
*/
string CCAtom::GetResName() const
{
	return ( m_residue ? m_residue->GetResName() : "" );
}

/**
* @brief set internal atom type as atomic symbol
*
* First use global variable gElementalTable to get atomic
* symbol from atomic number. Then set the internal atomic
* type as the atomic symbol.
*
* @see CCElementalTable::GetSymbol()
*/
void CCAtom::SetTypeFromSymbol()
{
    m_type = gElementalTable.GetSymbol( m_ele );
}

/**
* @brief set internal atom type as atomic symbol plus ".1"
*
* First use global variable gElementalTable to get atomic
* symbol from atomic number. Then set the internal atomic
* type as the atomic symbol plus ".1".
*
* @see CCElementalTable::GetSymbol()
*/
void CCAtom::SetSp1Type()
{
    m_type = GetSymbol() + ".1";
}

/**
* @brief set internal atom type as atomic symbol plus ".2"
*
* First use global variable gElementalTable to get atomic
* symbol from atomic number. Then set the internal atomic
* type as the atomic symbol plus ".2".
*
* @see CCElementalTable::GetSymbol()
*/
void CCAtom::SetSp2Type()
{
    m_type = GetSymbol() + ".2";
}

/**
* @brief set internal atom type as atomic symbol plus ".3"
*
* First use global variable gElementalTable to get atomic
* symbol from atomic number. Then set the internal atomic
* type as the atomic symbol plus ".3".
*
* @see CCElementalTable::GetSymbol()
*/
void CCAtom::SetSp3Type()
{
    m_type = GetSymbol() + ".3";
}

/**
* @brief get internal atom type
* @return internal atom type
*
* First check whether the parent molecule has the flag of
* MOL_INT_TYPE_PERCEIVED. If not, call molecular class
* function PerceiveInternalAtomBondTypes() to perceive
* internal atom and bond types. Then return the internal 
* atom type of the atom.
*
* @see CCMol::PerceiveInternalAtomBondTypes()
*/
string CCAtom::GetType() const
{
    if ( m_mol && !m_mol->HasFlag( MOL_INT_TYPE_PERCEIVED ) ) {
        m_mol->PerceiveInternalAtomBondTypes();
    }

    return m_type;
}

/**
* @brief aromatic atom ?
* @return true if the atom is aromatic.
*
* First check whether the parent molecule has aromaticity
* perceived (has MOL_AROMATICITY_PERCEIVED flag). If not, 
* call CCMol::PerceiveAromaticity(), then check whether 
* the atom has ATOM_IS_AROMATIC flag.
*
* @see CCMol::PerceiveAromaticity()
*/
bool CCAtom::IsAromatic()
{
    if ( m_mol && !m_mol->HasFlag( MOL_AROMATICITY_PERCEIVED ) ) {
        m_mol->PerceiveAromaticity();
    }

    return HasFlag( ATOM_IS_AROMATIC );
}

/**
* @brief check whether the atom is in ring
* @return whether the atom is in ring
*
* First check whether the atom has ATOM_IS_IN_RING flag,
* If the flag exist, return true. Else check whether it 
* needs to perceive ring atoms and bonds for the parent
* molecule. If the parent molecule had not assigned 
* MOL_RING_PERCEIVED flag, call PerceiveRingAtomsAndBonds()
* function to perceive ring atoms and bonds. Then check 
* whether the atom had ATOM_IS_IN_RING flag.
*
* @see CCMol::PerceiveRingAtomsAndBonds()
*/
bool CCAtom::IsInRing()
{
    if ( HasFlag( ATOM_IS_IN_RING ) ) {
        return true;
    }

    if ( m_mol && !m_mol->HasFlag( MOL_RING_PERCEIVED ) ) {
        m_mol->PerceiveRingAtomsAndBonds();
    }

    return HasFlag( ATOM_IS_IN_RING );
}

/**
* @brief hbond donor ?
* @return whether the atom is hydrogen bond donor atom.
*
* First check whether the atom has ATOM_IS_HBDONOR flag, If the 
* flag exist, return true. Else check whether it needs to perceive
* hydrogen bond donor and acceptor for the parent molecule. If the
* parent molecule has no MOL_HBDONORACCEPTOR_PERCEIVED flag, call 
* CCMol::PerceiveHBDonorAcceptor(), then check again whether the 
* atom has ATOM_IS_HBDONOR flag.
*
* @see CCMol::PerceiveHBDonorAcceptor()
*/
bool CCAtom::IsHBDonor()
{
    if ( HasFlag( ATOM_IS_HBDONOR ) ) {
        return true;
    }

    if ( m_mol && !m_mol->HasFlag( MOL_HBDONORACCEPTOR_PERCEIVED ) ) {
        m_mol->PerceiveHBDonorAcceptor();
    }

    return HasFlag( ATOM_IS_HBDONOR );
}

/**
* @brief hbond acceptor ?
* @return whether the atom is hydrogen bond donor atom.
*
* First check whether the atom has ATOM_IS_HBACCEPTOR flag, If the 
* flag exist, return true. Else check whether it needs to perceive
* hydrogen bond donor and acceptor for the parent molecule. If the
* parent molecule has no MOL_HBDONORACCEPTOR_PERCEIVED flag, call 
* CCMol::PerceiveHBDonorAcceptor(), then check again whether the 
* atom has ATOM_IS_HBACCEPTOR flag.
*
* @see CCMol::PerceiveHBDonorAcceptor()
*/
bool CCAtom::IsHBAcceptor()
{
    if ( HasFlag( ATOM_IS_HBACCEPTOR ) ) {
        return true;
    }

    if ( m_mol && !m_mol->HasFlag( MOL_HBDONORACCEPTOR_PERCEIVED ) ) {
        m_mol->PerceiveHBDonorAcceptor();
    }

    return HasFlag( ATOM_IS_HBACCEPTOR );
}

/** 
* @brief check whether the atom is donor hydrogen
* @return whether the atom is donor hydrogen
*
* whether the atom is hydrogen and it bonded to a
* hydrogen bond donor atom.
*/
bool CCAtom::IsDornorH()
{
    if ( IsHydrogen() ) {
        FOR_NBORS_OF_ATOM( nbr, *this ) {
            if ( nbr->IsHBDonor() ) {
                return true;
            }
        }
    }
    return false;
}

/** 
* @brief check whether the atom is acceptor hydrogen
* @return whether the atom is acceptor hydrogen
*
* whether the atom is hydrogen and it bonded to a
* hydrogen bond acceptor atom.
*/
bool CCAtom::IsAcceptorH()
{
	if ( IsHydrogen() ) {
		FOR_NBORS_OF_ATOM( nbr, *this ) {
			if ( nbr->IsHBAcceptor() ) {
				return true;
			}
		}
	}
	return false;
}

/**
* @brief get implicit valence (implicit bonded atoms)
* @return number of implicitly bonded valences
*
* First check whether the parent molecule has implicit valence 
* perceived (has MOL_IMPVALENCE_PERCEIVED flag). otherwise, call
* CCMol::PerceiveImplicitValence(). Then return implicit valences
* of the atom.
*
* @see CCMol::PerceiveImplicitValence()
*/
int CCAtom::GetImpVal() const
{
    if ( m_mol && !m_mol->HasFlag( MOL_IMPVALENCE_PERCEIVED ) ) {
        m_mol->PerceiveImplicitValence();
    }

    return m_impval;
}

/**
* @brief get the substructure id to which atom belongs
* @return the fragment part to which the atom belongs
*
* First check whether the parent molecule has fragment
* perceived (has MOL_FRAGMENT_PERCEIVED flag). If not, call 
* CCMol::PerceiveFragment(), Then return the fragment 
* part to which the atom belongs.
*
* @see CCMol::PerceiveFragment()
*/
int CCAtom::GetPart() const
{
    if ( m_mol && !m_mol->HasFlag( MOL_FRAGMENT_PERCEIVED ) ) {
        m_mol->PerceiveFragment();
    }

    return m_part;
}

/**
* @brief get hybridization state
* @return the hybridization state
*
* First check whether the parent molecule has hybridization
* perceived (has MOL_HYBRID_PERCEIVED flag). If not, call 
* %CCMol::PerceiveHybridization(), Then return the hybridization
* state.
*
* @see CCMol::PerceiveHybridization()
*/
int CCAtom::GetHyb() const
{
    if ( m_mol && !m_mol->HasFlag( MOL_HYBRID_PERCEIVED ) ) {
        m_mol->PerceiveHybridization();
    }

    return m_hyb;
}

/**
* @brief get the smallest bond angle of center atom
* @return the smallest bond angle of center atom
*
* Calculate all the bond angel whose center atom is this atom.
* Return the smallest bond angle. If no bond angle exist, return
* 360.0 degree.
*/
double CCAtom::SmallestBondAngle()
{
    CCAtom* a1, * a2;
    CCBondIter i, j;

    double degrees = 0.0, minDegrees = 360.0;

    for ( a1 = BeginNbrAtom( i ); a1; a1 = NextNbrAtom( i ) ) {
        j = i;
        for ( a2 = NextNbrAtom( j ); a2; a2 = NextNbrAtom( j ) ) {
            degrees = Angle( a1->GetXYZ(), GetXYZ(), a2->GetXYZ() );
            if ( degrees < minDegrees ) {
                minDegrees = degrees;
            }
        }
    }

    return minDegrees;
}

/**
* @brief get the average bond angle of center atom
* @return the average bond angle of center atom
*
* Calculate all the bond angel of center atom. Return the 
* average bond angle. If no bond angle exist, return 0.0 degree.
*/
double CCAtom::AverageBondAngle()
{
    CCAtom* a1, * a2;
    CCBondIter i, j;

    double avgDegrees = 0.0;
    int n = 0;

    for ( a1 = BeginNbrAtom( i ); a1; a1 = NextNbrAtom( i ) ) {
        j = i;
        for ( a2 = NextNbrAtom( j ); a2; a2 = NextNbrAtom( j ) ) {
            avgDegrees += Angle( a1->GetXYZ(), GetXYZ(), a2->GetXYZ() );
            n++;
        }
    }

    return ( n >= 1 ) ? avgDegrees / n : 0.0;
}

/**
* @brief check whether the atom is metal atom. 
* @return true is the atom is metal atom
*
* Using atomic number to judge whether the atom is metal atom.
*/
bool CCAtom::IsMetal() const
{
    switch ( m_ele ) {
    case 1:
        // H
    case 5:
        // B
    case 6:
        // C
    case 7:
        // N
    case 8:
        // O
    case 9:
        // F
    case 14:
        // Si
    case 15:
        // P
    case 16:
        // S
    case 17:
        // Cl
    case 33:
        // As
    case 34:
        // Se
    case 35:
        // Br
    case 51:
        // Sb
    case 52:
        // Te
    case 53:
        // I
    case 85:
        // At
        return false;
    default:
        return true;
    }
}

/**
* @brief get the min and max number of pi electrons 
* that contribute to aromaticity.
* @return a ranged pair<int,int>
*
* Get pi electrons contribution to aromaticity 
* acoording to atomic hybridization, return a
* ranged pair<int,int> 
*/
pair<int, int> CCAtom::GetPIelectron() const
{
    int fcharge = ( ( CCAtom* )this )->GetFormalCharge();
    if ( IsCarbon() && fcharge == -1 ) {
        // [C-]
        return make_pair( 2, 2 );
    }
    else if ( IsSp2Carbon() ) {
        // >C= or ~C~
        return make_pair( 1, 1 );
    }
    else if ( IsNitrogen() && fcharge == -1 ) {
        // [N-]
        return make_pair( 2, 2 );
    }
    else if ( IsNitrogen() && fcharge == 1 && GetValence() < 4 ) {
        // N-Oxide
        return make_pair( 1, 1 );
    }
    else if ( IsSp2Nitrogen() ) {
        // -N= or ~N~
        return make_pair( 1, 2 );
    }
    else if ( IsSp3Nitrogen() ) {
        // >N-
        return make_pair( 2, 2 );
    }
    else if ( IsOxygen() && fcharge == 1 ) {
        // -[O+]=
        return make_pair( 1, 1 );
    }
    else if ( IsSp3Oxygen() ) {
        // -O-
        return make_pair( 2, 2 );
    }
    else if ( IsSulfur() && fcharge == 1 ) {
        // -[S+]=
        return make_pair( 1, 1 );
    }
    else if ( IsSp3Sulfur() && GetValence() < 4 ) {
        // -S-
        return make_pair( 2, 2 );
    }
    else if ( IsSp2Sulfur() && GetValence() == 3 ) {
        //  >S=O
        return make_pair( 2, 2 );
    }
    else if ( IsSp2Selenium() && GetValence() == 3 ) {
        // >Se=O
        return make_pair( 2, 2 );
    }
    else if ( GetHyb() == 1 ) {
        // sp center
        return make_pair( 1, 1 );
    }
    else {
        return make_pair( 0, 0 );
    }
}

/**
* @brief check whether the atom bonded to a conjugate system
* such as X-a=b or X-a+=b.
* @param[in] includePandS whether X can be P or S
* @return true if atom in a system such as X-a=b
*
* In X-a=b, a and b represent the 'apha' and 'beta' atoms, 
* respectively and X indicates the current atom.
*/
bool CCAtom::HasAlphaBetaUnsat( bool includePandS )
{
    FOR_NBORS_OF_ATOM( a, this ) {
        if ( !includePandS && ( a->IsPhosphorus() || a->IsSulfur() ) ) {
            return false;
        }

        FOR_NBORS_OF_ATOM( b, &*a ) {
            if ( &*b != this && ( (CCBond*)~b )->GetOrder() > 1 ) {
                return true;
            }
        }
    }

    return false;
}

/**
* @brief get max bond order of neighboring bonds
* @return the max bond order of neighboring bonds
*
* If the atom had no neighboring bond, return 0. The
* bond order was achieved by GetOrder().
*
* @see CCBond::GetOrder()
*/
int CCAtom::GetMaxBO()
{
    int largestorder = 0;
    FOR_BONDS_OF_ATOM( bond, ( CCAtom * )this )
    if ( bond->GetOrder() > largestorder ) {
        largestorder = bond->GetOrder();
    }
    return largestorder; 
}

/**
* @brief get sum of neighboring bond orders
* @param[in] hydrogen whether count bond to hydrogen atom
* @return the sum of neighboring bond orders
*
* If the atom had no neighboring bond, return 0. The
* bond order was achieved by GetOrder().
*
* @see CCBond::GetOrder()
*/
int CCAtom::SumBondOrder( bool hydrogen ) {
	int count = 0;
	FOR_BONDS_OF_ATOM( bond, ( CCAtom * )this ) {
		if ( !hydrogen && bond->GetNbrAtom( this )->IsHydrogen() ) {
			continue;
		}
		count += bond->GetOrder();
	}
	return count;
}

/**
* @brief count bond with order
* @param[in] order bond order to count
* @return the number of neighboring bonds with expected order
*
* The bond order was achieved by GetOrder().
*
* @see CCBond::GetOrder()
*/
int CCAtom::CountBondWithOrder( int order ) const
{
    int count = 0;
    FOR_BONDS_OF_ATOM( bond, ( CCAtom * )this )
    if ( bond->GetOrder() == order ) {
        count++;
    }
    return count;
}

/**
* @brief count bond with larger order
* @param[in] order bond order for comparing
* @return the number of neighboring bonds whose order is
* larger than required order.
*
* The bond order was achieved by GetOrder().
*
* @see CCBond::GetOrder()
*/
int CCAtom::CountBondWithLargerOrder( int order ) const
{
    int count = 0;
    FOR_BONDS_OF_ATOM( bond, ( CCAtom * )this )
    if ( bond->GetOrder() > order ) {
        count++;
    }
    return count;
}

/**
* @brief check whether the atom has a neighboring bond whose
* order is equal to required order.
* @param[in] order bond order for comparing
* @return whether has neighboring bond with required order
*
* The bond order was achieved by GetOrder().
*
* @see CCBond::GetOrder()
*/
bool CCAtom::HasBondWithOrder( int order ) const
{
    FOR_BONDS_OF_ATOM( bond, ( CCAtom * )this )
    if ( bond->GetOrder() == order ) {
        return true;
    }
    return false;
}

/**
* @brief check whether the atom has aromatic neighboring
* bond ( ring bond with order of 5 ).
* @return whether has neighboring aromatic bond
*
* The aromatic bond should be in ring and with order of 5.
*
* @see CCBond::GetOrder()
* @see CCBond::IsInRing()
*/
bool CCAtom::HasAromaticBond() const
{
    FOR_BONDS_OF_ATOM( bond, ( CCAtom * )this )
    if ( bond->GetOrder() == 5 && bond->IsInRing() ) {
        return true;
    }
    return false;
}

/**
* @brief check whether the atom has a neighboring bond whose
* order is larger than required order.
* @param[in] order bond order for comparing
* @return whether has neighboring bond with order larger than
* required order.
*
* The bond order was achieved by GetOrder().
*
* @see CCBond::GetOrder()
*/
bool CCAtom::HasBondWithLargerOrder( int order ) const
{
    FOR_BONDS_OF_ATOM( bond, ( CCAtom * )this )
    if ( bond->GetOrder() > order ) {
        return true;
    }
    return false;
}

/**
* @brief get the bond connects this and na
* @param[in] na neighboring atom
* @return the bond between this atom and na
*
* If there is no bond connected this atom and na atom,
* return NULL. Else return the bond between this atom
* and na atom.
*/
CCBond* CCAtom::GetBond(CCAtom* na) const
{
	FOR_BONDS_OF_ATOM(nb, (CCAtom*)this) {
		if ( (~nb) == na ) {
			return (&*nb);
		}
	}
	return NULL;
}

/**
* @brief get the number of bonded heavy atoms
* @return the number of bonded heavy atoms.
*/
int CCAtom::GetHvyVal() const
{
    int count = 0;
    FOR_NBORS_OF_ATOM( na, ( CCAtom * )this )
    if ( na->IsHvyAtom() ) {
        count++;
    }

    return count;
}

/**
* @brief set atomic isotope state.
* @param[in] iso isotope state
*
* If the atom is H and iso is equal to 2, the internal
* atom type will also be changed as "D". If the atom is
* H and iso is equal to 1 or 0, change the internal atom
* type as "H". In other situation, just set m_isotope
* as iso.
*/
void CCAtom::SetIsotope( int iso )
{
	if ( m_ele == 1 ) {
		if ( iso == 2 ) {
			SetType( "D" );
		}
		else if ( iso == 1 || iso == 0 ) {
			SetType( "H" );
		}
    }

    m_isotope = iso;
}

/**
* @brief count implicit H
* @return the number of implicit hydrogens.
*
* Count the number of hydrogens needed to fill
* the implicit valence of this atom. A radical
* will modify this atom's implicit valence. eg.
* an unfilled methane C has implicit H of 4.
*
* @see CCAtom::GetImpVal(), CCAtom::GetValence()
*/
int CCAtom::ImpHydrogenCount()
{
    int impH = GetImpVal() - GetValence();

    // we need to modify this implicit valence if we have a radical
    int mult = GetSpin();
    if ( mult == 2 ) //radical
    {
        impH -= 1;
    }
    else if ( mult == 1 || mult == 3 ) //carbene
    {
        impH -= 2;
    }

    return ( impH > 0 ) ? impH : 0;
}

/**
* @brief count explicit H
* @return the number of explicit hydrogens.
*
* Count the number of hydrogens explicitly bound to 
* this atom, optionally excluding D,T and isotope 
* explicitly set to 1.<BR>
* If excludeIsotopes is true, H atoms with 
* m_isotope != 0 are not included. This excludes
* D, T and H when m_isotope explicitly set to 1 
* rather than the default 0. eg. an unfilled methane C
* has explicit H of 0.
*/
int CCAtom::ExpHydrogenCount( bool excludeIsotopes ) const
{
    // If excludeIsotopes is true, H atoms with 
	// m_isotope!=0 are not included.
    // This excludes D, T and H when m_isotope 
	// explicitly set to 1 rather than the default 0.
    int numH = 0;
    FOR_NBORS_OF_ATOM( nbr, ( CCAtom * )this ) {
        if ( nbr->IsHydrogen()
          && !( excludeIsotopes && nbr->GetIsotope() != 0 ) ) {
            numH++;
        }
    }
    return numH;
}

/**
* @brief sum of bond order
* @return the sum of neighboring bond orders
*
* sum of the bond orders of the neighboring bonds 
* (i.e. double bond = 2, aromatic bond = 3...). Note
* if the order is larger than or equal to 4, it will 
* contribute 3 (not 5, this is different from 
* CCAtom::SumBondOrder()), Otherwise it will contribute
* original bond order.
*
* @see CCAtom::SumBondOrder()
*/
int CCAtom::SumBO() const
{
    int bo;
    int bosum = 0;

    FOR_BONDS_OF_ATOM( nbrbond, ( CCAtom * )this ) {
        bo = nbrbond->GetOrder();
        bosum += ( bo < 4 ) ? 2 * bo : 3;
    }
    return ( bosum /= 2 );
}

/**
* @brief used by smi format to get chiral and stereo information.
*/
bool CCAtom::GetNewBondVector( CCVector& v, double length )
{
    // ***experimental code***

    CCAtom* atom;
    vector<CCBond*>::iterator i, j;
    v = VZero;

    if ( GetValence() == 0 ) {
        v = VX;
        v *= length;
        v += GetXYZ();
        return true;
    }

    if ( GetValence() == 1 ) {
        CCVector vtmp, v1, v2;
        atom = BeginNbrAtom( i );
        if ( atom ) {
            vtmp = GetXYZ() - atom->GetXYZ();
        }

        if ( GetHyb() == 2 || ( IsOxygen() && HasAlphaBetaUnsat() ) ) {
            bool quit = false;
            CCAtom* a1, * a2;
            v2 = VZero;
            for ( a1 = BeginNbrAtom( i );
                  a1 && !quit;
                  a1 = NextNbrAtom( i ) ) {
                for ( a2 = a1->BeginNbrAtom( j );
                      a2 && !quit;
                      a2 = a1->NextNbrAtom( j ) ) {
                    if ( a1 && a2 && a2 != this ) {
                        v2 = a1->GetXYZ() - a2->GetXYZ();
                        quit = true;
                    }
                }
            }

            if ( v2.IsApprox( VZero, 1.0e-8 ) ) {
                v1 = Cross( vtmp, VX );
                v2 = Cross( vtmp, VY );
                if ( v1.Length() < v2.Length() ) {
                    v1 = v2;
                }
            }
            else {
                v1 = Cross( vtmp, v2 );
            }

            CCMatrix m;
            m.RotAboutAxisByAngle( v1, 60.0 );
            v = m * vtmp;
            v.Normalize();
        }
        else if ( GetHyb() == 3 ) {
            v1 = Cross( vtmp, VX );
            v2 = Cross( vtmp, VY );
            if ( v1.Length() < v2.Length() ) {
                v1 = v2;
            }
            CCMatrix m;
            m.RotAboutAxisByAngle( v1, 70.5 );
            v = m * vtmp;
            v.Normalize();
        }
        if ( GetHyb() == 1 ) {
            v = vtmp;
        }

        v *= length;
        v += GetXYZ();
        return true;
    }

    if ( GetValence() == 2 ) {
        CCVector v1, v2, vtmp, vsum, vnorm;
        atom = BeginNbrAtom( i );
        if ( !atom ) {
            return false;
        }
        v1 = GetXYZ() - atom->GetXYZ();
        atom = NextNbrAtom( i );
        if ( !atom ) {
            return false;
        }
        v2 = GetXYZ() - atom->GetXYZ();
        v1.Normalize();
        v2.Normalize();
        vsum = v1 + v2;
        vsum.Normalize();

        if ( GetHyb() == 2 ) {
            v = vsum;
        }
        else if ( GetHyb() == 3 ) {
            vnorm = Cross( v2, v1 );
            vnorm.Normalize();

#ifndef ONE_OVER_SQRT3
#define ONE_OVER_SQRT3  0.57735026918962576451
#endif //SQRT_TWO_THIRDS
#ifndef SQRT_TWO_THIRDS
#define SQRT_TWO_THIRDS 0.81649658092772603272
#endif //ONE_OVER_SQRT3

            vsum *= ONE_OVER_SQRT3;
            vnorm *= SQRT_TWO_THIRDS;

            v = vsum + vnorm;
        }

        v *= length;

        v += GetXYZ();
        return true;
    }

    if ( GetValence() == 3 ) {
        CCVector vtmp, vsum;
        CCAtom* atom;
        vector<CCBond*>::iterator i;
        for ( atom = BeginNbrAtom( i ); atom; atom = NextNbrAtom( i ) ) {
            vtmp = GetXYZ() - atom->GetXYZ();
            vtmp.Normalize();
            vtmp /= 3.0;
            vsum += vtmp;
        }
        vsum.Normalize();
        v = vsum;
        v *= length;
        v += GetXYZ();
        return true;
    }

    return true;
}

/**
* @brief count the number of rings that contain this atom.
* @return the number of rings that contain this atom.
*
* @note If ring in the parent molecule is not perceived, it will
* touch off PerceiveRingAtomsAndBonds() function to perceive ring
* atoms and bonds.
*/
int CCAtom::MemberOfRingCount() const
{
    if ( !( ( CCAtom * ) this )->IsInRing() || !m_mol ) {
        return ( 0 );
    }

    int count = 0;
    FOR_RINGS_OF_MOL( ring, m_mol ) {
        if ( ring->IsAtomInRing( m_idx ) ) {
            count++;
        }
    }

    return count;
}

/**
* @brief return the size of the first ring that contains this
* atom.
* @return the size of the first ring that contains this atom.
*
* If no ring contains this atom, just return 0. Else return the
* size of the first ring that contains this atom.
*
* @note If ring in the parent molecule is not perceived, it will
* touch off PerceiveRingAtomsAndBonds() function to perceive ring
* atoms and bonds.
*/
int CCAtom::MemberOfRingSize() const
{
    if ( !( ( CCAtom * ) this )->IsInRing() || !m_mol ) {
        return 0;
    }

    FOR_RINGS_OF_MOL( ring, m_mol ) {
        if ( ring->IsAtomInRing( m_idx ) ) {
            return ring->GetSize();
        }
    }

    return 0;
}

/**
* @brief count the number of explicit ring connections to
* this atom.
* @return the number of explicit ring bond contain this atom.
*
* If no ring contains this atom, just return 0. Else return the
* number of explicit ring bond contain this atom.
*
* @note If ring in the parent molecule is not perceived, it will
* touch off PerceiveRingAtomsAndBonds() function to perceive ring
* atoms and bonds.
*/
int CCAtom::CountRingBonds() const
{
    int count = 0;

    FOR_BONDS_OF_ATOM( b, ( CCAtom * )this ) {
        if ( b->IsInRing() ) {
            count++;
        }
    }

    return count;
}

/**
* @brief in ring of size 'sz' ?
* @param[in] size ring size.
* @return whether the atom is in size membered ring.
* @note If ring in the parent molecule is not perceived, it will
*       trigger CCMol::PerceiveRingAtomsAndBonds() to perceive 
*       ring atoms and bonds.
*
* If this atom was not in ring, just return false. Else return
* whether the atom is in size membered ring.
*/
bool CCAtom::IsInRingSize( int size )
{
    if ( !IsInRing() || !m_mol ) {
        return false;
    }

    FOR_RINGS_OF_MOL( ring, m_mol ) {
        if ( ring->GetSize() == size && ring->IsAtomInRing( m_idx ) ) {
            return true;
        }
    }

    return false;
}

/**
* @brief get the smallest ring size this atom belonged to.
* @return the smallest ring size this atom belonged to.
*
* If this atom was not in ring, just return 0. Else return
* the smallest ring size this atom belonged to.
*
* @note If ring in the parent molecule is not perceived, it will
* touch off PerceiveRingAtomsAndBonds() function to perceive ring
* atoms and bonds.
*/
int CCAtom::InSmallestRing()
{
    int ringsize = 0;
    if ( !( ( CCAtom * ) this )->IsInRing() || !m_mol ) {
        return ringsize;
    }

#define USING_EXIST_RINGS 1
#if USING_EXIST_RINGS
    FOR_RINGS_OF_MOL( ring, m_mol ) {
        if ( ring->GetSize() < ringsize ) {
            if ( ring->IsAtomInRing( m_idx ) ) {
                ringsize = ring->GetSize();
            }
        }
        else if ( ringsize == 0 && ring->IsAtomInRing( m_idx ) ) {
            ringsize = ring->GetSize();
        }
    }
#else
    // percieve sssr include this atom
    int numatom = m_mol->GetNumAtom();
    vector<CCBitVec> avPaths( numatom + 1 ); // equals to number of atoms +1
    for ( int k = 1;
          k < avPaths.size();
          ++k ) // avPaths[0] was used to intersection and union
    {
        avPaths[k].Resize( numatom + 1 );
    }
    vector<vector<int> > aPaths( numatom + 1 );       // sequential path 

    CCBitVec avpath( numatom + 1 ), visit( numatom + 1 ), used( numatom + 1 );
    vector<int> prev( numatom + 1 );
    CCAtom* head, * nbr;
    vector<CCBond*>::iterator i;
    int rootIdx = m_idx, headIdx, nbrIdx;

    // initial queue
    queue<CCAtom*> q;
    q.push( this );
    used.Set( rootIdx );
    aPaths[rootIdx].push_back( rootIdx );
    avPaths[rootIdx].Set( rootIdx );

    while ( !q.empty() ) {
        head = q.front();
        q.pop();
        headIdx = head->GetIdx();
        visit.Set( headIdx );

        for ( nbr = head->BeginNbrAtom( i );
              nbr;
              nbr = head->NextNbrAtom( i ) ) {
            nbrIdx = nbr->GetIdx();
            if ( nbrIdx == prev[headIdx] || !nbr->IsInRing() ) {
                continue;
            }
            else if ( !visit[nbrIdx] ) {
                if ( !used[nbrIdx] ) // avoid duplicate push into queue
                {
                    q.push( nbr );
                    used.Set( nbrIdx );
                    prev[nbrIdx] = headIdx;
                    aPaths[nbrIdx] = aPaths[headIdx];
                    aPaths[nbrIdx].push_back( nbrIdx );
                    avPaths[nbrIdx].Set( nbrIdx );
                    avPaths[nbrIdx] |= avPaths[headIdx];
                }
            }
            else // collision detected
            {
                avPaths[0].Reset(); // clear
                avPaths[0] = avPaths[headIdx] & avPaths[nbrIdx];
				// a sssr was found including rootnode (singleton)
                if ( avPaths[0].Count() == 1 ) 
                {
                    ringsize = aPaths[headIdx].size()
                             + aPaths[nbrIdx].size()
                             - 1;
                    return ringsize;
                }
            }
        }
    }
#endif
    return ringsize;
}

/**
* @brief check whether the atom match the smarts pattern.
* @return whether the atom match the smarts pattern.
*
* This function will call CCSmartsPattern class to do
* substructure search and pattern match.
*
* @see CCSmartsPattern::Match()
*/
bool CCAtom::MatchesSMARTS( const char* pattern )
{
    CCMol* mol = GetMol();
    if ( !mol ) {
        return false;
    }

    vector<vector<int> > mlist;

    CCSmartsPattern sp;
    sp.Init( pattern );
    if ( sp.Match( *mol ) ) {
        mlist = sp.GetUMapList();
        for ( vector<vector<int> >::iterator l = mlist.begin();
              l != mlist.end();
              ++l ) {
            if ( GetIdx() == ( *l )[0] ) {
                return true;
            }
        }
    }
    return false;
}

bool CCAtom::IsAmideNitrogen() const
{
    if ( !IsNitrogen() ) {
        return false;
    }

    FOR_BONDS_OF_ATOM( nb, ( CCAtom * )this )
    if ( nb->IsAmide() ) {
        return true;
    }

    return false;
}

bool CCAtom::IsApolarHydrogen() const
{
    if ( !IsHydrogen() ) {
        return false;
    }

    FOR_NBORS_OF_ATOM( nbr, ( CCAtom * )this ) {
        if ( nbr->IsCarbon() ) {
            return true;
        }
    }

    return false;
}

bool CCAtom::IsInClosureBond()
{
    FOR_BONDS_OF_ATOM( bond, this ) {
        if ( bond->IsClosure() ) {
            return true;
        }
    }

    return false;
}

bool CCAtom::IsOneOf( const string& eleList )
{
    return ( eleList + "," ).find( GetSymbol() + "," ) != string::npos
        ? true : false;
}

//////////////////////////////////////////////////////////////////////////
/**
* @brief get the max weighted neighboring atom.
* @param[in] excludeatom atom which will be ignored.
* @return the max weighted neighboring atom.
*
* Here weight is defined as (nbor's atomic number) * (bond order).
*/
CCAtom* CCAtom::GetMaxWeightedNbr( CCAtom* excludeatom /* = NULL */ ) const
{
    CCAtom* atom( NULL );
    int maxweight = 0;
    int weight = 0;

    FOR_NBORS_OF_ATOM( nbr, ( CCAtom * )this ) {
        if ( ( &*nbr ) == excludeatom ) {
            continue;
        }

        weight = nbr->GetAtomicNum() * ( (CCBond*)~nbr )->GetOrder();
        if ( weight > maxweight ) {
            atom = &*nbr;
            maxweight = weight;
        }
    }
    return atom;
}

/**
* @brief add hydrogen atom to this atom.
* @param[in] polarHydrogenOnly add only polar hydrogen or not.
* @return the number of added hydrogen atoms.
*
* This function will add atoms, bonds and residues to parent molecule.
*/
int CCAtom::AddHydrogen( bool polarHydrogenOnly /* =false */ )
{
    static double bond_c_h = 1.09;
    static double bond_o_h = 0.96;
    static double bond_n_h = 1.02;
    static double bond_s_h = 1.33;

    if ( polarHydrogenOnly && !IsPolarToHydrogen() ) {
        return 0;
    }

    // ignore atoms with more than 3 links
    int val = GetValence();
    if ( val > 3 ) {
        return 0;
    }

    int numh = ImpHydrogenCount();
    if ( numh == 0 ) {
        return 0;
    }

    double bond_length;
    // gather the necessary information first
    if ( IsOxygen() ) {
        bond_length = bond_o_h;
    }
    else if ( IsNitrogen() ) {
        bond_length = bond_n_h;
    }
    else if ( IsSulfur() ) {
        bond_length = bond_s_h;
    }
    else {
        bond_length = bond_c_h;
    }

    CCVector v[4];
    CCVector candidate[4];
    CCVector vref;
    CCVector curcoor( GetXYZ() );
    // find the reference vector to determine where to add hydrogen
    for ( int i = 0; i < val; ++i ) {
        v[i] = GetNbrAtom( i + 1 )->GetXYZ() - curcoor;
    }

    // if it is a single atom, two of the hydrogen atom are put in X and 
    // Y axis.
    if ( val == 0 ) {
        v[0] = VX;
        v[1] = VY;
    }

    // if it is a terminal atom, use the reference vector to determine
    // the position of the first hydrogen atom.
    if ( val == 1 && GetHyb() > 1 ) {
        CCAtom* nbr, * aref;
        nbr = GetNbrAtom( 1 );

        // get the max weighted atom and its vector to nbr atom
        aref = nbr->GetMaxWeightedNbr( this );
        // nbr atom is not terminal atom
        if ( aref ) {
            vref = aref->GetXYZ() - nbr->GetXYZ();
        }
        // nbr atom is terminal atom, vref is set as VX or VY which is 
        // not in a line with root and nbr atoms
        else {
            vref = IsInALine( curcoor, nbr->GetXYZ(), VX ) ? VY : VX;
        }
    }

    // Get the coordinates of each added hydrogen atoms
    if ( GetHyb() == 3 ) {
        if ( vref != VZero ) {
            GetSp3Coordinates3( v[0], vref, v[1], v[2], v[3] );

            candidate[0] = curcoor + v[1] * bond_length;
            candidate[1] = curcoor + v[2] * bond_length;
            candidate[2] = curcoor + v[3] * bond_length;
        }
        else if ( numh == 2 ) {
            GetSp3Coordinates2( v[0], v[1], v[2], v[3] );

            candidate[0] = curcoor + v[2] * bond_length;
            candidate[1] = curcoor + v[3] * bond_length;
        }
        else if ( numh == 1 ) {
            if ( val == 2 ) {
                // *-N-* need to add only one hydrogen to N
                GetSp3Coordinates2( v[0], v[1], v[2], v[3] );
            }
            else {
                GetSp3Coordinates( v[0], v[1], v[2], v[3] );
            }

            candidate[0] = curcoor + v[3] * bond_length;
        }
        else if ( numh == 4 ) {
            GetSp3Coordinates2( v[0], v[1], v[2], v[3] );

            candidate[0] = curcoor + v[0] * bond_length;
            candidate[1] = curcoor + v[1] * bond_length;
            candidate[2] = curcoor + v[2] * bond_length;
            candidate[3] = curcoor + v[3] * bond_length;
        }
        else {
            return 0;
        }
    }
    else if ( GetHyb() == 2 ) {
        if ( vref != VZero ) {
            GetSp2Coordinates2( v[0], vref, v[1], v[2] );

            candidate[0] = curcoor + v[1] * bond_length;
            candidate[1] = curcoor + v[2] * bond_length;
        }
        else if ( numh == 1 ) {
            GetSp2Coordinates( v[0], v[1], v[2] );

            candidate[0] = curcoor + v[2] * bond_length;
        }
        else {
            return 0;
        }
    }
    else if ( GetHyb() == 1 ) {
        if ( numh == 1 ) {
            GetSp1Coordinates( v[0], v[1] );

            candidate[0] = curcoor + v[1] * bond_length;
        }
        else {
            return 0;
        }
    }

    for ( int i = 0; i < numh; ++i ) {
        CCAtom atom;
        atom.SetName( "H" );
        atom.SetXYZ( candidate[i] );
        atom.SetAtomicNum( 1 );
        atom.SetPart( GetPart() );
        atom.SetType( "H" );
		atom.SetImpVal(1);

        CCMol* mol = GetMol();
        CCAtom* lastatom( NULL );
        if ( !( lastatom = mol->AddAtom( atom ) ) ) {
            gErrorLog.Throw( __FUNCTION__, "failed to add hydrogen!\n");
            return 0;
        }

		CCBond *bnd;
        if ( !( bnd = mol->AddBond( GetIdx(), lastatom->GetIdx(), 1 ) ) ) {
            gErrorLog.Throw( __FUNCTION__, "failed to add bond!\n");
            return 0;
        }
		bnd->SetType( "SB" );

        CCResidue* res = GetResidue();
        if ( res ) {
            res->AddAtom( lastatom );
            res->SetResAtomName( lastatom, "H" );
            res->SetSerNum( lastatom, res->GetSerNum( this ) );
        }
		else {
			// set the SUBST_TYPE from current atom's information
			CCGenericData *gd = GetData( "SUBST_TYPE" );
			if ( gd ) {
				CCGenericData *hgd = new CCPairString( "SUBST_TYPE", 
					( dynamic_cast<CCPairString*>( gd ) )->GetValue().c_str() );
				AssertMemAlloc( hgd );
				lastatom->SetData( hgd );
			}
		}
    }

    return numh;
}

/**
* @brief add a hydrogen atom in certain position 
* which is bond to current atom.
* @param[in] coor the coordination of added hydrogen.
* @param[in] order bond order between current atom and added hydrogen.
* 
* @return true if added successfully.
*
* This function will add a hydrogen atom in specific position.
* It was used for generating fragment library.
*
* @see GenerateFragLib()
*/
bool CCAtom::AddSpecificHydrogen( const CCVector &coor,
							      int order /*= 1*/ ) {
	CCAtom atom;
	atom.SetName( "H" );
	atom.SetXYZ( coor );
	atom.SetAtomicNum( 1 );
	atom.SetPart( GetPart() );
	atom.SetType( "H" );

	CCMol* mol = GetMol();
	CCAtom* lastatom( NULL );
	if ( !( lastatom = mol->AddAtom( atom ) ) ) {
		gErrorLog.Throw( __FUNCTION__, "failed to add hydrogen!\n",
			ccWarning );
		return 0;
	}

	CCBond *bnd;
	if ( !( bnd = mol->AddBond( GetIdx(), lastatom->GetIdx(), 1 ) ) ) {
		gErrorLog.Throw( __FUNCTION__, "failed to add bond!\n");
		return false;
	}
	if ( order == 2 ) {
		bnd->SetType( "DB" );
	}
	else {
		bnd->SetType( "SB" );
	}

	CCResidue* res = GetResidue();
	if ( res ) {
		res->AddAtom( lastatom );
		res->SetResAtomName( lastatom, "H" );
		res->SetSerNum( lastatom, res->GetSerNum( this ) );
	}
	else {
		// set the SUBST_TYPE from current atom's information
		CCGenericData *gd = GetData( "SUBST_TYPE" );
		if ( gd ) {
			CCGenericData *hgd = new CCPairString( "SUBST_TYPE", 
				( dynamic_cast<CCPairString*>( gd ) )->GetValue().c_str() );
			AssertMemAlloc( hgd );
			lastatom->SetData( hgd );
		} 
	}
	return true;
}

/**
* @brief delete attached non-polar hydrogen/all hydrogen.
* @param[in] nonPolarHydrogenOnly delete only polar hydrogen or all hydrogen.
* @return the number of deleted hydrogen atoms.
*
* This function will record the hydrogen atoms that need to be deleted first.
* Then it call the molecular DeleteAtoms() function to delete hydrogen atoms
* and relavant bond.
*/
int CCAtom::DeleteHydrogen( bool nonPolarHydrogenOnly /*= false*/ )
{
    vector<CCAtom*> vh;
    FOR_NBORS_OF_ATOM( nbr, ( CCAtom * )this ) {
        if ( nbr->IsHydrogen() ) {
            if ( nonPolarHydrogenOnly && nbr->IsPolarToHydrogen() ) {
                continue;
            }
            vh.push_back( &*nbr );
        }
    }
    GetMol()->DeleteAtoms( vh );

    return ( int ) vh.size();
}

///HAS///HAS///HAS///HAS///HAS///HAS///HAS///HAS///HAS///HAS///HAS///HAS///HAS///
bool CCAtom::HasH( bool includeImpH, bool excludeIsotopes ) 
{
    // explicit hydrogen
    if ( ExpHydrogenCount(excludeIsotopes) ) {
        return true;
    }

    // implicit
    if ( includeImpH && ImpHydrogenCount() ) {
        return true;
    }

    return false;
}

bool CCAtom::HasPiElectron()
{
    // the atoms that has pi electron are defined in GMCharge.txt
    // these atoms include: C.1, C.2, C.ar, C.cat
    //                      N.1, N.2, N.3, N.4, N.am, N.ar, N.pl3
    //                      O.2, O.3, O.co2
    //                      S.2, S.3, S.o, S.o2
    //                      F, Cl, Br, I
    if ( m_ele == 7
      || m_ele == 8
      || m_ele == 9
      || m_ele == 16
      || m_ele == 17
      || m_ele == 35
      || m_ele == 53 ) {
        return true;
    }
    if ( m_ele == 6 && GetType() != "C.3" ) {
        return true;
    }
    return false;
}

//////////////////////////////////////////////////////////////////////////
//                       Arom related functions                         //
//////////////////////////////////////////////////////////////////////////
/**
* @brief check whether an atom is in carbonyl group.
* @param[in] atom atom for checking.
* @param[in] bit specify allowable atom type in this group
* @return int number of this atomic symbol if it match this group.
*
* bit = 1 ( default ): set allowable atom type as oxygen.<br>
* bit = 2 : set allowable atom type as sulfur.<br>
* bit = 4 : set allowable atom type as nitrogen.<br>
* bit = 3 : set allowable atom type as oxygen or sulfur.<br>
* bit = 7 : set allowable atom type as oxygen or sulfur or nitrogen.<br>
* return 0 if this atom does not match this group.
*/
int IsCarbonyl( CCAtom* atom, int bit )
{
    if ( atom->IsCarbon() && atom->GetValence() == 3 ) {
        int numO2 = 0;
        FOR_NBORS_OF_ATOM( nbr, atom ) {
            if ( ( (CCBond*)~nbr )->GetOrder() == 2
              && ( ( nbr->IsTerOxygen() && 1 & bit ) ||
                ( nbr->IsTerSulfur() && 2 & bit ) ||
                ( nbr->IsNitrogen() && 4 & bit && !atom->IsAromatic() ) ) ) {
                numO2++;
            }
        }
        return numO2 == 1 ? 'C' : 0;
    }
    else if ( atom->IsTerOxygen() && 1 & bit ) {
        return atom->GetNbrAtom( 1 )->IsCarbon()
            && atom->GetNbrAtom( 1 )->GetValence() == 3
            && IsCarbonyl( atom->GetNbrAtom( 1 ), bit ) == 'C' ? 'O' : 0;
    }
    else if ( atom->IsTerSulfur() && 2 & bit ) {
        return atom->GetNbrAtom( 1 )->IsCarbon()
            && atom->GetNbrAtom( 1 )->GetValence() == 3
            && IsCarbonyl( atom->GetNbrAtom( 1 ), bit ) == 'C' ? 'S' : 0;
    }
    else {
        return 0;
    }
}

/**
* @brief check whether an atom is in carboxyl group.
* @param[in] atom atom for checking.
* @param[in] bit specify allowable atom type in this group
* @return int number of this atomic symbol if it match this group.
*
* bit = 1 ( default ): set allowable atom type as oxygen.<br>
* bit = 2 : set allowable atom type as sulfur.<br>
* bit = 4 : set allowable atom type as nitrogen.<br>
* bit = 3 : set allowable atom type as oxygen or sulfur.<br>
* bit = 7 : set allowable atom type as oxygen or sulfur or nitrogen.<br>
* return 0 if this atom does not match this group.
*/
int IsCarboxyl( CCAtom* atom, int bit )
{
    return IsCarbonyl( atom, bit ) == 'C'
        && HasAdjacentHydroxyl( atom, bit ) ? 'C' : 0;
}

/**
* @brief check whether an atom is in carboxylate group.
* @param[in] atom atom for checking.
* @param[in] bit specify allowable atom type in this group
* @return int number of this atomic symbol if it match this group.
*
* bit = 1 ( default ): set allowable atom type as oxygen.<br>
* bit = 2 : set allowable atom type as sulfur.<br>
* bit = 4 : set allowable atom type as nitrogen.<br>
* bit = 3 : set allowable atom type as oxygen or sulfur.<br>
* bit = 7 : set allowable atom type as oxygen or sulfur or nitrogen.<br>
* return 0 if this atom does not match this group.
*/
int IsCarboxylate( CCAtom* atom, int bit )
{
    if ( atom->IsCarbon() && atom->GetValence() == 3 ) {
        int cntO2 = 0;
        FOR_NBORS_OF_ATOM( nbr, atom )
        if ( ( nbr->IsTerOxygen() && 1 & bit ) || 
            ( nbr->IsTerSulfur() && 2 & bit ) ) {
            cntO2++;
        }
        return ( cntO2 == 2 ) ? 'C' : 0;
    }
    else if ( atom->IsTerOxygen() && ( 1 & bit ) ) {
        return atom->GetNbrAtom( 1 )->IsCarbon()
            && atom->GetNbrAtom( 1 )->GetValence() == 3
            && IsCarboxylate( atom->GetNbrAtom( 1 ), bit ) == 'C' ? 'O' : 0;
    }
    else if ( atom->IsTerSulfur() && ( 2 & bit ) ) {
        return atom->GetNbrAtom( 1 )->IsCarbon()
            && atom->GetNbrAtom( 1 )->GetValence() == 3
            && IsCarboxylate( atom->GetNbrAtom( 1 ), bit ) == 'C' ? 'S' : 0;
    }
    else {
        return 0;
    }
}

/**
* @brief check whether an atom is in cyano group.
* @param[in] atom atom for checking.
* @return int number of this atomic symbol if it match this group.
*
* return 0 if this atom does not match this group.
*/
int IsCyano( CCAtom* atom )
{
    if ( atom->IsCarbon() && atom->GetValence() == 2 ) {
        int numN = 0;
        FOR_NBORS_OF_ATOM( nbr, atom ) {
            if ( ( (CCBond*)~nbr )->GetOrder() == 3
              && nbr->IsNitrogen()
              && nbr->GetValence() == 1 ) {
                numN++;
            }
        }
        return numN == 1 ? 'C' : 0;
    }
    else if ( atom->IsNitrogen() && atom->GetValence() == 1 ) {
        return atom->GetNbrAtom( 1 )->IsCarbon()
            && atom->GetNbrAtom( 1 )->GetValence() == 2
            && IsCyano( atom->GetNbrAtom( 1 ) ) == 'C' ? 'N' : 0;
    }
    else {
        return 0;
    }
}

/**
* @brief check whether an atom is in nitro group.
* @param[in] atom atom for checking.
* @return int number of this atomic symbol if it match this group.
*
* return 0 if this atom does not match this group.
*/
int IsNitro( CCAtom* atom )
{
    if ( atom->IsNitrogen() && atom->GetValence() == 3 ) {
        int numO2 = 0, bosum = atom->SumBO();

        FOR_NBORS_OF_ATOM( nbr, atom ) {
            if ( nbr->IsTerOxygen() ) {
                numO2++;
            }
        }
        return ( numO2 == 2 && bosum > 3 ) ? 'N' : 0;
    }
    else if ( atom->IsTerOxygen() ) {
        return atom->GetNbrAtom( 1 )->IsNitrogen()
            && atom->GetNbrAtom( 1 )->GetValence() == 3
            && IsNitro( atom->GetNbrAtom( 1 ) ) == 'N' ? 'O' : 0;
    }
    else {
        return 0;
    }
}

/**
* @brief check whether an atom is in nitroso group.
* @param[in] atom atom for checking.
* @return int number of this atomic symbol if it match this group.
*
* return 0 if this atom does not match this group.
*/
int IsNitroso( CCAtom* atom )
{
    if ( atom->IsNitrogen() && atom->GetValence() == 2 ) {
        int numO2 = 0;
        FOR_NBORS_OF_ATOM( nbr, atom ) {
            if ( ( (CCBond*)~nbr )->GetOrder() == 2 && nbr->IsTerOxygen() ) {
                numO2++;
            }
        }
        return numO2 == 1 ? 'N' : 0;
    }
    else if ( atom->IsTerOxygen() ) {
        return atom->GetNbrAtom( 1 )->IsNitrogen()
            && atom->GetNbrAtom( 1 )->GetValence() == 2
            && IsNitroso( atom->GetNbrAtom( 1 ) ) == 'N' ? 'O' : 0;
    }
    else {
        return 0;
    }
}

/**
* @brief check whether an atom is in nitro-oxide group.
* @param[in] atom atom for checking.
* @return int number of this atomic symbol if it match this group.
*
* return 0 if this atom does not match this group.
*/
int IsNitroOxide( CCAtom* atom )
{
	if ( atom->IsNitrogen() && atom->GetValence() == 3 ) {
		int numO2 = 0, bosum = atom->SumBO();
		FOR_NBORS_OF_ATOM( nbr, atom ) {
			if ( nbr->IsTerOxygen() ) {
				numO2++;
			}
		}
		return ( numO2 == 1 && bosum > 3 ) ? 'N' : 0;
	}
	else if ( atom->IsTerOxygen() ) {
		return atom->GetNbrAtom( 1 )->IsNitrogen()
			&& atom->GetNbrAtom( 1 )->GetValence() == 3
			&& IsNitroOxide( atom->GetNbrAtom( 1 ) ) == 'N' ? 'O' : 0;
	}
	else {
		return 0;
	}
}

/**
* @brief check whether an atom is in sulfone group.
* @param[in] atom atom for checking.
* @param[in] bit specify allowable atom type in this group
* @return int number of this atomic symbol if it match this group.
*
* bit = 1 ( default ): set allowable atom type as oxygen.<br>
* bit = 2 : set allowable atom type as sulfur.<br>
* bit = 4 : set allowable atom type as nitrogen.<br>
* bit = 3 : set allowable atom type as oxygen or sulfur.<br>
* bit = 7 : set allowable atom type as oxygen or sulfur or nitrogen.<br>
* return 0 if this atom does not match this group.
*/
int IsSulfone( CCAtom* atom, int bit )
{
    if ( atom->IsSulfur() && atom->GetValence() == 4 ) {
        int numO2 = 0;
        FOR_NBORS_OF_ATOM( nbr, atom ) {
            if ( ( (CCBond*)~nbr )->GetOrder() == 2
              && \
                ( nbr->IsTerOxygen() && 1 & bit )
              || ( nbr->IsTerSulfur() && 2 & bit )
              || ( nbr->IsNitrogen() && 4 & bit ) ) {
                numO2++;
            }
        }
        return numO2 == 2 ? 'S' : 0;
    }
    else if ( atom->IsTerOxygen() && 1 & bit ) {
        return atom->GetNbrAtom( 1 )->IsSulfur()
            && atom->GetNbrAtom( 1 )->GetValence() == 4
            && IsSulfone( atom->GetNbrAtom( 1 ), bit ) == 'S' ? 'O' : 0;
    }
    else if ( atom->IsTerSulfur() && 2 & bit ) {
        return atom->GetNbrAtom( 1 )->IsSulfur()
            && atom->GetNbrAtom( 1 )->GetValence() == 4
            && IsSulfone( atom->GetNbrAtom( 1 ), bit ) == 'S' ? 'S' + 1 : 0;
    }
    else if ( atom->IsPotentialSp2N() && 4 & bit ) {
        return atom->GetNbrAtom( 1 )->IsSulfur()
            && atom->GetNbrAtom( 1 )->GetValence() == 4
            && IsSulfone( atom->GetNbrHvyAtom( 1 ), bit ) == 'S' ? 'N' : 0;
    }
    else {
        return 0;
    }
}

/**
* @brief check whether an atom is in sulfonate group.
* @param[in] atom atom for checking.
* @param[in] bit specify allowable atom type in this group
* @return int number of this atomic symbol if it match this group.
*
* bit = 1 ( default ): set allowable atom type as oxygen.<br>
* bit = 2 : set allowable atom type as sulfur.<br>
* bit = 4 : set allowable atom type as nitrogen.<br>
* bit = 3 : set allowable atom type as oxygen or sulfur.<br>
* bit = 7 : set allowable atom type as oxygen or sulfur or nitrogen.<br>
* return 0 if this atom does not match this group.
*/
int IsSulfonate( CCAtom* atom, int bit )
{
	if ( atom->IsSulfur() && atom->GetValence() == 4 ) {
		int cntO2 = 0;
		FOR_NBORS_OF_ATOM( nbr, atom )
			if ( ( nbr->IsTerOxygen() && 1 & bit ) || \
				( nbr->IsTerSulfur() && 2 & bit ) ) {
					cntO2++;
			}
			return ( cntO2 == 3 ) ? 'S' : 0;
	}
	else if ( atom->IsTerOxygen() && ( 1 & bit ) ) {
		return atom->GetNbrAtom( 1 )->IsSulfur()
			&& atom->GetNbrAtom( 1 )->GetValence() == 4
			&& IsSulfonate( atom->GetNbrAtom( 1 ), bit ) == 'S' ? 'O' : 0;
	}
	else if ( atom->IsTerSulfur() && ( 2 & bit ) ) {
		return atom->GetNbrAtom( 1 )->IsSulfur()
			&& atom->GetNbrAtom( 1 )->GetValence() == 4
			&& IsSulfonate( atom->GetNbrAtom( 1 ), bit ) == 'S' ? 'S' + 1 : 0;
	}
	else {
		return 0;
	}
}

/**
* @brief check whether an atom is in sulfine group.
* @param[in] atom atom for checking.
* @param[in] bit specify allowable atom type in this group
* @return int number of this atomic symbol if it match this group.
*
* bit = 1 ( default ): set allowable atom type as oxygen.<br>
* bit = 2 : set allowable atom type as sulfur.<br>
* bit = 4 : set allowable atom type as nitrogen.<br>
* bit = 3 : set allowable atom type as oxygen or sulfur.<br>
* bit = 7 : set allowable atom type as oxygen or sulfur or nitrogen.<br>
* return 0 if this atom does not match this group.
*/
int IsSulfine( CCAtom* atom, int bit )
{
	if ( atom->IsSulfur() && atom->GetValence() == 3 ) {
		int numO2 = 0;
		FOR_NBORS_OF_ATOM( nbr, atom ) {
			if ( ( (CCBond*)~nbr )->GetOrder() == 2
				&& \
				( nbr->IsTerOxygen() && 1 & bit )
				|| ( nbr->IsTerSulfur() && 2 & bit ) ) {
					numO2++;
			}
		}
		return numO2 == 1 ? 'S' : 0;
	}
	else if ( atom->IsTerOxygen() && 1 & bit ) {
		return atom->GetNbrAtom( 1 )->IsSulfur()
			&& atom->GetNbrAtom( 1 )->GetValence() == 3
			&& IsSulfine( atom->GetNbrAtom( 1 ), bit ) == 'S' ? 'O' : 0;
	}
	else if ( atom->IsTerSulfur() && 2 & bit ) {
		return atom->GetNbrAtom( 1 )->IsSulfur()
			&& atom->GetNbrAtom( 1 )->GetValence() == 3
			&& IsSulfine( atom->GetNbrAtom( 1 ), bit ) == 'S' ? 'S' + 1 : 0;
	}
	else {
		return 0;
	}
}

/**
* @brief check whether an atom is in phosphone group.
* @param[in] atom atom for checking.
* @param[in] bit specify allowable atom type in this group
* @return int number of this atomic symbol if it match this group.
*
* bit = 1 ( default ): set allowable atom type as oxygen.<br>
* bit = 2 : set allowable atom type as sulfur.<br>
* bit = 4 : set allowable atom type as nitrogen.<br>
* bit = 3 : set allowable atom type as oxygen or sulfur.<br>
* bit = 7 : set allowable atom type as oxygen or sulfur or nitrogen.<br>
* return 0 if this atom does not match this group.
*/
int IsPhosphone( CCAtom* atom, int bit )
{
    if ( atom->IsPhosphorus() && atom->GetValence() == 4 ) {
        int numO2 = 0;
        FOR_NBORS_OF_ATOM( nbr, atom ) {
            if ( ( (CCBond*)~nbr )->GetOrder() == 2
              && \
                ( nbr->IsTerOxygen() && 1 & bit )
              || ( nbr->IsTerSulfur() && 2 & bit )
              || ( nbr->IsNitrogen() && 4 & bit ) ) {
                numO2++;
            }
        }
        return numO2 == 1 ? 'P' : 0;
    }
    else if ( atom->IsTerOxygen() && 1 & bit ) {
        return atom->GetNbrAtom( 1 )->IsPhosphorus()
            && atom->GetNbrAtom( 1 )->GetValence() == 4
            && IsPhosphone( atom->GetNbrAtom( 1 ), bit ) == 'P' ? 'O' : 0;
    }
    else if ( atom->IsTerSulfur() && 2 & bit ) {
        return atom->GetNbrAtom( 1 )->IsPhosphorus()
            && atom->GetNbrAtom( 1 )->GetValence() == 4
            && IsPhosphone( atom->GetNbrAtom( 1 ), bit ) == 'P' ? 'S' : 0;
    }
    else if ( atom->IsPotentialSp2N() && 4 & bit ) {
        return atom->GetNbrAtom( 1 )->IsPhosphorus()
            && atom->GetNbrAtom( 1 )->GetValence() == 4
            && IsPhosphone( atom->GetNbrHvyAtom( 1 ), bit ) == 'P' ? 'N' : 0;
    }
    else {
        return 0;
    }
}

/**
* @brief check whether an atom is in phosphate group.
* @param[in] atom atom for checking.
* @param[in] bit specify allowable atom type in this group
* @return int number of this atomic symbol if it match this group.
*
* bit = 1 ( default ): set allowable atom type as oxygen.<br>
* bit = 2 : set allowable atom type as sulfur.<br>
* bit = 4 : set allowable atom type as nitrogen.<br>
* bit = 3 : set allowable atom type as oxygen or sulfur.<br>
* bit = 7 : set allowable atom type as oxygen or sulfur or nitrogen.<br>
* return 0 if this atom does not match this group.
*/
int IsPhosphate( CCAtom* atom, int bit )
{
	if ( atom->IsPhosphorus() && atom->GetValence() == 4 ) {
		int cntO2 = 0;
		FOR_NBORS_OF_ATOM( nbr, atom )
			if ( ( nbr->IsTerOxygen() && 1 & bit ) || \
				( nbr->IsTerSulfur() && 2 & bit ) ) {
					cntO2++;
			}
			return ( cntO2 == 3 ) ? 'P' : 0;
	}
	else if ( atom->IsTerOxygen() && ( 1 & bit ) ) {
		return atom->GetNbrAtom( 1 )->IsPhosphorus()
			&& atom->GetNbrAtom( 1 )->GetValence() == 4
			&& IsPhosphate( atom->GetNbrAtom( 1 ), bit ) == 'P' ? 'O' : 0;
	}
	else if ( atom->IsTerSulfur() && ( 2 & bit ) ) {
		return atom->GetNbrAtom( 1 )->IsPhosphorus()
			&& atom->GetNbrAtom( 1 )->GetValence() == 4
			&& IsPhosphate( atom->GetNbrAtom( 1 ), bit ) == 'P' ? 'S' : 0;
	}
	else {
		return 0;
	}
}

/**
* @brief check whether an atom is in hydroxyl group.
* @param[in] atom atom for checking.
* @param[in] bit specify allowable atom type in this group
* @return int number of this atomic symbol if it match this group.
*
* bit = 1 ( default ): set allowable atom type as oxygen.<br>
* bit = 2 : set allowable atom type as sulfur.<br>
* bit = 4 : set allowable atom type as nitrogen.<br>
* bit = 3 : set allowable atom type as oxygen or sulfur.<br>
* bit = 7 : set allowable atom type as oxygen or sulfur or nitrogen.<br>
* return 0 if this atom does not match this group.
*/
int IsHydroxyl( CCAtom* atom, int bit )
{
    if ( atom->ExpHydrogenCount() + atom->ImpHydrogenCount() == 1 ) {
        if ( atom->IsSp3Oxygen() && 1 & bit ) {
            return 'O';
        }
        else if ( atom->IsSp3Sulfur() && 2 & bit ) {
            return 'S';
        }
        else {
            return 0;
        }
    }
    else {
        return 0;
    }
}

/**
* @brief check whether an atom is in azide group.
* @param[in] atom atom for checking.
* @return whether an atom is in azide group.
*
* return false if this atom does not match this group.
*/
bool IsAzide( CCAtom* atom )
{
    if ( !atom->IsSp1Nitrogen() ) {
        return false;
    }

    if ( atom->GetValence() == 1 ) {
        return atom->GetNbrAtom( 1 )->IsSp1Nitrogen()
            && atom->GetNbrAtom( 1 )->GetValence() == 2;
    }
    else {
        FOR_NBORS_OF_ATOM( nbr, atom ) {
            if ( nbr->IsSp1Nitrogen() && nbr->GetValence() == 1 ) {
                return true;
            }
        }
        return false;
    }
}

/**
* @brief check whether an atom is in acyl group.
* @param[in] atom atom for checking.
* @param[in] bit specify allowable atom type in this group
* @return whether an atom is in acyl group.
*
* bit = 1 ( default ): set allowable atom type as oxygen.<br>
* bit = 2 : set allowable atom type as sulfur.<br>
* bit = 4 : set allowable atom type as nitrogen.<br>
* bit = 3 : set allowable atom type as oxygen or sulfur.<br>
* bit = 7 : set allowable atom type as oxygen or sulfur or nitrogen.<br>
* return false if this atom does not match this group.
*/
bool IsAcyl( CCAtom* atom, int bit )
{
    if ( IsCarbonyl( atom, 7 ) == 'C' && 1 & bit ) {
        return true;
    }
    else if ( IsSulfone( atom, 3 ) == 'S' && 2 & bit ) {
        return true;
    }
    else if ( IsPhosphone( atom, bit ) == 'P' && 4 & bit ) {
        return true;
    }
    else {
        return false;
    }
}

/**
* @brief check whether an atom is in electron withdraw group.
* @param[in] atom atom for checking.
* @return whether an atom is in electron withdraw group.
*/
bool IsEWG( CCAtom* atom )
{
    if ( IsCarbonyl( atom, 7 ) == 'C' ) {
        return true;
    }
    else if ( IsCyano( atom ) == 'C' ) {
        return true;
    }
    else if ( IsNitro( atom ) == 'N' ) {
        return true;
    }
    else if ( IsNitroso( atom ) == 'N' ) {
        return true;
    }
    else if ( IsSulfine( atom, 3 ) == 'S' ) {
        return true;
    }
    else if ( IsSulfone( atom, 3 ) == 'S' ) {
        return true;
    }
    else if ( IsPhosphone( atom, 3 ) == 'P' ) {
        return true;
    }
    else {
        return false;
    }
}

/**
* @brief check whether an atom is in pi system.
* @param[in] atom atom for checking.
* @return whether an atom is in pi system.
*
* The current atom should be sp1 or sp2 hybridized and it
* bonded to an atom which is aromatic or has multi-order bond
* to another atom.
*/
bool IsPiSystem( CCAtom* atom ) // unfinished...
{
    if ( atom->GetHyb() > 2
        /*|| *//*IsSulfone(atom)*/ /*|| IsPhosphate(atom)*/ ) {
        return false;
    }

    FOR_BONDS_OF_ATOM( nbr, atom ) {
        if ( nbr->IsAromatic() || nbr->GetOrder() >= 2 ) {
            return true;
        }
    }
    return false;
}

/**
* @brief check whether an atom is a aliphatic nitrogen.
* @param[in] atom atom for checking.
* @return whether the atom is nitrogen in aliphatic chain or ring.
*
* The current atom should be nitrogen and none of its neighbor
* has muilti bond order.
*/
bool IsAliphaticNitrogen( CCAtom* atom )
{
    if ( !atom->IsNitrogen() ) {
        return false;
    }

    FOR_NBORS_OF_ATOM( nbr, atom ) {
        if ( nbr->HasBondWithLargerOrder( 1 ) ) {
            return false;
        }
    }
    return true;
}

/**
* @brief check whether an atom is in symmetric group.
* @param[in] atom atom for checking.
* @param[in] preT record neighboring atomic number.
* @return whether an atom is in symmetric group.
*
* symmetric group: -NH2 -N[+]H3 NH3 -NO2 -CH3 CH4 -CX3
* -PO3 -SO3 -CO2.
*/
bool IsATx( CCAtom* atom, int preT )
{
    if ( atom->GetHvyVal() == 1 ) {
        return true;
    }

    int markT = 0, markNT = 0;
    FOR_NBORS_OF_ATOM( nbr, atom ) {
        if ( nbr->IsTerAtom() ) {
            markT++;

            if ( preT == -1 ) {
                preT = nbr->GetAtomicNum();
            }

            if ( preT != nbr->GetAtomicNum() ) {
                return false;
            }
        }
        else {
            markNT++;
        }
    }

    return ( markNT <= 1 && markT == atom->GetImpVal() - 1 );
}

/**
* @brief check whether an atom is in t-butyl group.
* @param[in] atom atom for checking.
* @return whether an atom is in t-butyl group.
*/
bool IsTButyl( CCAtom* atom )
{
    if ( !atom->IsSp3Carbon() ) {
        return false;
    }

    FOR_NBORS_OF_ATOM( nbr, atom ) {
        if ( !nbr->IsSp3Carbon() || !IsATx( &*nbr, 1 ) ) // -CH3
        {
            return false;
        }
    }
    return true;
}

/**
* @brief check whether an atom is Positively Ionizable(PI) atom.
* @return whether an atom is Positively Ionizable(PI) atom.
*
* Positively Ionizable(PI) atoms: basic amines, basic secondary amidines, 
* basic primary amidines, basic guanidines, and positive charges not 
* adjacent to a negative charge.
*/
bool IsPositivelyIonizable( CCAtom *atom ) {
	// positive charges not adjacent to a negative charge
	if ( atom->GetFormalCharge() == 1 ) {
		FOR_NBORS_OF_ATOM( nbr, atom ) {
			if ( nbr->GetFormalCharge() == -1 ) {
				return false;
			}
		}
		return true;
	}
	// basic amines
	if ( atom->IsNitrogen() ) {
		FOR_BONDS_OF_ATOM( b, atom ) {
			if ( b->GetOrder() > 1 ) {
				return false;
			}
		}
		FOR_NBORS_OF_ATOM( nbr, atom ) {
			if ( nbr->GetMaxBO() > 1 ) {
				return false;
			}
		}
		return true;
	}
	// basic secondary amidines, basic primary amidines, basic guanidines
	if ( atom->IsCarbon() && !atom->IsInRing() ) {
		int terN = 0; // terminal nitrogen
		int DBN = 0; // double bonded Nitrogen
		FOR_NBORS_OF_ATOM( nbr, atom ) {
			if ( nbr->GetBond( atom )->GetOrder() == 2 &&
				 nbr->IsNitrogen() ) {
				++DBN;
			}
			else if ( nbr->IsTerNitrogen() ) {
				++terN;
			}
		}
		// amidines or guanidines
		if ( terN != 0 && DBN != 0 ) {
			return true;
		}
		return false;
	}
	return false;
}

/**
* @brief check whether an atom is Negatively Ionizable(NI) atom.
* @return whether an atom is Negatively Ionizable(NI) atom.
*
* Negatively Ionizable(NI) atoms: trifluoromethyl sulfonamide hydrogens,
* sulfonic acids, phosphonic acids, sulfinic, carboxylic or phosphinic 
* acids, tetrazoles, negative charges which are not neutralized by an 
* adjacency to a positive charge.
*/
bool IsNegativelyIonizable( CCAtom *atom ) {
	// negative charges not adjacent to a positive charge
	if ( atom->GetFormalCharge() == -1 ) {
		FOR_NBORS_OF_ATOM( nbr, atom ) {
			if ( nbr->GetFormalCharge() == 1 ) {
				return false;
			}
		}
		return true;
	}

	// trifluoromethyl sulfonamide
	if ( atom->IsTerNitrogen() ) {
		FOR_NBORS_OF_ATOM( nbr, atom ) {
			if ( nbr->IsSulfur() ) {
				int DBO = 0; // double bonded oxygen
				int numF = 0; // number of F
				FOR_NBORS_OF_ATOM( nbr2, &*nbr ) {
					if ( nbr2->GetBond( &*nbr )->GetOrder() == 2 &&
						 nbr2->IsOxygen() ) {
						++DBO;
					}
					else if ( nbr2->IsSp3Carbon() ) {
						FOR_NBORS_OF_ATOM( nbr3, &*nbr2 ) {
							if ( nbr3->IsFluorine() ) {
								++numF;
							}
						}
					}
				}
				if ( DBO == 2 && numF == 3 ) {
					return true;
				}
			}
		}
		return false;
	}

	// sulfonic acids, phosphonic acids, 
	// sulfinic, carboxylic or phosphinic acids
	if ( atom->IsOxygen() && 
		 atom->GetHvyVal() == 1 &&
		 atom->GetMaxBO() == 1 ) {
		FOR_NBORS_OF_ATOM( nbr, atom ) {
			if ( nbr->IsSulfur() || 
				 nbr->IsPhosphorus() ||
				 nbr->IsSp2Carbon() ) {
				int DBO = 0;
				int terO = 0;
				FOR_NBORS_OF_ATOM( nbr2, &*nbr ) {
					if ( nbr2->GetBond( &*nbr )->GetOrder() == 2 &&
						 nbr2->IsOxygen() ) {
						++DBO;
					}
					else if ( nbr2->IsOxygen() && 
						      nbr2->GetHvyVal() == 1 ) {
						++terO;
					}
				}
				if ( DBO != 0 && terO != 0 ) {
					return true;
				}
			}
		}
		return false;
	}

	// tetrazoles
	if ( atom->IsNitrogen() && atom->IsInRing() &&
		 atom->GetHvyVal() == 2 && 
		 atom->GetMaxBO() == 1 ) {
		CCAtom *nbrC(NULL);
		CCAtom *nbrN(NULL);
		FOR_NBORS_OF_ATOM( nbr, atom ) {
			if ( nbr->IsCarbon() ) {
				nbrC = &*nbr;
			}
			else if ( nbr->IsNitrogen() ) {
				nbrN = &*nbr;
			}
		}
		if ( nbrC && nbrN ) {
			CCAtom *nbrC2(NULL);
			CCAtom *nbrN2(NULL);
			FOR_NBORS_OF_ATOM( nbr2, nbrC ) {
				if ( nbr2->GetBond( nbrC )->GetOrder() == 2 &&
					 nbr2->IsNitrogen() ) {
					nbrC2 = &*nbr2;
				}
			}
			FOR_NBORS_OF_ATOM( nbr3, nbrN ) {
				if ( nbr3->GetBond( nbrN )->GetOrder() == 2 &&
					 nbr3->IsNitrogen() ) {
					nbrN2 = &*nbr3;
				}
			}
			if ( nbrC2 && nbrN2 && nbrC2->IsConnected( nbrN2 ) ) {
				return true;
			}
		}
		return false;
	}

	return false;
}

/**
* @brief check whether an atom has adjacent halogen atom.
* @param[in] atom atom for checking.
* @return whether an atom has adjacent halogen atom.
*/
int HasAdjacentHalogen( CCAtom* atom )
{
    int count = 0;

    FOR_NBORS_OF_ATOM( nbr, atom ) {
        if ( nbr->IsHalAtom() ) {
            count++;
        }
    }
    return count;
}

/**
* @brief check whether an atom has adjacent sulfone S atom.
* @param[in] atom atom for checking.
* @param[in] bit specify allowable atom type in sulfone group
* @return whether an atom has adjacent sulfone S atom.
*
* bit = 1 ( default ): set allowable atom type as oxygen.<br>
* bit = 2 : set allowable atom type as sulfur.<br>
* bit = 4 : set allowable atom type as nitrogen.<br>
* bit = 3 : set allowable atom type as oxygen or sulfur.<br>
* bit = 7 : set allowable atom type as oxygen or sulfur or nitrogen.<br>
*
* @see CCAtom::IsSulfone()
*/
bool HasAdjacentSulfone( CCAtom* atom, int bit )
{
    FOR_NBORS_OF_ATOM( nbr, atom ) {
        if ( IsSulfone( &*nbr, bit ) == 'S' ) {
            return true;
        }
    }
    return false;
}

/**
* @brief check whether an atom has adjacent sulfine S atom.
* @param[in] atom atom for checking.
* @param[in] bit specify allowable atom type in sulfine group
* @return whether an atom has adjacent sulfine S atom.
*
* bit = 1 ( default ): set allowable atom type as oxygen.<br>
* bit = 2 : set allowable atom type as sulfur.<br>
* bit = 4 : set allowable atom type as nitrogen.<br>
* bit = 3 : set allowable atom type as oxygen or sulfur.<br>
* bit = 7 : set allowable atom type as oxygen or sulfur or nitrogen.<br>
*
* @see CCAtom::IsSulfine()
*/
bool HasAdjacentSulfine( CCAtom* atom, int bit )
{
    FOR_NBORS_OF_ATOM( nbr, atom ) {
        if ( IsSulfine( &*nbr, bit ) == 'S' ) {
            return true;
        }
    }
    return false;
}

/**
* @brief check whether an atom has adjacent phosphone S atom.
* @param[in] atom atom for checking.
* @param[in] bit specify allowable atom type in phosphone group
* @return whether an atom has adjacent phosphone S atom.
*
* bit = 1 ( default ): set allowable atom type as oxygen.<br>
* bit = 2 : set allowable atom type as sulfur.<br>
* bit = 4 : set allowable atom type as nitrogen.<br>
* bit = 3 : set allowable atom type as oxygen or sulfur.<br>
* bit = 7 : set allowable atom type as oxygen or sulfur or nitrogen.<br>
*
* @see CCAtom::IsPhosphone()
*/
bool HasAdjacentPhosphone( CCAtom* atom, int bit )
{
    FOR_NBORS_OF_ATOM( nbr, atom ) {
        if ( IsPhosphone( &*nbr, bit ) == 'P' ) {
            return true;
        }
    }
    return false;
}

/**
* @brief check whether an atom has adjacent acyl group.
* @param[in] atom atom for checking.
* @param[in] bit specify allowable atom type in acyl group
* @return whether an atom has adjacent acyl group.
*
* bit = 1 ( default ): set allowable atom type as oxygen.<br>
* bit = 2 : set allowable atom type as sulfur.<br>
* bit = 4 : set allowable atom type as nitrogen.<br>
* bit = 3 : set allowable atom type as oxygen or sulfur.<br>
* bit = 7 : set allowable atom type as oxygen or sulfur or nitrogen.<br>
*
* @see CCAtom::IsAcyl()
*/
bool HasAdjacentAcyl( CCAtom* atom, int bit )
{
    FOR_NBORS_OF_ATOM( nbr, atom ) {
        if ( IsAcyl( &*nbr, bit ) ) {
            return true;
        }
        else {
            continue;
        }
    }
    return false;
}

/**
* @brief check whether an atom has adjacent hydroxyl group.
* @param[in] atom atom for checking.
* @param[in] bit specify allowable atom type in hydroxyl group
* @return whether an atom has adjacent hydroxyl group.
*
* bit = 1 ( default ): set allowable atom type as oxygen.<br>
* bit = 2 : set allowable atom type as sulfur.<br>
* bit = 4 : set allowable atom type as nitrogen.<br>
* bit = 3 : set allowable atom type as oxygen or sulfur.<br>
* bit = 7 : set allowable atom type as oxygen or sulfur or nitrogen.<br>
*
* @see CCAtom::IsHydroxyl()
*/
bool HasAdjacentHydroxyl( CCAtom* atom, int bit )
{
	FOR_NBORS_OF_ATOM( nbr, atom ) {
		if ( IsHydroxyl( &*nbr, bit ) ) {
			return true;
		}
	}
	return false;
}

/**
* @brief check whether an atom has adjacent EWG.
* @param[in] atom atom for checking.
* @return whether an atom has adjacent EWG.
*
* @see CCAtom::IsEWG()
*/
bool HasAdjacentEWG( CCAtom* atom )
{
	FOR_NBORS_OF_ATOM( nbr, atom ) {
		if ( IsEWG( &*nbr ) ) {
			return true;
		}
	}
	return false;
}

/**
* @brief check whether an atom has adjacent EWG.
* @param[in] atom atom for checking.
* @return whether an atom has adjacent EWG.
*
* @see CCAtom::IsEWG()
*/
bool HasAdjacentAromatic( CCAtom* atom )
{
    FOR_NBORS_OF_ATOM( nbr, atom ) {
        if ( nbr->IsAromatic() ) {
            return true;
        }
    }
    return false;
}

/**
* @brief check whether an atom has adjacent pi system.
* @param[in] atom atom for checking.
* @return whether an atom has adjacent pi system.
*
* @see CCAtom::IsPiSystem()
*/
bool HasAdjacentPiSystem( CCAtom* atom )
{
    FOR_NBORS_OF_ATOM( nbr, atom ) {
        if ( IsPiSystem( &*nbr ) ) {
            return true;
        }
    }
    return false;
}

/**
* @brief check whether an atom has adjacent ring.
* @param[in] atom atom for checking.
* @param[out] ring save adjacent ring if it exists.
* @return whether an atom has adjacent ring.
*
* It will call GetSSSR() function to perceive smallest set
* of smallest ring and then check whether its neighboring
* atoms is in ring. If find a neighboring ring atom, set ring
* as current sssr and return true.
*
* @see CCAtom::GetSSSR()
*/
bool HasAdjacentRing( CCAtom* atom, CCRing** ring ) // point to a pointer
{
    *ring = NULL;

    vector<CCRing*>& sssr = atom->GetMol()->GetSSSR();
    if ( sssr.empty() ) {
        return false;
    }

    CCRingIter itr;
    FOR_NBORS_OF_ATOM( nbr, atom ) {
        for ( itr = sssr.begin(); itr != sssr.end(); ++itr ) {
            if ( ( *itr )->IsAtomInRing( nbr->GetIdx() ) ) {
                *ring = *itr;
                return true;
            }
        }
    }
    return false;
}

/**
* @brief check whether two atoms have the same altloc data.
* @param[in] a1 the first atom for comparing.
* @param[in] a2 the second atom for comparing.
* @return whether an atom has adjacent ring.
*
* In pdb format, the 17 column is the altloc information of the atom.
* This function will check whether two given atoms have the same 
* altloc data. ( This is for pdb format only ).
*/
bool HasSameAltlocData( CCAtom* a1, CCAtom* a2 )
{
    // for generic data
    CCGenericData* gd( NULL );

    string altloca1 = "";
    string altloca2 = "";
    if ( ( gd = a1->GetData( "ALTLOC" ) ) ) {
        altloca1 = dynamic_cast<CCPairString*>( gd )->GetValue();
    }
    if ( ( gd = a2->GetData( "ALTLOC" ) ) ) {
        altloca2 = dynamic_cast<CCPairString*>( gd )->GetValue();
    }
    return ( altloca1 == altloca2 );
}

/**
* @brief search the shortest path length of the two atoms.
* @param[in] a1 the first atom.
* @param[in] a2 the second atom.
* @return the shortest path length of the two atoms.
*
* The path was counted in number of bonds.<br>
* e.g. a1-b1-b2-b3-b4-a2 --> return 5.
*/
int ShortestPathLength( CCAtom* a1, CCAtom* a2 )
{
    // get the shortest path between a1 and a2
    // BFS
    if ( !a1 || !a2 || 
		 a1->GetMol() != a2->GetMol() || 
		 a1->GetPart() != a2->GetPart() ) {
        return 0;
    }

    // allow quick detection if two atoms are bonded
    if ( a1->IsConnected( a2 ) ) {
        return 1;
    }

    int numAtom = a1->GetMol()->GetNumAtom();
    vector<vector<CCAtom*> > paths( numAtom + 1 );
    CCBitVec used( numAtom + 1 );
    CCAtom* head;
    int headIdx, nbrIdx;

    queue<CCAtom*> q;
    q.push( a1 );
    used.Set( a1->GetIdx() );
    paths[a1->GetIdx()].push_back( a1 );

    do {
        head = q.front();
        headIdx = head->GetIdx();
        q.pop();

        FOR_NBORS_OF_ATOM( nbr, head ) {
            nbrIdx = ( &*nbr )->GetIdx();
            if ( !used[nbrIdx] ) {
                paths[nbrIdx] = paths[headIdx];
                paths[nbrIdx].push_back( &*nbr );
                if ( &*nbr == a2 ) {
                    return paths[nbrIdx].size() - 1;
                }
                q.push( &*nbr );
                used.Set( nbrIdx );
            }
        }
    }
    while ( !q.empty() );

    return 0;
}

//////////////////////////////////////////////////////////////////////////
// similarity based on Atom Pair Descriptor (AP)
// a AP is defined as substructure composed of two non-hydrogen atoms 
// (but not necessarily bonded) with shortest inter-atomic distance of
// separation, like the follow:
// <atom 1 description> - <separation> - <atom 2 description>
// here we use the internal CCBGKit atom type as description of an atom
// Ref. J. Chem. Inf. Comput. Sci. 1985, 25, 82-85 
//////////////////////////////////////////////////////////////////////////
/**
* @brief search the shortest path length of the two atoms.
* @param[in] a1 the first atom.
* @param[in] a2 the second atom.
* @param[out] sPath record the atom path from a1 to a2.
* @param[in] includeHydrogen whether include hydrogen atom in searching.
* @return the shortest path length of the two atoms.
*
* Apply breadth-first search to retrieve the path.
* The path was counted in number of bonds.<br>
* e.g. a1-b1-b2-b3-b4-a2 --> return 5.
*/
int ShortestPath( CCAtom* a1, CCAtom* a2, vector<CCAtom*>* sPath/*=NULL*/,
                 bool includeHydrogen/*=false*/ )
{
    // get the shortest path between a1 and a2
    // BFS
    if ( !a1 || !a2 ) {
        return 0;
    }

    if ( !includeHydrogen && ( a1->IsHydrogen() || a2->IsHydrogen() ) ) {
        return 0;
    }

    // allow quick detection if two atoms are bonded
    if ( a1->IsConnected( a2 ) ) {
        if ( sPath ) {
            sPath->push_back( a1 );
            sPath->push_back( a2 );
        }
        return 1;
    }

    int numAtom = a1->GetMol()->GetNumAtom();
    vector<vector<CCAtom*> > paths( numAtom + 1 );
    CCBitVec used( numAtom + 1 );
    CCAtom* head;
    int headIdx, nbrIdx;

    queue<CCAtom*> q;
    q.push( a1 );
    used.Set( a1->GetIdx() );
    paths[a1->GetIdx()].push_back( a1 );

    do {
        head = q.front();
        headIdx = head->GetIdx();
        q.pop();

        FOR_NBORS_OF_ATOM( nbr, head ) {
            if ( !includeHydrogen && nbr->IsHydrogen() ) {
                continue;
            }

            nbrIdx = ( &*nbr )->GetIdx();
            if ( !used[nbrIdx] ) {
                paths[nbrIdx] = paths[headIdx];
                paths[nbrIdx].push_back( &*nbr );
                if ( &*nbr == a2 ) {
                    if ( sPath ) {
                        *sPath = paths[nbrIdx];
                    }
                    return paths[nbrIdx].size() - 1;
                }
                q.push( &*nbr );
                used.Set( nbrIdx );
            }
        }
    }
    while ( !q.empty() );

    return 0;
}

//////////////////////////////////////////////////////////////////////////
// similarity based Topological Torsion Descriptor (TT)
// a TT is defined as four consecutively bonded non-hydrogen atom, each
// described by:
//   its atom type (internal type used in CCBGKit), 
//   the number of non-hydrogen branches
//   ring feature: (0: not in a ring, 1: in normal ring, 2 in aromatic ring
// Ref. J. Chem. Inf. Comput. Sci. 1987, 27, 64-73
//////////////////////////////////////////////////////////////////////////
/**
* @brief use depth first search method to find all the atoms of nDepth.
* @param[in] root the root atom.
* @param[out] paths record the atom path of searching.
* @param[in] nDepth searching how many depth.
* @param[in] impH whether include hydrogen atom in searching.
* @return the shortest path length of the two atoms.
*
* Apply depth-first search to retrieve the path.
* The path was counted in number of bonds.<br>
* e.g. a1-b1-b2-b3-b4-a2 --> return 5.
*/
bool DFSnDepth( CCAtom* root, vector<vector<CCAtom*> >& paths, int nDepth,
               bool impH )
{
    if ( !root || ( root->IsHydrogen() && impH == false ) ) {
            return false;
    }
    if ( root->IsHydrogen() ) {
        impH = true;
    }

    // all will be clean when DFS finished
    static vector<CCAtom*> p;
    static CCBitVec visited;
    visited.Resize( root->GetMol()->GetNumAtom() );

    p.push_back( root );
    visited.Set( root->GetIdx() );

    if ( nDepth - 1 == 0 ) {
        paths.push_back( p );
        goto BACK;
    }

    FOR_NBORS_OF_ATOM( nbr, root ) {
        if ( nbr->IsHydrogen() && impH == false ) {
            continue;
        }
        if ( !visited[nbr->GetIdx()] ) {
            DFSnDepth( &*nbr, paths, nDepth - 1, impH );
        }
    }

BACK : p.pop_back();
    visited.Reset( root->GetIdx() );
    return paths.empty() ? false : true;
}

/**
* @brief compare two atoms in ascending order of their coordinates.
* @param[in] a1 the first atom for comparing.
* @param[in] a2 the second atom for comparing.
* @return whether a1's coordinates is less than a2's coordinates.
*
* The comparing order was from x to y to z coordinates.  
*/
bool CmpXYZ( const CCAtom* a1, const CCAtom* a2 )
{
	if ( !a1 || !a2 ) {
		return false;
	}

	double ax = a1-> GetX(), ay = a1->GetY(), az = a1->GetZ();
	double bx = a2-> GetX(), by = a2->GetY(), bz = a2->GetZ();

	if ( ax < bx ) {
		return true;
	}
	else if ( ax > bx ) {
		return false;
	}
	else {
		if ( ay < by ) {
			return true;
		}
		else if ( ay > by ) {
			return false;
		}
		else {
			if ( az < bz ) {
				return true;
			}
			else {
				return false;
			}
		}
	}
}
