/** @file 
* @brief Base classes to build a molecule
* @author Yuan ZHAO, Tiejun CHENG
* @date 2008/08/24
* @version 1.0.1 
*
*/
#include <ccbg/CCBase.h>

CCBase::CCBase()
    : // m_vdata,
      m_flags( 0 )
{
    //
}

void CCBase::Clear()
{
    m_vdata.clear();
    m_flags = 0;
}

/** 
* @brief destroy generic data.
*
* Release generic data in m_vdata.
*/
void CCBase::DestroyGD()
{
    vector<CCGenericData*>::iterator x = m_vdata.begin();
    for ( ; x != m_vdata.end(); ++x ) {
        if ( *x ) {
            delete * x;
            *x = NULL;
        }
    }
    m_vdata.clear();
}

/** 
* @brief copy CCBase data.
* @param[in] src source for data copy.
*
* Copy generic data and flag data from src.
*/
void CCBase::Clone( const CCBase& src )
{
    vector<CCGenericData*>::const_iterator it = src.m_vdata.begin();
    for ( ; it != src.m_vdata.end() && (*it); ++it ) {
        m_vdata.push_back( ( *it )->Clone() );
    }
    m_flags = src.m_flags;
}

/** 
* @brief get first generic data of certain attribute.
* @param[in] attr string type of attribute.
* @return first generic data of certain attribute.
*
* Get the first generic data whose attribute is equal to attr.
* If no generic data has required attribute, return NULL.
*/
CCGenericData* CCBase::GetData( const string& attr ) const
{
    vector<CCGenericData*>::const_iterator i = m_vdata.begin();
    for ( ; i != m_vdata.end() && (*i); ++i ) {
        if ( ( *i )->GetAttribute() == attr ) {
            return( *i );
        }
    }
    return( NULL );
}

/** 
* @brief get first generic data of certain data type.
* @param[in] type int type of data type.
* @return first generic data of certain data type.
*
* Get the first generic data whose data type is equal to type.
* If no generic data has required type, return NULL.
*/
CCGenericData* CCBase::GetData( const int type ) const
{
    vector<CCGenericData*>::const_iterator i = m_vdata.begin();
    for ( ; i != m_vdata.end() && (*i); ++i ) {
        if ( ( *i )->GetDataType() == type ) {
            return( *i );
        }
    }
    return( NULL );
}

/** 
* @brief get all generic data of certain type.
* @param[in] source DataOrigin type of data.
* @return all generic data of certain type.
*
* Get the first generic data whose DataOrigin type is 
* equal to source. If no generic data has required 
* DataOrigin type, return NULL.
*/
vector<CCGenericData*> CCBase::GetData( DataOrigin source ) const
{
    vector<CCGenericData*> filtered; // filtered data only from source

    vector<CCGenericData*>::const_iterator i = m_vdata.begin();
    for ( ; i != m_vdata.end() && (*i); ++i ) {
        if ( ( *i )->GetOrigin() == source ) {
            filtered.push_back( ( *i ) );
        }
    }
    return filtered;
}

/** 
* @brief check whether it has certain type of attribute.
* @param[in] attr string type of attribute.
* @return whether it has certain type of attribute.
*
* Returns true if the generic attribute/value pair exists.
*/
bool CCBase::HasData( const string& attr ) const
    //returns true if the generic attribute/value pair exists
{
    if ( m_vdata.empty() ) {
        return false;
    }

    vector<CCGenericData*>::const_iterator i = m_vdata.begin();
    for ( ; i != m_vdata.end() && (*i); ++i ) {
        if ( ( *i )->GetAttribute() == attr ) {
            return true;
        }
    }
    return false;
}

/** 
* @brief check whether it has certain type of data type.
* @param[in] type int type of data type.
* @return whether it has certain type of data type.
*
* Returns true if the generic attribute/value pair exists.
*/
bool CCBase::HasData( const int type ) const
{
    //returns true if the generic attribute/value pair exists
    if ( m_vdata.empty() ) {
        return false;
    }

    vector<CCGenericData*>::const_iterator i = m_vdata.begin();
    for ( ; i != m_vdata.end() && (*i); ++i ) {
        if ( ( *i )->GetDataType() == type ) {
            return true;
        }
    }
    return false;
}

