/** @file
* @brief Vector of bits.
* @author Yuan ZHAO, Tiejun CHENG
* @date 2008/08/24
* @version 1.0.1
*
*/

#include <ccbg/CCBitVec.h>
#include <ccbg/CCError.h>

/**
* @brief clear the data of m_set.
*
* Make all bit as 0.
*/
void CCBitVec::Clear()
{
    vector<int>::iterator i = m_set.begin();
    for ( ; i != m_set.end(); ++i ) {
        *i = 0;
    }
}

/**
* @brief CCBitVec constructor.
*
* Reset the bit vector size as 1.
*/
CCBitVec::CCBitVec()
{
    m_set.resize( 1 );
    Clear();
}

/**
* @brief initialize CCBitVec with an integer.
* @param[in] bits length for resize.
*
* Reset the bit vector size to bits length.
*/
CCBitVec::CCBitVec( int bits )
{
    Resize( bits );
    Clear();
}

/**
* @brief initialize CCBitVec with a string.
* @param[in] s string for initializing.
*
* Reset the bit vector size to the size of s and
* make each bit as the element of s.
* @note s should be like "011010".
*/
CCBitVec::CCBitVec( const string s )
{
    //s = "011010"
    Resize( s.size() );

    for ( int i = 0; i < s.size(); ++i ) {
        if ( s[i] - 48 ) {
            Set( i );
        }
    }
}

CCBitVec::CCBitVec( const CCBitVec& bv )
: m_set( bv.m_set )
{
}

/**
* @brief resize the size of m_set to bits length.
* @param[in] bits length for resize.
*
* Resize the bit vector's size to bits length.
* change the capability of m_set without changing data in it
*/
void CCBitVec::Resize( int bits )
{
    int word = bits / WORDSIZE;

    // increase only
    if ( word >= Size() ) {
        // bits=64 vs. 32
        m_set.resize( word + 1 );
    }
}

/**
* @brief set the pos bit on.
* @param[in] pos position for set bit on.
*
* Set the pos bit on without change the size of m_set.
*/
void CCBitVec::Set( int pos )
{
    if ( pos<0 || pos>Bits() - 1 ) {
        gErrorLog.Throw( __FUNCTION__, "pos out of range.", ccWarning );
        return;
    }
    m_set[pos / WORDSIZE] |= ( 1 << pos % WORDSIZE );
}

/**
* @brief set all bits on.
*
* Set all elements in m_set as 1.
*/
void CCBitVec::Set()
{
    vector<int>::iterator i = m_set.begin();
    for ( ; i != m_set.end(); ++i ) {
        *i = ~0;
    }
}

/**
* @brief set the pos bit off.
* @param[in] pos position for set bit off.
*
* Set the pos bit off without change the size of m_set.
*/
void CCBitVec::Reset( int pos )
{
    if ( pos<0 || pos>Bits() - 1 ) {
        gErrorLog.Throw( __FUNCTION__, "pos out of range.", ccWarning );
        return;
    }
    m_set[pos / WORDSIZE] &= ~( 1 << pos % WORDSIZE );
}

/**
* @brief set all bits off.
*
* Set all elements in m_set as 0.
*/
void CCBitVec::Reset()
{
    Clear();
}

/**
* @brief flip the pos bit.
* @param[in] pos position for flib.
*
* If the pos bit in on, set it as off. Otherwise
* set it as on.
*/
void CCBitVec::Flip( int pos )
{
    if ( pos<0 || pos>Bits() - 1 ) {
        gErrorLog.Throw( __FUNCTION__, "pos out of range.", ccWarning );
        return;
    }
    m_set[pos / WORDSIZE] ^= 1 << ( pos % WORDSIZE );
}

/**
* @brief flip all bits.
*
* If the pos bit in on, set it as off. Otherwise
* set it as on.
*/
void CCBitVec::Flip()
{
    vector<int>::iterator i = m_set.begin();
    for ( ; i != m_set.end(); ++i ) {
        *i = ~( *i );
    }
}

/**
* @brief initialize all bits from a string.
* @param[in] s string used to format bit vector.
*
* Call CCBitVec::CCBitVec( const string s ) to
* initialize a bit vector.
* @note s should be like "011010".
*/
void CCBitVec::FromString( const string& s )
{
    CCBitVec bv( s );
    *this = bv;
}

/**
* @brief output all bits to a string.
* @return all element of bit vector in string format
*
* The return value is like "011010".
*/
string CCBitVec::ToString() const
{
    stringstream ss;

    for ( int i = 0; i < Bits(); ++i ) {
        ss << int( this->operator[]( i ) );
    }

    return ss.str();
}

/**
* @brief initialize from an int vector of indices where bits are set to 1
* @param[in] v integer vector used to set bit vector.
*
* The size of the current bit vector will be resized as
* the max element of v.
*
*/
void CCBitVec::FromTrueIdx( vector<int>& v )
{
    Clear();

    int maxbit = *( std::max_element( v.begin(), v.end() ) );

    Resize( maxbit );

    vector<int>::iterator i = v.begin();
    for ( ; i != v.end(); ++i ) {
        Set( *i );
    }
}

/**
* @brief return indices where bits are set to 1
* @return indices of bit vector where bits are set to 1
*
*/
vector<int> CCBitVec::ToTureIdx() const
{
    vector<int> v;
    for ( int i = 0; i < Bits(); ++i ) {
        if ( Test( i ) ) {
            v.push_back( i );
        }
    }
    return v;
}

/**
* @brief initialize all bits from an integer vector.
* @param[in] v integer vector used to set bit vector.
* @note v should be like {0,1,1,0,1,0,1}.
*
* The size of the current bit vector will be resized as
* the size of v.
*
*/
void CCBitVec::FromVecInt( vector<int>& v )
{
	Clear();

	Resize( (int) v.size() );

	for ( int i=0; i < (int) v.size(); ++i ) {
		switch (v[i]) {
		case 1:
			Set(i); break;
		case 0:
			Reset(i); break;
		default:
			gErrorLog.Throw(__FUNCTION__, 
				"only element 0 or 1 is allowed.", ccFatal);
			exit( EXIT_FAILURE);
		}
	}
}

/**
* @brief output all bits an integer vector.
* @return all element of bit vector in an integer vector.
*
* The return value is like {0,1,1,0,1,0,1}.
*/
vector<int> CCBitVec::ToVecInt() const
{
	vector<int> v;
	for ( int i = 0; i < Bits(); ++i ) {
		v.push_back(this->operator [](i));
	}
	return v;
}

/**
* @brief return the total number of on bit.
* @return the total number of on bit.
*/
int CCBitVec::Count() const
{
    int count = 0;
    for ( int i = 0; i < Bits(); ++i ) {
        if ( Test( i ) ) {
            count++;
        }
    }
    return count;
}

/**
* @brief check whether at least one bit is on.
* @return whether at least one bit is on.
*
* It is opposite to None.
*
* @see CCBitVec::None()
*/
bool CCBitVec::Any() const
{
    return !None();
}

/**
* @brief check whether all bits is off.
* @return whether all bits is off.
*/
bool CCBitVec::None() const
{
    vector<int>::const_iterator i = m_set.begin();
    for ( ; i != m_set.end(); ++i ) {
        if ( *i ) {
            return false;
        }
    }
    return true;
}

/**
* @brief check whether all bits is on.
* @return whether all bits is on.
*/
bool CCBitVec::All() const
{
    vector<int>::const_iterator i = m_set.begin();
    for ( ; i != m_set.end(); ++i ) {
        if ( !( *i ) ) {
            return false;
        }
    }
    return true;
}

/**
* @brief test the state of pos bit.
* @return whether pos bit is on.
*
* It is the same as operator [].
*
* @see CCBitVec::operator[]()
*/
bool CCBitVec::Test( int pos ) const
{
    if ( pos<0 || pos>Bits() - 1 ) {
        gErrorLog.Throw( __FUNCTION__, "pos out of range.", ccWarning );
        return false;
    }
    return ( pos / WORDSIZE >= Size() )
         ? false
         : ( m_set[pos / WORDSIZE] >> ( pos % WORDSIZE ) & 1 );
}

/**
* @brief check whether pos bit is on.
* @return whether pos bit is on.
*
* It calls Test function.
*
* @see CCBitVec::Test()
*/
bool CCBitVec::operator[]( int pos ) const
{
    return Test( pos );
}

/** @brief CCBitVec copy assignment operator */
/**
* @brief CCBitVec copy assignment operator.
* @param[in] bv bit vector for copying.
* @return a bit vector with the same element as in bv.
*/
CCBitVec& CCBitVec::operator =( const CCBitVec& bv )
{
    if ( this != &bv ) {
        if ( Size() != bv.Size() ) {
            Resize( bv.Bits() );
        }

        int i;
        for ( i = 0; i < bv.Size(); ++i ) {
            m_set[i] = bv.m_set[i];
        }
        for ( ; i < Size(); ++i ) {
            m_set[i] = 0;
        }
    }
    return *this;
}

/**
* @brief bit and assignment operator.
* @param[in] bv bit vector for and operation.
* @return bit and result of current bit vector and bv.
*
* First calculate the min size of current bit vector and bv.
* Do and operation on min size bits of current bit vector and bv,
* then save the result in current bit vector. If the size of bv
* is larger than current bit vector, just return current bit vector.
* Otherwise, make the rest bits in current bit vector as 0.
* This operator will not change the size of current bit vector.
*/
CCBitVec& CCBitVec::operator &=( const CCBitVec& bv )
{
    int i;
    int min = ( bv.Size() < Size() ) ? bv.Size() : Size();

    for ( i = 0; i < min; i++ ) {
        m_set[i] &= bv.m_set[i];
    }
    // clear rest bits
    // 0 & a = 0
    for ( ; i < Size(); i++ ) {
        m_set[i] = 0;
    }

    return *this;
}

/**
* @brief bit or assignment operator.
* @param[in] bv bit vector for or operation.
* @return bit or result of current bit vector and bv.
*
* First calculate the min size of current bit vector and bv.
* Do or operation on min size bits of current bit vector and bv,
* then save the result in current bit vector. If the size of bv
* is larger than current bit vector, just return current bit vector.
* Otherwise, keep the rest bits in current bit vector.
* This operator will not change the size of current bit vector.
*/
CCBitVec& CCBitVec::operator |=( const CCBitVec& bv )
{
    int min = ( bv.Size() < Size() ) ? bv.Size() : Size();

    for ( int i = 0; i < min; i++ ) {
        m_set[i] |= bv.m_set[i];
    }
    // keep rest bits
    // 0 | a = a

    return *this;
}

/**
* @brief bit xor assignment operator.
* @param[in] bv bit vector for xor operation.
* @return bit xor result of current bit vector and bv.
*
* First calculate the min size of current bit vector and bv.
* Do xor operation on min size bits of current bit vector and bv,
* then save the result in current bit vector. If the size of bv
* is larger than current bit vector, just return current bit vector.
* Otherwise, keep the rest bits in current bit vector.
* This operator will not change the size of current bit vector.
*/
CCBitVec& CCBitVec::operator ^=( const CCBitVec& bv )
{
    int min = ( bv.Size() < Size() ) ? bv.Size() : Size();

    for ( int i = 0; i < min; i++ ) {
        m_set[i] ^= bv.m_set[i];
    }
    // keep rest bits
    // 0 ^ a = a

    return *this;
}

/**
* @brief equal operator of two CCBitVec.
* @param[in] bv1 the first bit vector for comparing.
* @param[in] bv2 the second bit vector for comparing.
* @return whether these two bit vector is equal.
*
* If bv1 and bv2 have different size, just return false.
* Else compare each element in bv1 and bv2.
*/
bool operator ==( const CCBitVec& bv1, const CCBitVec& bv2 )
{
    if ( bv1.Size() != bv2.Size() ) {
        return false;
    }

    for ( int i = 0; i < bv1.Size(); ++i ) {
        if ( bv1.m_set[i] != bv2.m_set[i] ) {
            return false;
        }
    }

    return true;
}

/**
* @brief unequal operator of two CCBitVec.
* @param[in] bv1 the first bit vector for comparing.
* @param[in] bv2 the second bit vector for comparing.
* @return whether these two bit vector is unequal.
*
* If bv1 and bv2 have different size, just return true.
* Else compare each element in bv1 and bv2.
*/
bool operator !=( const CCBitVec& bv1, const CCBitVec& bv2 )
{
    return !( bv1 == bv2 );
}

/**
* @brief bit and operator of two CCBitVec.
* @param[in] bv1 the first bit vector for and operation.
* @param[in] bv2 the second bit vector for and operation.
* @return bit and result of bv1 and bv2.
*
* It first use bv1 to new a bit vector. Then call bit and
* assignment operator.
*
* @see CCBitVec::operator&=()
*/
CCBitVec operator &( const CCBitVec& bv1, const CCBitVec& bv2 )
{
    CCBitVec bv( bv1 );
    bv &= bv2;

    return bv;
}

/**
* @brief bit or operator of two CCBitVec.
* @param[in] bv1 the first bit vector for or operation.
* @param[in] bv2 the second bit vector for or operation.
* @return bit or result of bv1 and bv2.
*
* It first use bv1 to new a bit vector. Then call bit or
* assignment operator.
*
* @see CCBitVec::operator|=()
*/
CCBitVec operator |( const CCBitVec& bv1, const CCBitVec& bv2 )
{
    CCBitVec bv( bv1 );;
    bv |= bv2;

    return bv;
}

/**
* @brief bit xor operator of two CCBitVec.
* @param[in] bv1 the first bit vector for xor operation.
* @param[in] bv2 the second bit vector for xor operation.
* @return bit xor result of bv1 and bv2.
*
* It first use bv1 to new a bit vector. Then call bit xor
* assignment operator.
*
* @see CCBitVec::operator^=()
*/
CCBitVec operator ^( const CCBitVec& bv1, const CCBitVec& bv2 )
{
    CCBitVec bv( bv1 );
    bv ^= bv2;

    return bv;
}

/**
* @brief overload input operator on CCBitVec.
* @param[in] is istream object.
* @param[in] bv CCBitVec for input.
* @return istream.
*
* This function read each element in bv to an istream
* object.
*/
istream& operator >>( istream& is, CCBitVec& bv )
{
    string s;
    is >> s;
	bv.FromString(s);

    /*bv.Resize( s.size() );

    for ( int i = 0; i < s.size(); ++i ) {
        if ( s[i] - 48 ) {
            bv.Set( i );
        }
    }*/

    return is;
}

/**
* @brief overload output operator on CCBitVec.
* @param[in] is ostream object.
* @param[in] bv CCBitVec for output.
* @return ostream.
*
* This function output each element in bv to an ostream
* object. The output format is like "[ 0 1 0 0 1 1 1 ] ".
*/
ostream& operator <<( ostream& os, const CCBitVec& bv )
{
    os << "[ ";
    for ( int i = 0; i < bv.Size(); ++i ) {
        for ( int j = 0; j < WORDSIZE; ++j ) {
            os << ( bv.m_set[i] >> j & 1 ) << ' ';
        }
    }
    os << "]";

    return os;
}

