/** @file 
* @brief CCBond class and relevant functions.
* @author Yuan ZHAO, Tiejun CHENG
* @date 2008/08/26
* @version 1.0.1 
*
*/

#include <ccbg/CCBond.h>
#include <ccbg/CCMol.h>

/** @brief CCBond constructor. */
CCBond::CCBond()
    : m_mol( NULL ), m_idx( 0 ), m_bgn( NULL ), m_end( NULL ), m_order( 0 )
                   //m_type

{
    CCBase::CCBase();
}

/** 
* @brief CCBond copy constructor.
* @param[in] src bond for copy.
*/
CCBond::CCBond( const CCBond& src )
    : m_mol( src.m_mol ), m_idx( src.m_idx ), m_bgn( src.m_bgn ),
      m_end( src.m_end ), m_order( src.m_order ), m_type( src.m_type )
{
    CCBase::Clone( src );
}

/** 
* @brief CCBond copy assignment operator.
* @param[in] src bond for copy assignment.
*/
CCBond& CCBond::operator =( const CCBond& src )
{
    if ( this != &src ) {
        CCBase::Clone( src );
        m_mol = src.m_mol;
        m_idx = src.m_idx;
        m_bgn = src.m_bgn;
        m_end = src.m_end;
        m_order = src.m_order;
        m_type = src.m_type;
    }
    return *this;
}

/** @brief CCBond destructor. */
CCBond::~CCBond()
{
	CCBase::DestroyGD();
}

/** @brief clear members of CCBond. */
void CCBond::Clear()
{
	CCBase::Clear();
	m_mol = NULL;
	m_idx = 0;
	m_bgn = NULL;
	m_end = NULL;
	m_order = 0;
	m_type.clear();
}

///GET///GET///GET///GET///GET///GET///GET///GET///GET///GET///GET///GET//
/**
* @brief get internal bond type.
* @return internal bond type.
*
* First check whether the parent molecule has the flag of
* MOL_INT_TYPE_PERCEIVED. If not, call molecular class
* function PerceiveInternalAtomBondTypes() to perceive
* internal atom and bond types. Then return the internal 
* bond type of current atom.
*
* @see CCMol::PerceiveInternalAtomBondTypes()
*/
string CCBond::GetType() const
{
	if ( m_mol && !m_mol->HasFlag( MOL_INT_TYPE_PERCEIVED ) ) {
		m_mol->PerceiveInternalAtomBondTypes();
	}

	return m_type;
}

/**
* @brief get the atom in current bond which is different from curr atom.
* @param[in] curr one of the atom in current bond.
* @return the atom in current bond which is different from curr atom.
*
* If curr atom is not in current bond, just return NULL. Else return
* the atom in current bond which is different from curr atom.
*/
CCAtom* CCBond::GetNbrAtom( CCAtom* curr ) const
{
	if ( curr == m_bgn ) {
		return m_end;
	}
	else if ( curr == m_end ) {
		return m_bgn;
	}
	else {
		gErrorLog.Throw( __FUNCTION__,
			"Atom is not in Bond or NULL atom pointer!",
			ccWarning );
		return NULL;
	}
}

/**
* @brief get the index of atom in current bond which is
* different from curr atom.
* @param[in] curr one of the atom in current bond.
* @return the index of atom in current bond which is
* different from curr atom.
*
* If curr atom is not in current bond, just return -1. 
* Else return the index of the atom in current bond which
* is different from curr atom.
*/
int CCBond::GetNbrAtomIdx( CCAtom* curr ) const
{
	if ( curr == m_bgn ) {
		return m_end->GetIdx();
	}
	else if ( curr == m_end ) {
		return m_bgn->GetIdx();
	}
	else {
		gErrorLog.Throw( __FUNCTION__,
			"Atom is not in Bond or NULL atom pointer!",
			ccWarning );
		return -1;
	}
}

/**
* @brief get the index of beginning atom in current bond.
* @return the index of beginning atom in current bond.
*/
int CCBond::GetBeginAtomIdx()           const
{
    return m_bgn->GetIdx();
}

/**
* @brief get the index of end atom in current bond.
* @return the index of end atom in current bond.
*/
int CCBond::GetEndAtomIdx()             const
{
    return m_end->GetIdx();
}

/**
* @brief get bond length
* @return bond length of current bond.
*/
double CCBond::GetLength() const
{
	return Distance( m_bgn->GetXYZ(), m_end->GetXYZ() );
}

/** @brief get bond length. */
/**
* @brief get standard bond length of current bond.
* @return standard bond length of current bond.
*
* @see gRBLTable::GetRefBondLength()
*/
double CCBond::GetSTLength() const
{
	return gRBLTable.GetRefBondLength( m_bgn->GetAtomicNum(),
		m_end->GetAtomicNum() );
}

/** @brief get the coordinates of the center of current bond. */
CCVector CCBond::GetCenter() const
{
	return ( m_bgn->GetXYZ() + m_end->GetXYZ() ) / 2.0;
}

///SET///SET///SET///SET///SET///SET///SET///SET///SET///SET///SET///SET//
/** 
* @brief set internal bond type from bond order.
*
* If current bond is aromatic bond, set bond type as "AR";
* If current bond is amide bond, set bond type as "AM";
* If current bond is tripos bond, set bond type as "TB";
* If current bond is double bond, set bond type as "DB";
* If current bond is single bond, set bond type as "SB";
* Else set bond type as "DU";
*/
void CCBond::SetTypeFromBO()
{
    if ( IsAromatic() ) {
        SetType( "AR" );
    }
    else if ( IsAmide() ) {
        SetType( "AM" );
    }
    else if ( IsTriple() ) {
        SetType( "TB" );
    }
    else if ( IsDouble() ) {
        SetType( "DB" );
    }
    else if ( IsSingle() ) {
        SetType( "SB" );
    }
    else {
        SetType( "DU" );
    }
}

/** @brief set bond index, beginning atom, end atom, order and flag. */
void CCBond::Set( int idx, CCAtom* bgn, CCAtom* end, int order,
                  unsigned long flag )
{
    SetIdx( idx );
    SetBegin( bgn );
    SetEnd( end );
    SetOrder( order );
    SetFlag( flag );
}

/** 
* @brief assign BOND_IS_KSINGLE flag to current bond.
*
* It will clear the BOND_IS_KSINGLE, BOND_IS_KDOUBLE and
* BOND_IS_KTRIPLE flag in current, then set BOND_IS_KSINGLE flag.
*/
void CCBond::SetKSingle()
{
	UnsetFlag( BOND_IS_KSINGLE | BOND_IS_KDOUBLE | BOND_IS_KTRIPLE );
	SetFlag( BOND_IS_KSINGLE );
}

/** 
* @brief assign BOND_IS_KDOUBLE flag to current bond.
*
* It will clear the BOND_IS_KSINGLE, BOND_IS_KDOUBLE and
* BOND_IS_KTRIPLE flag in current, then set BOND_IS_KDOUBLE flag.
*/
void CCBond::SetKDouble()
{
	UnsetFlag( BOND_IS_KSINGLE | BOND_IS_KDOUBLE | BOND_IS_KTRIPLE );
	SetFlag( BOND_IS_KDOUBLE );
}

/** 
* @brief assign BOND_IS_KTRIPLE flag to current bond.
*
* It will clear the BOND_IS_KSINGLE, BOND_IS_KDOUBLE and
* BOND_IS_KTRIPLE flag in current, then set BOND_IS_KTRIPLE flag.
*/
void CCBond::SetKTriple()
{
	UnsetFlag( BOND_IS_KSINGLE | BOND_IS_KDOUBLE | BOND_IS_KTRIPLE );
	SetFlag( BOND_IS_KTRIPLE );
}

////IS////IS////IS////IS////IS////IS////IS////IS////IS////IS////IS////IS///
/**
* @brief check whether current bond is RECAP bond.
* @param strict whether perceive strict RECAP bond
* @return whether current bond is RECAP bond.
*
* First check whether the parent molecule has the flag of
* MOL_RECAPBOND_PERCEIVED. If not, call molecular class
* function PerceiveRECAPBond() to perceive RECAP bonds.
* Then check whether current bond has BOND_IS_RECAP flag.
* If the parameter strict is ture, the RECAP bond will be the
* same as defined in ref:
* Lewell, X. Q.; Judd, D. B.; Watson, S. P.; Hann, M. M. 
* RECAP-Retrosynthetic Combinatorial Analysis Procedure: A 
* Powerful New Technique for Identifying Privileged Molecular
* Fragments with Useful Applications in Combinatorial Chemistry. 
* J. Chem. Inf. Comput. Sci. 1998, 38, (3), 511-522.
* else the RECAP bond is defined as *-N, *-O, C(sp2)=C(sp2) and
* C/N(ar)-C/N(ar).
*
* @see CCMol::PerceiveRECAPBond()
*/
bool CCBond::IsRECAP( bool strict /*= false*/ )
{
	if ( m_mol && !m_mol->HasFlag( MOL_RECAPBOND_PERCEIVED ) ) {
		m_mol->PerceiveRECAPBond( strict );
	}

	return HasFlag( BOND_IS_RECAP );
}

/**
* @brief check whether current bond is aromatic bond.
* @return whether current bond is aromatic bond.
*
* First check whether the parent molecule has the flag of
* MOL_AROMATICITY_PERCEIVED. If not, call molecular class
* function PerceiveAromaticity() to perceive aromaticity.
* Then check whether current bond has BOND_IS_AROMATIC flag.
*
* @see CCMol::PerceiveAromaticity()
*/
bool CCBond::IsAromatic()
{
	// ignore bond aromatic flag before detection
	if ( m_mol && !m_mol->HasFlag( MOL_AROMATICITY_PERCEIVED ) ) {
		m_mol->PerceiveAromaticity();
	}

	return HasFlag( BOND_IS_AROMATIC );
}

/**
* @brief check whether current bond is in ring.
* @return whether current bond is in ring.
*
* First check whether current bond has BOND_IS_IN_RING flag.
* If it has the flag, just return true. Else, check whether the
* parent molecule has the flag of MOL_AROMATICITY_PERCEIVED.
* If not, call molecular class function PerceiveRingAtomsAndBonds()
* to perceive ring atoms and bonds. Then check whether current bond
* has BOND_IS_IN_RING flag.
*
* @see CCMol::PerceiveRingAtomsAndBonds()
*/
bool CCBond::IsInRing()
{
	if ( HasFlag( BOND_IS_IN_RING ) ) {
		return true;
	}

	if ( m_mol && !m_mol->HasFlag( MOL_RING_PERCEIVED ) ) {
		m_mol->PerceiveRingAtomsAndBonds();
	}

	return HasFlag( BOND_IS_IN_RING );
}

/**
* @brief check whether current bond is closure ring bond.
* @return whether current bond is closure ring bond.
*
* First check whether current bond has BOND_IS_CLOSURE flag.
* If it has the flag, just return true. Else, check whether the
* parent molecule has the flag of MOL_CLOSURE_BOND_PERCEIVED.
* If not, call molecular class function PerceiveClosureBond()
* to perceive closure ring bonds. Then check whether current bond
* has BOND_IS_CLOSURE flag.
*
* @see CCMol::PerceiveClosureBond()
*/
bool CCBond::IsClosure()
{
	if ( HasFlag( BOND_IS_CLOSURE ) ) {
		return true;
	}

	if ( m_mol && !m_mol->HasFlag( MOL_CLOSURE_BOND_PERCEIVED ) ) {
		m_mol->PerceiveClosureBond();
	}

	return HasFlag( BOND_IS_CLOSURE );
}

/**
* @brief check whether current bond is rotatable bond.
* @return whether current bond is rotatable bond.
*
* First check whether current bond has BOND_IS_ROTATABLE flag.
* If it has the flag, just return true. Else, check whether the
* parent molecule has the flag of MOL_ROTATABLEBOND_PERCEIVED.
* If not, call molecular class function PerceiveRotatableBond()
* to perceive ratatable bonds. Then check whether current bond
* has BOND_IS_ROTATABLE flag.
*
* @see CCMol::PerceiveClosureBond()
*/
bool CCBond::IsRotatable()
{
	if ( HasFlag( BOND_IS_ROTATABLE ) ) {
		return true;
	}

	if ( m_mol && !m_mol->HasFlag( MOL_ROTATABLEBOND_PERCEIVED ) ) {
		m_mol->PerceiveRotatableBond();
	}

	return HasFlag( BOND_IS_ROTATABLE );
}

/**
* @brief check whether current bond is tautomer aromatic bond.
* @return whether current bond is tautomer aromatic bond.
*
* Tautomer aromatic bond is defined as the aromatic bond whose
* order can be 1 or 2, such as each bonds in benzene.<br>
* First check whether current bond has BOND_IS_TAUTOMERAR flag.
* If it has the flag, just return true. Else, check whether the
* parent molecule has the flag of MOL_TAUTOMERARBOND_PERCEIVED.
* If not, call molecular class function PerceiveTautomerARBond()
* to perceive tautomer aromatic bonds. Then check whether current
* bond has BOND_IS_TAUTOMERAR flag.
*
* @see CCMol::PerceiveTautomerARBond()
*/
bool CCBond::IsTautomerAR()
{
	if ( HasFlag( BOND_IS_TAUTOMERAR ) ) {
		return true;
	}

	if ( m_mol && !m_mol->HasFlag( MOL_TAUTOMERARBOND_PERCEIVED ) ) {
		m_mol->PerceiveTautomerARBond();
	}

	return HasFlag( BOND_IS_TAUTOMERAR );
}

/**
* @brief check whether current bond is single bond.
* @return whether current bond is single bond.
*
* First check whether the parent molecule has the flag of
* MOL_AROMATICITY_PERCEIVED. If not, call molecular class
* function PerceiveAromaticity() to perceive aromaticity of
* the parent molecule. Then check whether current bond is not
* aromatic bond and its bond order is 1.
*
* @see CCMol::PerceiveAromaticity()
*/
bool CCBond::IsSingle()
{
    if ( m_mol && !m_mol->HasFlag( MOL_AROMATICITY_PERCEIVED ) ) {
        m_mol->PerceiveAromaticity();
    }

    if ( GetOrder() == 1 && !HasFlag( BOND_IS_AROMATIC ) ) {
        return true;
    }

    return false;
}

/**
* @brief check whether current bond is double bond.
* @return whether current bond is double bond.
*
* First check whether the parent molecule has the flag of
* MOL_AROMATICITY_PERCEIVED. If not, call molecular class
* function PerceiveAromaticity() to perceive aromaticity of
* the parent molecule. Then check whether current bond is not
* aromatic bond and its bond order is 2.
*
* @see CCMol::PerceiveAromaticity()
*/
bool CCBond::IsDouble()
{
    if ( m_mol && !m_mol->HasFlag( MOL_AROMATICITY_PERCEIVED ) ) {
        m_mol->PerceiveAromaticity();
    }

    if ( GetOrder() == 2 && !HasFlag( BOND_IS_AROMATIC ) ) {
        return true;
    }

    return false;
}

/**
* @brief check whether current bond is triple bond.
* @return whether current bond is triple bond.
*
* First check whether the parent molecule has the flag of
* MOL_AROMATICITY_PERCEIVED. If not, call molecular class
* function PerceiveAromaticity() to perceive aromaticity of
* the parent molecule. Then check whether current bond is not
* aromatic bond and its bond order is 3.
*
* @see CCMol::PerceiveAromaticity()
*/
bool CCBond::IsTriple()
{
    if ( m_mol && !m_mol->HasFlag( MOL_AROMATICITY_PERCEIVED ) ) {
        m_mol->PerceiveAromaticity();
    }

    if ( GetOrder() == 3 && !HasFlag( BOND_IS_AROMATIC ) ) {
        return true;
    }

    return false;
}

/**
* @brief check whether current bond is amide bond.
* @return whether current bond is amide bond.
*
* Only for C-N bond in O=C-N.
*/
bool CCBond::IsAmide()
{
	if ( GetOrder() != 1 || !m_bgn || !m_end ) {
		return false;
	}

	CCAtom* pcarbon;

	if ( m_bgn->IsCarbon()
		&& ( m_end->IsNitrogen() && !m_end->HasBondWithLargerOrder( 1 ) ) ) {
			pcarbon = m_bgn;
	}
	else if ( m_end->IsCarbon()
		&& ( m_bgn->IsNitrogen() && !m_bgn->HasBondWithLargerOrder( 1 ) ) ) {
			pcarbon = m_end;
	}
	else {
		return false;
	}

	FOR_BONDS_OF_ATOM( nb, pcarbon )
		if ( nb->GetOrder() == 2 && ( (CCAtom*)~nb )->IsOxygen() ) {
			return true;
		}

	return false;
}

/**
* @brief check whether current bond is sulfone amide bond.
* @return whether current bond is sulfone amide bond.
*
* Only for N-S(=O)=O bond.
*/
bool CCBond::IsSulfoneAmide() {
	if ( GetOrder() != 1 || !m_bgn || !m_end ) {
		return false;
	}

	CCAtom* pcarbon;

	if ( m_bgn->IsSulfur()
		&& ( m_end->IsNitrogen() && !m_end->HasBondWithLargerOrder( 1 ) ) ) {
			pcarbon = m_bgn;
	}
	else if ( m_end->IsSulfur()
		&& ( m_bgn->IsNitrogen() && !m_bgn->HasBondWithLargerOrder( 1 ) ) ) {
			pcarbon = m_end;
	}
	else {
		return false;
	}

	int doublebond = 0;
	FOR_BONDS_OF_ATOM( nb, pcarbon ) {
		if ( nb->GetOrder() == 2 && ( (CCAtom*)~nb )->IsOxygen() ) {
			++doublebond;
		}
	}

	if ( doublebond >= 2 ) {
		return true;
	}

	return false;
}

/**
* @brief check whether current bond is ester bond.
* @return whether current bond is ester bond.
*
* Only for C-O bond in O=C-O.
*/
bool CCBond::IsEster() {
	if ( GetOrder() != 1 || !m_bgn || !m_end ) {
		return false;
	}

	CCAtom* pcarbon;

	if ( m_bgn->IsCarbon()
		&& ( m_end->IsOxygen() && !m_end->HasBondWithLargerOrder( 1 ) ) ) {
			pcarbon = m_bgn;
	}
	else if ( m_end->IsCarbon()
		&& ( m_bgn->IsOxygen() && !m_bgn->HasBondWithLargerOrder( 1 ) ) ) {
			pcarbon = m_end;
	}
	else {
		return false;
	}

	FOR_BONDS_OF_ATOM( nb, pcarbon )
		if ( nb->GetOrder() == 2 && ( (CCAtom*)~nb )->IsOxygen() ) {
			return true;
		}

	return false;
}

/**
* @brief check whether current bond is kekulized single bond.
* @return whether current bond is kekulized single bond.
*
* First check whether the parent molecule has the flag of
* MOL_KEKULE_PERCEIVED. If not, call molecular class
* function NewPerceiveKekuleBonds() to do kekulize on the
* parent molecule. Then check whether current bond has 
* BOND_IS_KSINGLE flag.<br>
*
* @see CCMol::NewPerceiveKekuleBonds()
*
* @note after kekulizing, the single bond not in aromatic ring
* will always has BOND_IS_KSINGLE flag. The kekulized single bond
* in aromatic ring will also has BOND_IS_KSINGLE flag.
*/
bool CCBond::IsKSingle()
{
    if ( m_mol && !m_mol->HasFlag( MOL_KEKULE_PERCEIVED ) ) {
        m_mol->NewPerceiveKekuleBonds();
    }

    return HasFlag( BOND_IS_KSINGLE );
}

/**
* @brief check whether current bond is kekulized double bond.
* @return whether current bond is kekulized double bond.
*
* First check whether the parent molecule has the flag of
* MOL_KEKULE_PERCEIVED. If not, call molecular class
* function NewPerceiveKekuleBonds() to do kekulize on the
* parent molecule. Then check whether current bond has 
* BOND_IS_KDOUBLE flag.
*
* @see CCMol::NewPerceiveKekuleBonds()
*
* @note after kekulizing, the double bond not in aromatic ring
* will always has BOND_IS_KDOUBLE flag. The kekulized double bond
* in aromatic ring will also has BOND_IS_KDOUBLE flag.
*/
bool CCBond::IsKDouble()
{
    if ( m_mol && !m_mol->HasFlag( MOL_KEKULE_PERCEIVED ) ) {
        m_mol->NewPerceiveKekuleBonds();
    }

    return HasFlag( BOND_IS_KDOUBLE );
}

/**
* @brief check whether current bond is kekulized triple bond.
* @return whether current bond is kekulized triple bond.
*
* First check whether the parent molecule has the flag of
* MOL_KEKULE_PERCEIVED. If not, call molecular class
* function NewPerceiveKekuleBonds() to do kekulize on the
* parent molecule. Then check whether current bond has 
* BOND_IS_KTRIPLE flag.
*
* @see CCMol::NewPerceiveKekuleBonds()
*
* @note after kekulizing, the triple bond not in aromatic ring
* will always has BOND_IS_KTRIPLE flag. The kekulized triple bond
* in aromatic ring will also has BOND_IS_KTRIPLE flag.
*/
bool CCBond::IsKTriple()
{
    if ( m_mol && !m_mol->HasFlag( MOL_KEKULE_PERCEIVED ) ) {
        m_mol->NewPerceiveKekuleBonds();
    }

    return HasFlag( BOND_IS_KTRIPLE );
}

/** 
* @brief check whether current bond is near planar.
* @param[in] cutoff the minimum dihedral angle for a plan.
* @return whether current bond is near planar.
*
* If all the dihedral of current bond is less than cutoff value
* ( default 30.0 degree ), then defines the bond as near-planar
* bond. All the near-planar should form a muitibond.
*/
bool CCBond::IsNearPlanar( double cutoff /*= 30.0 */ )
{
    if ( m_bgn->GetValence() < 2 || m_end->GetValence() < 2 ) {
        // no dihedral angle exist
        return false;
    }
    double MAXPLANARTORSION = cutoff;

    FOR_NBORS_OF_ATOM( bgnneib, m_bgn ) {
        if ( &*bgnneib == m_end ) {
            continue;
        }
        FOR_NBORS_OF_ATOM( endneib, m_end ) {
            if ( &*endneib == m_bgn ) {
                continue;
            }
            double torsion = Torsion( bgnneib->GetXYZ(), m_bgn->GetXYZ(),
                                      m_end->GetXYZ(), endneib->GetXYZ() );
            if ( MinTorsion( torsion ) > MAXPLANARTORSION ) {
                return false;
            }
        }
    }
    return true;
}

/**
* @brief check whether current bond is polar bond.
* @return whether current bond is polar bond.
* 
* Polar is defined as the electronegativities difference 
* of two bonded atoms >0.45.
* 
* @see CCAtom::GetEleNeg()
*/
bool CCBond::IsPolar()
{
    return abs( m_bgn->GetEleNeg() - m_end->GetEleNeg() ) > 0.45;
}

/************************************************************************/
/* Bond related functions                                               */
/************************************************************************/
/**
* @brief check whether two atoms a1 and a2 can form an acceptable bond.
* @param[in] a1 the first atom.
* @param[in] a2 the second atom
* @param[in] cutoff deviation value ( default 0.4A)
* @return whether two atoms a1 and a2 can form an acceptable bond.
*
* It will return true if <br>
*        sqrt(0.4) <= distance <= ( Rcov1 + Rcov2 + cutoff )<br>
* Here distance is the distance of atom a1 and a2. Rcov1 and Rcov2 are
* the covalent radius of atom a1 and a2 respectively.
*
* @see CCAtom::GetCovRad()
*/
bool IsAcceptableBond( CCAtom* a1, CCAtom* a2, double cutoff /*= 0.4*/ )
{
    if ( !a1 || !a2 ) {
        return false;
    }

    double d = Distance( a1->GetXYZ(), a2->GetXYZ() );
    if ( d > ( a1->GetCovRad() + a2->GetCovRad() + cutoff ) // too far away
        || d < sqrt( 0.4 ) ) { // too close
        /*gErrorLog.Throw( __FUNCTION__,
                "bond length is out of range:\n"
                "sqrt(0.4) < d < (Rcov1 + Rcov2 + 0.4).",
                ccWarning );*/
        return false;
    }

    return true;
}

/**
* @brief check whether the bond is in a conjugated system.
* @param[in] bond bond for checking.
* @return whether the bond is in a conjugated system.
*
* Check the neighboring atoms of beginning and end atoms of
* this bond. If one of the neighboring atom has a multi-bond
* which not bonded to the atoms in this bond, return true.
*/
bool HasConjugateBond( CCBond* bond )
{
    CCAtom* bgn = bond->GetBeginAtom();
    CCAtom* end = bond->GetEndAtom();
    FOR_NBORS_OF_ATOM( atom, bgn ) {
        if ( &*atom != end && atom->HasBondWithLargerOrder( 1 ) ) {
            return true;
        }
    }
    FOR_NBORS_OF_ATOM( atom, end ) {
        if ( &*atom != bgn && atom->HasBondWithLargerOrder( 1 ) ) {
            return true;
        }
    }
    return false;
}

/**
* @brief check whether b1 is more close to form a double bond than b2.
* @param[in] b1 the first bond for comparing.
* @param[in] b2 the second bond for comparing.
* @return whether b1 is more close to double bond than b2.
*
* If b1 is in a conjugated system but not b2, return true.
* If b2 is in a conjugated system but not b1, return false.
* Otherwise, compare the deviation value of b1's length to double bond
* and the deviation value of b2's length to double bond. Return which
* one is smaller.<br>
* This function is used to check which bond has more possibility to form
* a double bond.
*/
bool DoubleBondDev( CCBond* b1, CCBond* b2 )
{
    bool b1conju = HasConjugateBond( b1 );
    bool b2conju = HasConjugateBond( b2 );
    // b1 has conjugated bond but b2 does not
    if ( b1conju && !b2conju ) {
        return true;
    }

    // b2 has conjugated bond but b1 does not
    if ( !b1conju && b2conju ) {
        return false;
    }

    return ( ( b1->GetLength() - ( b1->GetSTLength() - 0.11 ) )
           < ( b2->GetLength() - ( b2->GetSTLength() - 0.11 ) ) );
}

