/** @file 
* @brief Perceive R/S chirality and E/Z stereo of each atom.
* @author Yuan ZHAO, Tiejun CHENG
* @date 2008/08/26
* @version 1.0.1 
* 
*/
#include <ccbg/CCChiral.h>
#include <ccbg/CCMol.h>

/** 
* @brief compare two <CCAtom*, int> pairs by ascending 
*        order of the second int
* @remark e.g. pair a and b, true if a.second < b.second
*/
lesspairsecond<CCAtom*, int> CCComparePairSecond;
//static bool CCComparePairSecond( const pair<CCAtom*, int>& a,
//								 const pair<CCAtom*, int>& b )
//{
//	return( a.second < b.second );
//}

/** 
* @brief compare two <CCAtom*, string> pairs by ascending
*        order of the second string
* @remark e.g. pair a and b, true if a.second < b.second
*/
lesspairsecond<CCAtom*, string> CCComparePairSecondStr;
//static bool CCComparePairSecondStr( const pair<CCAtom*, string>& a,
//								    const pair<CCAtom*, string>& b )
//{
//	return( a.second < b.second );
//}

/** 
* @brief compare an <CCAtom*, int> pair by the atom index.
* @param[in] a the first pair of <CCAtom*, int>
* @param[in] a the second pair of <CCAtom*, int>
* @return whether the index of CCAtom in a is less than 
* the index of CCAtom in b.
*/
static bool CCComparePairFirst( const pair<CCAtom*, int>& a,
							    const pair<CCAtom*, int>& b )
{
	return( a.first->GetIdx() < b.first->GetIdx() );
}

/** 
* @brief perceive each chiral atom's R/S chirality.
*
* First check whether current molecule has the flag of
* MOL_CHIRALITY_PERCEIVED. If not, set MOL_CHIRALITY_PERCEIVED
* flag to current molecule and perceive the chirality of each
* atom in current molecule. It only check the chirality of 4 
* connected(implicit valence = 4) atoms. The vector of graph
* invariants( gid vector ) are calculated, so each atom will
* give a weight, which will be used to judge the size of 
* neighboring groups. Then it calculate the signed volume to
* determine the chirality of chiral atoms.
*
* ref: P. Labute, An Efficient Algorithm For The Determination
* Of Topological RS Chirality. 1996
*
* @see GetGIDVector()
* @see CalcSignedVolume()
*/
void CCMol::PerceiveChirality()
{
    if ( HasFlag( MOL_CHIRALITY_PERCEIVED ) ) {
        return;
    }
    SetFlag( MOL_CHIRALITY_PERCEIVED );

    //do quick test to see if there are any possible chiral centers
    bool mayHaveChiralCenter = false;
    CCAtom* atom, * nbr;
    CCAtomIter ai;
    for ( atom = BeginAtom( ai ); atom; atom = NextAtom( ai ) ) {
		// only consider the chirality of 4 connected atom
        if ( atom->GetHvyVal() >= 3 &&
			 atom->GetImpVal() == 4 ) {
            mayHaveChiralCenter = true;
            break;
        }
    }

    if ( !mayHaveChiralCenter ) {
        return;
    }

    vector<int> gidv;
    GetGIDVector( gidv );

	// print gidv
	//FOR_ATOMS_OF_MOL( a, this ) {
	//	cout << a->GetIdx() << " -- " << vgid[a->GetIdx()] << endl;
	//}
	//cout << endl;

    vector<int> tlist;
    vector<int>::iterator k;
	CCBondIter bj;

    bool ischiral;
    for ( atom = BeginAtom( ai ); atom; atom = NextAtom( ai ) ) {
        if ( atom->GetHvyVal() >= 3 &&
			 atom->GetImpVal() == 4 ) {
            tlist.clear();
            ischiral = true;

			// chech whether it is chiral atom
            for ( nbr = atom->BeginNbrAtom( bj );
                  nbr;
                  nbr = atom->NextNbrAtom( bj ) ) {
                for ( k = tlist.begin(); k != tlist.end(); ++k ) {
                    if ( gidv[nbr->GetIdx()] == *k ) {
                        ischiral = false;
                    }
                }

                if ( ischiral ) {
                    tlist.push_back( gidv[nbr->GetIdx()] );
                }
                else {
                    break;
                }
            }

            if ( ischiral ) {
                atom->SetChiral();
            }
			else {
				atom->UnsetChiral();
			}
        }
    }

	// perceive chirality of each chiral atom
    for ( atom = BeginAtom( ai ); atom; atom = NextAtom( ai ) ) {
		if ( !atom->IsChiral() ) {
			continue;
		}

        double sv = CalcSignedVolume( atom, gidv );
        if ( sv < 0.0 ) {
			// S-chirality
			atom->UnsetChiral();
            atom->SetFlag( ATOM_IS_SCHIRAL | ATOM_IS_CHIRAL );
        }
        else if ( sv > 0.0 ) {
			// R-chirality
			atom->UnsetChiral();
            atom->SetFlag( ATOM_IS_RCHIRAL | ATOM_IS_CHIRAL );
        }
		else {
			atom->UnsetChiral();
		}
    }
}

/** 
* @brief perceive each stereo atom's E/Z stereo.
*
* First check whether current molecule has the flag of
* MOL_STEREO_PERCEIVED. If not, set MOL_STEREO_PERCEIVED
* flag to current molecule and perceive the stereo of each
* atom in current molecule. It only check the atoms in the 
* double bond which is not in aromatic ring. The vector of 
* graph invariants( gid vector ) are calculated, so each 
* atom will give a weight, which will be used to judge the
* size of neighboring groups. Then it calculate the torsion
* of whose center is the double bond to determine the stereo
* of the atoms.
*
* ref: P. Labute, An Efficient Algorithm For The Determination
* Of Topological RS Chirality. 1996
*
* @see GetGIDVector()
*/
void CCMol::PerceiveStereo()
{
	if ( HasFlag( MOL_STEREO_PERCEIVED ) ) {
		return;
	}
	SetFlag( MOL_STEREO_PERCEIVED );

	//do quick test to see if there are any possible chiral centers
	bool mayHaveStereo = false;
	CCAtom *bgn, *end, *nbrbgn, *nbrend, *nbr;
	CCBond *bond;
	CCAtomIter ai;
	CCBondIter bi;
	for ( bond = BeginBond( bi ); bond; bond = NextBond( bi ) ) {
		// only consider double bond
		if ( bond->GetOrder() == 2 && !bond->IsAromatic() ) {
			bgn = bond->GetBeginAtom();
			end = bond->GetEndAtom();

			if ( bgn->GetHvyVal() > 1 && bgn->GetHyb() == 2 &&
				 end->GetHvyVal() > 1 && end->GetHyb() == 2 ) {
				mayHaveStereo = true;
				break;
			}
		}
	}

	if ( !mayHaveStereo ) {
		return;
	}

	vector<int> gidv;
	GetGIDVector( gidv );
	CCBondIter bj;
	int weight(0);

	for ( bond = BeginBond( bi ); bond; bond = NextBond( bi ) ) {
		if ( bond->GetOrder() == 2 && !bond->IsAromatic() ) {
			bgn = bond->GetBeginAtom();
			end = bond->GetEndAtom();

			if ( bgn->GetHvyVal() > 1 && bgn->GetHyb() == 2 &&
				 end->GetHvyVal() > 1 && end->GetHyb() == 2 ) {
				weight = -1;
				for ( nbr = bgn->BeginNbrAtom( bj );
					  nbr;
					  nbr = bgn->NextNbrAtom( bj ) ) {
					if ( nbr == end || nbr->IsHydrogen() ) {
					  continue;
					}
					if ( gidv[nbr->GetIdx()] > weight ) {
						weight = gidv[nbr->GetIdx()];
						nbrbgn = nbr;
					}
					else if ( gidv[nbr->GetIdx()] == weight ) {
						nbrbgn = NULL;
						break;
					}
				}
				if ( !nbrbgn ) {
					bgn->UnsetStereo();
					end->UnsetStereo();
					continue;
				}

				weight = -1;
				for ( nbr = end->BeginNbrAtom( bj );
					  nbr;
					  nbr = end->NextNbrAtom( bj ) ) {
					if ( nbr == bgn || nbr->IsHydrogen() ) {
						continue;
					}
					if ( gidv[nbr->GetIdx()] > weight ) {
						weight = gidv[nbr->GetIdx()];
						nbrend = nbr;
					}
					else if ( gidv[nbr->GetIdx()] == weight ) {
						nbrend = NULL;
						break;
					}
				}
				if ( !nbrend ) {
					bgn->UnsetStereo();
					end->UnsetStereo();
					continue;
				}

				// perceive stereo
				double torsion = Torsion( nbrbgn->GetXYZ(), bgn->GetXYZ(),
					                      end->GetXYZ(), nbrend->GetXYZ() );
				torsion = fabs( torsion );
				if ( torsion < 90.0 ) {
					// Z-Stereo
					bgn->UnsetStereo();
					end->UnsetStereo();
					bgn->SetFlag( ATOM_IS_ZSTEREO | ATOM_IS_STEREO );
					end->SetFlag( ATOM_IS_ZSTEREO | ATOM_IS_STEREO );
				}
				else if ( torsion > 90.0 ) {
					// E-Stereo
					bgn->UnsetStereo();
					end->UnsetStereo();
					bgn->SetFlag( ATOM_IS_ESTEREO | ATOM_IS_STEREO );
					end->SetFlag( ATOM_IS_ESTEREO | ATOM_IS_STEREO );
				}
				else {
					bgn->UnsetStereo();
					end->UnsetStereo();
				}
			}
		}
	}
}

/** 
* @brief gets Atom4Refs used to define the chirality.
*
* Used in smi CorrectChirality.
*
* @see CorrectChirality()
*/
int GetParity4Ref( vector<int> pref )
{
    if ( pref.size() != 4 ) {
        return( -1 );
    } // should be given a vector of size 4.
    int parity = 0;
    for ( int i = 0; i < 3; ++i ) // do the bubble sort this many times
    {
        for ( int j = 0;
              j < 3;
              ++j ) // iterate across the array 4th element has no
        {
            // right hand neighbour so no need to sort
            if ( pref[j + 1] < pref[j] ) // compare the two neighbors
            {
                int tmp = pref[j];        // swap a[j] and a[j+1]  
                pref[j] = pref[j + 1];
                pref[j + 1] = tmp;
                parity++; // parity odd will invert stereochem
            }
        }
    } // End Bubble Sort
    return( parity % 2 );
}

/** 
* @brief correct chirality of patm.
*
* Used in smi format.
*/
bool CorrectChirality( CCAtom* patm, atomreftype i, atomreftype o )
{
	// if no chirality defined can't do any more for 0D
    if ( !patm->HasChiralitySpecified() )
    {
        return false;
    }

    int parityI = 0, parityO = 0;
    CCChiralData* cd = ( CCChiralData* )
                       patm->GetData( CCGenericDataType::ChiralData );
    if ( ( cd->GetAtom4Refs( input ) ).size() != 4 ) {
        return false;
    } // must have 4 refs

	// Gets Atom4Refs used to define the chirality
    parityI = GetParity4Ref( cd->GetAtom4Refs( i ) );
	//GetsOutput parity.
    parityO = GetParity4Ref( cd->GetAtom4Refs( o ) );
    /* switch (CHTYPE)
    {
    case SMILES: // SMILES always uses 1234 atom refs
    parityO=0; // if parityO==parityI then clockwise in input = clockwise in output
    break;
    case MOLV3000: // MOLV3000 uses 1234 unless an H then 123H
    if (atm->GetHvyVal()==3)
    {
    CCAtom *nbr;
    int Hid=1000;// max Atom ID +1 should be used here
    vector<int> nbr_atms;
    vector<CCBond*>::iterator i;
    for (nbr = atm->BeginNbrAtom(i);nbr;nbr = atm->NextNbrAtom(i))
    {
    if (nbr->IsHydrogen()){Hid=nbr->GetIdx();continue;}
    nbr_atms.push_back(nbr->GetIdx());
    }
    sort(nbr_atms.begin(),nbr_atms.end());
    nbr_atms.push_back(Hid);
    int tmp[4];
    for(int i=0;i<4;++i){tmp[i]=nbr_atms[i];}
    parityO=GetParity4Ref(tmp);    
    } 
    else if (atm->GetHvyVal()==4)
    parityO=0;   
    break;
    default:
    parityO=0;                               
    }*/
    if ( parityO == parityI ) {
        //cout << "Parity is the same"<<endl;
        return true;
    }
	// Need to invert the Chirality which has been set
    else if ( parityO != parityI )
    {
        //cout << "Parity is Opposite"<<endl;       
        if ( patm->IsRChiral() ) {
            patm->UnsetChiral();
            patm->SetFlag( ATOM_IS_SCHIRAL | ATOM_IS_CHIRAL );
        }
        else if ( patm->IsSChiral() ) {
            patm->UnsetChiral();
            patm->SetFlag( ATOM_IS_RCHIRAL | ATOM_IS_CHIRAL );
        }
        else {
            return false;
        }
        return true;
    }
    return false;
}

/** 
* @brief calculate the signed volume for an atom.
* @param[in] patm root atom for calculate signed volume
* @param[in] gidv vector of graph invariants
* @param[in] ReZeroZ only for bonds in 2D mol whose label is
*            WEDGE or HASH<br>
*            true(default) -- don't change z coords in 2D mol<br>
*            false -- make pseudo z coords and leave them in 2D mol
* @return the signed volume of the atom.
*
* Calculate the signed volume for an atom. If the atom has a valence
* of 3, the coordinates of an attached hydrogen are calculated.
* It first sorts the neighbor atoms of patm from smaller gidv value
* to larger gidv value. The larger gidv value indicate higher priority.
* Then it calculate the signed volume. The positive signed volume
* indicates a R-chirality, and the negative signed volume indicates a
* S-chirality.
*
* @Note the chirality result is independent of atom index.
* such as:     (1)B--A--C(2)   and    (2)B--A--C(1)
*                    |                      |
*                    D(3)                   D(3)
* will give the same result.
*
* @see signed_volume()
*/
double CalcSignedVolume( CCAtom* patm, vector<int> &gidv, bool ReZeroZ )
{
	CCMol *pmol = patm->GetMol();
	if ( pmol->GetNumAtom() + 1 != gidv.size() ) {
		gErrorLog.Throw( __FUNCTION__, 
			             "gvid does not match with mol." );
		return( 0.0 );
	}

	CCVector tmp_crd;
	vector<CCVector> nbr_crds;
	bool use_central_atom = false, is2D = false;
	vector<pair<CCAtom*, int> > vp;

	if ( !ReZeroZ || !pmol->Has3D() ) //give pseudo Z coords if mol is 2D
	{
		CCVector v, vz( 0.0, 0.0, 1.0 );
		is2D = true;
		CCAtom* nbr;
		CCBond* bond;
		vector<CCBond*>::iterator i;
		for ( bond = patm->BeginNbrBond( i ); 
			  bond;
			  bond = patm->NextNbrBond( i ) ) {
			nbr = bond->GetNbrAtom( patm );

			v = nbr->GetXYZ();
			if ( bond->IsWedge() ) {
				v += vz;
			}
			else if ( bond->IsHash() ) {
				v -= vz;
			}

			nbr->SetXYZ( v );
		}
	}

	if ( patm->GetHvyVal() < 3 ) {
		stringstream errorMsg;
		errorMsg
			<< "Molecular name: " << patm->GetMol()->GetName() << endl
			<< "Cannot calculate a signed volume for "
			<< patm->GetType() << endl
			<< "with a heavy atom valence of "
			<< patm->GetHvyVal()
			<< endl;
		gErrorLog.Throw( __FUNCTION__, errorMsg.str(), ccWarning );
		return( 0.0 );
	}

	// Create a vector with the coordinates of the neighbor atoms
	// Also make a vector with Atom IDs
	CCAtom* nbr;
	vector<CCBond*>::iterator bint;
	for ( nbr = patm->BeginNbrAtom( bint );
		  nbr;
		  nbr = patm->NextNbrAtom( bint ) ) {
		vp.push_back( make_pair( nbr, gidv[nbr->GetIdx()] ) );
	}

	// sort the neighbor atoms by gvid to insure a consistent ordering
	// note that larger gvid value indicate higher priority 
	// now is from largest fragment to small fragment
	sort( vp.rbegin(), vp.rend(), CCComparePairSecond );
	for ( int i = 0; i < vp.size(); ++i ) {
		CCAtom* tmp_atm = vp[i].first;
		nbr_crds.push_back( tmp_atm->GetXYZ() );
	}
	
	// Checks for a neighbour having 0 co-ords (added hydrogen etc)
	for ( int j = 0; j < nbr_crds.size(); ++j ) {
		// are the coordinates zero to 6 or more significant figures
		if ( nbr_crds[j].IsZeroVec() && use_central_atom == false ) {
			use_central_atom = true;
		}
		else if ( nbr_crds[j].IsZeroVec() ) {
			gErrorLog.Throw( __FUNCTION__,
				"More than 2 neighbours have 0 co-ords when attempting 3D chiral calculation",
				ccWarning );
		}
	}

	// If we have three heavy atoms we can use the chiral
	// center atom itself for the fourth will always give
	// same sign (for tetrahedron), magnitude will be smaller.
	if ( nbr_crds.size() == 3 || use_central_atom == true ) {
		nbr_crds.push_back( patm->GetXYZ() );
	}

	//re-zero psuedo-coords
	if ( is2D && ReZeroZ ) {
		CCVector v;
		CCAtom* atom;
		CCAtomIter k;
		for ( atom = pmol->BeginAtom( k ); 
			  atom;
			  atom = pmol->NextAtom( k ) ) {
			v = atom->GetXYZ();
			v.SetZ( 0.0 );
			atom->SetXYZ( v );
		}
	}

	return( signed_volume( nbr_crds[0], nbr_crds[1],
		                   nbr_crds[2], nbr_crds[3] ) );
}


/** 
* @brief calculate a signed volume given a set of four coordinates.
* @param[in] a the first coordinates
* @param[in] a the second coordinates
* @param[in] a the third coordinates
* @param[in] a the fourth coordinates
* @return signed volume of the four coordinates.
*/
double signed_volume( const CCVector& a, const CCVector& b, 
					  const CCVector& c, const CCVector& d )
{
    CCVector A, B, C;
    A = d - a;
    B = d - b;
    C = d - c;

    CCMatrix m( A, B, C );
    return( m.Determinant() );
}

/** 
* @brief calculate vector of graph invariants.
* @param[out] gidv the vector of graph invariants.
*
* @note The larger gidv value indicate higher priority of an atom.
*/

// #define SHOWGID
void CCMol::GetGIDVector( vector<int>& gidv ) {
	int i(0), j(0), k(0), idx(0);
	CCAtom *atom(NULL), *nbr(NULL);
	CCAtomIter ai;
	CCBondIter bi;
	char buff[BUFF_SIZE] = "";
	
	int L = 5;
	if ( m_natom > 10000 ) {
		L = 10;
	}

	// put a blank pair to make the vpgids's index
	// synchronize with atomic index
	vector<pair<CCAtom*, string> > vpgids;
	vpgids.clear();
	vpgids.push_back( make_pair( ( CCAtom* ) NULL, "") );
	for ( atom = BeginAtom( ai ); atom; atom = NextAtom( ai ) ) {
		if ( atom->IsHydrogen() ) {
			// ignore H
			vpgids.push_back( make_pair( atom, "" ) );
		}
		else {
			sprintf( buff, "%0*d", L, atom->GetAtomicNum() );
			vpgids.push_back( make_pair( atom, buff ) );
		}
	}

	vector<int> nbrcodes;
	vector<int>::iterator vi;
	vector<int> codes( m_natom + 1, 0 );
	vector<pair<CCAtom*, string> >::iterator vpi, vpj;
	// sort form the second element.
	// in linux, if sort from the first element, the first element
	// may be put back of Hydrogen atom and causes 
	// vpj->first->GetIdx() overflow.
	sort( vpgids.begin() + 1, vpgids.end(), CCComparePairSecondStr );

#ifdef SHOWGID
	for ( vpj = vpgids.begin() + 1; vpj != vpgids.end(); ++vpj ) {
		idx = vpj->first->GetIdx();
		cout << idx << " -- " << vpj->second << endl;
	}
#endif

	int size = vpgids.size();
	int numgourps = 0;
	for ( ; ; ) {
#ifdef SHOWGID
		cout << "numgroups = " << numgourps << endl;
#endif
		// Update codes vector and give each partition a new code
		int N = 1;  // vpgids.bengin->second is always empty
		for ( vpi = vpj = vpgids.begin() + 1; vpj != vpgids.end(); ++vpj ) {
			if ( vpj->first->IsHydrogen() ) {
				continue;
			}
			idx = vpj->first->GetIdx();
			if( vpj->second != vpi->second ) {
				++N;
				codes[idx] = N;
				vpi = vpj;
#ifdef SHOWGID
				cout << idx << " -- " << N << endl;
#endif
			}
			else {
				codes[idx] = N;
#ifdef SHOWGID
				cout << idx << " -- " << N << endl;
#endif
			}
		}
		if (  N == numgourps || N == m_natom ) {
			break;
		}
		numgourps = N;		

		// calculate new code for each atom in each partition
		int curcode = 0;
		for ( i = j = 1; j < size; ++j ) {
			atom = vpgids[j].first;
			if ( !atom || codes[atom->GetIdx()] == 0 ) {
				continue;
			}

			idx = atom->GetIdx();
			nbrcodes.clear();
			int maxcode( 0 ); // for alternative bond such as aromatic bond
			bool dbbond( false );
			for ( nbr = atom->BeginNbrAtom( bi );
				  nbr;
				  nbr = atom->NextNbrAtom( bi ) ) {
				if ( nbr->IsHydrogen() ) {
					continue;
				}

				// put the max code of neighboring aromatic
				// atom's code into codes twice
				if ( (*bi)->IsTautomerAR() ) {
					if ( codes[nbr->GetIdx()] > maxcode ) {
						maxcode = codes[nbr->GetIdx()];
					}
					nbrcodes.push_back( codes[nbr->GetIdx()] );
				}
				else {
					for ( k = 0; k < (*bi)->GetOrder(); ++k ) {
						nbrcodes.push_back( codes[nbr->GetIdx()] );
					}
				}
			}
			if ( maxcode != 0 ) {
				nbrcodes.push_back( maxcode );
			}
			// sort from larger to smaller
			if ( nbrcodes.size() > 1 ) {
				sort( nbrcodes.rbegin(), nbrcodes.rend() );
			}

			// put the root atom's code in the head of the new code
			sprintf( buff, "%0*d", L, codes[idx] );
			vpgids[j].second.clear();
			vpgids[j].second += buff;
			// attach the sorted neighboring atoms code
			for ( vi = nbrcodes.begin(); vi != nbrcodes.end(); ++vi ) {
				sprintf( buff, "%0*d", L, (*vi) );
				vpgids[j].second += buff;
			}

			if( codes[idx] != curcode ) {
				if ( !vpgids[j].second.empty() && j - i > 1 ) {
					sort( vpgids.begin() + i, 
						  vpgids.begin() + j,
						  CCComparePairSecondStr );
				}
				i = j;
				curcode = codes[idx];
			}
#ifdef SHOWGID
			cout << idx << " curcode = " << vpgids[j].second << endl;
#endif
		}

		if ( j - i > 1 ) {
			sort( vpgids.begin() + i,
				  vpgids.begin() + j,
				  CCComparePairSecondStr );
		}
	}
	
	// copy codes to gidv and release vpgids.
	gidv.clear();
	gidv.resize( m_natom + 1, 0 );
	for ( vpi = vpgids.begin() + 1; vpi != vpgids.end(); ++vpi ) {
		idx = vpi->first->GetIdx();
		gidv[idx] = codes[idx];
		vpi->first = NULL;
		vpi->second = "";
	}
}

