/** 
* @file 
* @brief interface for handling file under WIN32 and UNIX/Linux
* @author Yuan ZHAO, Tiejun CHENG
* @date 2008/09/09
* @version 1.0.1 
*
*/
#include <ccbg/CCFileHandler.h>

// Windows----------------------------------------------------------------------------------------
//typedef struct _WIN32_FIND_DATA {
//DWORD dwFileAttributes; //ļ
//FILETIME ftCreationTime; // ļʱ
//FILETIME ftLastAccessTime; // ļһηʱ
//FILETIME ftLastWriteTime; // ļһ޸ʱ
//DWORD nFileSizeHigh; // ļȸ32λ
//DWORD nFileSizeLow; // ļȵ32λ
//DWORD dwReserved0; // ϵͳ
//DWORD dwReserved1; // ϵͳ
//TCHAR cFileName[ MAX_PATH ]; // ļ
//TCHAR cAlternateFileName[ 14 ]; // 8.3ʽļ
//} WIN32_FIND_DATA, *PWIN32_FIND_DATA;

#ifdef WIN32 // for Windows
#include <Windows.h>
/**
* @brief find files with given name
* @param[out] file_list resulting files
* @param[in] filename to be searched
* @return number of files found
* @see FileFilesInDir()
*
* @remark for WIN32, perfect for finding files with certain pattern,
*         such as 'C:/folder/*.sdf' or more complex one 'C:/*a/*.sdf'
*/
int FindFiles( vector<string>& file_list, const string& filename )
{
    string::size_type pos = filename.find_first_of( "*?" );
    if ( pos == string::npos ) {
        //no wildcard in filename
        file_list.push_back( filename );
        return 1;
    }

    pos = filename.rfind( '\\', pos );
    if ( pos != string::npos ) {
        return FindFilesInDir( file_list, filename.substr( 0, pos + 1 ),
                               filename.substr( pos + 1 ) );
    }
    else {
        // find in current directory
        return FindFilesInDir( file_list, ".\\", filename );
    }
}

/**
* @brief find files with given pattern in directory
* @param[out] file_list resulting files
* @param[in] path path to be searched
* @param[in] pattern file pattern contain wildcard
* @return number of files found
* @note path must end with '\\'
* @see FileFiles()
*
* @remark for WIN32, typically used internally by FindFiles, but also 
*         can be used independently
*/
int FindFilesInDir( vector<string>& file_list, const string& path,
                    const string& pattern )
{
    string dirPattn, curPattn, nxtPattn;
    string::size_type p = pattern.find_first_of( '\\' );
    if ( p == string::npos ) {
        curPattn = path + pattern;
    }
    else {
        dirPattn = pattern.substr( 0, p );
        curPattn = path + dirPattn;
        nxtPattn = pattern.substr( p + 1 );
    }

    WIN32_FIND_DATA file_data;
    HANDLE handle = FindFirstFile( curPattn.c_str(), &file_data );
    if ( handle == INVALID_HANDLE_VALUE ) {
        return 0;
    }

    ULONG value;
    string s;

    do {
        value = ( file_data.dwFileAttributes ) & FILE_ATTRIBUTE_DIRECTORY;
        s = file_data.cFileName;

        if ( dirPattn.empty() ) {
            // pattern is file
            if ( value != FILE_ATTRIBUTE_DIRECTORY ) {
                file_list.push_back( path + s );
            }
        }
        else {
            // pattern contains directory
            if ( value == FILE_ATTRIBUTE_DIRECTORY && s != "." && s != ".." ) {
                FindFilesInDir( file_list, path + s + '\\', nxtPattn );
            }
        }
    }
    while ( FindNextFile( handle, &file_data ) );

    if ( !FindClose( handle ) ) {
        return 0;
    }// couldn't close search handler

    return file_list.size();
}

// UNIX/Linux-------------------------------------------------------------------------------------
#else
#include <sys/stat.h>
#include <sys/types.h>
#include <dirent.h>
#include <fnmatch.h>
/**
* @brief find files with given name
* @param[out] file_list resulting files
* @param[in] filename to be searched
* @return number of files found
* @see FileFilesInDir()
*
* @remark for UNIX/Linux, perfect for finding files with certain pattern,
*         such as '/folder/*.sdf' or more complex one '/folder/*a/*.sdf'
*/
int FindFiles( vector<string>& file_list, const string& filename )
{
    string::size_type pos = filename.find_first_of( "*?" );
    if ( pos == string::npos ) {
        //no wildcard in filename
        file_list.push_back( filename );
        return 1;
    }

    pos = filename.rfind( '/', pos );
    int depth( 0 );
    string::size_type p = ( pos != string::npos ) ? pos : 0;
    while ( ( p = filename.find_first_of( '/', p + 1 ) ) != string::npos ) {
        depth++;
    }

    if ( pos != string::npos ) {
        return FindFilesInDir( file_list, filename.substr( 0, pos + 1 ),
                               filename.substr( pos + 1 ), depth );
    }
    else {
        // find in current directory
        return FindFilesInDir( file_list, "./", filename, depth );
    }
}

/**
* @brief find files with given pattern in directory
* @param[out] file_list resulting files
* @param[in] path path to be searched
* @param[in] pattern file pattern contain wildcard
* @return number of files found
* @note path must end with '/'
* @see FileFiles()
*
* @remark for UNIX/Linux, typically used internally by FindFiles, 
*         but also can be used independently
*/
int FindFilesInDir( vector<string>& file_list, const string& path,
                    const string& pattern, int MAX_DEPTH )
{
    DIR* dhandle;
    struct dirent* entry;
    struct stat file_stat;

    if ( ( dhandle = opendir( path.c_str() ) ) == NULL ) {
        return 0;
    }

    string full_path, file_name, dirPattn, nxtPattn;

    string::size_type p = pattern.find_first_of( '/' );
    if ( p == string::npos ) {
        nxtPattn = pattern;
    }
    else {
        dirPattn = pattern.substr( 0, p );
        nxtPattn = pattern.substr( p + 1 );
    }

    while ( entry = readdir( dhandle ) ) {
        file_name = entry->d_name;
        if ( file_name == "." || file_name == ".." ) {
            continue;
        }

        stat( ( path + file_name ).c_str(), &file_stat );
        if ( S_ISDIR( file_stat.st_mode ) && \
            ( MAX_DEPTH == -1 || MAX_DEPTH - 1 >= 0 ) ) {
            if ( !dirPattn.empty()
              && fnmatch( dirPattn.c_str(), file_name.c_str(), 
			              FNM_PATHNAME | FNM_PERIOD ) == FNM_NOMATCH ) {
                continue;
            }
            int nxtDepth = ( MAX_DEPTH == -1 ) ? -1 : MAX_DEPTH - 1;
            FindFilesInDir( file_list, path + file_name + '/', nxtPattn,
                            nxtDepth );
        }
        else if ( S_ISREG( file_stat.st_mode ) ) {
            if ( fnmatch( pattern.c_str(), file_name.c_str(),
                          FNM_PATHNAME | FNM_PERIOD ) != FNM_NOMATCH ) {
                file_list.push_back( path + file_name );
            }
        }
    }
    closedir( dhandle );
    return file_list.size();
}

#endif

/** 
* @brief assert whether file opened successfully (C style)
* @param[in] fp file handle to be asserted
* @param[in] filename file name
* @remark exit program if failed
*/
void AssertOpenFile( FILE* fp, const string& filename )
{
    if ( fp == NULL ) {
        clog << "\nOoops: Unable to open the FILE '" << filename << "'\n"
             << "Please make sure it exist.\n";
        exit( EXIT_FAILURE );
    }
}

/** 
* @brief assert whether file opened successfully (C++ style)
* @param[in] fs file handle to be asserted
* @param[in] filename file name
* @remark exit program if failed
*/
void AssertOpenFile( ios& fs, const string& filename )
{
    if ( !fs.good() ) {
        clog << "\nOoops: Unable to open the FILE '" << filename << "'\n"
             << "Please make sure it exist.\n";
        exit( EXIT_FAILURE );
    }
}

/** 
* @brief parse full path into pieces 
* @param[in] fullpath path to be parsed
* @param[out] dfbe pieces are in the following order:<BR>
*         \li d directory with '/' or '\\'
*         \li f full filename with extension
*         \li b base filename without extension
*         \li e file extension without '.'
* @note if fullpath contains no '/' or '\\', directory will
*       be './' or '.\\' (platform-dependent)
*/
void FileParse( const string& fullpath, vector<string>& dfbe )
{
    dfbe.resize( 4 ); dfbe.assign( 4, "" );
    string::size_type p1 = 0, p2 = 0, s_size = fullpath.size();

    p1 = fullpath.find_last_of( "/\\" ); // position of '/' or '\\'
    p2 = fullpath.find_last_of( "." ); // position of '.';

    if ( p1 < s_size ) // has directory
    {
        dfbe[0] = fullpath.substr( 0, p1 + 1 ); // directory
        if ( p1 + 1 < s_size ) {
            dfbe[1] = fullpath.substr( p1 + 1 ); // full filename
            if ( p1 < p2 && p2 < s_size ) {
                dfbe[2] = fullpath.substr( p1 + 1, p2 - ( p1 + 1 ) ); // base filename
                if ( p2 + 1 < s_size ) {
                    dfbe[3] = fullpath.substr( p2 + 1 ); // file extension
                }
            }
        }
    }
    else // no directory
    {
#ifdef WIN32
		dfbe[0] = ".\\"; // directory
#else
		dfbe[0] = "./"; // directory
#endif
        dfbe[1] = fullpath; // full filename
        if ( p2 < s_size ) {
            dfbe[2] = fullpath.substr( 0, p2 ); // base filename
            if ( p2 + 1 < s_size ) {
                dfbe[3] = fullpath.substr( p2 + 1 ); // extension
            }
        }
    }
}

/** 
* @brief return directory name (with '/' or '\\') of a full path
* @see FileParse()
*/
string DirName( const string& fullpath )
{
    vector<string> dfbe;
    FileParse( fullpath, dfbe );
    return dfbe[0];
}

/** 
* @brief return full file name (i.e. with extension) of a full path 
* @see FileParse()
*/
string FileName( const string& fullpath )
{
    vector<string> dfbe;
    FileParse( fullpath, dfbe );
    return dfbe[1];
}

/** 
* @brief return base file name (i.e. without extension) of a full path 
* @see FileParse()
*/
string BaseName( const string& fullpath )
{
    vector<string> dfbe;
    FileParse( fullpath, dfbe );
    return dfbe[2];
}

/** 
* @brief return extension name (without '.') of a full path 
* @see FileParse()
*/
string ExtName( const string& fullpath )
{
    vector<string> dfbe;
    FileParse( fullpath, dfbe );
    return dfbe[3];
}
