/** @file
* @brief 3*3 matrix and its operators.
* @author Yuan ZHAO, Tiejun CHENG
* @date 2008/08/30
* @version 1.0.1
*
*/
#include <ccbg/CCMatrix.h>
#include <ccbg/CCError.h>
#include <ccbg/CCVector.h>
#include <ccbg/CCUtil.h>

/** @brief CCMatrix constructor. */
CCMatrix::CCMatrix()
{
    m_ele[0][0] = 0.0;
    m_ele[0][1] = 0.0;
    m_ele[0][2] = 0.0;
    m_ele[1][0] = 0.0;
    m_ele[1][1] = 0.0;
    m_ele[1][2] = 0.0;
    m_ele[2][0] = 0.0;
    m_ele[2][1] = 0.0;
    m_ele[2][2] = 0.0;
}

/** 
* @brief Constructs a matrix from row vectors.
* @param[in] row1 data of the first row in the matrix
* @param[in] row2 data of the second row in the matrix
* @param[in] row3 data of the third row in the matrix
*
*/
CCMatrix::CCMatrix( const CCVector& row1, const CCVector& row2,
                    const CCVector& row3 )
{
    m_ele[0][0] = row1.x();
    m_ele[0][1] = row1.y();
    m_ele[0][2] = row1.z();
    m_ele[1][0] = row2.x();
    m_ele[1][1] = row2.y();
    m_ele[1][2] = row2.z();
    m_ele[2][0] = row3.x();
    m_ele[2][1] = row3.y();
    m_ele[2][2] = row3.z();
}

/**
* @brief Constructs a matrix from a 3x3-array of doubles.
* @param[in] d source of the 3x3-array
*
* The first index represents the row, the second index the column.
*/
CCMatrix::CCMatrix( const double d[3][3] )
{
    m_ele[0][0] = d[0][0];
    m_ele[0][1] = d[0][1];
    m_ele[0][2] = d[0][2];
    m_ele[1][0] = d[1][0];
    m_ele[1][1] = d[1][1];
    m_ele[1][2] = d[1][2];
    m_ele[2][0] = d[2][0];
    m_ele[2][1] = d[2][1];
    m_ele[2][2] = d[2][2];
}

/**
* @brief CCMatrix copy constructor.
* @param[in] src matrix for copy
*/
CCMatrix::CCMatrix( const CCMatrix& src )
{
	m_ele[0][0] = src.m_ele[0][0];
	m_ele[0][1] = src.m_ele[0][1];
	m_ele[0][2] = src.m_ele[0][2];
	m_ele[1][0] = src.m_ele[1][0];
	m_ele[1][1] = src.m_ele[1][1];
	m_ele[1][2] = src.m_ele[1][2];
	m_ele[2][0] = src.m_ele[2][0];
	m_ele[2][1] = src.m_ele[2][1];
	m_ele[2][2] = src.m_ele[2][2];
}

/**
* @brief CCMatrix copy assignment operator.
* @param[in] src matrix for copy assignment
*/
CCMatrix& CCMatrix::operator=( const CCMatrix& src )
{
	if ( this != &src ) {
		m_ele[0][0] = src.m_ele[0][0];
		m_ele[0][1] = src.m_ele[0][1];
		m_ele[0][2] = src.m_ele[0][2];
		m_ele[1][0] = src.m_ele[1][0];
		m_ele[1][1] = src.m_ele[1][1];
		m_ele[1][2] = src.m_ele[1][2];
		m_ele[2][0] = src.m_ele[2][0];
		m_ele[2][1] = src.m_ele[2][1];
		m_ele[2][2] = src.m_ele[2][2];
	}   
	return *this;
}

/** 
* @brief operator for get the value of certain element.
* @param[in] row the row number of the element who want to get
* @param[in] column the column number of the element who want to get
* @return the value of m_ele[row][column] element.
*
* If row or column is out of range, it throw an exception and return
* m_ele[0][0], Otherwise return the value of m_ele[row][column].
*
* @note This operator can not change the value of certain element
*       in the matrix.
*/
const double& CCMatrix::operator()( int row, int column ) const
{
    if ( row<0 || row>2 || column<0 || column>2 ) {
        gErrorLog.Throw( __FUNCTION__, "pos out of range." );
		return m_ele[0][0];
    }
    return m_ele[row][column];
}

/** 
* @brief operator for get the reference of certain element.
* @param[in] row the row number of the element who want to get
* @param[in] column the column number of the element who want to get
* @return the reference of m_ele[row][column] element.
*
* If row or column is out of range, it throw an exception and return
* m_ele[0][0]. Otherwise return the reference of m_ele[row][column].
*
* @note This operator can change the value of certain element
*       in the matrix.
*/
double& CCMatrix::operator()( int row, int column )
{
    if ( row<0 || row>2 || column<0 || column>2 ) {
        gErrorLog.Throw( __FUNCTION__, "pos out of range." );
		return m_ele[0][0];
    }
    return m_ele[row][column];
}

/** 
* @brief calculate the determinant of current matrix.
* @return the determinant of current matrix.
*
* This function is used by signed_volume. The determinant
* can determine the rotation direction of three vectors.
* Determinant larger than 0 represent clockwise of the vector
* of m_ele[0], m_ele[1] and m_ele[2].
*
* @see signed_volume()
*/
double CCMatrix::Determinant() const
{
    return( m_ele[0][0] * ( m_ele[1][1] * m_ele[2][2]
                          - m_ele[1][2] * m_ele[2][1] )
          + m_ele[0][1] * ( m_ele[1][2] * m_ele[2][0]
                          - m_ele[1][0] * m_ele[2][2] )
          + m_ele[0][2] * ( m_ele[1][0] * m_ele[2][1]
                          - m_ele[1][1] * m_ele[2][0] ) );
}

/** 
* @brief calculate the inverse of current matrix.
* @return the inverse matrix.
*
* This function first checks if the absolute value of the determinant 
* is smaller than 1e-6. If so, return current matrix and throw an 
* exception. Otherwise, the inverse matrix is calculated and returned.
* this is not changed.<br>
* warning If the determinant is close to zero, but not == 0.0,
* this method may behave in unexpected ways and return almost
* random results; details may depend on your particular floating
* point implementation. The use of this method is therefore highly
* discouraged, unless you are certain that the determinant is in a
* reasonable range, away from 0.0 (Stefan Kebekus).
*/
CCMatrix CCMatrix::Inverse() const
{
	double det = Determinant();

	if ( fabs( det ) <= 1e-6 ) {
		gErrorLog.Throw( __FUNCTION__,
			"can not be called on matrix with |determinant| <= 1e-6.\n",
			ccWarning );
		return *this;
	}

	CCMatrix inverse;
	inverse.m_ele[0][0] = m_ele[1][1] * m_ele[2][2]
	- m_ele[1][2] * m_ele[2][1];
	inverse.m_ele[1][0] = m_ele[1][2] * m_ele[2][0]
	- m_ele[1][0] * m_ele[2][2];
	inverse.m_ele[2][0] = m_ele[1][0] * m_ele[2][1]
	- m_ele[1][1] * m_ele[2][0];
	inverse.m_ele[0][1] = m_ele[2][1] * m_ele[0][2]
	- m_ele[2][2] * m_ele[0][1];
	inverse.m_ele[1][1] = m_ele[2][2] * m_ele[0][0]
	- m_ele[2][0] * m_ele[0][2];
	inverse.m_ele[2][1] = m_ele[2][0] * m_ele[0][1]
	- m_ele[2][1] * m_ele[0][0];
	inverse.m_ele[0][2] = m_ele[0][1] * m_ele[1][2]
	- m_ele[0][2] * m_ele[1][1];
	inverse.m_ele[1][2] = m_ele[0][2] * m_ele[1][0]
	- m_ele[0][0] * m_ele[1][2];
	inverse.m_ele[2][2] = m_ele[0][0] * m_ele[1][1]
	- m_ele[0][1] * m_ele[1][0];

	inverse /= det;

	return inverse;
}

/** 
* @brief calculate the transpose of current matrix.
* @return the transpose of current matrix.
*
* This method returns the transpose of a matrix. The original
* matrix remains unchanged.
*/
CCMatrix CCMatrix::Transpose() const
{
	CCMatrix transpose;

	for ( int i = 0; i < 3; i++ ) {
		for ( int j = 0; j < 3; j++ ) {
			transpose.m_ele[i][j] = m_ele[j][i];
		}
	}

	return transpose;
}

/** 
* @brief calculates a rotation matrix.
* @param[in] v the rotation axis
* @param[in] angle rotation angle in degree
*
* Calculates a rotation matrix, rotating current matrix around 
* axis v by angle degrees. The result will used to update current
* matrix.
*/
void CCMatrix::RotAboutAxisByAngle( const CCVector& v, double angle )
{
	double theta = angle* DEG_TO_RAD;
	double s = sin( theta );
	double c = cos( theta );
	double t = 1 - c;

	CCVector vtmp = v;
	vtmp.Normalize();

	double x = vtmp. x(), y = vtmp.y(), z = vtmp.z();

	m_ele[0][0] = t * x * x + c;
	m_ele[0][1] = t * x * y + s * z;
	m_ele[0][2] = t * x * z - s * y;

	m_ele[1][0] = t * x * y - s * z;
	m_ele[1][1] = t * y * y + c;
	m_ele[1][2] = t * y * z + s * x;

	m_ele[2][0] = t * x * z + s * y;
	m_ele[2][1] = t * y * z - s * x;
	m_ele[2][2] = t * z * z + c;
}

/** 
* @brief matrix-vector multiplication m(3*3) * v(3*1).
* @param[in] m the matrix productor
* @param[in] v the vector productor
* @return the product m*v.
*
* Calculates the product m*v of the matrix m and the column
* vector represented by v.
*/
CCVector operator*( const CCMatrix& m, const CCVector& v )
{
    CCVector vv;

    vv.x() = v.x() * m.m_ele[0][0]
           + v.y() * m.m_ele[0][1]
           + v.z() * m.m_ele[0][2];
    vv.y() = v.x() * m.m_ele[1][0]
           + v.y() * m.m_ele[1][1]
           + v.z() * m.m_ele[1][2];
    vv.z() = v.x() * m.m_ele[2][0]
           + v.y() * m.m_ele[2][1]
           + v.z() * m.m_ele[2][2];

    return vv;
}

/** 
* @brief vector-matrix multiplication v(1*3) * m(3*3).
* @param[in] v the vector productor
* @param[in] m the matrix productor
* @return the product v*m.
*
* Calculates the product v*m of the column vector represented
* by v and the matrix m.
*/
CCVector operator*( const CCVector& v, const CCMatrix& m )
{
    CCVector vv;

    vv.x() = v.x() * m.m_ele[0][0]
           + v.y() * m.m_ele[1][0]
           + v.z() * m.m_ele[2][0];
    vv.y() = v.x() * m.m_ele[0][1]
           + v.y() * m.m_ele[1][1]
           + v.z() * m.m_ele[2][1];
    vv.z() = v.x() * m.m_ele[0][2]
           + v.y() * m.m_ele[1][2]
           + v.z() * m.m_ele[2][2];

    return vv;
}

/** 
* @brief matrix-matrix multiplication..
* @param[in] m1 the first matrix productor
* @param[in] m2 the second matrix productor
* @return the product m1*m2.
*
* Calculates the product two matrix m1 and m2.
*/
CCMatrix operator *( const CCMatrix& m1, const CCMatrix& m2 )
{
    CCMatrix result;

    result.m_ele[0][0] = m1.m_ele[0][0] * m2.m_ele[0][0]
                       + m1.m_ele[0][1] * m2.m_ele[1][0]
                       + m1.m_ele[0][2] * m2.m_ele[2][0];
    result.m_ele[0][1] = m1.m_ele[0][0] * m2.m_ele[0][1]
                       + m1.m_ele[0][1] * m2.m_ele[1][1]
                       + m1.m_ele[0][2] * m2.m_ele[2][1];
    result.m_ele[0][2] = m1.m_ele[0][0] * m2.m_ele[0][2]
                       + m1.m_ele[0][1] * m2.m_ele[1][2]
                       + m1.m_ele[0][2] * m2.m_ele[2][2];

    result.m_ele[1][0] = m1.m_ele[1][0] * m2.m_ele[0][0]
                       + m1.m_ele[1][1] * m2.m_ele[1][0]
                       + m1.m_ele[1][2] * m2.m_ele[2][0];
    result.m_ele[1][1] = m1.m_ele[1][0] * m2.m_ele[0][1]
                       + m1.m_ele[1][1] * m2.m_ele[1][1]
                       + m1.m_ele[1][2] * m2.m_ele[2][1];
    result.m_ele[1][2] = m1.m_ele[1][0] * m2.m_ele[0][2]
                       + m1.m_ele[1][1] * m2.m_ele[1][2]
                       + m1.m_ele[1][2] * m2.m_ele[2][2];

    result.m_ele[2][0] = m1.m_ele[2][0] * m2.m_ele[0][0]
                       + m1.m_ele[2][1] * m2.m_ele[1][0]
                       + m1.m_ele[2][2] * m2.m_ele[2][0];
    result.m_ele[2][1] = m1.m_ele[2][0] * m2.m_ele[0][1]
                       + m1.m_ele[2][1] * m2.m_ele[1][1]
                       + m1.m_ele[2][2] * m2.m_ele[2][1];
    result.m_ele[2][2] = m1.m_ele[2][0] * m2.m_ele[0][2]
                       + m1.m_ele[2][1] * m2.m_ele[1][2]
                       + m1.m_ele[2][2] * m2.m_ele[2][2];

    return result;
}

