/** @file 
* @brief CCMol class and relevant functions.
* @author Yuan ZHAO, Tiejun CHENG
* @date 2008/08/23
* @version 1.0.1
*
*/
#include <ccbg/CCMol.h>

/** @brief CCMol constructor */
CCMol::CCMol()
    : m_natom( 0 ), m_nbond( 0 ), m_nsubs( 0 ),
      m_dimension( 0 )
{
    CCBase::CCBase();
}

/** @brief CCMol destructor */
CCMol::~CCMol()
{
    CCBase::DestroyGD();
	m_name = "";
	m_natom = 0;
	m_nbond = 0;
	m_nsubs = 0;
	m_dimension = 0;
	Destroy();
}

/** @brief clear members of CCMol */
void CCMol::Clear()
{
    CCBase::Clear();
    m_name = "";
    m_natom = 0;
    m_nbond = 0;
    m_nsubs = 0;
    m_vatom.clear();
    m_vbond.clear();
    m_vring.clear();
    m_vresidue.clear();
    m_vchain.clear();
    m_dimension = 0;
}

/** @brief destroy members of CCMol */
void CCMol::Destroy() {
	m_name = "";
	m_natom = 0;
	m_nbond = 0;
	m_nsubs = 0;

	CCAtomIter i = m_vatom.begin();
	for ( ; i != m_vatom.end(); ++i ) {
		if ( *i ) {
			delete * i; *i = NULL;
		}
	}
	m_vatom.clear();

	CCBondIter j = m_vbond.begin();
	for ( ; j != m_vbond.end(); ++j ) {
		if ( *j ) {
			delete * j; *j = NULL;
		}
	}
	m_vbond.clear();

	CCRingIter k = m_vring.begin();
	for ( ; k != m_vring.end(); ++k ) {
		if ( *k ) {
			delete * k; *k = NULL;
		}
	}
	m_vring.clear();

	CCChainIter m = m_vchain.begin();
	for ( ; m != m_vchain.end(); ++m ) {
		if ( *m ) {
			delete * m; *m = NULL;
		}
	}
	m_vchain.clear();

	CCResidueIter l = m_vresidue.begin();
	for ( ; l != m_vresidue.end(); ++l ) {
		if ( *l ) {
			delete * l; *l = NULL;
		}
	}
	m_vresidue.clear();
}

/** 
* @brief CCMol copy constructor.
* @param[in] src molecule for copy.
*
* Note that this funtion is different from CCMol::Clone().
* Clone will get a copy of src molecule and the parent of
* each atom and bond was point to the new created molecule.
* However, in CCMol copy constructor, the parent of each
* atom and bond was point to the src molecule.
* 
* @see CCMol::Clone()
*/
CCMol::CCMol( const CCMol& src )
    : m_name( src.m_name ), m_natom( src.m_natom ), m_nbond( src.m_nbond ),
      m_nsubs( src.m_nsubs ), m_vatom( src.m_vatom ), m_vbond( src.m_vbond ),
      m_vring( src.m_vring ), m_vresidue( src.m_vresidue ),
      m_vchain( src.m_vchain ), m_dimension( src.m_dimension )
{
    CCBase::Clone( src );
}

/** 
* @brief CCMol copy assignment operator.
* @param[in] src molecule for copy assignment.
* 
* Note that this funtion is different from CCMol::Clone().
* Clone will get a copy of src molecule and the parent of
* each atom and bond was point to the new created molecule.
* However, in CCMol assignment operator, the parent of each
* atom and bond was point to the src molecule.
* 
* @see CCMol::Clone()
*/
CCMol& CCMol::operator=( const CCMol& src )
{
    if ( this != &src ) {
        CCBase::Clone( src );
        m_name = src.m_name;
        m_natom = src.m_natom;
        m_nbond = src.m_nbond;
        m_nsubs = src.m_nsubs;
        m_vatom = src.m_vatom;
        m_vbond = src.m_vbond;
        m_vring = src.m_vring;
        m_vresidue = src.m_vresidue;
        m_vchain = src.m_vchain;
        m_dimension = src.m_dimension;
    }

    return *this;
}

/** 
* @brief Clone an instance of src.
* @param[in] src molecule for copy.
*
* Note that this funtion is different from 
* CCMol::CCMol( const CCMol& src ).
* Clone will get a copy of src molecule and the parent of
* each atom and bond was point to the new created molecule.
* However, in CCMol copy constructor, the parent of each
* atom and bond was point to the src molecule.
* 
* @see CCMol::CCMol( const CCMol& src )
*/
void CCMol::Clone( const CCMol& src )
{
    m_name = src.m_name;
    m_natom = src.m_natom;
    m_nbond = src.m_nbond;
    m_nsubs = src.m_nsubs;
    m_dimension = src.m_dimension;

    //****** clone generic data ******//
    CCBase::Clone( src );

    //****** clone atm ******//
    map<CCAtom*, CCAtom*> aa;// record [srcatom, atm] pair
    FOR_ATOMS_OF_MOL( srcatom, ( CCMol * ) &src ) {
        CCAtom* atm = new CCAtom( *srcatom ); AssertMemAlloc( atm );
        atm->SetMol( this );
        m_vatom.push_back( atm );
        aa.insert( make_pair( &*srcatom, atm ) );
    }
    //****** clone bnd ******//
    map<CCBond*, CCBond*> bb;// record [srcbond, bnd] pair
    FOR_BONDS_OF_MOL( srcbond, ( CCMol * ) &src ) {
        CCBond* bnd = new CCBond( *srcbond ); AssertMemAlloc( bnd );
        CCAtom* srcbgn = srcbond->GetBeginAtom();
        CCAtom* srcend = srcbond->GetEndAtom();
        bnd->SetMol( this );
        bnd->SetBegin( aa[srcbgn] );
        bnd->SetEnd( aa[srcend] );
        m_vbond.push_back( bnd );
        bb.insert( make_pair( &*srcbond, bnd ) );
    }

    //****** update connection table,                 ******//
	//****** note atom already has bond nbrs from src ******//
    CCBond* nbr; CCBondIter nbrIter;
    FOR_ATOMS_OF_MOL( atom, ( CCMol * )this ) {
        for ( nbr = atom->BeginNbrBond( nbrIter );
              nbr;
              nbr = atom->NextNbrBond( nbrIter ) ) {
            *nbrIter = bb[nbr];
        }
    }

    //****** clone ring data ******//
    // will not trigger ring detection
    for ( vector<CCRing*>::const_iterator ringIter = src.m_vring.begin();
          ringIter != src.m_vring.end() && *ringIter;
          ++ringIter ) {
        CCRing* ring = new CCRing( **ringIter );
        AssertMemAlloc( ring );
        ring->SetMol( this );
        vector<CCAtom*>& vatom = ring->GetAtoms();
        for ( int m = 0; m < vatom.size(); ++m ) {
            vatom[m] = aa[vatom[m]];
        }
        m_vring.push_back( ring );
    }

    //****** clone residue data ******//
    map<CCResidue*, CCResidue*> rr;
    FOR_RESIDUES_OF_MOL( srcres, ( CCMol * ) &src ) {
        CCResidue* residue = new CCResidue( *srcres );
        AssertMemAlloc( residue );
        residue->SetMol( this );
        vector<CCAtom*>& vatom = residue->GetAtoms();
        for ( int m = 0; m < vatom.size(); ++m ) {
            vatom[m] = aa[vatom[m]];
            vatom[m]->SetResidue( residue );
        }
        m_vresidue.push_back( residue );
        rr.insert( make_pair( &*srcres, residue ) );
    }

    //****** clone chain data ******//
    FOR_CHAINS_OF_MOL( srcchain, ( CCMol * ) &src ) {
        CCChain* chn = new CCChain( *srcchain );
        AssertMemAlloc( chn );
        chn->SetMol( this );
        vector<CCResidue*>& vres = chn->GetResidues();
        for ( int m = 0; m < vres.size(); ++ m ) {
            vres[m] = rr[vres[m]];
            vres[m]->SetChain( chn );
        }
    }
}

/** 
* @brief Tailor from the bond tbond to get two fragments.
* @param[in] tbond tailoring bond.
* @param[out] bgnpart the fragment contain beginning atom of tbond.
* @param[out] endpart the fragment contain end atom of tbond.
* @param[out] bgnpoint atom in bgnpart which corresponds to 
*             beginning atom of tbond.
* @param[out] endpoint atom in endpart which corresponds to
*             end atom of tbond.
*
* Tailor molecule mol from the bond tbond, the result is two fragments
* ( bgnpart and endpart ) and their corresponding junction atom.
*/
void CCMol::Tailoring( CCBond *tbond, CCMol &bgnpart, CCMol &endpart,
                       CCAtom **bgnpoint, CCAtom **endpoint ) {
	CCMol *mol = tbond->GetMol();
	if ( this != mol ) {
		gErrorLog.Throw( __FUNCTION__, 
			             "Tailoring bond was not in the molecule.",
						 ccWarning );
		return;
	}
    CCAtom *bgn = tbond->GetBeginAtom();
    CCAtom *end = tbond->GetEndAtom();

	if ( tbond->IsInRing() ) {
		gErrorLog.Throw( __FUNCTION__, 
			             "Can't tailoring a ring bond. Just break it.",
						 ccWarning );
		mol->DeleteBond( tbond );
		mol->ClearFlag();
		mol->SetFlag( MOL_CONNECTIVITY_PERCEIVED );

		// copy the molecule to bgnpart and endpart molecules,
		// then modify their name as mname_bgn and mname_end
		bgnpart.Clone( *mol );
		bgnpart.SetName( bgnpart.GetName() + "_bgn" );
		endpart.Clone( *mol );
		endpart.SetName( endpart.GetName() + "_end" );
		(*bgnpoint) = bgnpart.GetAtom( bgn->GetIdx() );
		(*endpoint) = endpart.GetAtom( end->GetIdx() );

		return;
	}

    // gether the atoms that link to different parts
    // used[i] = 1 are the bgn part
    // used[i] = 0 are the end part
    int numAtoms    = mol->GetNumAtom();
    CCBitVec used   ( numAtoms + 1 );

    queue<CCAtom*> q;
    CCAtom *atom;

    q.push( bgn );
    used.Set( bgn->GetIdx() );

    // set the neighbors of begin atom ( exclude end atom )
    // to be used.
    FOR_NBORS_OF_ATOM( nbr, bgn ) {
        if ( &*nbr != end ) {
            q.push( &*nbr );
            used.Set( nbr->GetIdx() );
        }
    }
    q.pop();

    // set all the atoms in bgn part to be used
    while ( !q.empty() ) {
        atom = q.front();
        q.pop();

        FOR_NBORS_OF_ATOM( nbr, atom ) {
            if ( !used[nbr->GetIdx()] ) {
                q.push( &*nbr );
                used.Set( nbr->GetIdx() );
            }
        }
    }

    // copy the molecule to bgnpart and endpart molecules,
    // then modify their name as mname_bgn and mname_end
    bgnpart.Clone( *mol );
    bgnpart.SetName( bgnpart.GetName() + "_bgn" );
    endpart.Clone( *mol );
    endpart.SetName( endpart.GetName() + "_end" );

    (*bgnpoint) = bgnpart.GetAtom( bgn->GetIdx() );
    (*endpoint) = endpart.GetAtom( end->GetIdx() );
	(*bgnpoint)->UnsetChiral();
	(*endpoint)->UnsetChiral();

    // gather the atoms that link to bgn atom in endpart molecule
    vector<CCAtom*> bgnpartdelatoms;
    vector<CCAtom*> endpartdelatoms;
    for ( int i = 1; i <= numAtoms; ++i ) {
        if ( used[i] ) {
            endpartdelatoms.push_back( endpart.GetAtom( i ) );
        }
        else {
            bgnpartdelatoms.push_back( bgnpart.GetAtom( i ) );
        }
    }

    // delete the atoms that link to end atom to get bgnpart molecule
    bgnpart.DeleteAtoms( bgnpartdelatoms );
    // delete the atoms that link to bgn atom to get endpart molecule
    endpart.DeleteAtoms( endpartdelatoms );
}

void CCMol::AddMol( CCMol& mol )
{
    int curnatom = m_natom;  // backup origin number of atoms
    int curnbond = m_nbond;  // backup origin number of bonds

    m_name = m_name + "||" + mol.m_name;
    m_natom += mol.m_natom;
    m_nbond += mol.m_nbond;
    m_nsubs += mol.m_nsubs;
    m_dimension = ( m_dimension < mol.m_dimension )
                ? mol.m_dimension
                : m_dimension;

    // Destroy generic data and set the flag
    // as the union set of the two molecules.
    // CCBase::Clone( mol );
	CCBase::DestroyGD();
	int flag = GetFlag() & mol.GetFlag();
    SetFlag( flag );

    // delete NULL atom point in m_vatom
	if ( m_vatom.size() > curnatom ) {
		m_vatom.erase( m_vatom.begin() + curnatom, m_vatom.end() );
	}
	// clone atom
    map<CCAtom*, CCAtom*> aa;// record [srcatom, atm] pair
    FOR_ATOMS_OF_MOL( srcatom, ( CCMol * ) &mol ) {
        CCAtom* atm = new CCAtom( *srcatom );
        AssertMemAlloc( atm );
        atm->SetMol( this );
        m_vatom.push_back( atm );
        aa.insert( make_pair( &*srcatom, atm ) );
    }
    // reset new added atom id and update connection table
    for ( int idx = curnatom; idx < m_natom; ++idx ) {
        m_vatom[idx]->SetIdx( idx + 1 );
    }

    // delete NULL bond point in m_vbond
	if ( m_vbond.size() > curnbond ) {
		m_vbond.erase( m_vbond.begin() + curnbond, m_vbond.end() );
	}
	// clone bond
    map<CCBond*, CCBond*> bb;// record [srcbond, bnd] pair
    FOR_BONDS_OF_MOL( srcbond, ( CCMol * ) &mol ) {
        CCBond* bnd = new CCBond( *srcbond );
		AssertMemAlloc( bnd );
        CCAtom* srcbgn = srcbond->GetBeginAtom();
        CCAtom* srcend = srcbond->GetEndAtom();
        bnd->SetMol( this );
        bnd->SetBegin( aa[srcbgn] );
        bnd->SetEnd( aa[srcend] );
        m_vbond.push_back( bnd );
        bb.insert( make_pair( &*srcbond, bnd ) );
    }
    // reset new added bond id and update connection table
    for ( int idx = curnbond; idx < m_nbond; ++idx ) {
        m_vbond[idx]->SetIdx( idx + 1 );
    }

    // update connection table, note atom already has bond nbrs from src
    CCAtom* atom;
    CCBond* nbr;
    CCBondIter nbrIter;
    for ( int idx = curnatom + 1; idx <= m_natom; ++idx ) {
        atom = GetAtom( idx );
        for ( nbr = atom->BeginNbrBond( nbrIter );
              nbr;
              nbr = atom->NextNbrBond( nbrIter ) ) {
            *nbrIter = bb[nbr];
        }
    }

    // clone ring data
    // will not trigger ring detection
    for ( vector<CCRing*>::const_iterator ringIter = mol.m_vring.begin();
          ringIter != mol.m_vring.end() && *ringIter;
          ++ringIter ) {
        CCRing* ring = new CCRing( **ringIter );
        AssertMemAlloc( ring );
        ring->SetMol( this );
        vector<CCAtom*>& vatom = ring->GetAtoms();
        for ( int m = 0; m < vatom.size(); ++m ) {
            vatom[m] = aa[vatom[m]];
        }
        // update rings bvpath
        ring->UpdateBVPath();
        m_vring.push_back( ring );
    }

    // clone residue data
	int resIdx = (int)m_vresidue.size();
    map<CCResidue*, CCResidue*> rr;
    FOR_RESIDUES_OF_MOL( srcres, ( CCMol * ) &mol ) {
        CCResidue* residue = new CCResidue( *srcres );
        AssertMemAlloc( residue );
        residue->SetMol( this );
		residue->SetIdx( ++resIdx );
        vector<CCAtom*>& vatom = residue->GetAtoms();
        for ( int m = 0; m < vatom.size(); ++m ) {
            vatom[m] = aa[vatom[m]];
            vatom[m]->SetResidue( residue );
        }
        m_vresidue.push_back( residue );
        rr.insert( make_pair( &*srcres, residue ) );
    }

    // clone chain data
	int chnIdx = (int)m_vchain.size();
    FOR_CHAINS_OF_MOL( srcchain, ( CCMol * ) &mol ) {
        CCChain* chn = new CCChain( *srcchain );
        AssertMemAlloc( chn );
        chn->SetMol( this );
		chn->SetIdx( ++chnIdx );
        vector<CCResidue*>& vres = chn->GetResidues();
        for ( int m = 0; m < vres.size(); ++ m ) {
            vres[m] = rr[vres[m]];
            vres[m]->SetChain( chn );
        }
		m_vchain.push_back( chn );
    }
}

void CCMol::MergeMol( CCMol& mol )
{
    int curnatom = m_natom;  // backup origin number of atoms
    int curnbond = m_nbond;  // backup origin number of bonds

    m_name = m_name + "||" + mol.m_name;
    m_natom += mol.m_natom;
    m_nbond += mol.m_nbond;
    m_nsubs += mol.m_nsubs;
    m_dimension = ( m_dimension < mol.m_dimension )
                ? mol.m_dimension
                : m_dimension;

    // Destroy generic data and set the flag
    // as the union set of the two molecules.
    // CCBase::Clone( mol );
	CCBase::DestroyGD();
	int flag = GetFlag() & mol.GetFlag();
	SetFlag( flag );

    // delete NULL atom point in m_vatom
	if ( m_vatom.size() > curnatom ) {
		m_vatom.erase( m_vatom.begin() + curnatom, m_vatom.end() );
	}
	// merge atoms
    FOR_ATOMS_OF_MOL( srcatom, ( CCMol * ) &mol ) {
        srcatom->SetMol( this );
        m_vatom.push_back( &*srcatom );
    }
	mol.m_vatom.clear();
    // reset new added atom id and update connection table
    for ( int idx = curnatom; idx < m_natom; ++idx ) {
        m_vatom[idx]->SetIdx( idx + 1 );
    }

    // delete NULL bond point in m_vbond
	if ( m_vbond.size() > curnbond ) {
		m_vbond.erase( m_vbond.begin() + curnbond, m_vbond.end() );
	}
	// merge bonds
    FOR_BONDS_OF_MOL( srcbond, ( CCMol * ) &mol ) {
        srcbond->SetMol( this );
        m_vbond.push_back( &*srcbond );
    }
	mol.m_vbond.clear();
    // reset new added bond id and update connection table
    for ( int idx = curnbond; idx < m_nbond; ++idx ) {
        m_vbond[idx]->SetIdx( idx + 1 );
    }

	// merge ring data
	// will not trigger ring detection
	vector<CCRing*>::const_iterator ringIter;
    for ( ringIter = mol.m_vring.begin();
          ringIter != mol.m_vring.end() && *ringIter;
          ++ringIter ) {
        (*ringIter)->SetMol( this );
        // update rings bvpath
        (*ringIter)->UpdateBVPath();
        m_vring.push_back( *ringIter );
    }
	mol.m_vring.clear();

    // merge residue data
	int resIdx = (int)m_vresidue.size();
    FOR_RESIDUES_OF_MOL( srcres, ( CCMol * ) &mol ) {
        srcres->SetMol( this );
		srcres->SetIdx( ++resIdx );
        m_vresidue.push_back( &*srcres );
    }
	mol.m_vresidue.clear();

    // merge chain data
	int chnIdx = (int)m_vchain.size();
    FOR_CHAINS_OF_MOL( srcchain, ( CCMol * ) &mol ) {
        srcchain->SetMol( this );
		srcchain->SetIdx( ++chnIdx );
		m_vchain.push_back( &*srcchain );
    }
	mol.m_vchain.clear();

	delete &mol;
}

/**
* @brief Integrate parent molecule of point1 and parent
*        molecule of point2.
* @param[in] point1 linking atom of the first molecule
* @param[in] point2 linking atom of the second molecule
* @param[in] order bond order to link point1 and point2 ( default 1 )
* @param[in] flag flag for point1 and point2 bond ( default 0 )
* @param[out] newpoint1 corresponding atom of point1 in integrated molecule
* @param[out] newpoint2 corresponding atom of point2 in integrated molecule
* @return Integrated molecule
*
*/
CCMol& CCMol::Integrate( CCAtom* point1, CCAtom* point2,
						 int order /*= 1*/, int flag /*= 0*/,
						 CCAtom **newpoint1 /*= NULL*/,
						 CCAtom **newpoint2 /*= NULL*/ )
{
	(*newpoint1) = point1;
	(*newpoint2) = point2;
    if ( point1 == NULL || point2 == NULL || point1 == point2 ) {
        gErrorLog.Throw( __FUNCTION__,
                         "Transpoint atom has something wrong!\n",
                         ccWarning );
        return *this;
    }

	CCMol *m1 = point1->GetMol();
	CCMol *m2 = point2->GetMol();
	if ( m1 == this || m2 == this ) {
        gErrorLog.Throw( __FUNCTION__,
                         "Transpoint atom was in resultant molecule!\n",
                         ccWarning );
        return *this;
	}
	if ( m1 == m2 ) {
        gErrorLog.Throw( __FUNCTION__,
                         "Please use AddBond() instead of Integrate()!\n",
                         ccWarning );
        return *this;
	}

	if ( !this->IsEmpty() ) {
		Destroy();
	}

	// junction atom can not take the bond order.
    if ( point1->ImpHydrogenCount() < order
      || point2->ImpHydrogenCount() < order
      || order < 1 ) {
		string msg = "Transpoint atom or bond order was not suitable!\n";
		msg += "M1: " + m1->GetName() +
			   " -- Atom: " + toSTR( point1->GetIdx() ) + "\n";
		msg += "M2: " + m2->GetName() +
			   " -- Atom: " + toSTR( point2->GetIdx() ) + "\n";
        gErrorLog.Throw( __FUNCTION__, msg, ccWarning );
        return *this;
    }

	Clone( *( point1->GetMol() ) );
    AddMol( *( point2->GetMol() ) );
    CCAtom* junction1 = GetAtom( point1->GetIdx() );
    CCAtom* junction2 = GetAtom( point2->GetIdx() +
		                         m1->GetNumAtom() );

	if ( junction1 && junction2 ) {
		// add bond between two points and set bond order and flag
		AddBond( junction1->GetIdx(), junction2->GetIdx(), order, flag );
		// modify the formal charge of junction atoms, which is necessary
		// for adding phosphate and sulphate groups. 
		if ( junction1->GetFormalCharge() < 0 ) {
			junction1->SetFormalCharge( junction1->GetFormalCharge() + 1 );
		}
		if ( junction2->GetFormalCharge() < 0 ) {
			junction2->SetFormalCharge( junction2->GetFormalCharge() + 1 );
		}
		(*newpoint1) = junction1;
		(*newpoint2) = junction2;
		m_nsubs--;

		ClearFlag(); // clear all flags; different from SetFlag( 0 );
		SetFlag( MOL_CONNECTIVITY_PERCEIVED );
	}
	else {
		string msg = "Junction atoms were not exist!\n";
		msg += m1->GetName() + " -- Atom: " +
			   toSTR( point1->GetIdx() ) + "\n" +
			   m2->GetName() + " -- Atom: " +
			   toSTR( point2->GetIdx() );
		gErrorLog.Throw( __FUNCTION__, msg, ccWarning );
	}
    return *this;
}

/**
* @brief Integrate parent molecule of points1 and each parent
*        molecule of atoms in points2.
* @param[in] points1 linking atom of the first molecule
* @param[in] points2 linking atom of the second molecule
* @param[in] orders bond order to link point1 and point2 ( default 1 )
* @param[in] flags flag for point1 and point2 bond ( default 0 )
* @return whether successfully integrated.
*
* @Note all atoms in points1 should be 
*/
bool CCMol::MultiIntegrate( vector<CCAtom*> &points1,
							vector<CCAtom*> &points2,
						    vector<int> &orders,
							vector<int> &flags )
{
	int num = (int)points1.size();
    if ( num <= 0 ||
		 points2.size() != num || 
		 orders.size() != num ||
		 flags.size() != num ) {
        gErrorLog.Throw( __FUNCTION__,
                         "Transpoint atoms have something wrong!\n",
                         ccWarning );
        return false;
    }

	FOR_ATOMS_IN_VECTOR( atm, points1 ) {
		if ( *atm == NULL || atm->GetMol() != this ) {
			gErrorLog.Throw( __FUNCTION__,
				"Not all the transpoint root atoms was in the molecule!\n",
				ccWarning );
			return false;
		}
	}

	map<CCMol*, bool> molPoints;
	FOR_ATOMS_IN_VECTOR( atm, points2 ) {
		if ( *atm == NULL || molPoints.count( atm->GetMol() ) ) {
			gErrorLog.Throw( __FUNCTION__,
				"Transpoint fragment atoms was in the same molecule!\n",
				ccWarning );
			return false;
		}
		molPoints[atm->GetMol()] = true;
	}

	for ( int i = 0; i < num; ++i ) {
		CCAtom *point1 = points1[i];
		CCAtom *point2 = points2[i];
		int order = orders[i];
		int flag  = flags[i];
		CCMol *m1 = point1->GetMol();
		CCMol *m2 = point2->GetMol();
		// junction atom can not take the bond order.
		if ( point1->ImpHydrogenCount() < order
		  || point2->ImpHydrogenCount() < order
		  || order < 1 ) {
			string msg = "Transpoint atom or bond order was not suitable!\n";
			msg += "M1: " + m1->GetName() +
				   " -- Atom: " + toSTR( point1->GetIdx() ) + "\n";
			msg += "M2: " + m2->GetName() +
				   " -- Atom: " + toSTR( point2->GetIdx() ) + "\n";
			gErrorLog.Throw( __FUNCTION__, msg, ccWarning );
			return false;
		}

		int nAtom = m_natom;
		AddMol( *m2 );
		CCAtom* junction1 = point1;
		CCAtom* junction2 = GetAtom( point2->GetIdx() +
									 nAtom );

		if ( junction1 && junction2 ) {
			// add bond between two points and set bond order and flag
			AddBond( junction1->GetIdx(), junction2->GetIdx(), order, flag );
			// modify the formal charge of junction atoms, which is necessary
			// for adding phosphate and sulphate groups. 
			if ( junction1->GetFormalCharge() < 0 ) {
				junction1->SetFormalCharge( junction1->GetFormalCharge() + 1 );
			}
			if ( junction2->GetFormalCharge() < 0 ) {
				junction2->SetFormalCharge( junction2->GetFormalCharge() + 1 );
			}
			m_nsubs--;

			ClearFlag(); // clear all flags; different from SetFlag( 0 );
			SetFlag( MOL_CONNECTIVITY_PERCEIVED );
		}
		else {
			string msg = "Junction atoms were not exist!\n";
			msg += m1->GetName() + " -- Atom: " +
				   toSTR( point1->GetIdx() ) + "\n" +
				   m2->GetName() + " -- Atom: " +
				   toSTR( point2->GetIdx() );
			gErrorLog.Throw( __FUNCTION__, msg, ccWarning );
			return false;
		}
	}
    return true;
}

int CCMol::GetNumHvyAtom() {
	int hvy = 0;
	FOR_ATOMS_OF_MOL( a, this ) {
		if ( a->IsHvyAtom() ) {
			++hvy;
		}
	}
	return hvy;
}

int CCMol::GetNumChiralAtom() {
	PerceiveChirality();
	int n = 0;
	FOR_ATOMS_OF_MOL( a, this ) {
		if ( a->HasFlag( ATOM_IS_CHIRAL ) ) {
			++n;
		}
	}
	return n;
}

int CCMol::GetNumBond()
{
    if ( !HasFlag( MOL_CONNECTIVITY_PERCEIVED ) ) {
        PerceiveInternalAtomBondTypes();
    }
    return m_nbond;
}

CCAtom* CCMol::AddAtom( CCAtom& atom )
{
    CCAtom* atm = new CCAtom( atom );
    AssertMemAlloc( atm );
    atm->SetMol( this );
    atm->SetIdx( m_natom + 1 );

    /**
	* @brief atom creation increment
	* @remark since alloc memory once is much more effective than 10000
	*         times for a mol with 10000 atoms, let's save some time by
	*          allocating memory for 100 atoms once.
	*/
#define CCAtomIncrement 100
    if ( m_natom + 1 > m_vatom.size() ) {
        m_vatom.resize( m_natom + CCAtomIncrement );
    }
#undef CCAtomIncrement
    m_vatom[m_natom] = atm;
    m_natom++;

    return atm;
}

CCBond* CCMol::AddBond( CCBond& bond )
{
    return AddBond( bond.GetBeginAtomIdx(), bond.GetEndAtomIdx(),
                    bond.GetOrder(), bond.GetFlag() );
}

CCBond* CCMol::AddBond( int first, int second, int order, unsigned long flag,
                        int insertpos )
{
    if ( first == second
      || first > m_natom
      || second > m_natom ) {

        return NULL;
    }

    CCBond* bnd(NULL);
    if ( bnd = GetBond( first, second ) ) {
        // bond already exists
        bnd->SetOrder( order );
        bnd->SetFlag( flag );
        return bnd;
    }


    CCAtom* bgn, * end;
    bgn = GetAtom( first );
    end = GetAtom( second );

    if ( !bgn || !end ) {
        return NULL;
    }

    bnd = new CCBond;
	AssertMemAlloc( bnd );

    bnd->Set( m_nbond + 1, bgn, end, order, flag );
    bnd->SetMol( this );

    /**
	* @brief bond creation increment
	* @remark since alloc memory once is much more effective than 10000
	*         times for a mol with 10000 bonds, let's save some time by
	*         alloc memory for 100 bonds once.
	*/
#define CCBondIncrement 100
    if ( m_nbond + 1 > m_vbond.size() ) {
        m_vbond.resize( m_nbond + CCBondIncrement );
    }
#undef CCBondIncrement
    m_vbond[m_nbond] = bnd;
    m_nbond++;

    // set aromatic info if possible
    if ( order == 5 ) {
        bnd->SetAromatic();
        bgn->SetAromatic();
        end->SetAromatic();
    }

    bgn->AddBond( bnd );
    end->AddBond( bnd );

    return bnd;
}

CCResidue* CCMol::AddResidue( CCResidue& residue )
{
    CCResidue* res = new CCResidue( residue );
    AssertMemAlloc( res );
    res->SetMol( this );
    res->SetIdx( m_vresidue.size() + 1 );
    m_vresidue.push_back( res );

    return res;
}

CCChain* CCMol::AddChain( CCChain& chain )
{
    CCChain* chn = new CCChain( chain );
    AssertMemAlloc( chn );
    chn->SetMol( this );
    chn->SetIdx( m_vchain.size() + 1 );
    m_vchain.push_back( chn );

    return chn;
}

CCAtom* CCMol::NewAtom()
{
    CCAtom* patom = new CCAtom;
    AssertMemAlloc( patom );
    patom->SetMol( this );
    patom->SetIdx( m_natom + 1 );

#define CCAtomIncrement 100
    if ( m_natom + 1 > m_vatom.size() ) {
        m_vatom.resize( m_natom + CCAtomIncrement );
    }
#undef CCAtomIncrement
    m_vatom[m_natom] = patom;
    m_natom++;
    return patom;
}

CCBond* CCMol::NewBond()
{
    CCBond* pbond = new CCBond; AssertMemAlloc( pbond );
    pbond->SetMol( this );
    pbond->SetIdx( m_nbond + 1 );

    // need further consideration ?!
#define CCBondIncrement 100
    if ( m_nbond + 1 > m_vbond.size() ) {
        m_vbond.resize( m_nbond + CCBondIncrement );
    }
#undef CCBondIncrement
    m_vbond[m_nbond] = pbond;
    m_nbond++;
    return pbond;
}

CCResidue* CCMol::NewResidue()
{
    CCResidue* residue = new CCResidue; AssertMemAlloc( residue );
    residue->SetMol( this );
    residue->SetIdx( m_vresidue.size() + 1 );

#define CCResidueIncrement 100
    m_vresidue.push_back( residue );
#undef CCResidueIncrement
    return( residue );
}

CCChain* CCMol::NewChain()
{
    CCChain* chain = new CCChain; AssertMemAlloc( chain );
    chain->SetMol( this );
    chain->SetIdx( m_vchain.size() + 1 );

#define CCResidueIncrement 100
    m_vchain.push_back( chain );
#undef CCResidueIncrement
    return( chain );
}

CCAtom* CCMol::BeginAtom( CCAtomIter& i )
{
    i = m_vatom.begin();
    return ( i == m_vatom.end() ? NULL : *i );
}

CCAtom* CCMol::NextAtom( CCAtomIter& i )
{
    i++;
    return ( i == m_vatom.end() ? NULL : *i );
}

CCBond* CCMol::BeginBond( CCBondIter& i )
{
    i = m_vbond.begin();
    return ( i == m_vbond.end() ? NULL : *i );
}

CCBond* CCMol::NextBond( CCBondIter& i )
{
    ++i;
    return ( i == m_vbond.end() ? NULL : *i );
}

CCRing* CCMol::BeginRing( CCRingIter& i )
{
    i = m_vring.begin();
    return ( i == m_vring.end() ? NULL : *i );
}

CCRing* CCMol::NextRing( CCRingIter& i )
{
    ++i;
    return ( i == m_vring.end() ? NULL : *i );
}

CCResidue* CCMol::BeginResidue( CCResidueIter& i )
{
    i = m_vresidue.begin();
    return ( ( i == m_vresidue.end() ) ? NULL : *i );
}

CCResidue* CCMol::NextResidue( CCResidueIter& i )
{
    i++;
    return ( ( i == m_vresidue.end() ) ? NULL : *i );
}

CCChain* CCMol::BeginChain( CCChainIter& i )
{
    i = m_vchain.begin();
    return ( ( i == m_vchain.end() ) ? NULL : *i );
}

CCChain* CCMol::NextChain( CCChainIter& i )
{
    i++;
    return ( ( i == m_vchain.end() ) ? NULL : *i );
}

CCAtom* CCMol::GetAtom( int idx ) const
{
    return ( idx <1 || idx> m_natom ) ? NULL : m_vatom[idx - 1];
}

CCBond* CCMol::GetBond( int idx ) const
{
    return ( idx <1 || idx> m_nbond ) ? NULL : m_vbond[idx - 1];
}

CCBond* CCMol::GetBond( int start, int end ) const
{
    return GetBond( GetAtom( start ), GetAtom( end ) );
}

CCBond* CCMol::GetBond( CCAtom* bgn, CCAtom* end ) const
{
    if ( !bgn || !end ) {
        return NULL;
    }

    CCBondIter i;
    CCAtom* nbr( NULL );
    for ( nbr = bgn->BeginNbrAtom( i ); nbr; nbr = bgn->NextNbrAtom( i ) ) {
        if ( nbr == end ) {
            return ( *i );
        }
    }

    return NULL;
}

CCResidue* CCMol::GetResidue( int idx ) const
{
    return ( idx <1 || idx> GetNumRes() )
         ? NULL
         : m_vresidue[idx - 1];
}

vector<CCResidue*> CCMol::GetResOfName( string resname ) const
{
    vector<CCResidue*> vres;
    FOR_RESIDUES_OF_MOL( res, ( CCMol * ) this ) {
        if ( res->GetResName() == resname ) {
            vres.push_back( &*res );
        }
    }
    return vres;
}

CCChain* CCMol::GetChain( int idx ) const
{
    return ( idx <1 || idx> GetNumChain() )
         ? NULL
         : m_vchain[idx - 1];
}

double CCMol::GetDistance( int a1, int a2 ) const
{
    return( GetDistance( GetAtom( a1 ), GetAtom( a2 ) ) );
}

double CCMol::GetDistance( CCAtom* a1, CCAtom* a2 ) const
{
    return( Distance( a1->GetXYZ(), a2->GetXYZ() ) );
}

double CCMol::GetAngle( int a1, int a2, int a3 ) const
{
    return( GetAngle( GetAtom( a1 ), GetAtom( a2 ), GetAtom( a3 ) ) );
}

double CCMol::GetAngle( CCAtom* a1, CCAtom* a2, CCAtom* a3 ) const
{
    return ( Angle( a1->GetXYZ(), a2->GetXYZ(), a3->GetXYZ() ) );
}

double CCMol::GetTorsion( int a1, int a2, int a3, int a4 ) const
{
    return( GetTorsion( GetAtom( a1 ), GetAtom( a2 ), GetAtom( a3 ),
                        GetAtom( a4 ) ) );
}

double CCMol::GetTorsion( CCAtom* a1, CCAtom* a2, CCAtom* a3, CCAtom* a4 ) const
{
    return( Torsion( a1->GetXYZ(), a2->GetXYZ(), a3->GetXYZ(), a4->GetXYZ() ) );
}

//////////////////////////////////////////////////////////////////////////
// Handle PDB file, implemented in I-interpret by Steven
// Ref. J. Chem. Inf. Model. 2007, 47, 1379-1385
void CCMol::HandlePDB()
{
    if ( IsEmpty() ) {
        return;
    }

    // first perceive connectivity from coordinates
    PerceiveConnectivityFromXYZ();

    // then perceive putative hybridization from geometry
    // may be revised in later
    PerceivePutativeHybFromXYZ();

    // perceive functional group (and revise hybridization)
    PerceiveFGroupFromXYZ();

    // perceive aromaticity
    PerceiveAromaticityFromXYZ();
    //SetFlag( MOL_AROMATICITY_PERCEIVED );

    // perceive bond order
    PerceiveBondOrderFromXYZ();

    // now is ready for assign internal atom and bond types
    PerceiveAtomTypesFromXYZ();
}

void CCMol::PerceiveConnectivityFromXYZ()
{
    if ( HasFlag( MOL_CONNECTIVITY_PERCEIVED ) ) {
        return;
    }
    SetFlag( MOL_CONNECTIVITY_PERCEIVED );

    vector<double> radiu;
    vector<int> maxneib;
    CCMolAtomIter ai;
    CCAtom* nbrAtom;
    int rootIdx, nbrIdx;

    // make radiu[i] is the covalent radiu of ith atom
    radiu.push_back( 0.0 );
    maxneib.push_back( 0 );
    FOR_ATOMS_OF_MOL( atom, this ) {
        radiu.push_back( gElementalTable.GetCovalentRad( atom->GetAtomicNum() ) );
        maxneib.push_back( gElementalTable.GetMaxBond( atom->GetAtomicNum() ) );
    }

    // produce the candidate list by equation 0.1 < rij < Ri + Rj + 0.4
    // vector<vector<double> > length( m_natom, vector<double>( m_natom ) );
    FOR_ATOMS_OF_MOL( atom, this ) {
        for ( ai = atom, ai++; ai; ++ai ) {
            nbrAtom = &*ai;

            // use alternate location indicator in PDB file to handle
            // alternative atom
            //if ( !HasSameAltlocData( &*atom, nbrAtom) ) {
            //  continue;
            //}
            //if ( atom->GetResidue() && nbrAtom->GetResidue() &&
            //   atom->GetResidue()->GetAltLoc() !=
            //   nbrAtom->GetResidue()->GetAltLoc() ) {
            //  continue;
            //}
            if ( atom->GetResidue()
              && nbrAtom->GetResidue()
              && atom->GetResidue()->GetAltLoc()
              != nbrAtom->GetResidue()->GetAltLoc() ) {
                continue;
            }

            // bonded if closer than elemental Rcov + tolerance
            rootIdx = atom->GetIdx();
            nbrIdx = nbrAtom->GetIdx();
            double d = GetDistance( &*atom, nbrAtom );
            double dref = radiu[rootIdx] + radiu[nbrIdx] + 0.4;
            if ( d > 0.1 && d < dref ) {
                AddBond( rootIdx, nbrIdx, 1 );
            }
        }
    }

    // check the candidate list to solve the conflicts
    // whose atomic links exceed its max neighbor
    FOR_ATOMS_OF_MOL( atom, this ) {
        rootIdx = atom->GetIdx();
        int conflict = atom->GetValence() - maxneib[rootIdx];
        if ( conflict <= 0 ) {
            continue;
        }

        vector<CCBond*> delbonds;
        vector<CCBond*>::iterator dbi;
        // first: link with both vatom exceed their max neighbor is broken
        FOR_NBORS_OF_ATOM( nbr, &*atom ) {
            nbrIdx = nbr->GetIdx();
            if ( ( (CCBond*)~nbr )->HasFlag( BOND_IS_INPUT ) ) {
                continue;
            }

            if ( nbr->GetValence() - maxneib[nbrIdx] > 0 ) {
                delbonds.push_back( (CCBond*)~nbr );

                if ( --conflict == 0 ) {
                    break;
                }
            }
        }
        for ( dbi = delbonds.begin(); dbi != delbonds.end(); ++dbi ) {
			// should not call DeleteBond() because it would modify the
			// flag of mol

			CCBond *bond = (*dbi);
			// remove this bond from its component
			bond->GetBeginAtom()->RemoveBond( bond );
			bond->GetEndAtom()->RemoveBond( bond );

			// remove this bond from connection table
			m_vbond.erase( m_vbond.begin() + ( bond->GetIdx() - 1 ) );
			m_nbond--;

			// reset bond id and update connection table
			for ( int bidx = 1; bidx <= m_nbond; ++bidx ) {
				m_vbond[bidx - 1]->SetIdx( bidx );
			}

			// destroy this bond
			DestroyIt( bond );
        }
        delbonds.clear();

        // then: link between two vatom which have different res code is broken
        // finally: link with largest bond length is broken
        while ( conflict > 0 ) {
            CCBond* maxLengthDevBond( NULL );
            double maxLengthDev = 0.0;  // maximal bond length deviation
            double lengthDev = 0.0;
            bool diffRes = false;

            FOR_NBORS_OF_ATOM( nbr, &*atom ) {
                nbrIdx = nbr->GetIdx();
                if ( ( (CCBond*)~nbr )->HasFlag( BOND_IS_INPUT ) ) {
                    continue;
                }

                lengthDev = fabs( ( (CCBond*)~nbr )->GetLength()
                                - ( (CCBond*)~nbr )->GetSTLength() );
                if ( atom->GetResidue() != nbr->GetResidue() ) {
                    if ( !diffRes ) {
                        diffRes = true;
                        maxLengthDevBond = (CCBond*)~nbr;
                        maxLengthDev = lengthDev;
                    }
                    else if ( lengthDev > maxLengthDev ) {
                        maxLengthDevBond = (CCBond*)~nbr;
                        maxLengthDev = lengthDev;
                    }
                }
                else if ( diffRes ) {
                    continue;
                }
                else if ( lengthDev > maxLengthDev ) {
                    maxLengthDevBond = (CCBond*)~nbr;
                    maxLengthDev = lengthDev;
                }
            }

            if ( maxLengthDevBond != NULL ) {
				// should not call DeleteBond() because it would modify the
				// flag of mol

				CCBond *bond = maxLengthDevBond;
				// remove this bond from its component
				bond->GetBeginAtom()->RemoveBond( bond );
				bond->GetEndAtom()->RemoveBond( bond );

				// remove this bond from connection table
				m_vbond.erase( m_vbond.begin() + ( bond->GetIdx() - 1 ) );
				m_nbond--;

				// reset bond id and update connection table
				for ( int bidx = 1; bidx <= m_nbond; ++bidx ) {
					m_vbond[bidx - 1]->SetIdx( bidx );
				}

				// destroy this bond
				DestroyIt( bond );
                //DeleteBond( maxLengthDevBond );
                --conflict;
            }
            else {
                gErrorLog.Throw( __FUNCTION__,
                                 "atomic links exceed max neighbor!\n",
                                 ccWarning );

                FOR_NBORS_OF_ATOM( nbr, &*atom ) {
                    ( (CCBond*)~nbr )->UnsetFlag( BOND_IS_INPUT );
                }
            }
        }
    }
}

void CCMol::PerceivePutativeHybFromXYZ()
{
    if ( HasFlag( MOL_PUTATIVE_HYBRID_PERCEIVED ) ) {
        return;
    }
    SetFlag( MOL_PUTATIVE_HYBRID_PERCEIVED );

    if ( !HasFlag( MOL_CONNECTIVITY_PERCEIVED ) ) {
        PerceiveConnectivityFromXYZ();
    }

    static double SP1CUTOFFANGLE = 155.0;
    static double SP2CUTOFFANGLE = 115.0;
    static double FLATRINGFIVE = 7.5;
    static double FLATRINGSIX = 12.0;

    // Step 1: Assign putative hybridization state based on average angles
    //for ( atom = BeginAtom(i); atom; atom = NextAtom(i))
    FOR_ATOMS_OF_MOL( atom, this ) {
        // val is the number of explicit neighboring vatom
        int val = atom->GetValence();
        double angle;  // average bond angle

        if ( val == 1 ) {
            if ( atom->IsHydrogen() ) {
                atom->SetHyb( 3 );
                continue;
            }
            else {
                atom->SetHyb( 0 );
            }
        }
        else if ( val == 2 ) {
            if ( atom->IsOxygen() ) {
                atom->SetHyb( 3 );
                continue;
            }
            //if ( atom->IsNitrogen() && atom->GetHvyVal() == 1 ) {
            //    atom->SetHyb( 0 );
            //    continue;
            //}

            angle = atom->AverageBondAngle();
            if ( angle > SP1CUTOFFANGLE ) {
                atom->SetHyb( 1 );
                continue;
            }
            else if ( angle > SP2CUTOFFANGLE ) {
                atom->SetHyb( 2 );
                continue;
            }
            //****** there is a special case in original I-interpret ******//
            else {
                atom->SetHyb( 3 );
                continue;
            }
        }
        else if ( val == 3 ) {
            if ( atom->IsSulfur() ) {
                atom->SetHyb( 2 );
                continue;
            }
            //if ( atom->IsNitrogen() && atom->GetHvyVal() == 1 ) {
            //    atom->SetHyb( 0 );
            //    continue;
            //}

            angle = atom->AverageBondAngle();
            if ( angle > SP2CUTOFFANGLE ) {
                atom->SetHyb( 2 );
                continue;
            }
            else {
                atom->SetHyb( 3 );
                continue;
            }
        }
        //else if ( val == 4 && atom->IsPhosphorus() ) {
        //    atom->SetHyb( 2 );
        //}
        else {
            atom->SetHyb( 3 );
        }
    }

    // Step 2: look for 5 or less membered rings with torsions < 7.5 degrees
    //         and 6 or more membered rings with torsions < 12 degrees
    //         (set all vatom with at least two bonds to sp2)
    FOR_RINGS_OF_MOL( ring, this ) {
        if ( ring->IsNearPlanar() ) {
            ring->SetFlag( RING_IS_PLANAR );
            FOR_ATOMS_OF_RING( atom, &*ring ) {
                if ( atom->IsOxygen() ) {
                    atom->SetHyb( 3 );
                }
                else if ( atom->GetValence() == 2 ) {
                    atom->SetHyb( 2 );
                }
            }
        }
        else {
            ring->SetFlag( RING_IS_NORMAL );
        }
    }

    // Step 3 : If an sp-hybridized atom doesn't have a sp-hybridized or
    //          terminal neighbor with unfilled valence, it is reassigned
    //          as sp2-hybridized. Similarly, if an sp2-hybridized atom
    //          doesn't have a sp2-hybridized or terminal neighbor with
    //          unfilled valence, it is reassigned as sp3-hybridized

    // temporarily set MOL_HYBRID_PERCEIVED to avoid GetHyb()
    // calling on HandlePDB()
    SetFlag( MOL_HYBRID_PERCEIVED );
    FOR_ATOMS_OF_MOL( atom, this ) {
        if ( atom->GetHyb() == 1 ) {
            bool mark = true;
            FOR_NBORS_OF_ATOM( nbr, &*atom ) {
                if ( nbr->GetHyb() < 2 ) {
                    mark = false;
                    break;
                }
            }
            if ( mark ) {
                atom->SetHyb( 2 );
            }
        }

        if ( atom->GetHyb() == 2 ) {
            bool mark = true;
            FOR_NBORS_OF_ATOM( nbr, &*atom ) {
                if ( nbr->GetHyb() < 3 ) {
                    mark = false;
                    break;
                }
            }
            if ( mark ) {
                atom->SetHyb( 3 );
            }
        }
    }
    UnsetFlag( MOL_HYBRID_PERCEIVED );
}

// CompareAtoms() was used to sort neighboring vatom by
// their type and bond length to the root atom.
// The less weight means that the atom is more likely
// to be sp2 hybridizatom.
bool CompareAtoms( const pair<CCAtom*, CCBond*>& T1,
                   const pair<CCAtom*, CCBond*>& T2 )
{
    CCAtom* atom1 = T1.first;
    CCAtom* atom2 = T2.first;
    int weight1 = 99, weight2 = 99;

    if ( atom1->IsTerOxygen() ) {
        weight1 = 1;
    }
    else if ( atom1->IsTerSulfur() ) {
        weight1 = 2;
    }
    else if ( atom1->IsOxygen() && atom1->GetHvyVal() == 1 ) {
        weight1 = 3;
    }
    else if ( atom1->IsSulfur() && atom1->GetHvyVal() == 1 ) {
        weight1 = 4;
    }
    else if ( atom1->IsTerNitrogen() ) {
        weight1 = 5;
    }
    else if ( atom1->IsPotentialSp2N() ) {
        weight1 = 6;
    }
    else if ( atom1->IsNitrogen() && atom1->GetHvyVal() == 1 ) {
        weight1 = 7;
    }
    else if ( atom1->IsOxygen() ) {
        weight1 = 8;
    }
    else if ( atom1->IsSulfur() ) {
        weight1 = 9;
    }
    else if ( atom1->IsNitrogen() ) {
        weight1 = 10;
    }

    if ( atom2->IsTerOxygen() ) {
        weight2 = 1;
    }
    else if ( atom2->IsTerSulfur() ) {
        weight2 = 2;
    }
    else if ( atom2->IsOxygen() && atom2->GetHvyVal() == 1 ) {
        weight2 = 3;
    }
    else if ( atom2->IsSulfur() && atom2->GetHvyVal() == 1 ) {
        weight2 = 4;
    }
    else if ( atom2->IsTerNitrogen() ) {
        weight2 = 5;
    }
    else if ( atom2->IsPotentialSp2N() ) {
        weight2 = 6;
    }
    else if ( atom2->IsNitrogen() && atom2->GetHvyVal() == 1 ) {
        weight1 = 7;
    }
    else if ( atom2->IsOxygen() ) {
        weight2 = 8;
    }
    else if ( atom2->IsSulfur() ) {
        weight2 = 9;
    }
    else if ( atom2->IsNitrogen() ) {
        weight2 = 10;
    }

    if ( weight1 == weight2 ) {
        return ( T1.second->GetLength() < T2.second->GetLength() );
    }
    else {
        return ( weight1 < weight2 );
    }
}

void CCMol::PerceiveFGroupFromXYZ()
{
    if ( HasFlag( MOL_FGROUP_PERCEIVED ) ) {
        return;
    }
    SetFlag( MOL_FGROUP_PERCEIVED );

    if ( !HasFlag( MOL_PUTATIVE_HYBRID_PERCEIVED ) ) {
        PerceivePutativeHybFromXYZ();
    }

    // temporarily set MOL_HYBRID_PERCEIVED to avoid GetHyb()
    // calling on HandlePDB(); The flag will be unset at the end.
    SetFlag( MOL_HYBRID_PERCEIVED );

    FOR_ATOMS_OF_MOL( atom, this ) {
        if ( atom->IsValid()
          || !( atom->IsSp2Carbon()
             || atom->IsSp1Carbon()
             || atom->IsSp2Nitrogen()
             || atom->IsSp1Nitrogen()
             || atom->IsPhosphorus()
             || atom->IsSulfur()
             || atom->IsSp2Selenium() ) ) {
            continue;
        }

        // gather the information of neighboring vatom
        vector<pair<CCAtom*, CCBond*> > vp;
        vector<pair<CCAtom*, CCBond*> >::iterator vi;
        int possSp2OS = 0, possSp2N = 0;
        int possSp3OS = 0, possSp3N = 0;
        FOR_NBORS_OF_ATOM( nbr, &*atom ) {
            if ( nbr->IsTerOxygen() || nbr->IsTerSulfur() ) {
                ++possSp2OS;
            }
            else if ( nbr->IsPotentialSp2N() ) {
                ++possSp2N;
            }
            else if ( nbr->IsOxygen() || nbr->IsSulfur() ) {
                ++possSp3OS;
            }
            else if ( nbr->IsNitrogen() ) {
                ++possSp3N;
            }
            else {
                continue;
            }
            vp.push_back( make_pair( &*nbr, (CCBond*)~nbr ) );
        }

        // handle sp1 carbon and nitrogen which may contain no terminal vatom
        // in functional groups
        if ( atom->IsSp1Carbon() && atom->GetHvyVal() == 2 ) {
            bool cyanoroup = false;
            int sp2mark = 0;
            FOR_NBORS_OF_ATOM( nbr, &*atom ) {
                if ( nbr->IsValid() ) {
                    continue;
                }
                if ( nbr->IsTerNitrogen() && ( (CCBond*)~nbr )->IsLengthInTB() ) {
                    // # cyano -C#N
                    atom->SetType( "C.CN" );
                    atom->SetHyb( 1 );
                    atom->SetValidFGroup();
                    nbr->SetType( "N.CN" );
                    nbr->SetHyb( 1 );
                    nbr->SetValidFGroup();

                    ( (CCBond*)~nbr )->SetType( "TB.CN" );
                    ( (CCBond*)~nbr )->SetValidFGroup();

                    cyanoroup = true;
                    break;
                }
                // !------- check [C,O,S,N]=C=[C,O,S,N]
                else if ( ( nbr->IsCarbon() && nbr->GetHyb() < 3 )
                       || nbr->IsSp2Nitrogen()
                       || nbr->IsPotentialSp2N()
                       || nbr->IsTerOxygen()
                       || nbr->IsTerSulfur() ) {
                    ++sp2mark;
                }
            }

            if ( cyanoroup ) {
                continue;
            }
            else if ( sp2mark == 2 ) {
                // [C,O,S,N]=C=[C,O,S,N]
                atom->SetType( "C.1" );
                atom->SetHyb( 1 );
                atom->SetValidFGroup();
                FOR_NBORS_OF_ATOM( nbr, &*atom ) {
                    if ( nbr->IsSp1Carbon() ) {
                        nbr->SetFGroup();
                    }
                    else {
                        nbr->SetType( nbr->GetSymbol() + ".2" );
                        nbr->SetHyb( 2 );
                        nbr->SetValidFGroup();
                    }

                    ( (CCBond*)~nbr )->SetType( "DB" );
                    ( (CCBond*)~nbr )->SetValidFGroup();
                }
                continue;
            }
        }

        // !------- check azide -N=[N+]=[N-] and diazomethane >C=[N+]=[N-]
        else if ( atom->IsSp1Nitrogen() && atom->GetHvyVal() == 2 ) {
            CCAtom* aterN( NULL );
            CCBond* bterN( NULL );
            CCAtom* anterNC( NULL );
            CCBond* bnterNC( NULL );
            FOR_NBORS_OF_ATOM( nbr, &*atom ) {
                if ( nbr->IsTerNitrogen() ) {
                    if ( aterN == NULL ) {
                        aterN = &*nbr;
                        bterN = (CCBond*)~nbr;
                    }
                    else if ( ( (CCBond*)~nbr )->GetLength() < bterN->GetLength() ) {
                        anterNC = aterN;
                        bnterNC = bterN;
                        aterN = &*nbr;
                        bterN = (CCBond*)~nbr;
                    }
                    else {
                        anterNC = &*nbr;
                        bnterNC = (CCBond*)~nbr;
                    }
                }
                else if ( nbr->IsSp2Nitrogen() || nbr->IsSp2Carbon() ) {
                    anterNC = &*nbr;
                    bnterNC = (CCBond*)~nbr;
                }
                else {
                    break;
                }
            }

            if ( aterN != NULL && anterNC != NULL ) {
                // -N=[N+]=[N-]
                if ( anterNC->IsNitrogen() ) {
                    atom->SetType( "N+.NNN" );
                    atom->SetHyb( 1 );
                    atom->SetFormalCharge( 1 );
                    atom->SetValidFGroup();
                    aterN->SetType( "N-.NNN" );
                    aterN->SetHyb( 1 );
                    aterN->SetFormalCharge( -1 );
                    aterN->SetValidFGroup();
                    anterNC->SetType( "N.NNN" );
                    anterNC->SetHyb( 1 );
                    anterNC->SetValidFGroup();

                    bterN->SetType( "DB.NNN" );
                    bterN->SetValidFGroup();
                    bnterNC->SetType( "DB.NNN" );
                    bnterNC->SetValidFGroup();
                }
                // >C=[N+]=[N-]
                else if ( anterNC->IsCarbon() ) {
                    atom->SetType( "N+.CNN" );
                    atom->SetHyb( 1 );
                    atom->SetFormalCharge( 1 );
                    atom->SetValidFGroup();
                    aterN->SetType( "N-.CNN" );
                    aterN->SetHyb( 1 );
                    aterN->SetFormalCharge( -1 );
                    aterN->SetValidFGroup();
                    anterNC->SetType( "N.CNN" );
                    anterNC->SetHyb( 1 );
                    anterNC->SetValidFGroup();

                    bterN->SetType( "DB.CNN" );
                    bterN->SetValidFGroup();
                    bnterNC->SetType( "DB.CNN" );
                    bnterNC->SetValidFGroup();
                }
            }
        }


        if ( possSp2OS == 0 && ( possSp2N + possSp3N ) < 2 ) {
            continue;
        }
        sort( vp.begin(), vp.end(), CompareAtoms );

        if ( atom->IsSp2Carbon() ) {
            // O(ter)-C: standard C-O bond length = 1.43 (-0.11)
            CCAtom* fstatom = vp[0].first;
            CCBond* fstbond = vp[0].second;
            if ( fstatom->IsTerOxygen() || fstatom->IsTerSulfur() ) {
                if ( vp.size() > 1 ) {
                    CCAtom* secatom = vp[1].first;
                    CCBond* secbond = vp[1].second;
                    if ( secatom->IsTerOxygen() ) {
                        // C(=O)-O[-]
                        atom->SetType( "C.CO2" );
                        atom->SetHyb( 2 );
                        atom->SetValidFGroup();
                        fstatom->SetType( "O.CO2" );
                        fstatom->SetHyb( 2 );
                        fstatom->SetValidFGroup();
                        secatom->SetType( "O.CO2" );
                        secatom->SetHyb( 2 );
                        secatom->SetFormalCharge( -1 );
                        secatom->SetValidFGroup();

                        fstbond->SetType( "DB.AR.CO2" );
                        fstbond->SetValidFGroup();
                        secbond->SetType( "SB.AR.CO2" );
                        secbond->SetValidFGroup();

                        // !------- check N-C(=O)-O[-]
                        if ( vp.size() > 2 ) {
                            CCAtom* thratom = vp[2].first;
                            CCBond* thrbond = vp[2].second;
                            if ( thratom->IsNitrogen() ) {
                                if ( !thratom->HasFlag( ATOM_IS_VALID ) ) {
                                    // only set fgroup because it can
                                    // also be N.2.
                                    thratom->SetType( "N.am" );
                                    thratom->SetHyb( 2 );
                                    thratom->SetFGroup();
                                }
                                thrbond->SetType( "AM" );
                                thrbond->SetValidFGroup();
                            }
                        }
                    }
                    else if ( secatom->IsOxygen() || secatom->IsSulfur() ) {
                        // C(=O/S)-OH/S(H)
                        if ( !atom->HasFlag( ATOM_IS_VALID ) ) {
                            atom->SetType( "C.2" );
                            atom->SetHyb( 2 );
                            atom->SetValidFGroup();
                            fstatom->SetSp2Type();
                            fstatom->SetHyb( 2 );
                            fstatom->SetValidFGroup();

                            fstbond->SetType( "DB" );
                            fstbond->SetValidFGroup();
                        }

                        if ( !secatom->HasFlag( ATOM_IS_VALID ) ) {
                            secatom->SetSp3Type();
                            secatom->SetHyb( 3 );
                            secatom->SetValidFGroup();
                        }

                        secbond->SetType( "SB" );
                        secbond->SetValidFGroup();

                        // !------- check N-C(=O/S)-OH/S(H)
                        if ( vp.size() > 2 ) {
                            CCAtom* thratom = vp[2].first;
                            CCBond* thrbond = vp[2].second;
                            if ( thratom->IsNitrogen() ) {
                                if ( !thratom->HasFlag( ATOM_IS_VALID ) ) {
                                    thratom->SetType( "N.am" );
                                    thratom->SetHyb( 2 );
                                    thratom->SetFGroup();
                                }
                                thrbond->SetType( "AM" );
                                thrbond->SetValidFGroup();
                            }
                        }
                    }
                    else if ( secatom->IsNitrogen() ) {
                        // >N-C(=O/S)- (C is not in ring or
                        // C is in ring but C=O/S bond length in DB)
                        if ( !atom->IsInRing() || fstbond->IsLengthInDB() ) {
                            atom->SetType( "C.2" );
                            atom->SetHyb( 2 );
                            atom->SetValidFGroup();
                            fstatom->SetSp2Type();
                            fstatom->SetHyb( 2 );
                            fstatom->SetValidFGroup();
                            if ( !secatom->HasFlag( ATOM_IS_VALID ) ) {
                                secatom->SetType( "N.am" );
                                secatom->SetHyb( 2 );
                                secatom->SetFGroup();
                            }

                            fstbond->SetType( "DB" );
                            fstbond->SetValidFGroup();
                            secbond->SetType( "AM" );
                            secbond->SetValidFGroup();

                            if ( vp.size() > 2 ) {
                                CCAtom* thratom = vp[2].first;
                                CCBond* thrbond = vp[2].second;
                                if ( !thratom->HasFlag( ATOM_IS_VALID ) ) {
                                    thratom->SetType( "N.am" );
                                    thratom->SetHyb( 2 );
                                    thratom->SetFGroup();
                                }

                                thrbond->SetType( "AM" );
                                thrbond->SetValidFGroup();
                            }
                        }
                    }
                }
                else if ( fstbond->IsLengthInDB() ) {
                    atom->SetType( "C.2" );
                    atom->SetHyb( 2 );
                    atom->SetValidFGroup();
                    fstatom->SetSp2Type();
                    fstatom->SetHyb( 2 );
                    fstatom->SetValidFGroup();

                    fstbond->SetType( "DB" );
                    fstbond->SetValidFGroup();
                }
            }

            // !------- check N-C(=N)-(N)
            else if ( fstatom->IsPotentialSp2N() ) {
                if ( vp.size() > 1 ) {
                    CCAtom* secatom = vp[1].first;
                    CCBond* secbond = vp[1].second;
                    if ( secatom->IsNitrogen() && secatom->GetHvyVal() == 1 ) {
                        // # guanyl N-C(=N)
                        atom->SetType( "C.Ng+" );
                        atom->SetHyb( 2 );
                        atom->SetValidFGroup();
                        fstatom->SetType( "N.Ng+" );
                        fstatom->SetHyb( 2 );
                        fstatom->SetValidFGroup();
                        // it may be set as N.am or N.2 before.
                        if ( !secatom->HasFlag( ATOM_IS_IN_FGROUP ) ) {
                            secatom->SetType( "N.Ng+" );
                            secatom->SetHyb( 2 );
                            secatom->SetFGroup();
                        }

                        fstbond->SetType( "DB.AR.Ng+" );
                        fstbond->SetValidFGroup();
                        secbond->SetType( "SB.AR.Ng+" );
                        secbond->SetValidFGroup();

                        if ( vp.size() > 2 ) {
                            CCAtom* thratom = vp[2].first;
                            CCBond* thrbond = vp[2].second;
                            if ( thratom->IsNitrogen() ) {
                                // # guandium N-C(=N)-N
                                if ( !thratom->HasFlag( ATOM_IS_VALID ) ) {
                                    // only set fgroup because it can
                                    // also be N.2.
                                    thratom->SetType( "N.Ng+" );
                                    thratom->SetHyb( 2 );
                                    thratom->SetFGroup();
                                }
                                thrbond->SetType( "SB.AR.Ng+" );
                                thrbond->SetValidFGroup();
                            }
                        }
                    }
                }
            }
        }
        else if ( atom->IsSp2Nitrogen() ) {
            CCAtom* fstatom = vp[0].first;
            CCBond* fstbond = vp[0].second;

            if ( possSp2OS == 1
              && fstatom->IsTerOxygen()
              && fstbond->IsLengthInDB() ) {
                if ( atom->GetValence() == 2 ) {
                    //# nitroso -N=O
                    atom->SetType( "N.NO" );
                    atom->SetHyb( 2 );
                    atom->SetValidFGroup();
                    fstatom->SetType( "O.NO" );
                    fstatom->SetHyb( 2 );
                    fstatom->SetValidFGroup();

                    fstbond->SetType( "DB.NO" );
                    fstbond->SetValidFGroup();
                }
                else if ( atom->GetValence() == 3 ) {
                    // # nitro-oxide >N[+]=O[-]
                    atom->SetType( "N.NOx" );
                    atom->SetHyb( 2 );
                    atom->SetFormalCharge( 1 );
                    atom->SetValidFGroup();
                    fstatom->SetType( "O.NOx" );
                    fstatom->SetHyb( 2 );
                    fstatom->SetFormalCharge( -1 );
                    fstatom->SetValidFGroup();

                    fstbond->SetType( "SB.AR.NOx" );
                    fstbond->SetValidFGroup();
                }
            }
            else if ( possSp2OS == 2 ) {
                CCAtom* secatom = vp[1].first;
                CCBond* secbond = vp[1].second;
                if ( fstatom->IsTerOxygen() && secatom->IsTerOxygen() ) {
                    // # nitro -N[+](=O)-O[-]
                    atom->SetType( "N.NO2" );
                    atom->SetHyb( 2 );
                    atom->SetFormalCharge( 1 );
                    atom->SetValidFGroup();
                    fstatom->SetType( "O.NO2" );
                    fstatom->SetHyb( 2 );
                    fstatom->SetValidFGroup();
                    secatom->SetType( "O.NO2" );
                    secatom->SetHyb( 2 );
                    secatom->SetFormalCharge( -1 );
                    secatom->SetValidFGroup();

                    fstbond->SetType( "DB.AR.NO2" );
                    fstbond->SetValidFGroup();
                    secbond->SetType( "SB.AR.NO2" );
                    secbond->SetValidFGroup();
                }
            }
        }
        else if ( atom->IsPhosphorus() ) {
            CCAtom* fstatom = vp[0].first;
            CCBond* fstbond = vp[0].second;
            if ( fstatom->IsTerOxygen() ) {
                if ( possSp2OS > 1 ) {
                    CCAtom* secatom = vp[1].first;
                    CCBond* secbond = vp[1].second;
                    if ( secatom->IsTerOxygen() ) {
                        if ( possSp2OS > 2 ) {
                            CCAtom* thratom = vp[2].first;
                            CCBond* thrbond = vp[2].second;
                            if ( thratom->IsTerOxygen() ) {
                                if ( possSp2OS > 3 ) {
                                    CCAtom* fthatom = vp[3].first;
                                    CCBond* fthbond = vp[3].second;
                                    if ( fthatom->IsTerOxygen() ) {
                                        // # phosphate P(=O)(-O[-])(-O[-])(-O[-])
                                        fthatom->SetType( "O.PO3" );
                                        fthatom->SetHyb( 2 );
                                        fthatom->SetFormalCharge( -1 );
                                        fthatom->SetValidFGroup();

                                        fthbond->SetType( "SB.AR.PO3" );
                                        fthbond->SetValidFGroup();
                                    }
                                }

                                // # phosphate -P(=O)(-O[-])(-O[-])
                                atom->SetType( "P.PO3" );
                                atom->SetHyb( 3 );
                                atom->SetValidFGroup();
                                fstatom->SetType( "O.PO3" );
                                fstatom->SetHyb( 2 );
                                fstatom->SetFormalCharge( -1 );
                                fstatom->SetValidFGroup();
                                secatom->SetType( "O.PO3" );
                                secatom->SetHyb( 2 );
                                secatom->SetFormalCharge( -1 );
                                secatom->SetValidFGroup();
                                thratom->SetType( "O.PO3" );
                                thratom->SetHyb( 2 );
                                thratom->SetFormalCharge( -1 );
                                thratom->SetValidFGroup();

                                fstbond->SetType( "DB.AR.PO3" );
                                fstbond->SetValidFGroup();
                                secbond->SetType( "SB.AR.PO3" );
                                secbond->SetValidFGroup();
                                thrbond->SetType( "SB.AR.PO3" );
                                thrbond->SetValidFGroup();
                                continue;  // skip set phosphinate type
                            }
                        }

                        // # phosphinate -P(=O)(-O[-])(-O[-])
                        atom->SetType( "P.PO2" );
                        atom->SetHyb( atom->GetValence() < 4 ? 2 : 3 );
                        atom->SetValidFGroup();
                        fstatom->SetType( "O.PO2" );
                        fstatom->SetHyb( 2 );
                        fstatom->SetFormalCharge( -1 );
                        fstatom->SetValidFGroup();
                        secatom->SetType( "O.PO2" );
                        secatom->SetHyb( 2 );
                        secatom->SetFormalCharge( -1 );
                        secatom->SetValidFGroup();

                        fstbond->SetType( "DB.AR.PO2" );
                        fstbond->SetValidFGroup();
                        secbond->SetType( "SB.AR.PO2" );
                        secbond->SetValidFGroup();
                        continue;  // skip set phospine type
                    }
                }


                if ( fstbond->IsLengthInDB() ) {
                    // # phospine -P=O
                    atom->SetType( "P.PO" );
                    atom->SetHyb( atom->GetValence() < 4 ? 2 : 3 );
                    atom->SetValidFGroup();
                    fstatom->SetType( "O.PO" );
                    fstatom->SetHyb( 2 );
                    fstatom->SetValidFGroup();

                    fstbond->SetType( "DB.PO" );
                    fstbond->SetValidFGroup();
                    continue;
                }
            }
            else if ( fstbond->IsLengthInDB() ) {
                // # phospine -P=S/N
                atom->SetType( "P.3" );
                atom->SetHyb( atom->GetValence() < 4 ? 2 : 3 );
                atom->SetValidFGroup();
                fstatom->SetSp2Type();
                fstatom->SetHyb( 2 );
                fstatom->SetValidFGroup();

                fstbond->SetType( "DB" );
                fstbond->SetValidFGroup();
                continue;
            }
            //else {
            //  // # phospine -P-O/S
            //  atom->SetType( "P.3" );
            //  // atom->SetHyb( atom->GetValence() < 3 ? 2 : 3 );
            //  //atom->SetValidFGroup();
            //  fstatom->SetSp3Type();
            //  fstatom->SetHyb( 3 );
            //  fstatom->SetValidFGroup();

            //  fstbond->SetType( "SB" );
            //  fstbond->SetValidFGroup();
            //  continue;
            //}

        }
        else if ( atom->IsSulfur() ) {
            CCAtom* fstatom = vp[0].first;
            CCBond* fstbond = vp[0].second;
            if ( fstatom->IsTerOxygen() ) {
                if ( possSp2OS > 1 ) {
                    CCAtom* secatom = vp[1].first;
                    CCBond* secbond = vp[1].second;
                    if ( secatom->IsTerOxygen() ) {
                        if ( possSp2OS > 2 ) {
                            CCAtom* thratom = vp[2].first;
                            CCBond* thrbond = vp[2].second;
                            if ( thratom->IsTerOxygen() ) {
                                if ( possSp2OS > 3 ) {
                                    CCAtom* fthatom = vp[3].first;
                                    CCBond* fthbond = vp[3].second;
                                    if ( fthatom->IsTerOxygen() ) {
                                        // # sulphate S(=O)(=O)(-O[-])(-O[-])
                                        fthatom->SetType( "O.SO3" );
                                        fthatom->SetHyb( 2 );
                                        fthatom->SetFormalCharge( -1 );
                                        fthatom->SetValidFGroup();

                                        fthbond->SetType( "SB.AR.SO3" );
                                        fthbond->SetValidFGroup();
                                    }
                                }

                                // # sulphate S(=O)(=O)(-O[-])
                                atom->SetType( "S.SO3" );
                                atom->SetHyb( 3 );
                                atom->SetValidFGroup();
                                fstatom->SetType( "O.SO3" );
                                fstatom->SetHyb( 2 );
                                fstatom->SetValidFGroup();
                                secatom->SetType( "O.SO3" );
                                secatom->SetHyb( 2 );
                                secatom->SetFormalCharge( -1 );
                                secatom->SetValidFGroup();
                                thratom->SetType( "O.SO3" );
                                thratom->SetHyb( 2 );
                                thratom->SetFormalCharge( -1 );
                                thratom->SetValidFGroup();

                                fstbond->SetType( "DB.AR.SO3" );
                                fstbond->SetValidFGroup();
                                secbond->SetType( "DB.AR.SO3" );
                                secbond->SetValidFGroup();
                                thrbond->SetType( "SB.AR.SO3" );
                                thrbond->SetValidFGroup();
                                continue;  // skip set sulphone type
                            }
                        }

                        // # sulphone -S(=O)(=O)
                        atom->SetType( "S.SO2" );
                        atom->SetHyb( atom->GetValence() < 3 ? 1 : 3 );
                        atom->SetValidFGroup();
                        fstatom->SetType( "O.SO2" );
                        fstatom->SetHyb( 2 );
                        fstatom->SetValidFGroup();
                        secatom->SetType( "O.SO2" );
                        secatom->SetHyb( 2 );
                        secatom->SetValidFGroup();

                        fstbond->SetType( "DB.SO2" );
                        fstbond->SetValidFGroup();
                        secbond->SetType( "DB.SO2" );
                        secbond->SetValidFGroup();
                        continue;  // skip set sulphine type
                    }
                }


                if ( fstbond->IsLengthInDB() ) {
                    // # sulphine -S=O
                    atom->SetType( "S.SO" );
                    atom->SetHyb( atom->GetValence() < 3 ? 2 : 3 );
                    atom->SetValidFGroup();
                    fstatom->SetType( "O.SO" );
                    fstatom->SetHyb( 2 );
                    fstatom->SetValidFGroup();

                    fstbond->SetType( "DB.SO" );
                    fstbond->SetValidFGroup();
                    continue;
                }
            }
            else if ( possSp2OS == 0 && possSp2N == 2 ) {
                // !------- set >S(=N)(=N)
                atom->SetType( "S.3" );
                atom->SetHyb( atom->GetValence() < 3 ? 1 : 3 );
                atom->SetValidFGroup();
                FOR_NBORS_OF_ATOM( nbr, &*atom ) {
                    if ( nbr->IsPotentialSp2N() ) {
                        nbr->SetType( "N.2" );
                        nbr->SetHyb( 2 );
                        nbr->SetValidFGroup();

                        ( (CCBond*)~nbr )->SetType( "DB" );
                        ( (CCBond*)~nbr )->SetValidFGroup();
                    }
                }
            }
        }
        else if ( atom->IsSp2Selenium() ) {
            CCAtom* fstatom = vp[0].first;
            CCBond* fstbond = vp[0].second;
            if ( fstbond->IsLengthInDB() ) {
                // Se=O/S/N
                atom->SetType( "Se" );
                atom->SetHyb( atom->GetValence() < 4 ? 2 : 3 );
                atom->SetValidFGroup();
                fstatom->SetSp2Type();
                fstatom->SetHyb( 2 );
                fstatom->SetValidFGroup();

                fstbond->SetType( "DB" );
                fstbond->SetValidFGroup();
            }
        }
    }

    // temporarily set MOL_INT_TYPE_PERCEIVED to avoid GetType()
    // calling on PerceiveInternalAtomBondTypes(); The flag will
    // be unset at the end.
    SetFlag( MOL_INT_TYPE_PERCEIVED );

    // set bond order on functional groups
    FOR_BONDS_OF_MOL( bond, this ) {
        if ( bond->IsInFGroup() ) {
            string bondtype = bond->GetType();
            bondtype = bondtype.substr( 0, 2 );
            if ( bondtype == "TB" ) {
                bond->SetOrder( 3 );
            }
            else if ( bondtype == "DB" ) {
                bond->SetOrder( 2 );
            }
            else {
                bond->SetOrder( 1 );
            }
        }
    }
    UnsetFlag( MOL_INT_TYPE_PERCEIVED );

    UnsetFlag( MOL_HYBRID_PERCEIVED );
}

void CCMol::PerceiveAromaticityFromXYZ()
{
    if ( HasFlag( MOL_AROMATICITY_PERCEIVED ) ) {
        return;
    }
    SetFlag( MOL_AROMATICITY_PERCEIVED );

    if ( !HasFlag( MOL_FGROUP_PERCEIVED ) ) {
        PerceiveFGroupFromXYZ();
    }

    // unset aromaticity for all vatom and bonds
    FOR_ATOMS_OF_MOL( a, this )
    a->UnsetAromatic();
    FOR_BONDS_OF_MOL( b, this )
    b->UnsetAromatic();

    // temporarily set MOL_HYBRID_PERCEIVED to avoid GetHyb()
    // calling on HandlePDB(); The flag will be unset at the end.
    // The other two flag is much the same
    SetFlag( MOL_HYBRID_PERCEIVED );
    SetFlag( MOL_INT_TYPE_PERCEIVED );
    SetFlag( MOL_AUTOFCHARGE_PERCEIVED );

    CCAtom* ratom;
    CCBond* bond, * fbond, * bbond;

    DFSFindAllRings();

    FOR_RINGS_OF_MOL( ring, this ) {
        if ( !ring->IsNearPlanar() ) {
            ring->SetFlag( RING_IS_NORMAL );
            continue;
        }

        int ringsize = ring->GetSize();
        if ( ringsize < 5 ) {
            continue;
        }

        vector<CCAtom*> ringatoms = ring->GetAtoms();
        vector<CCBond*> ringbonds = ring->GetBonds();
        // count the number of pi electrons
        int sume = 0;
        vector<int> electron;
        for ( int i = 0; i < ringsize; ++i ) {
            electron.push_back( 1 );
        }

        // even membered rings
        if ( ringsize % 2 == 0 ) {
            bool single = false;
            for ( int i = 0; i < ringsize; ++i ) {
                // test the length of ring bond, if the ring contains
                // two bonds whose length are larger than single bond,
                // the ring will not be considered as aromatic
                if ( !ringbonds[i]->IsLengthInSB() ) {
                    if ( single ) {
                        ring->SetFlag( RING_IS_NORMAL );
                        break;
                    }
                    single = true;
                }

                // More strict aromatic criterion to make sure at least one
                // of each adjacent bonds should contain a double bond
                //if( !fbond->IsLengthInDB() &&
                //    !bbond->IsLengthInDB() )
                //{
                //  ring->SetFlag(RING_IS_NORMAL);
                //  break;
                //}
            }
            if ( ring->HasFlag( RING_IS_NORMAL ) ) {
                continue;
            }

            for ( int i = 0; i < ringsize; ++i ) {
                ratom = ringatoms[i];
                fbond = ringbonds[i];
                bbond = ringbonds[( i + ringsize - 1 ) % ringsize];

                if ( ratom->IsSp2Carbon() ) {
                    if ( ratom->IsInFGroup() ) {
                        ring->SetFlag( RING_IS_NORMAL );
                        break;
                    }
                    else if ( ratom->GetFormalCharge() == -1 ) {
                        electron[i] = 2;
                    }
                    else if ( ratom->GetFormalCharge() == 1 ) {
                        electron[i] = 0;
                    }
                    sume += electron[i];
                    continue;
                }
                else if ( ratom->IsSp2Nitrogen() && ratom->GetValence() == 3 ) {
                    if ( ratom->GetType() == "N.NOx" ) {
                        sume += electron[i];
                        continue;
                    }
                    else if ( fbond->IsLengthInDB() && bbond->IsLengthInDB() ) {
                        sume += electron[i];
                        continue;
                    }
                    // three special cases that nitrogen with three links
                    // in six membered ring is oxidative.
                    else if ( ratom->GetResName() == "NAD"
                           || ratom->GetResName() == "NAL"
                           || ratom->GetResName() == "NAP" ) {
                        sume += electron[i];
                        continue;
                    }
                    else {
                        electron[i] = 2;
                        sume += electron[i];
                        continue;
                    }
                }
                else if ( ratom->IsSp2Nitrogen() ) {
                    sume += electron[i];
                    continue;
                }
                else if ( ratom->GetType() == "S.SO" ) {
                    sume += electron[i];
                    continue;
                }
                else {
                    ring->SetFlag( RING_IS_NORMAL );
                    break;
                }
            }

            if ( ring->HasFlag( RING_IS_NORMAL ) ) {
                continue;
            }

            if ( sume % 4 == 2 ) {
                ring->SetFlag( RING_IS_AROMATIC );
                for ( int i = 0; i < ringsize; ++i ) {
                    ratom = ringatoms[i];
                    if ( ratom->IsCarbon() ) {
                        ratom->SetType( "C.ar" );
                    }
                    else if ( ratom->IsNitrogen() ) {
                        ratom->SetType( "N.ar" );
                    }
                    else if ( ratom->IsSulfur() ) {
                        ;}
                    else {
                        ratom->SetTypeFromSymbol();
                    }
                    ratom->SetAromatic();
                    ratom->SetValid();

                    bond = ringbonds[i];
                    bond->SetType( "AR" );
                    bond->SetOrder( 5 );
                    bond->SetAromatic();
                    bond->SetValid();
                }
            }
            else {
                ring->SetFlag( RING_IS_NORMAL );
            }
        }

        // odd ring only if ODD_RING_AROMATICITY was set
        else {
            bool single = false;
            for ( int i = 0; i < ringsize; ++i ) {
                // test the length of ring bond, if the ring contains
                // two bonds whose length are larger than single bond,
                // the ring will not be considered as aromatic
                if ( ringatoms[i]->IsOneOf( "O,S,Se,Te" )
                  || ringatoms[( i + 1 ) % ringsize]->IsOneOf( "O,S,Se,Te" ) ) {
                    continue;
                }
                else if ( !ringbonds[i]->IsLengthInSB() ) {
                    if ( single ) {
                        ring->SetFlag( RING_IS_NORMAL );
                        break;
                    }
                    single = true;
                }

                // More strict aromatic criterion to make sure at least one
                // of each adjacent bonds should contain a double bond
                //if( !fbond->IsLengthInDB() &&
                //    !bbond->IsLengthInDB() )
                //{
                //  ring->SetFlag(RING_IS_NORMAL);
                //  break;
                //}
            }
            if ( ring->HasFlag( RING_IS_NORMAL ) ) {
                continue;
            }

            for ( int i = 0; i < ringsize; ++i ) {
                ratom = ringatoms[i];

                if ( ratom->IsSp2Carbon() ) {
                    if ( ratom->IsInFGroup() ) {
                        ring->SetFlag( RING_IS_NORMAL );
                        break;
                    }
                    else if ( ratom->GetFormalCharge() == -1 ) {
                        electron[i] = 2;
                    }
                    else if ( ratom->GetFormalCharge() == 1 ) {
                        electron[i] = 0;
                    }
                    sume += electron[i];
                    continue;
                }
                else if ( ratom->IsSp2Nitrogen() && ratom->GetValence() == 3 ) {
                    if ( ratom->GetFormalCharge() != 1 ) {
                        electron[i] = 2;
                    }
                    sume += electron[i];
                    continue;
                }
                else if ( ratom->IsSp2Nitrogen() && ratom->GetValence() == 2 ) {
                    sume += electron[i];
                    continue;
                }
                else if ( ratom->IsOneOf( "O,S,Se,Te" ) ) {
                    if ( ratom->IsSulfur() && ratom->GetHyb() == 3 ) {
                        ring->SetFlag( RING_IS_NORMAL );
                        break;
                    }
                    electron[i] = 2;
                    sume += electron[i];
                    continue;
                }
                else {
                    ring->SetFlag( RING_IS_NORMAL );
                    break;
                }
            }

            if ( ring->HasFlag( RING_IS_NORMAL ) ) {
                continue;
            }

            int minde;
            for ( int i = 1; 1; ++i ) {
                minde = 4 * i + 2 - sume;
                if ( minde >= 0 ) {
                    break;
                }
            }

            while ( minde != 0 ) {
                int orden, bestatom;
                int bestorden = 99;
                for ( int i = 0; i < ringsize; ++i ) {
                    if ( electron[i] == 1 ) {
                        orden = getorden( ringatoms[i] );
                        if ( orden < bestorden ) {
                            bestorden = orden;
                            bestatom = i;
                        }
                    }
                }
                if ( bestorden == 99 ) {
                    break;      // 4n+2 rule is not satisfied.
                }
                else {
					ratom = ringatoms[bestatom];
					if ( ratom->IsNitrogen() && 
						 ratom->GetValence() == 2 &&
						 ratom->GetFormalCharge() == 0 ) {
						CCAtom *ratom(NULL);
						// whether the molecule has explicit hydrogen atom
						if ( HasExplicitH() ) {
							// make the atom not be treated in next cycle
							electron[bestatom] += 1;
							continue;
						}
					}
                    electron[bestatom] += 1;
                    minde--;
                }
            }

            if ( minde != 0 ) {
                ring->SetFlag( RING_IS_NORMAL );
                continue;
            }

            // aromatic ring
            else {
                ring->SetFlag( RING_IS_AROMATIC );
                for ( int i = 0; i < ringsize; ++i ) {
                    ratom = ringatoms[i];
                    if ( ratom->IsCarbon() ) {
                        ratom->SetType( "C.ar" );
                    }
                    else if ( ratom->IsNitrogen() ) {
                        if ( electron[i] == 2 ) {
                            ratom->SetType( "N.pl3" );
                        }
                        else {
                            ratom->SetType( "N.ar" );
                        }
                    }
                    else if ( ratom->GetType() != "S.SO" ) {
                        ratom->SetSp3Type();
                    }
                    ratom->SetAromatic();
                    ratom->SetValid();

                    bond = ringbonds[i];
                    bond->SetType( "AR" );
                    bond->SetOrder( 5 );
                    bond->SetAromatic();
                    bond->SetValid();
                }
            }
        }
    }

    UnsetFlag( MOL_HYBRID_PERCEIVED );
    UnsetFlag( MOL_INT_TYPE_PERCEIVED );
    UnsetFlag( MOL_AUTOFCHARGE_PERCEIVED );
}

void CCMol::PerceiveBondOrderFromXYZ()
{
    if ( HasFlag( MOL_BONDORDER_PERCEIVED ) ) {
        return;
    }
    SetFlag( MOL_BONDORDER_PERCEIVED );

    if ( !HasFlag( MOL_AROMATICITY_PERCEIVED ) ) {
        PerceiveAromaticityFromXYZ();
    }

    SetFlag( MOL_HYBRID_PERCEIVED );
    SetFlag( MOL_INT_TYPE_PERCEIVED );

    CCAtom* bgn, * end;
    CCBitVec bvatom( GetNumAtom() + 1 );
    vector<CCBond*> potendbs; // potential double bond

    FOR_ATOMS_OF_MOL( atom, this ) {
        if ( atom->IsValid() ) {
            bvatom.Set( atom->GetIdx() );
        }
    }

    FOR_BONDS_OF_MOL( bond, this ) {
        if ( bond->IsValid() ) {
            continue;
        }

        bool singlebond = false;
        bgn = bond->GetBeginAtom();
        end = bond->GetEndAtom();
        bool planarbond = bond->IsNearPlanar();

        if ( !planarbond ) {
            if ( bgn->GetValence() > 2 && end->GetValence() > 2 ) {
                bond->SetOrder( 1 );
                bond->SetType( "SB" );
                bond->SetValid();
                continue;
            }
        }

        if ( bvatom[bgn->GetIdx()] || bvatom[end->GetIdx()] ) {
            singlebond = true;
        }
        else if ( bgn->GetValence() > 3 || end->GetValence() > 3 ) {
            singlebond = true;
        }
        else if ( ( bgn->IsNitrogen() && bgn->GetValence() == 3 )
               || ( end->IsNitrogen() && end->GetValence() == 3 ) ) {
                 singlebond = true;
        }
        else if ( ( bgn->GetHyb() == 3 && end->GetHyb() == 3 )
               || ( bgn->GetHyb() == 3 && end->GetHyb() == 0 )
               || ( bgn->GetHyb() == 0 && end->GetHyb() == 3 ) ) {
                 singlebond = true;
        }

        if ( singlebond ) {
            bond->SetOrder( 1 );
            bond->SetType( "SB" );
            bond->SetValid();
            continue;
        }

        if ( bond->IsLengthInTB() ) {
            // left for the following test
            if ( bgn->IsOxygen()
              || bgn->IsSulfur()
              || bgn->GetHyb() > 1
              || end->IsOxygen()
              || end->IsSulfur()
              || end->GetHyb() > 1 ) {
                ;}
            else {
                bond->SetOrder( 3 );
                bond->SetType( "TB" );
                bond->SetValid();
                bvatom.Set( bgn->GetIdx() );
                bvatom.Set( end->GetIdx() );
                continue;
            }
        }
        if ( bond->IsLengthInDB() ) {
            // left for the following test
            if ( bgn->IsSp3Oxygen() || end->IsSp3Oxygen() ) {
                ;}
            else {
                //bond->SetOrder( 2 );
                //bond->SetType( "DB" );
                //bond->SetValid();
                potendbs.push_back( &*bond );
                //bvbond.Set( bond->GetIdx() );
                continue;
            }
        }

        // bond is near planar and its length is between
        // single and double bond, but close to double bond.
        // Length - ( STLength - 0.11 ) < STLength - Length
        if ( bond->IsNearPlanar( 5.0 )
          && bond->GetLength() < bond->GetSTLength() - 0.055 ) {
            // one case: 6upj_ligand
            // left for set single bond
            if ( bgn->IsSp3Oxygen()
              || end->IsSp3Oxygen()
              || ( bgn->IsNitrogen() && bgn->GetValence() > 2 )
              || ( end->IsNitrogen() && end->GetValence() > 2 ) ) {
                ;}
            else {
                //bond->SetOrder( 2 );
                //bond->SetType( "DB" );
                //bond->SetValid();
                potendbs.push_back( &*bond );
                //bvbond.Set( bond->GetIdx() );
                continue;
            }
        }

        bond->SetOrder( 1 );
        bond->SetType( "SB" );
        bond->SetValid();
        continue;
    }

    // the potential double bonds were first sorted by whether they
    // were in a conjugated system. Then sort by the bond length
    // deviated from the standard double bond length.
    // the bond which was in a conjugated system and bond length is
    // close to the double bond would be set as double bond.
    // e.g.  1a4m_ligand
    sort( potendbs.begin(), potendbs.end(), DoubleBondDev );
    // determine where to assign double bonds
    CCBond* bond;
    vector<CCBond*>::iterator vi = potendbs.begin();
    for ( ; vi != potendbs.end(); ++vi ) {
        bond = *vi;
        bgn = bond->GetBeginAtom();
        end = bond->GetEndAtom();
        if ( bvatom[bgn->GetIdx()] || bvatom[end->GetIdx()] ) {
            bond->SetOrder( 1 );
            bond->SetType( "SB" );
            bond->SetValid();
        }
        else {
            bond->SetOrder( 2 );
            bond->SetType( "DB" );
            bond->SetValid();
            bvatom.Set( bgn->GetIdx() );
            bvatom.Set( end->GetIdx() );
        }
    }

    UnsetFlag( MOL_HYBRID_PERCEIVED );
    UnsetFlag( MOL_INT_TYPE_PERCEIVED );
}

void CCMol::PerceiveAtomTypesFromXYZ()
{
    if ( HasFlag( MOL_INT_TYPE_PERCEIVED ) ) {
        return;
    }
    SetFlag( MOL_INT_TYPE_PERCEIVED );

    SetFlag( MOL_HYBRID_PERCEIVED );

    if ( !HasFlag( MOL_BONDORDER_PERCEIVED ) ) {
        PerceiveBondOrderFromXYZ();
    }

    // set atom types form hybridization state
    FOR_ATOMS_OF_MOL( atom, this ) {
        if ( atom->IsValid() ) {
            continue;
        }
        if ( atom->GetType() != "" &&
			 atom->GetType() != atom->GetSymbol() ) {
            continue;
        }

        atom->SetValid();
        int numDB = 0, numTB = 0;

        FOR_BONDS_OF_ATOM( bond, &*atom ) {
            if ( bond->GetOrder() == 2 ) {
                numDB++;
            }
            else if ( bond->GetOrder() == 3 ) {
                numTB++;
            }
        }

        if ( atom->IsCarbon() || atom->IsNitrogen() || atom->IsOxygen() ) {
            if ( numTB != 0 || numDB == 2 ) {
                atom->SetType( atom->GetSymbol() + ".1" );
                atom->SetHyb( 1 );
            }
            else if ( numDB == 1 ) {
                atom->SetType( atom->GetSymbol() + ".2" );
                atom->SetHyb( 2 );
            }
            else {
                atom->SetType( atom->GetSymbol() + ".3" );
                atom->SetHyb( 3 );

                if ( atom->IsNitrogen() ) {
                    FOR_NBORS_OF_ATOM( nbr, &*atom ) {
                        if ( nbr->HasBondWithLargerOrder( 1 ) ) {
                            atom->SetType( "N.pl3" );
                            atom->SetHyb( 2 );
                            break;
                        }
                    }
                }
            }
        }
        else if ( atom->IsPhosphorus() ) {
            atom->SetType( "P.3" );
            if ( numDB == 1 && atom->GetValence() < 4 ) {
                atom->SetHyb( 2 );
            }
            else {
                atom->SetHyb( 3 );
            }
        }
        else if ( atom->IsSulfur() ) {
            if ( numDB == 1 ) {
                atom->SetType( "S.2" );
                atom->SetHyb( 2 );
            }
            else {
                atom->SetType( "S.3" );
                atom->SetHyb( 3 );
            }
        }
        else {
            atom->SetTypeFromSymbol();
        }
    }
}


//////////////////////////////////////////////////////////////////////////
void CCMol::SetAutoFormalCharge( bool val )
{
    if ( val ) {
        SetFlag( MOL_AUTOFCHARGE_TRUE );
    }
    else {
        UnsetFlag( MOL_AUTOFCHARGE_TRUE );
    }
}

void CCMol::SetAutoPartialCharge( bool val )
{
    if ( val ) {
        SetFlag( MOL_AUTOPCHARGE_TRUE );
    }
    else {
        UnsetFlag( MOL_AUTOPCHARGE_TRUE );
    }
}

int CCMol::GetNumSubs()
{
    if ( !HasFlag( MOL_FRAGMENT_PERCEIVED ) ) {
        PerceiveFragment();
    }

    return m_nsubs;
}

double CCMol::GetMW() const
{
    double mw = 0.0;
    FOR_ATOMS_OF_MOL( atom, ( CCMol * )this )
    mw += atom->GetMass();
    return mw;
}

string CCMol::GetFormula()
{
	// if molecular formula exists, just get it.
	// Otherwise perceive it.
	CCGenericData* gd( NULL );
	gd = GetData( "FORMULA" );
	if ( gd != NULL ) {
		return ( dynamic_cast<CCPairString*>( gd )->GetValue() );
	}

    map<string, int> fm; map<string, int>::iterator it;
    FOR_ATOMS_OF_MOL( atom, ( CCMol * )this ) {
        if ( atom->IsHydrogen() ) {
            continue;
        }

        fm[atom->GetSymbol()]++;
        fm["H"] += atom->ExpHydrogenCount() + atom->ImpHydrogenCount();
    }

    stringstream formula;
    // formated formula: C,H,O,N,S,P,F,Cl,Br,I,A
    if ( fm.find( "C" ) != fm.end() ) {
        formula << "C" << fm["C"];
        fm["C"] = -1;
    }
    if ( fm.find( "H" ) != fm.end() ) {
        formula << "H" << fm["H"];
        fm["H"] = -1;
    }
    if ( fm.find( "O" ) != fm.end() ) {
        formula << "O" << fm["O"];
        fm["O"] = -1;
    }
    if ( fm.find( "N" ) != fm.end() ) {
        formula << "N" << fm["N"];
        fm["N"] = -1;
    }
    if ( fm.find( "S" ) != fm.end() ) {
        formula << "S" << fm["S"];
        fm["S"] = -1;
    }
    if ( fm.find( "P" ) != fm.end() ) {
        formula << "P" << fm["P"];
        fm["P"] = -1;
    }
    if ( fm.find( "F" ) != fm.end() ) {
        formula << "F" << fm["F"];
        fm["F"] = -1;
    }
    if ( fm.find( "Cl" ) != fm.end() ) {
        formula << "Cl" << fm["Cl"];
        fm["Cl"] = -1;
    }
    if ( fm.find( "Br" ) != fm.end() ) {
        formula << "Br" << fm["Br"];
        fm["Br"] = -1;
    }
    if ( fm.find( "I" ) != fm.end() ) {
        formula << "I" << fm["I"];
        fm["I"] = -1;
    }

    for ( it = fm.begin(); it != fm.end(); ++it ) {
        if ( it->second != -1 ) {
            formula << it->first << it->second;
        }
    }

	// add formula string to molecular data
	gd = new CCPairString( "FORMULA", formula.str() );
	AssertMemAlloc( gd );
	SetData( gd );

    return formula.str();
}

/*
string CCMol::GetFormula() const
{
    map<string, int> fm; map<string, int>::iterator it;
    FOR_ATOMS_OF_MOL( atom, ( CCMol * )this ) {
        if ( atom->IsHydrogen() ) {
            continue;
        }

        fm[atom->GetSymbol()]++;
        fm["H"] += atom->ExpHydrogenCount() + atom->ImpHydrogenCount();
    }

    stringstream formula;
    // C goes first, H second
    if ( fm.find( "C" ) != fm.end() ) {
        formula << "C" << fm["C"];
    }
    if ( fm.find( "H" ) != fm.end() ) {
        formula << "H" << fm["H"];
    }

    for ( it = fm.begin(); it != fm.end(); ++it ) {
        if ( it->first != "C" && it->first != "H" ) {
            formula << it->first << it->second;
        }
    }

    return formula.str();
}*/


vector<CCRing*>& CCMol::GetSSSR()
{
    if ( !HasFlag( MOL_SSSR_PERCEIVED ) ) {
        BFSFindSSSR();
    }

    return m_vring;
}

vector<CCRing*>& CCMol::GetAllRings()
{
    if ( !HasFlag( MOL_ALL_RING_PERCEIVED ) ) {
        DFSFindAllRings();
    }

    return m_vring;
}

// !------------- deprecated, use AssignIntType() instead
bool ComparePP( pair<CCAtom*, CCBond*>p1, pair<CCAtom*, CCBond*>p2 )
{
    if ( p1.first->GetHyb() < p2.first->GetHyb() ) {
        return true;
    }
    else if ( p1.first->GetHyb() == p2.first->GetHyb() ) {
        return p1.first->GetAtomicNum() < p2.first->GetAtomicNum();
    }
    else {
        return false;
    }
}

void CCMol::PerceiveInternalAtomBondTypes()
{
    if ( HasFlag( MOL_INT_TYPE_PERCEIVED ) ) {
        return;
    }
    SetFlag( MOL_INT_TYPE_PERCEIVED );

    if ( HasFlag( MOL_CONNECTIVITY_PERCEIVED ) )  // MDL or MOL2 or SMI format
    {
        // atom types are initialized from elemental type
        FOR_ATOMS_OF_MOL( atom, this ) {
            atom->SetTypeFromSymbol();
        }

        // bond types are initialized according to bond order
        FOR_BONDS_OF_MOL( bond, this ) {
            bond->SetTypeFromBO();
        }

        // fix complicated functional groups
        gIntTyper.AssignIntType( *this );
    }
    else {
        // PDB format
        ClearFlag(); // clear all flags
        HandlePDB();
        // already assign internal atom and bond types
        FOR_BONDS_OF_MOL( bond, this ) {
            if ( bond->GetType() != "" ) {
                continue;
            }
            else {
                bond->SetTypeFromBO();
            }
        }
    }
}

void CCMol::PerceiveImplicitValence()
{
    if ( HasFlag( MOL_IMPVALENCE_PERCEIVED ) ) {
        return;
    }
    SetFlag( MOL_IMPVALENCE_PERCEIVED );

    FOR_ATOMS_OF_MOL( atom, this ) {
        // initialized as exp. valence
        atom->SetImpVal( atom->GetValence() );
        switch ( atom->GetAtomicNum() ) {
        case  1:
            // H
        case  3:
            // Li
        case 11:
            // Na
        case 19:
            // K
        case  9:
            // F
        case 17:
            // Cl
        case 35:
            // Br
        case 53:
            // I
            if ( atom->GetFormalCharge() == 1 ||    // H+, Li+, Na+, K+
                atom->GetFormalCharge() == -1 ) {
                // F-, Cl-, Br-, I-
                atom->SetImpVal( 0 );
            }
            else {
                atom->SetImpVal( 1 );
            }
            break;
        case  5:
            // B
            atom->SetImpVal( 3 );
            break;
        case  6:
            // C
            if ( atom->GetHyb() == 3 ) {
                if ( atom->GetFormalCharge() == 1
                  || atom->GetFormalCharge() == -1 ) {
                    // C+ or C-
                    atom->SetImpVal( 3 );
                }
                else {
                    atom->SetImpVal( 4 );
                }
            }
            else if ( atom->GetHyb() == 2 ) {
                if ( atom->GetFormalCharge() == -1 ) {
                    atom->SetImpVal( 2 );
                }
                else {
                    atom->SetImpVal( 3 );
                }
            }
            else if ( atom->GetHyb() == 1 ) {
                if ( atom->GetFormalCharge() == -1 ) {
                    // --N+##C-
                    atom->SetImpVal( 1 );
                }
                else {
                    atom->SetImpVal( 2 );
                }
            }
            else {
                atom->SetImpVal( 0 );
            }
            break;
        case 7:
            // N
            if ( atom->GetHyb() == 3 ) {
                if ( atom->GetFormalCharge() == 1 ) {
                    atom->SetImpVal( 4 );
                }
                else if ( atom->GetFormalCharge() == -1 ) {
                    atom->SetImpVal( 2 );
                }
                else if ( atom->GetValence() == 4 ) {
                    // NX4
                    atom->SetImpVal( 4 );
                }
                else {
                    atom->SetImpVal( 3 );
                }
            }
            else if ( atom->GetHyb() == 2 ) {
                if ( atom->GetFormalCharge() == 1 || atom->GetValence() == 3 ) {
                    // >[N+]-[O-]
                    atom->SetImpVal( 3 );
                }
                else if ( !atom->HasBondWithLargerOrder( 1 ) ) {
                    // for N-Pi
                    atom->SetImpVal( 3 );
                }
                else if ( atom->HasBondWithOrder( 5 ) && !atom->IsInRing() ) {
                    // like in guandinium
                    atom->SetImpVal( 3 );
                }
                else {
                    atom->SetImpVal( 2 );
                }
                // problem exists for aromatic nH with implicit H
            }
            else if ( atom->GetHyb() == 1 ) {
                if ( atom->GetValence() > 1 ) {
                    // -N=[N]=N
                    atom->SetImpVal( 2 );
                }
                else {
                    atom->SetImpVal( 1 );
                }
            }
            break;
        case 8:
            // O
            if ( atom->GetHyb() == 3 ) {
                if ( atom->GetFormalCharge() == 1 ) {
                    // >[O+]-
                    atom->SetImpVal( 3 );
                }
                else if ( atom->GetFormalCharge() == -1 ) {
                    // [O-]
                    atom->SetImpVal( 1 );
                }
                else {
                    atom->SetImpVal( 2 );
                }
            }
            else if ( atom->GetHyb() == 2 ) {
                if ( atom->GetFormalCharge() == 1 ) {
                    // [=O+] in aromatic ring
                    atom->SetImpVal( 2 );
                }
                else {
                    atom->SetImpVal( 1 );
                }
            }
            break;
        case 14:
            // Si
            if ( atom->GetHyb() == 3 ) {
                // >Si<
                atom->SetImpVal( 4 );
            }
            else if ( atom->GetHyb() == 1 ) {
                // O=Si=O
                atom->SetImpVal( 2 );
            }
            break;
        case 15:
            // P
            if ( atom->GetHyb() == 3 ) {
                if ( atom->GetValence() == 4 ) {
                    // >(P=O)-
                    atom->SetImpVal( 4 );
                }
                else if ( atom->HasBondWithOrder( 2 ) ) {
                    atom->SetImpVal( 4 );
                }
                else {
                    // PA3
                    atom->SetImpVal( 3 );
                }
            }
            else if ( atom->GetHyb() == 2 ) {
                atom->SetImpVal( 2 );
            }
            break;
        case 16:
            // S
            if ( atom->GetHyb() == 3 ) {
                if ( atom->GetFormalCharge() == 1 ) {
                    // >[S+]-
                    atom->SetImpVal( 3 );
                }
                else if ( atom->GetValence() >= 3 ) {
                    // >[S](=O)2-
                    atom->SetImpVal( 4 );
                }
                else {
                    // A-S-A
                    atom->SetImpVal( 2 );
                }
            }
            else if ( atom->GetHyb() == 2 ) {
                if ( atom->GetFormalCharge() == 1 ) {
                    // =[S+]-
                    atom->SetImpVal( 2 );
                }
                else if ( atom->GetValence() == 3 ) {
                    // >S=O
                    atom->SetImpVal( 3 );
                }
                else {
                    atom->SetImpVal( 1 );
                }
            }
            break;
        case 33:
            // As
            atom->SetImpVal( 3 );
            break;
        default:
            atom->SetImpVal( atom->GetValence() );
            break;
        }
    }

    // reset m_impval if GetImpVal() < GetValence()
    FOR_ATOMS_OF_MOL( atom, this ) {
        if ( atom->GetImpVal() < atom->GetValence() ) {
            atom->SetImpVal( atom->GetValence() );
        }
    }
}

void CCMol::PerceiveHybridization()
{
    if ( HasFlag( MOL_HYBRID_PERCEIVED ) ) {
        return;
    }
    SetFlag( MOL_HYBRID_PERCEIVED );


    FOR_ATOMS_OF_MOL( atom, this ) {
        int hyb = 0; // all set to 0 intially
        int valence = atom->GetValence();
        CCAtom* nbr = NULL;

        if ( valence > 3 ) {
            // vatom with >3 valence are set to sp3 eg. -(S(=O)2- >(P=O)<
            atom->SetHyb( 3 );
            continue;
        }

        if ( atom->HasBondWithOrder( 5 ) ) // Aromatic bond
        {
            // eg. bond in carboxylate may be marked as aromatic
            if ( atom->IsTerOxygen() || atom->IsTerSulfur() ) {
                hyb = ( atom->GetFormalCharge() == -1 ) ? 2 : 2; // O- or S-
            }
            else if ( ( atom->IsOxygen() || atom->IsSulfur() )
                   && valence == 2 ) {
                // i.e. aromatic O in 5-ring
                hyb = ( atom->GetFormalCharge() == 1 ) ? 2 : 3; // =O+ or =S+
            }
            else {
                hyb = 2;
            }
        }
        else if ( atom->HasBondWithOrder( 3 ) ) // Triple bond
        {
            hyb = 1;
        }
        else if ( atom->HasBondWithOrder( 2 ) ) // Double bond
        {
            if ( atom->CountBondWithOrder( 2 ) > 1 ) {
                // sp1 in *=[C, N]=*, sp2 in carboxylate group
                // sp3 in *-S(=O)(=O)-*
                hyb = ( valence == 2 ) ? 1 : ( atom->IsSulfur() ? 3 : 2 );
            }
            else {
                if ( atom->IsNitrogen() && valence == 1 ) // for nitrogen in *=N=(N)
                {
                    nbr = atom->GetNbrAtom( 1 );
                    hyb = ( nbr->IsNitrogen()
                         && nbr->GetValence() == 2
                         && nbr->SumBO() >= 4 )
                        ? 1
                        : 2;
                }
                else if ( atom->IsPhosphorus() && valence > 2 ) {
                    hyb = 3;
                }
                else {
                    hyb = 2;
                }
            }
        }
        else                              // Single bond
        {
            if ( atom->IsTerOxygen() ) {
                if ( atom->GetFormalCharge() == -1 ) {
                    hyb = 2;
                }
                else // for oxygen in nitro-oxide and nitro group
                {
                    nbr = atom->GetNbrAtom( 1 );
                    hyb = ( nbr->IsNitrogen()
                         && nbr->GetValence() == 3
                         && nbr->SumBO() > 3 )
                        ? 2
                        : 3;
                }
            }
            else if ( atom->IsNitrogen() && atom->HasAlphaBetaUnsat() )
                // for nitrogen in a conjugated system, sp2
            {
                hyb = 2;
            }
            else {
                hyb = 3;
            }
        }

        // set hyb
        atom->SetHyb( hyb );
    }
}

void CCMol::PerceiveAromaticity()
{
    if ( HasFlag( MOL_AROMATICITY_PERCEIVED ) ) {
        return;
    }
    SetFlag( MOL_AROMATICITY_PERCEIVED );

    gAromaticTyper.AssignAromaticity( *this );
}

void CCMol::PerceiveClosureBond()
{
    if ( HasFlag( MOL_CLOSURE_BOND_PERCEIVED ) ) {
        return;
    }
    SetFlag( MOL_CLOSURE_BOND_PERCEIVED );

    CCBitVec usedAtoms( GetNumAtom() + 1 ), usedBonds( GetNumBond() + 1 );
    CCAtom* cur = NULL;

    FOR_ATOMS_OF_MOL( atom, this ) {
        if ( !usedAtoms[atom->GetIdx()] ) {
            queue<CCAtom*> q;
            q.push( &*atom );
            usedAtoms.Set( atom->GetIdx() );

            while ( !q.empty() ) {
                cur = q.front();
                q.pop();

                FOR_NBORS_OF_ATOM( nbr, cur ) {
                    if ( !usedAtoms[nbr->GetIdx()] ) {
                        usedAtoms.Set( nbr->GetIdx() );
                        usedBonds.Set( ( (CCBond*)~nbr )->GetIdx() );
                        q.push( &*nbr );
                    }
                }
            }
        }

        if ( usedAtoms.Count() >= GetNumAtom() ) {
            break;
        } // all have been perceived previously
    }

    FOR_BONDS_OF_MOL( bond, this )
    if ( !usedBonds[bond->GetIdx()] ) {
        bond->SetFlag( BOND_IS_CLOSURE );
    }
}


/** 
* @brief perceive RECAP bond in molecule.
* @param strict whether use strict RECAP definition.
*
* RECAP is the abbreviation of Retrosynthetic Combinatorial 
* Analysis Procedure, which electronically fragments molecules
* based on chemical knowledge.<br>
* If the parameter strict is ture, the RECAP bond will be the
* same as defined in ref:
* Lewell, X. Q.; Judd, D. B.; Watson, S. P.; Hann, M. M. 
* RECAP-Retrosynthetic Combinatorial Analysis Procedure: A 
* Powerful New Technique for Identifying Privileged Molecular
* Fragments with Useful Applications in Combinatorial Chemistry. 
* J. Chem. Inf. Comput. Sci. 1998, 38, (3), 511-522.
* else the RECAP bond is defined as *-N, *-O, C(sp2)=C(sp2) and
* C/N(ar)-C/N(ar).
* 
*/
void CCMol::PerceiveRECAPBond( bool strict /*= false*/ )
{
	if ( HasFlag( MOL_RECAPBOND_PERCEIVED ) ) {
		return;
	}
	SetFlag( MOL_RECAPBOND_PERCEIVED );

	if ( strict ) {
		CCBitVec bv( GetNumBond() + 1 );
		FOR_ATOMS_OF_MOL( atm, this ) {
			if ( atm->HasBondWithLargerOrder( 1 ) ) {
				// C/N(ar)-C/N(ar) bond
				if ( atm->IsAromatic() ) {
					FOR_BONDS_OF_ATOM( bnd, &*atm ) {
						int idx = bnd->GetIdx();
						if ( !bv[idx] ) {
							bv.Set( idx );
							if ( !bnd->IsInRing() && 
								((CCAtom*)~bnd)->IsAromatic() &&
								 bnd->GetOrder() == 1 ) {
								bnd->SetFlag( BOND_IS_RECAP );
							}
						}
					}
				}
				// C=C bond
				else if ( atm->IsSp2Carbon() ) {
					FOR_BONDS_OF_ATOM( bnd, &*atm ) {
						int idx = bnd->GetIdx();
						if ( !bv[idx] ) {
							bv.Set( idx );
							if ( !bnd->IsInRing() && 
								((CCAtom*)~bnd)->IsSp2Carbon() &&
								 bnd->GetOrder() == 2 ) {
								bnd->SetFlag( BOND_IS_RECAP );
							}
						}
					}
				}
			}

			else {
				// nitrogen atom, check whether it has non-ring amide bond,
				// if so, only set the non-ring amide bond as RECAP bond.
				// else, set all the neighboring bonds as RECAP bonds.
				if ( atm->IsNitrogen() ) {
					bool amide = false;  // non-ring amide or sulfone amide bond
					FOR_BONDS_OF_ATOM( bnd, &*atm ) {
						int idx = bnd->GetIdx();
						if ( !bv[idx] ) {
							bv.Set( idx );
							if ( !bnd->IsInRing() ) {
								if ( bnd->IsAmide() || bnd->IsSulfoneAmide() ) {
									// non-ring amide bond exist
									amide = true;
									bnd->SetFlag( BOND_IS_RECAP );
								}
							}
						}
					}

					// no non-ring amide bond
					if ( !amide ) {
						FOR_BONDS_OF_ATOM( bnd, &*atm ) {
							// ignore N-H bond
							if ( bnd->GetBeginAtom()->IsHydrogen() ||
								 bnd->GetEndAtom()->IsHydrogen() ) {
								continue;
							}
							if ( !bnd->IsInRing() && bnd->GetOrder() == 1 ) {
								bnd->SetFlag( BOND_IS_RECAP );
							}
						}
					}
				}

				// oxygen atom, check whether it has non-ring ester bond,
				// if so, only set the non-ring ester bond as RECAP bond.
				// else, set all the neighboring bonds as RECAP bonds.
				else if ( atm->IsOxygen() ) {
					bool ester = false;  // non-ring amide bond
					FOR_BONDS_OF_ATOM( bnd, &*atm ) {
						int idx = bnd->GetIdx();
						if ( !bv[idx] ) {
							bv.Set( idx );
							if ( !bnd->IsInRing() && bnd->IsEster() ) {
								// non-ring ester bond exist
								ester = true;
								bnd->SetFlag( BOND_IS_RECAP );
							}
						}
					}

					// no non-ring ester bond
					if ( !ester ) {
						FOR_BONDS_OF_ATOM( bnd, &*atm ) {
							// ignore O-H bond
							if ( bnd->GetBeginAtom()->IsHydrogen() ||
								 bnd->GetEndAtom()->IsHydrogen() ) {
								continue;
							}
							if ( !bnd->IsInRing() && bnd->GetOrder() == 1 ) {
								bnd->SetFlag( BOND_IS_RECAP );
							}
						}
					}
				}
			}
		}
	}

	else {
		FOR_BONDS_OF_MOL( bnd, this ) {
			// ignore all of the ring bonds
			if ( bnd->IsInRing() ) {
				continue;
			}
			CCAtom *bgn = bnd->GetBeginAtom();
			CCAtom *end = bnd->GetEndAtom();
			if ( bgn->IsNitrogen() || end->IsNitrogen() ) {
				bnd->SetFlag( BOND_IS_RECAP );
			}
			else if ( bgn->IsSp3Oxygen() || end->IsSp3Oxygen() ) {
				bnd->SetFlag( BOND_IS_RECAP );
			}
			else if ( bgn->IsSp2Carbon() && end->IsSp2Carbon() &&
					  bnd->GetOrder() == 2 ) {
				bnd->SetFlag( BOND_IS_RECAP );
			}
			else if ( bgn->IsAromatic() && end->IsAromatic() ) {
				bnd->SetFlag( BOND_IS_RECAP );
			}
		}
	}
}

//////////////////////////////////////////////////////////////////////////
bool CCMol::HasNonZeroCoords()
{
    FOR_ATOMS_OF_MOL( atom, this ) {
		if ( Distance( atom->GetXYZ(), VZero ) < EPSILON ) {
			return true;
		}
	}

    return false;
}

bool CCMol::HasExplicitH()
{
	// if explicit hydrogen data exist, just get it.
	// Otherwise perceive it.
	CCGenericData* gd( NULL );
	gd = GetData( "EXPLICITH" );
	if ( gd != NULL ) {
		return ( dynamic_cast<CCPairString*>( gd )->GetValue() == "true" );
	}

	bool explicitH = false;
	string data = "false";
	FOR_ATOMS_OF_MOL( atom, this ) {
		if ( atom->IsHydrogen() ) {
			explicitH = true;
			data = "true";
			break;
		}
	}

	// add explicit hydrogen data to molecular data
	gd = new CCPairString( "EXPLICITH", data );
	AssertMemAlloc( gd );
	SetData( gd );

	return explicitH;
}

bool CCMol::Has2D()
{
    bool hasX( false ), hasY( false );
    FOR_ATOMS_OF_MOL( atom, ( CCMol * )this ) {
        if ( !hasX && !IsApprox( atom->GetX(), 0.0 ) ) {
            hasX = true;
        }
        if ( !hasY && !IsApprox( atom->GetY(), 0.0 ) ) {
            hasY = true;
        }

        if ( hasX && hasY ) {
            return true;
        }
    }

    return false;
}

bool CCMol::Has3D()
{
    bool hasX( false ), hasY( false ), hasZ( false );
    FOR_ATOMS_OF_MOL( atom, ( CCMol * )this ) {
        if ( !hasX && !IsApprox( atom->GetX(), 0.0 ) ) {
            hasX = true;
        }
        if ( !hasY && !IsApprox( atom->GetY(), 0.0 ) ) {
            hasY = true;
        }
        if ( !hasZ && !IsApprox( atom->GetZ(), 0.0 ) ) {
            hasZ = true;
        }

        if ( hasX && hasY && hasZ ) {
            return true;
        }
    }

    return false;
}

int CCMol::GetDimension()
{
    if ( m_dimension != 0 ) {
        ; // already has dimension info
    }
    else { // perceive dimension from coordinates
        SetDimensionFromXYZ();
    }
    
    return m_dimension;
}

void CCMol::SetDimensionFromXYZ()
{
    if ( m_dimension != 0 ) {
        ; // already has dimension info
    }
    else if ( Has3D() ) {
        m_dimension = 3;
    }
    else if ( Has2D() ) {
        m_dimension = 2;
    }
    else {
        m_dimension = 0;
    }
}

void CCMol::PerceiveRotatableBond()
{
    if ( HasFlag( MOL_ROTATABLEBOND_PERCEIVED ) ) {
        return;
    }
    SetFlag( MOL_ROTATABLEBOND_PERCEIVED );

    FOR_BONDS_OF_MOL( bond, this ) {
        if ( !bond->IsSingle() || bond->IsInRing() ) {
            continue;
        }

        CCAtom* bgn = bond->GetBeginAtom();
        CCAtom* end = bond->GetEndAtom();

        if ( bgn->IsTerAtom() || end->IsTerAtom() ) {
			// terminal bonds (leaf) are not rotatable 
            continue;
        }
        else if ( IsHydroxyl( bgn, 3 ) || IsHydroxyl( end, 3 ) ) {
			// exception for -OH, -SH, they are rotatable
            bond->SetFlag( BOND_IS_ROTATABLE );
        }
		else if ( IsATx( bgn ) || IsATx( end ) ) {
			// bonds connection with symmetric group such as 
			// -NH2 -NH3 -NO2 -CH3 -CX3 -PO3 -SO3 -CO2 
            continue;
        }
        else if ( IsTButyl( bgn ) || IsTButyl( end ) ) {
			// and also t-butyl group
            continue;
        }
        else if ( IsCyano( bgn ) == 'C' || IsCyano( end ) == 'C' ) {
			// and also cyano group are not rotatable
            continue;
        }
		else if ( bond->IsAmide() ) {
			// amide bonds are not rotatable
			continue;
		}
        else {
			// otherwise, bond is rotatable
            bond->SetFlag( BOND_IS_ROTATABLE );
        }
    }
}

//! perceive each atom's R/S chirality.
void CCMol::PerceiveTautomerARBond()
{
	if ( HasFlag( MOL_TAUTOMERARBOND_PERCEIVED ) ) {
		return;
	}
	SetFlag( MOL_TAUTOMERARBOND_PERCEIVED );

	CCAtom *atom1, *atom2;
	CCBond *bond;
	FOR_RINGS_OF_MOL( r, this ) {
		int size = r->GetSize();
		if ( r->IsAromatic() && size % 2 == 0 ) {
			vector<CCAtom*> ratoms = r->GetAtoms();
			for ( int i = 0; i < size; ++i ) {
				atom1 = ratoms[i];
				atom2 = ratoms[(i+1)%size];
				bond  = atom1->GetBond( atom2 );
				bond->SetFlag( BOND_IS_TAUTOMERAR );
			}
		}
	}
}

void CCMol::PerceiveHBDonorAcceptor()
{
    if ( HasFlag( MOL_HBDONORACCEPTOR_PERCEIVED ) ) {
        return;
    }
    SetFlag( MOL_HBDONORACCEPTOR_PERCEIVED );

    FOR_ATOMS_OF_MOL( atom, this ) {
        if ( atom->IsOxygen() && atom->HasH() ) {
            atom->SetFlag( ATOM_IS_HBDONOR );
            atom->SetFlag( ATOM_IS_HBACCEPTOR );
        }
        else if ( ( atom->IsNitrogen() /*&& !atom->HasBondWithLargerOrder(1)*/ )
               && atom->HasH() ) {
            atom->SetFlag( ATOM_IS_HBDONOR );
        }
        else if ( atom->IsSp2Oxygen() && !atom->IsInRing() ) {
            atom->SetFlag( ATOM_IS_HBACCEPTOR );
        }
        else if ( atom->IsSp3Oxygen() ) {
            atom->SetFlag( ATOM_IS_HBACCEPTOR );
        }
        else if ( atom->IsNitrogen()
               && atom->IsAromatic()
               && atom->GetValence() == 2 ) // like in pyridine
        {
                 atom->SetFlag( ATOM_IS_HBACCEPTOR );
        }
        else if ( IsCyano( &*atom ) == 'N' ) {
            atom->SetFlag( ATOM_IS_HBACCEPTOR );
        }    // -C#N
    }
}

void CCMol::UpdateRingBVPath()
{
    FOR_RINGS_OF_MOL( rng, *this ) {
        rng->UpdateBVPath();
    }
}

bool SortAtombyIdx( CCAtom* a1, CCAtom* a2 )
{
    return ( a1->GetIdx() < a2->GetIdx() );
}

// sort atoms by index
void CCMol::SortAtoms()
{
	std::sort( m_vatom.begin(), m_vatom.begin() + m_natom, SortAtombyIdx );
}

bool sortAtombyPDBOrder(CCAtom* a1, CCAtom* a2) 
{
	CCResidue* res1 = a1->GetResidue();
	CCResidue* res2 = a2->GetResidue();

	if (!res1) { return false; }
	if (!res2) { return false; }

	if (res1->GetChainID()<res2->GetChainID()) { return true; }
	if (res1->GetChainID()<res2->GetChainID()) { return false; }
	if (res1->GetResNum()<res2->GetResNum()) { return true; }
	if (res1->GetResNum()>res2->GetResNum()) { return false; }

	// for simplicity, hydrogen atoms are still in an unknown behavior
	static map<string, int> *order;
	if (!order) {
		order = new map<string,int>;
		AssertMemAlloc(order);
		order->insert(make_pair("N" ,0));
		order->insert(make_pair("CA" ,1));
		order->insert(make_pair("O" ,2));
		order->insert(make_pair("C" ,3));
		order->insert(make_pair("CB" ,4));
		order->insert(make_pair("SG" ,5));
		order->insert(make_pair("OG" ,6));
		order->insert(make_pair("CG" ,7));
		order->insert(make_pair("SD" ,8));
		order->insert(make_pair("OD" ,9));
		order->insert(make_pair("ND" ,10));
		order->insert(make_pair("CD" ,11));
		order->insert(make_pair("OE" ,12));
		order->insert(make_pair("NE" ,13));
		order->insert(make_pair("CE" ,14));
		order->insert(make_pair("NZ" ,15));
		order->insert(make_pair("CZ" ,16));
		order->insert(make_pair("OH" ,17));
		order->insert(make_pair("NH" ,18));
		order->insert(make_pair("CH" ,19));
		order->insert(make_pair("HA" ,20));
		order->insert(make_pair("HB" ,21));
		order->insert(make_pair("HG" ,22));
		order->insert(make_pair("HD" ,23));
		order->insert(make_pair("HE" ,24));
		order->insert(make_pair("HZ" ,25));
		order->insert(make_pair("HH" ,26));
		order->insert(make_pair("H" ,27));
	}

	string name1 = a1->GetName(); Trim(name1, "0123456789");
	string name2 = a2->GetName(); Trim(name2, "0123456789");
	if (order->find(name1)!=order->end() &&
		order->find(name2)!=order->end()) {
			return (*order)[name1] < (*order)[name2];
	}

	return false;
}

void CCMol::SortPDBAtoms()
{
	// sort atoms (H) by the order:
	// N, CA, C, O, Alpha, Beta, Gamma, Epsilon, Zeta, H(Eta), H
	// for simplicity, hydrogen atoms are still in an unknown behavior
	std::sort(m_vatom.begin(), m_vatom.begin()+m_natom, sortAtombyPDBOrder);
	int idx(1);
	FOR_ATOMS_OF_MOL(a, (CCMol*)this) {
		a->SetIdx(idx++);
	}
}

bool SortBondbyIdx( CCBond* b1, CCBond* b2 )
{
    return ( b1->GetIdx() < b2->GetIdx() );
}

bool SortBondbyAtomIdx( CCBond* b1, CCBond* b2 )
{
    if ( b1->GetBeginAtomIdx() < b2->GetBeginAtomIdx() ) {
        return true;
    }
    if ( b1->GetBeginAtomIdx() == b2->GetBeginAtomIdx()
      && b1->GetEndAtomIdx() < b2->GetEndAtomIdx() ) {
        return true;
    }
    return false;
}

// sort bonds by begin and end atoms
void CCMol::SortBonds(bool updateRingBVPath)
{
    FOR_BONDS_OF_MOL( bi, this ) {
        if ( bi->GetBeginAtomIdx() > bi->GetEndAtomIdx() ) {
            CCAtom* tmp = bi->GetBeginAtom();
            bi->SetBegin( bi->GetEndAtom() );
            bi->SetEnd( tmp );
        }
    }

	std::sort( m_vbond.begin(), m_vbond.begin() + m_nbond, SortBondbyAtomIdx );

    // reset bond index
	int idx(1);
    FOR_BONDS_OF_MOL( bi, this ) {
        bi->SetIdx( idx++ );
    }
	if (updateRingBVPath) {
		// update bond id in ring
		UpdateRingBVPath();
	}
}

//////////////////////////////////////////////////////////////////////////
CCVector CCMol::GetCentroid( bool includeHydrogen ) const
{
    CCVector centroid;
    int count( 0 );

    FOR_ATOMS_OF_MOL( atom, ( CCMol * )this ) {
        if ( !includeHydrogen && atom->IsHydrogen() ) {
            continue;
        }

        centroid += atom->GetXYZ();
        count++;
    }

    return centroid / double( count );
}

// remove an atom from connection
bool CCMol::DeleteAtom( CCAtom* atom )
{
    if ( !atom ) {
        return false;
    }

    // find bonds connected to this atom that is to be removed
    // and check whether it is necessary to modify the flags
    vector<CCBond*> vb;
    bool resetflag = false;
    bool resetringflag = false;
    FOR_BONDS_OF_ATOM( bond, *atom ) {
        vb.push_back( &*bond );
        if ( bond->GetType() != "SB" ) {
            resetflag = true;
        }
    }
    if ( atom->IsInRing() ) {
        resetringflag = true;
    }

    for ( int i = 0; i < vb.size(); ++i ) {
        DeleteBond( vb[i] );
    }

    // remove this atom from its parent residue
    if ( atom->GetResidue() ) {
        atom->GetResidue()->RemoveAtom( atom );
    }

    // remove this atom from connection table
    m_vatom.erase( m_vatom.begin() + ( atom->GetIdx() - 1 ) );
    m_natom--;

    // reset atom id and update connection table
    for ( int idx = 1; idx <= m_vatom.size(); ++idx ) {
        m_vatom[idx - 1]->SetIdx( idx );
    }

    // destroy this atom
    DestroyIt( atom );

    if ( resetflag ) {
        ClearFlag(); // unset all flags
        SetFlag( MOL_CONNECTIVITY_PERCEIVED );
    }
    else if ( resetringflag ) {
        UnsetFlag( MOL_SSSR_PERCEIVED );
        BFSFindSSSR();
    }

    return true;
}

bool CCMol::DeleteAtom( int idx )
{
    return DeleteAtom( GetAtom( idx ) );
}

bool CCMol::DeleteAtoms( vector<CCAtom*> atoms )
{
    if ( atoms.empty() ) {
        return false;
    }

    vector<CCAtom*>::iterator ai;
    vector<CCAtom*>::iterator delai;

    // first check whether it is necessary to modify the flags
    bool resetflag = false;
    bool resetringflag = false;
    for ( ai = atoms.begin(); ai != atoms.end(); ++ai ) {
        FOR_BONDS_OF_ATOM( bond, ( *ai ) ) {
            if ( bond->GetType() != "SB" ) {
                resetflag = true;
                break;
            }
        }
        if ( ( *ai )->IsInRing() ) {
            resetringflag = true;
        }
        if ( resetflag ) {
            break;
        }
    }

    sort( atoms.begin(), atoms.end(), SortAtombyIdx );

    vector<CCBond*> delvbs;
    for ( delai = atoms.begin(); delai != atoms.end(); ++delai ) {
        FOR_BONDS_OF_ATOM( bond, *delai ) {
            if ( find( delvbs.begin(), delvbs.end(), &*bond ) == delvbs.end() ) {
                delvbs.push_back( &*bond );
            }
        }
    }
    DeleteBonds( delvbs );

    delai = atoms.begin();
    for ( ai = m_vatom.begin(); ai != m_vatom.end(); ) {
        if ( ( *ai ) == ( *delai ) ) {
            // remove this atom from its parent residue
            if ( ( *ai )->GetResidue() ) {
                ( *ai )->GetResidue()->RemoveAtom( *ai );
            }

            // remove atoms from connection table
            ai = m_vatom.erase( ai );
            m_natom--;

            // destroy this atom
            DestroyIt( *delai );

            if ( ++delai == atoms.end() ) {
                break;
            }
            continue;
        }
        ++ai;
    }

    // reset atom id and update connection table
    for ( int idx = 1; idx <= m_natom; ++idx ) {
        m_vatom[idx - 1]->SetIdx( idx );
    }

    if ( resetflag ) {
        ClearFlag(); // unset all flags
        SetFlag( MOL_CONNECTIVITY_PERCEIVED );
    }
    else if ( resetringflag ) {
        UnsetFlag( MOL_SSSR_PERCEIVED );
        BFSFindSSSR();
    }

    return true;
}

bool CCMol::DeleteBond( CCBond* bond )
{
    if ( !bond || !bond->GetBeginAtom() || !bond->GetEndAtom() ) {
        return false;
    }

    // first check whether it is necessary to modify the flags
    bool resetflag = false;
    if ( bond->GetType() != "SB" ) {
        resetflag = true;
    }

	// check whether it is necessary to re-detect rings
	bool resetring = false;
	if ( bond->IsInRing() ) {
		resetring = true;
	}

    // remove this bond from its component
    bond->GetBeginAtom()->RemoveBond( bond );
    bond->GetEndAtom()->RemoveBond( bond );

    // remove this bond from connection table
    m_vbond.erase( m_vbond.begin() + ( bond->GetIdx() - 1 ) );
    m_nbond--;

    // reset bond id and update connection table
    for ( int idx = 1; idx <= m_nbond; ++idx ) {
        m_vbond[idx - 1]->SetIdx( idx );
    }

    // destroy this bond
    DestroyIt( bond );

    if ( resetflag ) {
        ClearFlag(); // unset all flags
        SetFlag( MOL_CONNECTIVITY_PERCEIVED );
    }
	if ( resetring ) {
		CCRingIter k = m_vring.begin();
		for ( ; k != m_vring.end(); ++k ) {
			if ( *k ) {
				delete * k; *k = NULL;
			}
		}
		m_vring.clear();
	}

    return true;
}

bool CCMol::DeleteBonds( vector<CCBond*> bonds )
{
    if ( bonds.empty() ) {
        return false;
    }

    Kekulize();
    vector<CCBond*>::iterator bi;
    vector<CCBond*>::iterator delbi;
    sort( bonds.begin(), bonds.end(), SortBondbyIdx );

    // first check whether it is necessary to modify the flags
    bool resetflag = false;
    for ( delbi = bonds.begin(); delbi != bonds.end(); ++delbi ) {
        if ( ( *delbi )->GetType() != "SB" ) {
            resetflag = true;
            break;
        }
    }

	// check whether it is necessary to re-detect rings
	bool resetring = false;
	for ( delbi = bonds.begin(); delbi != bonds.end(); ++delbi ) {
		if ( ( *delbi )->IsInRing() ) {
			resetring = true;
			break;
		}
	}

    // remove bonds from connection table
    delbi = bonds.begin();
    for ( bi = m_vbond.begin(); bi != m_vbond.end(); ) {
        if ( ( *bi ) == ( *delbi ) ) {
            // remove this bond from its component
            ( *bi )->GetBeginAtom()->RemoveBond( *bi );
            ( *bi )->GetEndAtom()->RemoveBond( *bi );

            bi = m_vbond.erase( bi );
            m_nbond--;

			// destroy this bond
            DestroyIt( *delbi );

            if ( ++delbi == bonds.end() ) {
                break;
            }
            continue;
        }
        ++bi;
    }


    // reset bond id and update connection table
    for ( int idx = 1; idx <= m_nbond; ++idx ) {
        m_vbond[idx - 1]->SetIdx( idx );
    }

    if ( resetflag ) {
        ClearFlag(); // unset all flags
        SetFlag( MOL_CONNECTIVITY_PERCEIVED );
    }
	if ( resetring ) {
		CCRingIter k = m_vring.begin();
		for ( ; k != m_vring.end(); ++k ) {
			if ( *k ) {
				delete * k; *k = NULL;
			}
		}
		m_vring.clear();
	}

    return true;
}

void CCMol::DeleteHydrogen( bool apolarHydrogenOnly/* =false */ )
{
    // find the hydrogen to be deleted
    vector<CCAtom*> vh;
    FOR_ATOMS_OF_MOL( atom, this ) {
        if ( apolarHydrogenOnly ) {
            if ( atom->IsApolarHydrogen() ) {
                vh.push_back( &*atom );
            }
        }
        else {
            if ( atom->IsHydrogen() ) {
                vh.push_back( &*atom );
            }
        }
    }

    DeleteAtoms( vh );
    //for ( CCAtomIter it = vh.begin(); it != vh.end(); ++it ) {
    //    DeleteAtom( *it );
    //}
}

int CCMol::AddHydrogen( bool polarHydrogenOnly /* =false */ )
{
    int numh = 0;

    for ( int i = 0; i < GetNumAtom(); ++i ) {
        CCAtom* atom = GetAtom( i + 1 );
        numh += atom->AddHydrogen( polarHydrogenOnly );
    }
    return numh;
}

int CCMol::AddBioHydrogen( bool polarHydrogenOnly /* = true */ )
{
	int numH(0), totalH(0);
	for (int i=0; i<GetNumAtom(); ++i) {
		CCAtom* atom=GetAtom(i+1);
		numH = atom->AddHydrogen(polarHydrogenOnly);
		totalH += numH;
		// standardize Hydrogen name
		if (numH<=0) { continue; }
		CCAtom* hyd(NULL); CCResidue* res(NULL);
		// get the pos. of parent atom, i.e. alpha, beta, galma
		string hName, posName = atom->GetName().substr(1);
		for (int j=0; j<numH; ++j) {
			// get the hydrogen
			hyd = atom->GetNbrAtom(atom->GetValence()-(numH-j)+1);
			if (hyd && hyd->IsHydrogen()) { // must be
				hName = numH==1 ? "H"+posName : toSTR(j+1)+"H"+posName;
				hyd->SetName(hName);
				res = hyd->GetResidue();
				if (res) {
					res->SetResAtomName(hyd, hName);
				}
			}
		}
	}

	// sort atoms according to PDB rule, this will gather the newly added
	// hydrogens to their parent residues, more human readable
	SortPDBAtoms();

	return totalH;
}

bool CCMol::ReadMolecule( const string& inFilename, unsigned long flag )
{
    CCFormat* fmt = FormatFromExt( inFilename );
    fmt->SetOption( flag );

    ifstream ifs( inFilename.c_str(),ios::binary );
    AssertOpenFile( ifs, inFilename );

    bool success = fmt->ReadMolecule( *this, ifs );
    ifs.close();
    return success;
}

bool CCMol::WriteMolecule( const string& outFilename )
{
    CCFormat* fmt = FormatFromExt( outFilename );

    ofstream ofs( outFilename.c_str(),ios::binary );
    AssertOpenFile( ofs, outFilename );

    bool success = fmt->WriteMolecule( *this, ofs );
    ofs.close();
    return success;
}

void CCMol::ShowConnectivity()
{
    FOR_ATOMS_OF_MOL( atom, *this ) {
        cout << atom->GetIdx() << '\t';
        FOR_NBORS_OF_ATOM( nbr, &*atom ) {
            cout << nbr->GetIdx() << '\t';
        }
        cout << endl;
    }
}

bool CompareBondID( CCBond* b1, CCBond* b2 )
{
    int bgn1id = b1->GetBeginAtomIdx();
    int end1id = b1->GetEndAtomIdx();
    int bgn2id = b2->GetBeginAtomIdx();
    int end2id = b2->GetEndAtomIdx();
    if ( bgn1id < bgn2id ) {
        return true;
    }
    if ( bgn1id > bgn2id ) {
        return false;
    }
    if ( end1id < end2id ) {
        return true;
    }
    return false;
}

/*
void CCMol::SortBondID()
{
    FOR_BONDS_OF_MOL( bond, *this ) {
        CCAtom* bgn = bond->GetBeginAtom();
        CCAtom* end = bond->GetEndAtom();
        if ( bgn->GetIdx() > end->GetIdx() ) {
            bond->SetBegin( end );
            bond->SetEnd( bgn );
        }
    }
    sort( m_vbond.begin(), m_vbond.end(), CompareBondID );
    int i = 1;
    FOR_BONDS_OF_MOL( bond, *this ) {
        bond->SetIdx( i++ );
    }
    UpdateRingBVPath();
}*/


// find all the atoms in angles and save them to mol data
void CCMol::FindAngles()
{
    //if already has data return
    if ( HasData( CCGenericDataType::AngleAtomData ) ) {
        return;
    }

    CCAtom* atom1, * atom2, * atom3;
    vector<vector<CCAtom*> > angles;
    vector<CCAtom*> angle;
    for ( int idx = 1; idx <= GetNumAtom(); ++idx ) {
        atom2 = GetAtom( idx );
        int numNbr = atom2->GetValence();
        if ( numNbr < 2 ) {
            continue;
        }

        for ( int i = 1; i < numNbr; ++i ) {
            atom1 = atom2->GetNbrAtom( i );
            for ( int j = i + 1; j <= numNbr; ++j ) {
                atom3 = atom2->GetNbrAtom( j );
                angle.clear();
                angle.push_back( atom1 );
                angle.push_back( atom2 );
                angle.push_back( atom3 );
                angles.push_back( angle );
            }
        }
    }

    // add angle atoms to molecular data
    CCGenericData* gd = new CCMultiAtomsData( angles, "ANGLES",
                                              CCGenericDataType::AngleAtomData );

    AssertMemAlloc( gd );
    SetData( gd );
}

// find all the atoms in torsions and save them to mol data
void CCMol::FindTorsions()
{
    //if already has data return
    if ( HasData( CCGenericDataType::TorsionAtomData ) ) {
        return;
    }

    CCBond* bond;
    CCAtom* atom1, * atom2, * atom3, * atom4;
    vector<CCAtom*> torsion;
    vector<vector<CCAtom*> > torsions;
    for ( int idx = 1; idx <= GetNumBond(); ++idx ) {
        bond = GetBond( idx );
        atom2 = bond->GetBeginAtom();
        atom3 = bond->GetEndAtom();

        int numNbr2 = atom2->GetValence();
        int numNbr3 = atom3->GetValence();

        if ( numNbr2 < 2 || numNbr3 < 2 ) {
            continue;
        }

        for ( int i = 1; i <= numNbr2; ++i ) {
            atom1 = atom2->GetNbrAtom( i );
            if ( atom1 == atom3 ) {
                continue;
            }

            for ( int j = 1; j <= numNbr3; ++j ) {
                atom4 = atom3->GetNbrAtom( j );
                if ( atom4 == atom2 ) {
                    continue;
                }
                torsion.clear();
                torsion.push_back( atom1 );
                torsion.push_back( atom2 );
                torsion.push_back( atom3 );
                torsion.push_back( atom4 );
                torsions.push_back( torsion );
            }
        }
    }

    // add torsion atoms to molecular data
    CCGenericData* gd = new CCMultiAtomsData( torsions, "TORSIONS",
                                              CCGenericDataType::AngleAtomData );

    AssertMemAlloc( gd );
    SetData( gd );
}

// find all pair of atoms( 1-N(N>4) ) and save them to mol data
void CCMol::FindPairs()
{
    //if already has data return
    if ( HasData( CCGenericDataType::PairAtomData ) ) {
        return;
    }

    CCAtom* atom1, * atom2, * atom3, * atom4;
    vector<CCAtom*> apair;
    vector<vector<CCAtom*> > apairs;
    bool not14;
    for ( int idx1 = 1; idx1 < GetNumAtom(); ++idx1 ) {
        atom1 = GetAtom( idx1 );
        for ( int idx2 = idx1 + 1; idx2 <= GetNumAtom(); ++idx2 ) {
            atom2 = GetAtom( idx2 );
            not14 = false;

            for ( int i = 1; !not14 && i <= atom1->GetValence(); ++i ) {
                atom3 = atom1->GetNbrAtom( i );
                if ( atom3 == atom2 ) {
                    // 1-2 connected
                    not14 = true;
                }
                for ( int j = 1; !not14 && j <= atom3->GetValence(); ++j ) {
                    atom4 = atom3->GetNbrAtom( j );
                    if ( atom4 == atom2 ) {
                        // 1-3 connected
                        not14 = true;
                    }
                    for ( int k = 1;
                          !not14 && k <= atom4->GetValence();
                          ++k ) {
                        if ( atom4->GetNbrAtom( k ) == atom2 ) {
                            // 1-4 connected
                            not14 = true;
                        }
                    }
                }
            }

            if ( !not14 ) {
                apair.clear();
                apair.push_back( atom1 );
                apair.push_back( atom2 );
                apairs.push_back( apair );
            }
        }
    }

    // add pair atoms to molecular data
    CCGenericData* gd = new CCMultiAtomsData( apairs, "PAIRS",
                                              CCGenericDataType::PairAtomData );

    AssertMemAlloc( gd );
    SetData( gd );
}


/************************************************************************/
/*                    Molecular related functions                       */
/************************************************************************/
bool SameFormula( CCMol& mol1, CCMol& mol2 )
{
    // compare the their formula
    return ( mol1.GetFormula() == mol2.GetFormula() );
}

bool SameFormula( CCMol* mol1, CCMol* mol2 ) {
	return SameFormula( *mol1, *mol2 );
}

bool HasHydrophobicRegion( CCMol& mol )
{
	// algorithm according to bond polarity
	CCBitVec ub( mol.GetNumBond()+1 );
	FOR_BONDS_OF_MOL( b, mol ) {
		if ( ub[b->GetIdx()]
		|| b->IsPolar()
			|| b->GetBeginAtom()->IsHydrogen()
			|| b->GetEndAtom()->IsHydrogen() ) {
				continue;
		}

		// collection
		queue<CCAtom*> q; CCAtom* root = b->GetBeginAtom();
		q.push( root );
		int cnt( 0 ); CCVector al;

		while ( !q.empty() ) {
			root = q.front();
			q.pop();
			bool nonpolar( true );
			FOR_BONDS_OF_ATOM( nb, root ) {
				if ( ub[nb->GetIdx()] ) {
					continue;
				}
				ub.Set( nb->GetIdx() );
				if ( nb->IsPolar() ) {
					nonpolar = false;
					continue;
				}
				if ( nb->GetBeginAtom()->IsHydrogen()
					|| nb->GetEndAtom()->IsHydrogen() ) {
						continue;
				}
				q.push( ( (CCAtom*)~nb ) );
			}
			if ( nonpolar ) {
				cnt++;
				al += root->GetXYZ();
			}
		}

		if ( cnt >= 3 ) {
			return true;
		}
	}

	return false;
}

/** 
* @brief check whether two molecules have collision.
* @param[in] m1 first molecule
* @param[in] m2 second moelcule
* @param[in] dist collision distance
* @return true if m1 has an heavy atom whose distance is less than
* COLLISIONDIST(A) to an heavy atom in m2.
*/
bool OuterCollisionCheck( CCMol &m1, CCMol &m2 ) {
	double dist = COLLISION_DIST * COLLISION_DIST;
	FOR_ATOMS_OF_MOL( a1, m1 ) {
		if ( a1->IsHydrogen() ) {
			continue;
		}
		FOR_ATOMS_OF_MOL( a2, m2 ) {
			if ( a2->IsHydrogen() ) {
				continue;
			}
			if ( Distance2( a1->GetXYZ(), a2->GetXYZ() ) < dist ) {
				return true;
			}
		}
	}
	return false;
}

/** 
* @brief check whether two molecules have collision.
* @param[in] mol molecule for checking collision
* @return true if there is two heavy atoms in mol whose distance
* is less than COLLISIONDIST(A).
*/
bool InnerCollisionCheck( CCMol &mol ) {
	double d = 0.0;
	double d0 = 0.0, d1 = 0.0, d2 = 0.0;
	int natom = mol.GetNumAtom();
	for ( int i = 1; i <= natom; ++i ) {
		CCAtom *patom1 = mol.GetAtom( i );
		d1 = gFFVDWTable.GetVDWRadius( patom1->GetType() );
		for ( int j = i + 1; j <= natom; ++j ) {
			CCAtom *patom2 = mol.GetAtom( j );
			d = Distance2( patom1->GetXYZ(), patom2->GetXYZ() );
			d2 = gFFVDWTable.GetVDWRadius( patom2->GetType() );
			d0 = d1 + d2;
			if ( d > d0 - VDW_BUMP_RANGE ) {
				continue;
			}
			// 1-2, 1-3, 1-4 collection
			if ( ShortestPathLength( patom1, patom2 ) < 4 ) {
				continue;
			}
			return true;
		}
	}
	return false;
}

/** 
* @brief delete small fragments and only keep the largest one.
* @param[in] mol molecule for handling
* @return the largest fragments
*
* @note this function will delete the small fragments in molecule
*/
void KeepLargestFragment( CCMol &mol ) {
	int numAtoms = mol.GetNumAtom();
	CCBitVec used( numAtoms + 1 );
	int head, part = 0;
	vector<CCAtom*> prefrags;
	vector<CCAtom*> delatoms;

	// group fragments
	int maxsize = 0;
	for ( int i = 1; i <= numAtoms && used.Count() != numAtoms; ++i ) {
		if ( !used[i] ) {
			queue<int> q;
			CCAtom* atom;

			q.push( i );
			used.Set( i );

			vector<CCAtom*> curfrag;

			while ( !q.empty() ) {
				head = q.front();
				q.pop();
				
				atom = mol.GetAtom( head );
				curfrag.push_back( atom );

				FOR_NBORS_OF_ATOM( nbr, atom ) {
					if ( !used[nbr->GetIdx()] ) {
						q.push( nbr->GetIdx() );
						used.Set( nbr->GetIdx() );
					}
				}
			}
			
			int fragsize = (int)curfrag.size();
			if ( fragsize > maxsize ) {
				maxsize = fragsize;
				delatoms.insert( delatoms.end(),
					             prefrags.begin(), prefrags.end() );
				prefrags.clear();
				prefrags = curfrag;
			}
			else {
				delatoms.insert( delatoms.end(),
					             curfrag.begin(), curfrag.end() );
			}
		}
	}
	mol.DeleteAtoms( delatoms );
}
