#include <ccbg/CCResidue.h>
#include <ccbg/CCMol.h>

///////////////////////////////////////////////////////////////////////////////
// Natural Amino Acid Definitions
///////////////////////////////////////////////////////////////////////////////

enum
{
    AA_ALA      = 1 << 1,
    AA_GLY      = 1 << 2,
    AA_LEU      = 1 << 3,
    AA_SER      = 1 << 4,
    AA_VAL      = 1 << 5,
    AA_THR      = 1 << 6,
    AA_LYS      = 1 << 7,
    AA_ASP      = 1 << 8,
    AA_ILE      = 1 << 9,
    AA_ASN      = 1 << 10,
    AA_GLU      = 1 << 11,
    AA_PRO      = 1 << 12,
    AA_ARG      = 1 << 13,
    AA_PHE      = 1 << 14,
    AA_GLN      = 1 << 15,
    AA_TYR      = 1 << 16,
    AA_HIS      = 1 << 17,
    AA_CYS      = 1 << 18,
    AA_MET      = 1 << 19,
    AA_TRP      = 1 << 20,
    AA_HOH      = 1 << 21,
};

/////////////////////////////////////////////////////////////////////////////
// Amino Acid Property Definitions
/////////////////////////////////////////////////////////////////////////////
#define IS_ACIDIC(x)      ((x) & ((AA_ASP)|(AA_GLU)))
#define IS_ACYCLIC(x)     ((x) & ((AA_ALA)|(AA_GLY)|(AA_LEU)|(AA_SER)|  \
                                 (AA_VAL)|(AA_THR)|(AA_LYS)|(AA_ASP)|  \
                                 (AA_ILE)|(AA_ASN)|(AA_GLU)|(AA_GLN)|  \
                                 (AA_CYS)|(AA_MET)))
#define IS_ALIPHATIC(x)   ((x) & ((AA_ALA)|(AA_GLY)|(AA_ILE)|(AA_LEU)|  \
                                 (AA_VAL)))
#define IS_AROMATIC(x)    ((x) & ((AA_HIS)|(AA_PHE)|(AA_TRP)|(AA_TYR)))
#define IS_BASIC(x)       ((x) & ((AA_ARG)|(AA_HIS)|(AA_LYS)))
#define IS_BURIED(x)      ((x) & ((AA_ALA)|(AA_CYS)|(AA_ILE)|(AA_LEU)|  \
                                 (AA_MET)|(AA_PHE)|(AA_TRP)|(AA_VAL)))
#define IS_CHARGED(x)     ((x) & ((AA_ASP)|(AA_GLU)|(AA_ARG)|(AA_HIS)|  \
                                 (AA_LYS)))
#define IS_CYCLIC(x)      ((x) & ((AA_HIS)|(AA_PHE)|(AA_PRO)|(AA_TRP)|  \
                                 (AA_TYR)))
#define IS_HYDROPHOBIC(x) ((x) & ((AA_ALA)|(AA_LEU)|(AA_VAL)|(AA_ILE)|  \
                                 (AA_PRO)|(AA_PHE)|(AA_MET)|(AA_TRP)))
#define IS_LARGE(x)       ((x) & ((AA_ARG)|(AA_PHE)|(AA_GLN)|(AA_TYR)|  \
                                 (AA_HIS)|(AA_LEU)|(AA_LYS)|(AA_ILE)|  \
                                 (AA_GLU)|(AA_MET)|(AA_TRP)))
#define IS_MEDIUM(x)      ((x) & ((AA_VAL)|(AA_THR)|(AA_ASP)|(AA_ASN)|  \
                                 (AA_PRO)|(AA_CYS)))
#define IS_NEGATIVE(x)    ((x) & ((AA_ASP)|(AA_GLU)))
#define IS_NEUTRAL(x)     ((x) & ((AA_ALA)|(AA_GLY)|(AA_LEU)|(AA_SER)|  \
                                 (AA_VAL)|(AA_THR)|(AA_PHE)|(AA_GLN)|  \
                                 (AA_TYR)|(AA_HIS)|(AA_CYS)|(AA_MET)|  \
                                 (AA_TRP)|(AA_ILE)|(AA_ASN)|(AA_PRO)))
#define IS_POLAR(x)       ((x) & ((AA_ASP)|(AA_ILE)|(AA_ASN)|(AA_GLU)|  \
                                 (AA_SER)|(AA_THR)|(AA_ARG)|(AA_GLN)|  \
                                 (AA_CYS)|(AA_HIS)))
#define IS_POSITIVE(x)    ((x) & ((AA_ARG)|(AA_HIS)|(AA_LYS)))
#define IS_SMALL(x)       ((x) & ((AA_ALA)|(AA_GLY)|(AA_SER)))
#define IS_SURFACE(x)     ((x) & ((AA_THR)|(AA_LYS)|(AA_ASP)|(AA_ILE)|  \
                                 (AA_ASN)|(AA_GLU)|(AA_PRO)|(AA_ARG)|  \
                                 (AA_GLY)|(AA_SER)|(AA_GLN)|(AA_TYR)|  \
                                 (AA_HIS)))

//! Residue property definitions
namespace AminoAcidProperty
{
enum
{
    ACIDIC              = 0,
    ACYCLIC             = 1,
    ALIPHATIC           = 2,
    AROMATIC            = 3,
    BASIC               = 4,
    BURIED              = 5,
    CHARGED             = 6,
    CYCLIC              = 7,
    HYDROPHOBIC         = 8,
    LARGE               = 9,
    MEDIUM              = 10,
    NEGATIVE            = 11,
    NEUTRAL             = 12,
    POLAR               = 13,
    POSITIVE            = 14,
    SMALL               = 15,
    SURFACE             = 16,
};
}

// Residue Index
namespace ResidueIndex
{
enum
{
    // begin Natural Amino Acid
    ALA                                 = 0,
    GLY                                 = 1,
    LEU                                 = 2,
    SER                                 = 3,
    VAL                                 = 4,
    THR                                 = 5,
    LYS                                 = 6,
    ASP                                 = 7,
    ILE                                 = 8,
    ASN                                 = 9,
    GLU                                 = 10,
    PRO                                 = 11,
    ARG                                 = 12,
    PHE                                 = 13,
    GLN                                 = 14,
    TYR                                 = 15,
    HIS                                 = 16,
    CYS                                 = 17,
    MET                                 = 18,
    TRP                                 = 19,
        // end Natural Amino Acid
    ASX                                 = 20,
    GLX                                 = 21,
    PCA                                 = 22,
    HYP                                 = 23,
    A                                   = 24,
    C                                   = 25,
    G                                   = 26,
    T                                   = 27,
    U                                   = 28,
    UPLUS                               = 29,
    I                                   = 30,
    _1MA                                = 31,
    _5MC                                = 32,
    OMC                                 = 33,
    _1MG                                = 34,
    _2MG                                = 35,
    M2G                                 = 36,
    _7MG                                = 37,
    OMG                                 = 38,
    YG                                  = 39,
    H2U                                 = 40,
    _5MU                                = 41,
    PSU                                 = 42,
    UNK                                 = 43,
    ACE                                 = 44,
    FOR                                 = 45,
    HOH                                 = 46,
    DOD                                 = 47,
    SO4                                 = 48,
    PO4                                 = 49,
    NAD                                 = 50,
    COA                                 = 51,
    NAP                                 = 52,
    NDP                                 = 53,
};
}

CCResidue::CCResidue()
    : m_mol( NULL ), m_chain( NULL ), m_idx( 0 ), m_resNum( 0 ), 
	//m_resName
	//m_hetatm
	//m_vatom
	//m_atmname
	//m_sernum
	m_altLoc( ' ' ), m_iCode(' ')
	//m_occupancy
	//m_tempFactor
{
    CCBase::CCBase();
}

CCResidue::~CCResidue()
{
    CCBase::DestroyGD();

    CCAtomIter i = m_vatom.begin();
    for ( ; i != m_vatom.end(); ++i ) {
        if ( *i ) {
            ( *i )->SetResidue( NULL );
        }
    }
    m_vatom.clear();
}

void CCResidue::Clear()
{
    CCBase::Clear();
    m_mol = NULL;
    m_chain = NULL;
    m_idx = 0;
    m_resNum = 0;
    m_resName = "";
    m_vatom.clear();
	m_hetatm.clear();
    m_atmName.clear();
    m_serial.clear();
    m_altLoc = ' ';
	m_iCode = ' ';
	m_occupancy.clear();
	m_tempFactor.clear();
	m_ss.clear();
}

CCResidue::CCResidue( const CCResidue& src )
    : m_mol( src.m_mol ), m_chain( src.m_chain ), m_idx( src.m_idx ),
      m_resNum( src.m_resNum ), m_resName( src.m_resName ),
      m_vatom( src.m_vatom ), m_hetatm( src.m_hetatm ),
      m_atmName( src.m_atmName ), m_serial( src.m_serial ),
      m_altLoc( src.m_altLoc ), m_iCode(src.m_iCode),
	  m_occupancy( src.m_occupancy ), m_tempFactor( src.m_tempFactor ),
	  m_ss( src.m_ss )
{
    CCBase::Clone( src );
}

CCResidue& CCResidue::operator =( const CCResidue& src )
{
    if ( this != &src ) {
        CCBase::Clone( src );
        m_mol = src.m_mol;
        m_chain = src.m_chain;
        m_idx = src.m_idx;
        m_resNum = src.m_resNum;
        m_resName = src.m_resName;
        m_vatom = src.m_vatom;
        m_hetatm = src.m_hetatm;
        m_atmName = src.m_atmName;
        m_serial = src.m_serial;
        m_altLoc = src.m_altLoc;
		m_iCode = src.m_iCode;
		m_occupancy = src.m_occupancy;
		m_tempFactor = src.m_tempFactor;
		m_ss = src.m_ss;
    }
    return *this;
}

char CCResidue::GetChainID() const
{
    return m_chain ? m_chain->GetChainID()
		: ' ';
}

string CCResidue::GetResAtomName( CCAtom* atom ) const
{
    for ( int i = 0 ; i < m_vatom.size() ; i++ ) {
        if ( m_vatom[i] == atom ) {
            return m_atmName[i];
        }
    }
    return "";
}

CCAtom* CCResidue::GetAtomOfName( string atom_name )
{
    for ( int i = 0 ; i < m_vatom.size() ; i++ ) {
        if ( m_vatom[i]->GetName() == atom_name ) {
            return m_vatom[i];
        }
    }
    return NULL;
}

int CCResidue::GetSerNum( CCAtom* atom ) const
{
    for ( int i = 0 ; i < m_vatom.size() ; i++ ) {
        if ( m_vatom[i] == atom ) {
            return m_serial[i];
        }
    }
    return 0;
}

double CCResidue::GetOccupancy( CCAtom* atom ) const
{
	for ( int i = 0 ; i < m_vatom.size() ; i++ ) {
		if ( m_vatom[i] == atom ) {
			return m_occupancy[i];
		}
	}
	return 0.0;
}

double CCResidue::GetTempFactor( CCAtom* atom ) const
{
	for ( int i = 0 ; i < m_vatom.size() ; i++ ) {
		if ( m_vatom[i] == atom ) {
			return m_tempFactor[i];
		}
	}
	return 0.0;
}

void CCResidue::SetHetAtom( CCAtom* atom, bool hetatm )
{
    for ( int i = 0 ; i < m_vatom.size() ; i++ ) {
        if ( m_vatom[i] == atom ) {
            m_hetatm[i] = hetatm;
        }
    }
}

bool CCResidue::IsHetAtom( CCAtom* atom ) const
{
    for ( int i = 0 ; i < m_vatom.size() ; i++ ) {
        if ( m_vatom[i] == atom ) {
            return m_hetatm[i];
        }
    }
    return false;
}

void CCResidue::SetResAtomName( CCAtom* atom, const string& atmname )
{
    for ( int i = 0 ; i < m_vatom.size() ; i++ ) {
        if ( m_vatom[i] == atom ) {
            m_atmName[i] = atmname;
        }
    }
}

void CCResidue::SetSerNum( CCAtom* atom, int sernum )
{
    for ( int i = 0 ; i < m_vatom.size() ; i++ ) {
        if ( m_vatom[i] == atom ) {
            m_serial[i] = sernum;
        }
    }
}

void CCResidue::SetOccupancy( CCAtom* atom, double occupancy )
{
	for ( int i = 0 ; i < m_vatom.size() ; i++ ) {
		if ( m_vatom[i] == atom ) {
			m_occupancy[i] = occupancy;
		}
	}
}

void CCResidue::SetTempFactor( CCAtom* atom, double tempFactor )
{
	for ( int i = 0 ; i < m_vatom.size() ; i++ ) {
		if ( m_vatom[i] == atom ) {
			m_tempFactor[i] = tempFactor;
		}
	}
}

CCAtom* CCResidue::BeginAtom( CCAtomIter& i )
{
    i = m_vatom.begin();
    return ( i == m_vatom.end() ? NULL : *i );
}

CCAtom* CCResidue::NextAtom( CCAtomIter& i )
{
    i++;
    return ( i == m_vatom.end() ? NULL : *i );
}

void CCResidue::AddAtom( CCAtom* atom )
{
    if ( atom != NULL ) {
        atom->SetResidue( this );
        m_hetatm.push_back( false );
        m_vatom.push_back( atom );
        m_atmName.push_back( "" );
        m_serial.push_back( 0 );
		m_occupancy.push_back(0.0);
		m_tempFactor.push_back(0.0);
    }
}

void CCResidue::RemoveAtom( CCAtom* atom )
{
    if ( atom != NULL ) {
        for ( int i = 0 ; i < m_vatom.size() ; ++i ) {
            if ( m_vatom[i] == atom ) {
                atom->SetResidue( NULL );
                m_hetatm.erase( m_hetatm.begin() + i );
                m_vatom.erase( m_vatom.begin() + i );
                m_atmName.erase( m_atmName.begin() + i );
                m_serial.erase( m_serial.begin() + i );
                break;
            }
        }
    }
}

/************************************************************************/
/* Residue Related Functions                                            */
/************************************************************************/
void SetResKeys( const string& resname, unsigned int& reskey,
                 unsigned int& aakey )
{
    reskey = GetResKey( resname );
    if ( reskey > 19 ) {
        aakey = 0;
        return;
    }

    switch ( reskey ) {
    case ResidueIndex::ALA:
        aakey = AA_ALA;
        break;
    case ResidueIndex::GLY:
        aakey = AA_GLY;
        break;
    case ResidueIndex::LEU:
        aakey = AA_LEU;
        break;
    case ResidueIndex::SER:
        aakey = AA_SER;
        break;
    case ResidueIndex::VAL:
        aakey = AA_VAL;
        break;
    case ResidueIndex::THR:
        aakey = AA_THR;
        break;
    case ResidueIndex::LYS:
        aakey = AA_LYS;
        break;
    case ResidueIndex::ASP:
        aakey = AA_ASP;
        break;
    case ResidueIndex::ILE:
        aakey = AA_ILE;
        break;
    case ResidueIndex::ASN:
        aakey = AA_ASN;
        break;
    case ResidueIndex::GLU:
        aakey = AA_GLU;
        break;
    case ResidueIndex::PRO:
        aakey = AA_PRO;
        break;
    case ResidueIndex::ARG:
        aakey = AA_ARG;
        break;
    case ResidueIndex::PHE:
        aakey = AA_PHE;
        break;
    case ResidueIndex::GLN:
        aakey = AA_GLN;
        break;
    case ResidueIndex::TYR:
        aakey = AA_TYR;
        break;
    case ResidueIndex::HIS:
        aakey = AA_HIS;
        break;
    case ResidueIndex::CYS:
        aakey = AA_CYS;
        break;
    case ResidueIndex::MET:
        aakey = AA_MET;
        break;
    case ResidueIndex::TRP:
        aakey = AA_TRP;
        break;
    default:
        aakey = 0;
        break;
    }
}

unsigned int GetResKey( const string& res )
{
    if ( !res.empty() && res.size() == 3 ) {
        int ch1 = toupper( res[0] );
        int ch2 = toupper( res[1] );
        int ch3 = toupper( res[2] );

        switch ( ch1 ) {
        case(' '):
            if ( ch2 == ' ' ) {
                switch ( ch3 ) {
                case('A'):
                    return( ResidueIndex::A );
                case('C'):
                    return( ResidueIndex::C );
                case('G'):
                    return( ResidueIndex::G );
                case('T'):
                    return( ResidueIndex::T );
                case('U'):
                    return( ResidueIndex::U );
                case('I'):
                    return( ResidueIndex::I );
                }
            }
            else if ( ch2 == '+' ) {
                if ( ch3 == 'U' ) {
                    return( ResidueIndex::UPLUS );
                }
            }
            else if ( ch2 == 'Y' ) {
                if ( ch3 == 'G' ) {
                    return( ResidueIndex::YG );
                }
            }

            break;

        case('0'):
            if ( ch2 == 'M' ) {
                if ( ch3 == 'C' ) {
                    return( ResidueIndex::OMC );
                }
                else if ( ch3 == 'G' ) {
                    return( ResidueIndex::OMG );
                }
            }

            break;

        case('1'):
            if ( ch2 == 'M' ) {
                if ( ch3 == 'A' ) {
                    return( ResidueIndex::_1MA );
                }
                else if ( ch3 == 'G' ) {
                    return( ResidueIndex::_1MG );
                }
            }

            break;

        case('2'):
            if ( ch2 == 'M' ) {
                if ( ch3 == 'G' ) {
                    return( ResidueIndex::_2MG );
                }
            }

            break;

        case('5'):
            if ( ch2 == 'M' ) {
                if ( ch3 == 'C' ) {
                    return( ResidueIndex::_5MC );
                }
                else if ( ch3 == 'U' ) {
                    return( ResidueIndex::_5MU );
                }
            }

            break;

        case('7'):
            if ( ch2 == 'M' ) {
                if ( ch3 == 'G' ) {
                    return( ResidueIndex::_7MG );
                }
            }

            break;

        case('A'):
            if ( ch2 == 'L' ) {
                if ( ch3 == 'A' ) {
                    return( ResidueIndex::ALA );
                }
            }
            else if ( ch2 == 'S' ) {
                if ( ch3 == 'P' ) {
                    return( ResidueIndex::ASP );
                }
                else if ( ch3 == 'N' ) {
                    return( ResidueIndex::ASN );
                }
                else if ( ch3 == 'X' ) {
                    return( ResidueIndex::ASX );
                }
            }
            else if ( ch2 == 'R' ) {
                if ( ch3 == 'G' ) {
                    return( ResidueIndex::ARG );
                }
            }
            else if ( ch2 == 'C' ) {
                if ( ch3 == 'E' ) {
                    return( ResidueIndex::ACE );
                }
            }
            else if ( ch2 == 'D' ) {
                if ( ch3 == 'E' ) {
                    return( ResidueIndex::A );
                }    /* "ADE" -> "  A" */
            }

            break;

        case('C'):
            if ( ch2 == 'Y' ) {
                if ( ch3 == 'S' ) {
                    return( ResidueIndex::CYS );
                }
                else if ( ch3 == 'H' ) {
                    return( ResidueIndex::CYS );
                }    /* "CYH" -> "CYS" */
                else if ( ch3 == 'T' ) {
                    return( ResidueIndex::C );
                }    /* "CYT" -> "  C" */
            }
            else if ( ch2 == 'O' ) {
                if ( ch3 == 'A' ) {
                    return( ResidueIndex::COA );
                }
            }
            else if ( ch2 == 'P' ) {
                if ( ch3 == 'R' ) {
                    return( ResidueIndex::PRO );
                }    /* "CPR" -> "PRO" */
            }
            else if ( ch2 == 'S' ) {
                if ( ch3 == 'H' ) {
                    return( ResidueIndex::CYS );
                }    /* "CSH" -> "CYS" */
                else if ( ch3 == 'M' ) {
                    return( ResidueIndex::CYS );
                }    /* "CSM" -> "CYS" */
            }

            break;

        case('D'):
            if ( ch2 == 'O' ) {
                if ( ch3 == 'D' ) {
                    return( ResidueIndex::DOD );
                }
            }
            else if ( ch2 == '2' ) {
                if ( ch3 == 'O' ) {
                    return( ResidueIndex::DOD );
                }    /* "D2O" -> "DOD" */
            }

            break;

        case('F'):
            if ( ch2 == 'O' ) {
                if ( ch3 == 'R' ) {
                    return( ResidueIndex::FOR );
                }
            }

            break;

        case('G'):
            if ( ch2 == 'L' ) {
                if ( ch3 == 'Y' ) {
                    return( ResidueIndex::GLY );
                }
                else if ( ch3 == 'U' ) {
                    return( ResidueIndex::GLU );
                }
                else if ( ch3 == 'N' ) {
                    return( ResidueIndex::GLN );
                }
                else if ( ch3 == 'X' ) {
                    return( ResidueIndex::GLX );
                }
            }
            else if ( ch2 == 'U' ) {
                if ( ch3 == 'A' ) {
                    return( ResidueIndex::G );
                }    /* "GUA" -> "  G" */
            }

            break;

        case('H'):
            if ( ch2 == 'I' ) {
                if ( ch3 == 'S' ) {
                    return( ResidueIndex::HIS );
                }
            }
            else if ( ch2 == 'O' ) {
                if ( ch3 == 'H' ) {
                    return( ResidueIndex::HOH );
                }
            }
            else if ( ch2 == 'Y' ) {
                if ( ch3 == 'P' ) {
                    return( ResidueIndex::HYP );
                }
            }
            else if ( ch2 == '2' ) {
                if ( ch3 == 'O' ) {
                    return( ResidueIndex::HOH );
                }    /* "H20" -> "HOH" */
                else if ( ch3 == 'U' ) {
                    return( ResidueIndex::H2U );
                }
            }

            break;

        case('I'):
            if ( ch2 == 'L' ) {
                if ( ch3 == 'E' ) {
                    return( ResidueIndex::ILE );
                }
            }

            break;

        case('L'):
            if ( ch2 == 'E' ) {
                if ( ch3 == 'U' ) {
                    return( ResidueIndex::LEU );
                }
            }
            else if ( ch2 == 'Y' ) {
                if ( ch3 == 'S' ) {
                    return( ResidueIndex::LYS );
                }
            }

            break;

        case('M'):
            if ( ch2 == 'E' ) {
                if ( ch3 == 'T' ) {
                    return( ResidueIndex::MET );
                }
            }
            else if ( ch2 == '2' ) {
                if ( ch3 == 'G' ) {
                    return( ResidueIndex::M2G );
                }
            }

            break;

        case('N'):
            if ( ch2 == 'A' ) {
                if ( ch3 == 'D' ) {
                    return( ResidueIndex::NAD );
                }
                else if ( ch3 == 'P' ) {
                    return( ResidueIndex::NAP );
                }
            }
            else if ( ch2 == 'D' ) {
                if ( ch3 == 'P' ) {
                    return( ResidueIndex::NDP );
                }
            }

            break;

        case('P'):
            if ( ch2 == 'R' ) {
                if ( ch3 == 'O' ) {
                    return( ResidueIndex::PRO );
                }
            }
            else if ( ch2 == 'H' ) {
                if ( ch3 == 'E' ) {
                    return( ResidueIndex::PHE );
                }
            }
            else if ( ch2 == 'C' ) {
                if ( ch3 == 'A' ) {
                    return( ResidueIndex::PCA );
                }
            }
            else if ( ch2 == 'O' ) {
                if ( ch3 == '4' ) {
                    return( ResidueIndex::PO4 );
                }
            }
            else if ( ch2 == 'S' ) {
                if ( ch3 == 'U' ) {
                    return( ResidueIndex::PSU );
                }
            }

            break;

        case('S'):
            if ( ch2 == 'E' ) {
                if ( ch3 == 'R' ) {
                    return( ResidueIndex::SER );
                }
            }
            else if ( ch2 == 'O' ) {
                if ( ch3 == '4' ) {
                    return( ResidueIndex::SO4 );
                }
                else if ( ch3 == 'L' ) {
                    return( ResidueIndex::HOH );
                }    /* "SOL" -> "HOH" */
            }
            else if ( ch2 == 'U' ) {
                if ( ch3 == 'L' ) {
                    return( ResidueIndex::SO4 );
                }    /* "SUL" -> "SO4" */
            }

            break;

        case('T'):
            if ( ch2 == 'H' ) {
                if ( ch3 == 'R' ) {
                    return( ResidueIndex::THR );
                }
                else if ( ch3 == 'Y' ) {
                    return( ResidueIndex::T );
                }    /* "THY" -> "  T" */
            }
            else if ( ch2 == 'Y' ) {
                if ( ch3 == 'R' ) {
                    return( ResidueIndex::TYR );
                }
            }
            else if ( ch2 == 'R' ) {
                if ( ch3 == 'P' ) {
                    return( ResidueIndex::TRP );
                }
                else if ( ch3 == 'Y' ) {
                    return( ResidueIndex::TRP );
                }    /* "TRY" -> "TRP" */
            }
            else if ( ch2 == 'I' ) {
                if ( ch3 == 'P' ) {
                    return( ResidueIndex::HOH );
                }    /* "TIP" -> "HOH" */
            }

            break;

        case('U'):
            if ( ch2 == 'N' ) {
                if ( ch3 == 'K' ) {
                    return( ResidueIndex::UNK );
                }
            }
            else if ( ch2 == 'R' ) {
                if ( ch3 == 'A' ) {
                    return( ResidueIndex::U );
                }    /* "URA" -> "  U" */
                else if ( ch3 == 'I' ) {
                    return( ResidueIndex::U );
                }    /* "URI" -> "  U" */
            }

            break;

        case('V'):
            if ( ch2 == 'A' ) {
                if ( ch3 == 'L' ) {
                    return( ResidueIndex::VAL );
                }
            }

            break;

        case('W'):
            if ( ch2 == 'A' ) {
                if ( ch3 == 'T' ) {
                    return( ResidueIndex::HOH );
                }
            }    /* "WAT" -> "HOH" */

            break;
        }

        /*unsigned int refno;
            for( refno = MINRES; refno < ResNo ; refno++ )
                if( !strncmp(Residue[refno],res,3) )
                    return refno;
            if ( ResNo < MAXRES - 1 )
            {
                ResNo++;
                Residue[refno][0] = (char) ch1;
                Residue[refno][1] = (char) ch2;
                Residue[refno][2] = (char) ch3;
                return refno;
            }*/
    }

    return ResidueIndex::UNK;
}

bool IsNaturalAA( const string& resname )
{
    if ( resname.size() != 3 ) { return false; }
    static set<string, CharPtrLess> *nat(NULL);
    if ( nat==NULL ) {
        nat = new set<string, CharPtrLess>; AssertMemAlloc(nat);
        nat->insert( "ALA" );
        nat->insert( "ARG" );
        nat->insert( "ASP" );
        nat->insert( "ASN" );
        nat->insert( "CYS" );
        nat->insert( "GLN" );
        nat->insert( "GLU" );
        nat->insert( "GLY" );
        nat->insert( "HIS" );
        nat->insert( "ILE" );
        nat->insert( "LEU" );
        nat->insert( "LYS" );
        nat->insert( "MET" );
        nat->insert( "PHE" );
        nat->insert( "PRO" );
        nat->insert( "SER" );
        nat->insert( "THR" );
        nat->insert( "TRP" );
        nat->insert( "TYR" );
        nat->insert( "VAL" );
    }
    return nat->find(resname ) != nat->end();
}

