/** 
* @file 
* @brief interface for utilities
* @author Yuan ZHAO, Tiejun CHENG
* @date 2008/08/23
* @version 1.0.1 
*
*/

#include <ccbg/CCUtil.h>
#include <string.h>

/* -------------- MISC -------------- */

/** @brief memory allocation error */
void MemAllocErr()
{
	clog << "\nOoops: Memory allocation error!\n"
		<< "This is usually caused by memory overflow.\n";
	exit( EXIT_FAILURE );
}
/** 
* @brief assert whether memory allocated successfully 
* @param[in] p pointer to be asserted
* @remark exit program if failed 
*/
void AssertMemAlloc( void* p )
{
    if ( p == NULL ) {
        MemAllocErr();
    }
}

/** 
* @brief comment line ?
* @param[in] line line to be asserted
* @param[in] delimstr first char indicating a comment line '#' by default
* @return true if the leading char of line is one of delimstr<BR>
*         false otherwise
*/
bool CommentLine( const string& line, char* delimstr )
{
    return ( line.find_first_of( delimstr ) == 0 );
}

/** 
* @brief blank line ?
* @param[in] line line to be asserted
* @param[in] delimstr blank chars
* @return true if the line is empty or contains only chars defined in delimstr
*         false otherwise
*/
bool BlankLine( const string& line, char* delimstr )
{
    if ( line.empty() ) {
        return true;
    }
    else {
        return ( ( line.find_first_not_of(delimstr) ) == string::npos )
             ? true
             : false;
    }
}

/** 
* @brief return current time in ASCII mode
*
* e.g. Wed Aug 22 10:38:05 2007
*/
char* GetTime()
{
    char* timestring;
    struct tm* ptr;
    time_t lt;

    lt = time( NULL );
    ptr = localtime( &lt );
    timestring = asctime( ptr );

    RTrim( timestring, "\n" );

    return timestring;
}

/** 
* @brief print a line
*
*/
void PrintLine(std::ostream& os, char c, int num)
{
	for (int i=0; i<num; ++i) {
		os << c;
	}
	os << endl;
}

/* -------------- STRING -------------- */

//# trim leading and tailing character defined in delimstr
void Trim( string& s, char* delimstr )
{
	// Trim right delimiter string
	string::size_type pos = s.find_last_not_of( delimstr );
	if ( pos != string::npos ) {
		s.erase( pos + 1 );
	}
	else {
		s.erase();
	}

	// Trim left delimiter string
	pos = s.find_first_not_of( delimstr );
	if ( pos != string::npos ) {
		s.erase( 0, pos );
	}
	else {
		s.erase();
	}
}
//# trim leading and tailing character defined in delimstr
void Trim( char* s, char* delimstr )
{
	string str = s;
	Trim( str, delimstr );
	strcpy( s, str.c_str() );
}
//# trim tailing character defined in delimstr
void RTrim( string& s, char* delimstr )
{
	// Trim right delimiter string
	string::size_type pos = s.find_last_not_of( delimstr );
	if ( pos != string::npos ) {
		s.erase( pos + 1 );
	}
	else {
		s.erase();
	}
}
//# trim tailing character defined in delimstr
void RTrim( char* s, char* delimstr )
{
	string str = s;
	RTrim( str, delimstr );
	strcpy( s, str.c_str() );
}
//# trim leading character defined in delimstr
void LTrim( string& s, char* delimstr )
{
	// Trim left delimiter string
	string::size_type pos = s.find_first_not_of( delimstr );
	if ( pos != string::npos ) {
		s.erase( 0, pos );
	}
	else {
		s.erase();
	}
}
//# trim leading character defined in delimstr
void LTrim( char* s, char* delimstr )
{
	string str = s;
	LTrim( str, delimstr );
	strcpy( s, str.c_str() );
}
//# Shift the supplied string to uppercase
void ToUpper( std::string& s )
{
	if ( s.empty() ) {
		return;
	}
	unsigned int i;
	for ( i = 0; i < s.size(); ++i ) {
		if ( isalpha( s[i] ) && !isdigit( s[i] ) ) {
			s[i] = toupper( s[i] );
		}
	}
}
//# Shift the supplied char* to uppercase
void ToUpper( char* cptr )
{
	char* c;
	for ( c = cptr; *c != '\0'; ++c ) {
		if ( isalpha( *c ) && !isdigit( *c ) ) {
			*c = toupper( *c );
		}
	}
}
//# Shift the supplied string to lowercase
void ToLower( std::string& s )
{
	if ( s.empty() ) {
		return;
	}
	unsigned int i;
	for ( i = 0; i < s.size(); ++i ) {
		if ( isalpha( s[i] ) && !isdigit( s[i] ) ) {
			s[i] = tolower( s[i] );
		}
	}
}
//# Shift the supplied char* to lowercase
void ToLower( char* cptr )
{
	char* c;
	for ( c = cptr; *c != '\0'; ++c ) {
		if ( isalpha( *c ) && !isdigit( *c ) ) {
			*c = tolower( *c );
		}
	}
}
//# Shift the supplied string: lowercase to upper, and upper to lower
//# s - The string to switch case
//# start - The position to start inverting case
void InvertCase( std::string& s, unsigned int start )
{
	if ( s.empty() || start >= s.size() ) {
		return;
	}
	unsigned int i;
	for ( i = start; i < s.size(); ++i ) {
		if ( isalpha( s[i] ) && !isdigit( s[i] ) ) {
			if ( isupper( s[i] ) ) {
				s[i] = tolower( s[i] );
			}
			else {
				s[i] = toupper( s[i] );
			}
		}
	}
}
//# Shift the supplied char*: lowercase to upper, and upper to lower
void InvertCase( char* cptr )
{
	char* c;
	for ( c = cptr; *c != '\0'; ++c ) {
		if ( isalpha( *c ) && !isdigit( *c ) ) {
			if ( isupper( *c ) ) {
				*c = tolower( *c );
			}
			else {
				*c = toupper( *c );
			}
		}
	}
}

bool iCaseEQ( const string& s1, const string& s2 )
{
	if ( s1.size() != s2.size() ) {
		return false;
	}

	string::const_iterator i = s1. begin(), j = s2.begin();
	for ( ; i != s1.end() && j != s2.end(); ++i,++j ) {
		if ( *i
			!= *j
			&& ( isalpha( *i )
			&& isalpha( *j )
			&& toupper( *i ) != toupper( *j ) ) ) {
				return false;
		}
	}

	return true;
}

int Tokenize( vector<string>& vstr, const string& s, const char* delimstr,
			 int limit )
{
	vstr.clear();
	string::size_type startpos = 0, endpos = 0;

	int matched = 0;
	int s_size = s.size();
	for ( ; ; ) {
		startpos = s.find_first_not_of( delimstr, startpos );
		endpos = s.find_first_of( delimstr, startpos );
		if ( endpos <= s_size && startpos <= s_size ) {
			vstr.push_back( s.substr( startpos, endpos - startpos ) );
			matched++;
			if ( matched == limit ) {
				startpos = endpos + 1;
				vstr.push_back( s.substr( startpos ) );
				break;
			}
		}
		else {
			if ( startpos < s_size ) {
				vstr.push_back( s.substr( startpos, s_size - startpos ) );
			}
			break;
		}

		startpos = endpos + 1;
	}

	return vstr.size();
}

bool IsOneOf( const string& substr, const string& str )
{
	return (str+",").find(substr+",") != string::npos
		? true : false;
}

/** 
* @brief join several c-style string with the "\\" or "/"
* @note this is function with variable number of arguments, all 
*       arguments must be type of char*, the last one must be 
*       "\\" or "/", and will be used to join the previous ones.
*
* eg. Concat("D:", "test", "a.txt", "\\") -> "D:\\test\\a.txt"
*/
#include <cstdarg>
string Concat(char *msg, ...) 
{ 
	va_list argp; // 屣溯Ľṹ
	char *para; // ȡַ

	// argpָĵһɱmsgһȷĲ
	va_start( argp, msg ); 

	string path;
	while (1) {
		// ȡǰĲΪchar * 
		para = va_arg( argp, char *);
		if ( !strcmp(para, "\\") || !strcmp(para, "/") ) 
			// "\\" or "/" ָʾ
			break; 
		if (path.empty()) { path = para; }
		else { path += "\\"; path += para; }
	} 
	// argpΪNULL
	va_end( argp ); 
	return path; 
}

/* -------------- MATH -------------- */

const double pINF = +9.9e99;
const double nINF = -9.9e99;
const double EPSILON = 1.0e-11;
const double PI = 3.1415926535898;
const double RAD_TO_DEG = 180.0 / PI;
const double DEG_TO_RAD = PI / 180.0;

//# return whether a is squarable
bool CanBeSquared( const double& a )
{
	if ( a == 0 ) {
		return true;
	}
	const double max_squarable = 1e150;
	const double min_squarable = 1e-150;
	double abs_a = fabs( a );
	return( abs_a <max_squarable && abs_a> min_squarable );
}

//////////////////////////////////////////////////////////////////////////
// get a random number that is average distributed between min and max  //
// the default precision is 4. [min, max]                               //
//////////////////////////////////////////////////////////////////////////
double AverageRandom( double min /* = 0.0 */, double max /* = 1.0 */,
					 int precision /* = 4 */ )
{
	// To avoid calling this function too frequently, srand() must
	// be call just before calling AverageRandom. 
	// If srand was called inside AverageRandom, (unsigned)time( NULL )
	// will be the same which will cause adjacent random numbers
	// have the same seed. So they will generation the same number.
	// srand( (unsigned)time( NULL ) );
	if ( max == min ) {
		return max;
	}
	if ( max < min ) {
		double tmp;
		tmp = min;
		min = max;
		max = tmp;
	}

	int PNUM = 1;
	for ( int i = 0; i < precision; ++i ) {
		PNUM *= 10;
	}

	int minInteger = int( min* PNUM );
	int maxInteger = int( max* PNUM );
	int randInteger = rand() * rand();
	int diffInterger = maxInteger - minInteger;
	int resultInteger = randInteger % diffInterger + minInteger;
	return resultInteger / ( double ) PNUM;
}

int BinnedDistance( double d, 
				   double flr, double cil, double spacing )
{
	if ( d < flr || d > cil ) {
		return 0;
	}

	int n = (int) ceil( ( d - flr ) / spacing );

	return n == 0 ? 1 : n;
}

//////////////////////////////////////////////////////////////////////////
// Hash functions
//////////////////////////////////////////////////////////////////////////
// This algorithm was first reported by Dan Bernstein many years ago 
// in comp.lang.c. It is one of the most efficient hash functions ever
// published. 
unsigned long DJB2Hash( string key ) {
	unsigned long hash = 5381;
	int c;
	string::iterator i;

	for ( int i = 0; i < key.size(); ++i ) {
		c = key[i];
		hash = ( ( hash << 5 ) + hash ) + c; /* hash * 33 + c */
	}

	return hash;
}

unsigned long DJB2Hash( vector<unsigned int> key ) {
	unsigned long hash = 5381;
	int c;
	vector<unsigned int>::iterator i;

	for ( i = key.begin(); i != key.end(); ++i ) {
		c = ( *i );
		hash = ( ( hash << 5 ) + hash ) + c; /* hash * 33 + c */
	}

	return hash;
}

unsigned long SDBMHash( string key ) {
	unsigned long hash = 0;
	int c;

	for ( int i = 0; i < key.size(); ++i ) {
		c = key[i];
		hash = c + ( hash << 6 ) + ( hash << 16 ) - hash;
	}

	return hash;
}

unsigned long SDBMHash( vector<unsigned int> key ) {
	unsigned long hash = 0;
	int c;

	vector<unsigned int>::iterator i;

	for ( i = key.begin(); i != key.end(); ++i ) {
		c = ( *i );
		hash = c + ( hash << 6 ) + ( hash << 16 ) - hash;
	}

	return hash;
}
