#include <ccbg/CCVector.h>
#include <ccbg/CCError.h>
#include <ccbg/CCUtil.h>
#include <ccbg/CCMatrix.h>

const CCVector VZero( 0.0, 0.0, 0.0 );
const CCVector VX( 1.0, 0.0, 0.0 );
const CCVector VY( 0.0, 1.0, 0.0 );
const CCVector VZ( 0.0, 0.0, 1.0 );

double CCVector::operator[]( int i )
{
    if ( i< 0 || i > 2 ) {
        gErrorLog.Throw( __FUNCTION__, "pos out of range." );
        return 0.0;
    }
    if ( i == 0 ) {
        return m_x;
    }
    if ( i == 1 ) {
        return m_y;
    }
    else {
        return m_z;
    }
}

double Distance2( const CCVector& p1, const CCVector& p2 )
{
    return ( p1.x() - p2.x() ) * ( p1.x() - p2.x() )
         + ( p1.y() - p2.y() ) * ( p1.y() - p2.y() )
         + ( p1.z() - p2.z() ) * ( p1.z() - p2.z() );
}

double Distance2( const CCVector& p1, double x, double y, double z )
{
    return ( p1.x() - x ) * ( p1.x() - x )
         + ( p1.y() - y ) * ( p1.y() - y )
         + ( p1.z() - z ) * ( p1.z() - z );
}

double Distance( const CCVector& p1, const CCVector& p2 )
{
    return sqrt( Distance2( p1, p2 ) );
}

double Distance( const CCVector& p1, double c[] )
{
    return sqrt( ( p1.x() - c[0] ) * ( p1.x() - c[0] )
               + ( p1.y() - c[1] ) * ( p1.y() - c[1] )
               + ( p1.z() - c[2] ) * ( p1.z() - c[2] ) );
}

double Distance( const CCVector& p1, double x, double y, double z )
{
    return sqrt( ( p1.x() - x ) * ( p1.x() - x )
               + ( p1.y() - y ) * ( p1.y() - y )
               + ( p1.z() - z ) * ( p1.z() - z ) );
}

// return (v1 . v2)
double Dot( const CCVector& v1, const CCVector& v2 )
{
    return ( v1.x() * v2.x() + v1.y() * v2.y() + v1.z() * v2.z() );
}

// result = (v1 * v2)
CCVector Cross( const CCVector& v1, const CCVector& v2 )
{
    CCVector result;
    result.x() = v1.y() * v2.z() - v2.y() * v1.z();
    result.y() = v1.z() * v2.x() - v2.z() * v1.x();
    result.z() = v1.x() * v2.y() - v2.x() * v1.y();
    return result;
}

// return the angle two vectors in degree
// v1.v2 = |v1| * |v2| * cos<v1, v2>
double AngleOfTwoVectors( const CCVector& v1, const CCVector& v2 )
{
    double mag;
    double dp;

    mag = v1.Length() * v2.Length();
    if ( fabs( mag ) < EPSILON ) {
#ifdef _DEBUG
        gErrorLog.Throw( __FUNCTION__, "zero vector!\n", ccWarning );
#endif
        return 0.0;
    }
    dp = Dot( v1, v2 ) / mag;

    if ( dp < -1.0 ) {
        return 180.0;
    }

    if ( dp > 1.0 ) {
        return 0.0;
    }

    return( RAD_TO_DEG * acos( dp ) );
}

double DistanceOfPointToPlane( const CCVector& p1, const CCVector& p2,
                               const CCVector& p3, const CCVector& p4 )
{
    CCVector v1, v2, v3;
    CCVector rd;  // direction of point a to the plane b-c-d
    double angle; // angle of vector a-b to the plane b-c-d
    double d;   // distance of point a to the plane b-c-d

    v1 = p1 - p2;
    v2 = p3 - p2;
    v3 = p4 - p2;

    // determine the direction of the plane (b-c-d) first
    // |    i         j           k    |
    // | c.x-b.x    c.y-b.y    c.z-b.z |
    // | d.x-b.x    d.y-b.y    d.z-b.z |
    rd = Cross( v2, v3 );

    angle = AngleOfTwoVectors( v1, rd );
    d = v1.Length() * fabs( cos( DEG_TO_RAD * angle ) );
    return d;
}

// This function calculates the angle of three points, represented by p1-p2-p3
double Angle( const CCVector& p1, const CCVector& p2, const CCVector& p3 )
{
    CCVector v12, v23;

    v12 = p2 - p1;
    v23 = p2 - p3;
    return AngleOfTwoVectors( v12, v23 );
}

/*!  This function calculates the torsion angle of three vectors, represented
by four points A--B--C--D, i.e. B and C are vertexes, but none of A--B,
B--C, and C--D are colinear.  A "torsion angle" is the amount of "twist"
or torsion needed around the B--C axis to bring A--B into the same plane
as B--C--D.  The torsion is measured by "looking down" the vector B--C so
that B is superimposed on C, then noting how far you'd have to rotate
A--B to superimpose A over D.  Angles are + in the anticlockwise
direction.  The operation is symmetrical in that if you reverse the image
(look from C to B and rotate D over A), you get the same answer.
*/
// This function calculates the torsion angle of four points that represented
// by p1--p2--p3--p4
double Torsion( const CCVector& p1, const CCVector& p2, const CCVector& p3,
                const CCVector& p4 )
{
    double torsion;
    CCVector v12, v23, v34, v1, v2, v3;

    v12 = p2 - p1;
    v23 = p3 - p2;
    v34 = p4 - p3;

    v1 = Cross( v12, v23 );
    v2 = Cross( v23, v34 );
    v3 = Cross( v1, v2 );

    if ( v1.Length() * v2.Length() < 0.001 ) {
        return 0.0;
    }
    else {
        torsion = AngleOfTwoVectors( v1, v2 );
        if ( Dot( v23, v3 ) > 0.0 ) {
            torsion *= -1.0;
        }
    }
    return torsion;
}

// return true if p1-p2-p3 is in a line
bool IsInALine( const CCVector& p1, const CCVector& p2, const CCVector& p3 )
{
    double angle = Angle( p1, p2, p3 );
    return ( fabs( angle ) < EPSILON || fabs( 180.0 - angle ) < EPSILON );
}

// axis is the vector which indicates the line's orientation
// origin is the start point of the line
// given two point of a line, origin = start point; axis = end - start
// return the distance of point to the line
double DistanceFromPointToLine( const CCVector& axis, const CCVector& origin,
                                const CCVector& point )
{
    CCVector v, vtmp;
    double dt, d2;

    vtmp = axis;
    vtmp.Normalize();

    v = point - origin;
    dt = Dot( v, vtmp );
    d2 = v.Length_2() - dt * dt;

    return ( d2 > 0.000 ) ? sqrt( d2 ) : 0.000;
}

// return the angle of the plane ( p1, p2, p3) and (q1, q2, q3)
double AngleOfTwoPlanes( const CCVector& p1, const CCVector& p2,
                         const CCVector& p3, const CCVector& q1,
                         const CCVector& q2, const CCVector& q3 )
{
    CCVector pv, qv, p12, p13, q12, q13;
    double angle, tmp1, tmp2, theta;

    //|     i               j               k       |
    //| p2[0]-p1[0]     p2[1]-p1[1]     p2[2]-p1[2] |
    //| p3[0]-p1[0]     p3[1]-p1[1]     p3[2]-p1[2] |

    p12 = p2 - p1;
    p13 = p3 - p1;
    pv = Cross( p12, p13 );

    q12 = q2 - q1;
    q13 = q3 - q1;
    qv = Cross( q12, q13 );

    tmp1 = Dot( pv, qv );
    tmp2 = pv.Length() * qv.Length();

    theta = acos( tmp1 / tmp2 ); 
    angle = theta * RAD_TO_DEG;

    if ( angle > 90.0 ) {
        angle = fabs( 180.0 - angle );
    }

    return angle;  // returned angle is in degree
}

// axis is the vector which indicates the orientation
// origin is the start point of the rotation axis
// start is the point which needs to be rotated
// angle is the rotation angle in degree
//            o(start)
//           /
//          / (angle)
// (origin)o---------------> (axis')
//     ^    \                  ^
//     |     \                 | translate
//     |      o(end)           |
//  (0,0,0)o---------------> (axis)
//
CCVector RotatePoint( const CCVector& axis, const CCVector& origin,
                      const CCVector& start, double angle )
{
	// make angle between [0, 360] degree.
	while ( angle < 0.0 ) {
		angle += 360.0;
	}
	while ( angle > 360.0 ) {
		angle -= 360.0;
	}

	// angle = 0.0, return start coordinates
	if ( fabs( angle ) < 1e-11 ) {
		return start;
	}
	// length = 0.0
	if ( fabs(start.Length()) < EPSILON ) {
		return start;
	}
	// origin is in the line of axis, return start coordinates
	if ( IsInALine( start, start + axis, origin ) ) {
		return start;
	}
	
    CCMatrix mat;
    CCVector p, end;
    double s, c, t;

    double theta = angle * DEG_TO_RAD;
    s = sin( theta );
    c = cos( theta );
    t = 1 - c;

    CCVector vtmp = axis;
    vtmp.Normalize();

    double x = vtmp.x(), y = vtmp.y(), z = vtmp.z();

    end = start - origin;

    mat( 0, 0 ) = x * x + ( 1 - x * x ) * c;
    mat( 1, 0 ) = x * y * t - z * s;
    mat( 2, 0 ) = x * z * t + y * s;
    mat( 0, 1 ) = x * y * t + z * s;
    mat( 1, 1 ) = y * y + ( 1 - y * y ) * c;
    mat( 2, 1 ) = y * z * t - x * s;
    mat( 0, 2 ) = x * z * t - y * s;
    mat( 1, 2 ) = y * z * t + x * s;
    mat( 2, 2 ) = z * z + ( 1 - z * z ) * c;

    p = end * mat;
    end = p + origin;

    return end;
}

// ******************************************************************
// v1 is known vectors while v2, v3 and v4 are desired vectors.
// v1 is the primary neighbor vector
// ref is the reference vector which was used to determine v2
// angle is the desired angle between v1 and v2 (or v3, v4) 
// v2, v3, and v4 are returned as unit vectors
// ******************************************************************
void GetSp3Coordinates3( const CCVector& v1, const CCVector& vref,
                         CCVector& v2, CCVector& v3, CCVector& v4,
                         double angle /* = 109.40 */ )
{
    CCVector tmp1( v1 );
    CCVector tmp2( -vref );
    tmp1.Normalize();
    tmp2.Normalize();

    CCVector axis;
    double diff;

    axis = Cross( tmp1, tmp2 );
    diff = angle - AngleOfTwoVectors( tmp1, tmp2 );
    v2 = RotatePoint( axis, VZero, tmp2, diff );

    axis = -tmp1;
    v3 = RotatePoint( axis, VZero, v2, 120.0 );
    v4 = RotatePoint( axis, VZero, v2, 240.0 );
}

// ******************************************************************
// v1 and v2 are known vectors while v3 and v4 are desired vectors.
// v1 is the primary neighbor vector
// v2 is the secondary neighbor vector
// v3,v4 are returned as unit vectors
// ******************************************************************
void GetSp3Coordinates2( const CCVector& v1, const CCVector& v2, CCVector& v3,
                         CCVector& v4 )
{
    CCVector tmp1( v1 );
    CCVector tmp2( v2 );
    tmp1.Normalize();
    tmp2.Normalize();

    CCVector axis;
    axis = tmp1 + tmp2;

    v3 = -RotatePoint( axis, VZero, tmp1, 90.0 );
    v4 = -RotatePoint( axis, VZero, tmp2, 90.0 );
}

// ******************************************************************
// v1, v2 and v3 are known vectors while v4 is the desired vector.
// v1 is the primary neighbor vector 
// v2 is the secondary neighbor vector
// v3 is the tertiary neighbor vector
// v4 is returned as unit vector
// ******************************************************************
void GetSp3Coordinates( const CCVector& v1, const CCVector& v2,
                        const CCVector& v3, CCVector& v4 )
{
    CCVector tmp1( v1 );
    CCVector tmp2( v2 );
    CCVector tmp3( v3 );

    tmp1.Normalize();
    tmp2.Normalize();
    tmp3.Normalize();

	// check whether v1 v2 and v3 are in a plane. if it is,
	// return the normalized perpendicular vector of the plane.
	// else return v4 = Normalize(-( tmp1 + tmp2 + tmp3 ))
	v4 = Cross( tmp1, tmp2 );
	if ( !IsZero( Dot( tmp3, v4 ) ) ) {
		v4 = -( tmp1 + tmp2 + tmp3 );
	}
}

// ******************************************************************
// v1 is known vectors while v2 and v3 are desired vectors.
// v1 is the primary neighbor vector
// ref is the reference vector which was used to determine v2
// v2 and v3 are returned as unit vectors
// ******************************************************************
void GetSp2Coordinates2( const CCVector& v1, const CCVector& vref,
                         CCVector& v2, CCVector& v3 )
{
    CCVector tmp1( v1 );
    tmp1.Normalize();

    CCVector axis = Cross( tmp1, vref );

    v2 = RotatePoint( axis, VZero, tmp1, 120.0 );
    v3 = RotatePoint( axis, VZero, tmp1, 240.0 );
}

// ******************************************************************
// v1 and v2 are known vectors and v3 is the desired vector.
// v1 is the primary neighbor vector
// v2 is the secondary neighbor vector
// v3 is returned as unit vector.
// ******************************************************************
void GetSp2Coordinates( const CCVector& v1, const CCVector& v2, CCVector& v3 )
{
    CCVector tmp1( v1 );
    CCVector tmp2( v2 );

    tmp1.Normalize();
    tmp2.Normalize();

    v3 = -( v1 + v2 );
    v3.Normalize();
}

// ******************************************************************
// v1 is known vectors while v2 is desired vectors.
// v1 is the primary neighbor vector
// v2 is returned as unit vectors
// ******************************************************************
void GetSp1Coordinates( const CCVector& v1, CCVector& v2 )
{
    v2 = -v1;
    v2.Normalize();
}

// return the minimum value of {angle, 180-angle}
double AcuteAngle( double angle )
{
    return angle < 90.0 ? angle : 180.0 - angle;
}

// return the minimum value of {|torsion|,|180-torsion|,|-180-torsion|}
double MinTorsion( double torsion )
{
    if ( torsion < -180.0 || torsion > 180.0 ) {
        gErrorLog.Throw( __FUNCTION__, "torsion must be [-180,180].",
                         ccWarning );
        return torsion;
    }
    if ( torsion > 0.0 ) {
        return torsion < 180.0 - torsion ? torsion : 180.0 - torsion;
    }
    else {
        return -torsion < 180.0 + torsion ? -torsion : 180.0 + torsion;
    }
}

//////////////////////////////////////////////////////////////////////////
bool CCVector::IsApprox( const CCVector& other, double precision ) const
{
    return( distSq( other )
         <= precision * precision * std::min( Length_2(),
                                              other.Length_2() ) );
}

bool CCVector::IsZeroVec() const
{
	if ( fabs( m_x ) < EPSILON && fabs( m_y ) < EPSILON && fabs( m_z ) < EPSILON ) {
		return true;
	}
	return false;
}

bool CCVector::CanBeNormalized() const
{
    if ( fabs( m_x ) < EPSILON && fabs( m_y ) < EPSILON && fabs( m_z ) < EPSILON ) {
        return false;
    }
    return ( CanBeSquared( m_x )
          && CanBeSquared( m_y )
          && CanBeSquared( m_z ) );
}

CCVector& CCVector::Normalize()
{
    if ( CanBeNormalized() ) {
        ( *this ) /= Length();
    }
    else {
        gErrorLog.Throw( __FUNCTION__, "can not be normalized!" );
    }

    return( *this );
}

void CCVector::RandomUnitVector()
{
    // obtain a random vector with 0.001 <= length^2 <= 1.0, normalize
    // the vector to obtain a random vector of length 1.0.
    double l;
    srand( ( unsigned ) time( NULL ) );
    do {
        SetXYZ( AverageRandom(), AverageRandom(), AverageRandom() );
        l = Length_2();
    }
    while ( ( l > 1.0 ) || ( l < 1e-4 ) );
    Normalize();
}

/////////////////////////////////////////////////////////////////////////
// Vector Analysis (used by CCXXFFYYYY)                                //
/////////////////////////////////////////////////////////////////////////
double VectorLengthDerivative( CCVector& a, CCVector& b )
{
    CCVector vab, drab;
    double rab;

    vab = a - b;
    rab = vab.Length();
    if ( rab < 0.1 ) // atoms are too close to each other
    {
        vab.RandomUnitVector();
        vab *= 0.1; // move the atoms a small distance apart
        rab = 0.1;
    }
    drab = vab / rab;

    a = -drab; // -drab/da
    b = drab; //  drab/db

    return rab;
}

double VectorAngleDerivative( CCVector& a, CCVector& b, CCVector& c )
{
    CCVector vab, vcb;
    double theta, rab, rab2, rcb, rcb2, abcb, abcb2;

    vab = a - b;
    vcb = c - b;
    rab = vab.Length();
    rcb = vcb.Length();
    rab2 = rab * rab;
    rcb2 = rcb * rcb;

    abcb = Dot( vab, vcb ) / ( rab * rcb );
    abcb2 = 1.0 - abcb * abcb;
    theta = acos( abcb ) * RAD_TO_DEG;

    if ( fabs( abcb2 ) < EPSILON ) {
        a = VZero;
        b = VZero;
        c = VZero;
        return 0.0;
    }

    a = ( vcb * rab * rcb - ( vab / rab ) * Dot( vab, vcb ) * rcb )
      / ( sqrt( abcb2 ) * rab2 * rcb2 );
    c = -( ( vcb / rcb ) * Dot( vab, vcb ) * rab - vab * rab * rcb )
      / ( sqrt( abcb2 ) * rab2 * rcb2 );
    b = -a - c;

    a *= ( 1.0 / DEG_TO_RAD );
    b *= ( 1.0 / DEG_TO_RAD );
    c *= ( 1.0 / DEG_TO_RAD );

    return theta;
}

double VectorOOPDerivative( CCVector& a, CCVector& b, CCVector& c, CCVector& d )
{
    CCVector vab, vcb, vdb;
    CCVector rd;  // direction of point a to the plane b-c-d
    double angle; // angle of vector a-b to the plane b-c-d
    double rap;   // distance of point a to the plane b-c-d

    vab = a - b;
    vcb = c - b;
    vdb = d - b;

    // determine the direction of the plane (b-c-d) first
    // |    i         j           k    |
    // | c.x-b.x    c.y-b.y    c.z-b.z |
    // | d.x-b.x    d.y-b.y    d.z-b.z |
    rd = Cross( vcb, vdb );

    angle = AngleOfTwoVectors( vab, rd );
    rap = vab.Length() * fabs( cos( DEG_TO_RAD * angle ) );
    if ( rap < EPSILON ) {
        a = VZero;
    }
    else if ( angle < 90.0 ) {
        a = -rd / rap;
    }
    else {
        a = rd / rap;
    }

    return rap;
}

double VectorTorsionDerivative( CCVector& a, CCVector& b, CCVector& c,
                                CCVector& d )
{
    CCVector vab, vbc, vcd, vac, vbd, grada, gradb, gradc, gradd;
    CCVector abbc, bccd, bcabbc, bcbccd, cdabbc, cdbccd, acabbc, acbccd,
             ababbc, abbccd, bdabbc, bdbccd;
    double tor, rabbc, rbccd, rabbc2, rbccd2, rabbc3, rbccd3;
    double abbc_bccd, rabbc_rbccd, abbc_bccd2;

    vab = a - b;
    vbc = b - c;
    vcd = c - d;
    vac = a - c;
    vbd = b - d;
    abbc = Cross( vab, vbc );
    bccd = Cross( vbc, vcd );

    tor = RAD_TO_DEG * acos( Dot( abbc, bccd )
                           / ( abbc.Length() * bccd.Length() ) );
    if ( Dot( abbc, bccd ) > 0.0 ) {
        tor = -tor;
    }

    bcabbc = Cross( vbc, abbc );
    bcbccd = Cross( vbc, bccd );
    cdabbc = Cross( vcd, abbc );
    cdbccd = Cross( vcd, bccd );
    acabbc = Cross( vac, abbc );
    acbccd = Cross( vac, bccd );
    ababbc = Cross( vab, abbc );
    abbccd = Cross( vab, bccd );
    bdabbc = Cross( vbd, abbc );
    bdbccd = Cross( vbd, bccd );
    rabbc = abbc.Length();
    rbccd = bccd.Length();
    rabbc2 = rabbc * rabbc;
    rbccd2 = rbccd * rbccd;
    rabbc3 = rabbc2 * rabbc;
    rbccd3 = rbccd2 * rbccd;
    abbc_bccd = Dot( abbc, bccd );
    rabbc_rbccd = abbc_bccd / ( rabbc * rbccd );
    abbc_bccd2 = sqrt( 1.0 - rabbc_rbccd * rabbc_rbccd );

    if ( fabs( abbc_bccd2 ) < EPSILON ) {
        abbc_bccd2 = 1e-7;
    }

    a = ( bcbccd
        / ( rabbc * rbccd )
        - ( bcabbc * abbc_bccd )
        / ( rabbc3 * rbccd ) )
      / abbc_bccd2;
    d = ( bcabbc
        / ( rabbc * rbccd )
        - ( bcbccd * abbc_bccd )
        / ( rabbc * rbccd3 ) )
      / abbc_bccd2;

    b = ( -( cdbccd * abbc_bccd )
        / ( rabbc * rbccd3 )
        + ( cdabbc - acbccd )
        / ( rabbc * rbccd )
        + ( acabbc * abbc_bccd )
        / ( rabbc3 * rbccd ) )
      / abbc_bccd2;

    c = ( ( bdbccd * abbc_bccd )
        / ( rabbc * rbccd3 )
        + ( abbccd - bdabbc )
        / ( rabbc * rbccd )
        + -( ababbc * abbc_bccd )
        / ( rabbc3 * rbccd ) )
      / abbc_bccd2;

    if ( IsNan( a.x() ) || IsNan( a.y() ) || IsNan( a.z() ) ) {
        a = VZero;
    }
    if ( IsNan( b.x() ) || IsNan( b.y() ) || IsNan( b.z() ) ) {
        b = VZero;
    }
    if ( IsNan( c.x() ) || IsNan( c.y() ) || IsNan( c.z() ) ) {
        c = VZero;
    }
    if ( IsNan( d.x() ) || IsNan( d.y() ) || IsNan( d.z() ) ) {
        d = VZero;
    }

    if ( Dot( abbc, bccd ) > 0.0 ) {
        a = -a;
        b = -b;
        c = -c;
        d = -d;
    }

    return tor;
}

// Random Vector whose x, y, z are all in [0,1]
CCVector RandomVector()
{
    return CCVector( AverageRandom(), AverageRandom(), AverageRandom() );
}

