/**********************************************************************
kekulize.cpp - Alternate algorithm to kekulize a molecule.

Copyright (C) 2004-2006 by Fabien Fontaine
Some portions Copyright (C) 2005-2006 by Geoffrey R. Hutchison

This file is part of the Open Babel project.
For more information, see <http://openbabel.sourceforge.net/>

This program is free software; you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation version 2 of the License.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
***********************************************************************/

#include <ccbg/CCMol.h>

#include <sstream>

const int SINGLE = 1;
const int DOUBLE = 2;

using namespace std;

//namespace OpenBabel 
//{

///////////////////////////////////////////////////////////////////////////////
//! \brief Kekulize aromatic rings without using implicit valence

//! This new perceive kekule bonds function has been especifically designed to 
//! handle molecule files without explicit hydrogens such as pdb or xyz.
//! The function does not rely on GetImplicitValence function
//! The function looks for groups of aromatic cycle 
//! For each group it tries to guess the number of electrons given by each atom
//! in order to satisfy the huckel (4n+2) rule
//! If the huckel rule cannot be satisfied the algorithm try with its best alternative guess
//! Then it recursively walk on the atoms of the cycle and assign single and double bonds
void CCMol::Kekulize()
{
    FOR_BONDS_OF_MOL( bond, this ) {
        if ( bond->IsKSingle() ) {
            bond->SetOrder( 1 );
        }
        else if ( bond->IsKDouble() ) {
            bond->SetOrder( 2 );
        }
        else if ( bond->IsKTriple() ) {
            bond->SetOrder( 3 );
        }
    }
}

void CCMol::NewPerceiveKekuleBonds()
{
    if ( HasFlag( MOL_KEKULE_PERCEIVED ) ) {
        return;
    }
    SetFlag( MOL_KEKULE_PERCEIVED );

    CCAtom* atom;
    std::vector<CCAtom*> cycle;
    CCBitVec avisit, cvisit;
    avisit.Resize( GetNumAtom() + 1 );
    cvisit.Resize( GetNumAtom() + 1 );
    //CCBond *bond;
    //std::vector<CCBond*>::iterator bi;
    std::vector<int> electron;
    int sume, orden, bestatom;
    int bestorden = 99;
    // Init the kekulized bonds
    unsigned i;
    FOR_BONDS_OF_MOL( bond, *this ) {
        switch ( bond->GetOrder() ) {
        case 1:
            bond->SetKSingle(); break;
        case 2:
            bond->SetKDouble(); break;
        case 3:
            bond->SetKTriple(); break;
        }
    }

    // Find all the groups of aromatic cycle
    for ( i = 1; i <= GetNumAtom(); i++ ) {
        atom = GetAtom( i );
        if ( atom->HasAromaticBond() && !cvisit[i] ) {
            // must in ring
            // is new aromatic atom of an aromatic cycle ?

            avisit.Clear();
            electron.clear();
            cycle.clear();

            avisit.Set( i );
            expandcycle( atom, avisit );
            //store the atoms of the cycle(s)
            unsigned int j;
            for ( j = 1; j <= GetNumAtom(); ++j ) {
                if ( avisit[j] ) {
                    atom = GetAtom( j );
                    cycle.push_back( atom );
                }
            }

			// for carboxyl, sulphonate, phosphinate, guandium, nitro, etc.
			if ( cycle.size() < 5 ) {
				FOR_BONDS_OF_ATOM( bond, atom ) {
					string type = bond->GetType();
					if ( type.substr( 0, 5 ) == "DB.AR" ) {
						bond->SetKDouble();
						bond->SetOrder( 2 );
					}
					else if ( type.substr( 0, 5 ) == "SB.AR" ) {
						bond->SetKSingle();
						bond->SetOrder( 1 );
					}
				}
				continue;
			}
            // At the begining each atom give one electron to the cycle
            for ( j = 0; j < cycle.size(); ++j ) {
                electron.push_back( 1 );
            }

            // remove one electron if the atom make a double bond out of the cycle
            sume = 0;
            for ( j = 0; j < cycle.size(); ++j ) {
                atom = cycle[j];
                FOR_BONDS_OF_ATOM( bond, atom ) {
                    if ( bond->GetOrder() == 2 ) {
                        CCAtom* atom2 = bond->GetNbrAtom( atom );
                        int fcharge = atom->GetFormalCharge();
                        int fcharge2 = atom2->GetFormalCharge();
                        if ( atom->IsNitrogen()
                          && atom2->IsOxygen()
                          && fcharge == 0
                          && fcharge2 == 0 ) {
                            //n=O to [n+][O-]
                            atom->SetFormalCharge( 1 );
                            atom2->SetFormalCharge( -1 );
                            bond->SetKSingle();
                            bond->SetOrder( 1 );
                        }
                        else if ( atom->IsCarbon() || atom->IsNitrogen() ) {
                            electron[j] = 0;
                        }
                    }
                }
                // count the number of electrons
                sume += electron[j];
            }


            // Save the electron state in case huckel rule is not satisfied
            vector<int> previousElectron = electron;

            // find the ideal number of electrons according to the huckel 4n+2 rule
            int minde;
            for ( int i = 1; 1; ++i ) {
                minde = 4 * i + 2 - sume;
                if ( minde >= 0 ) {
                    break;
                }
            }

            stringstream errorMsg;

            //cout << "minde before:" << minde << endl;
            // if huckel rule not satisfied some atoms must give more electrons
            //cout << "minde " << minde << endl;
            while ( minde != 0 ) {
                bestorden = 99;
                for ( j = 0; j < cycle.size(); ++j ) {
                    if ( electron[j] == 1 ) {
                        orden = getorden( cycle[j] );
                        if ( orden < bestorden ) {
                            bestorden = orden;
                            bestatom = j;
                        }
                    }
                }
                if ( bestorden == 99 ) {
                    // no electron giving atom found
                    errorMsg
                 << "Kekulize: Huckel rule not satisfied for molecule "
                 << GetName()
                 << endl;
                    gErrorLog.Throw( __FUNCTION__, errorMsg.str(), ccWarning );
                    break;             // Huckel rule cannot be satisfied
                }                    // try to kekulize anyway
                else {
                    electron[bestatom] += 1;
                    minde--;
                }
            }

            if ( bestorden == 99 ) {
                // Huckel rule not satisfied, just try to get an even number of electron before kekulizing

                electron = previousElectron; // restore electon's state

                int odd = sume % 2; 
                //cout << "odd:" << odd << endl;
                if ( odd ) {
                    // odd number of electrons try to add an electron to the best possible atom
                    for ( j = 0; j < cycle.size(); ++j ) {
                        if ( electron[j] == 1 ) {
                            orden = getorden( cycle[j] );
                            if ( orden < bestorden ) {
                                bestorden = orden;
                                bestatom = j;
                            }
                        }
                    }
                    if ( bestorden == 99 ) {
                        // no electron giving atom found
                        errorMsg
                     << "Kekulize: Cannot get an even number of electron for molecule "
                     << GetName()
                     << "\n";
                        gErrorLog.Throw( __FUNCTION__, errorMsg.str(),
                                         ccWarning );
                        break;           // impossible to choose an atom to obtain an even number of electron
                    }                    // try to kekulize anyway
                    else {
                        electron[bestatom] += 1;
                    }
                }
            }

            //cout << "minde after:" << minde <<endl;
            //for(j=0; j < cycle.size(); ++j) {
            //CCAtom *cycleAtom = cycle[j];
            //cout << "\t" << cycleAtom->GetIdx();
            //}
            //cout << endl;

            //  for(j=0; j < electron.size(); ++j) {
            //  cout << "\t" << electron[j];
            //  }
            //  cout << endl;

            // kekulize the cycle(s)
            start_kekulize( cycle, electron );

            // Set the kekulized cycle(s) as visited
            for ( j = 1; j <= GetNumAtom(); ++j ) {
                if ( avisit[j] ) {
                    cvisit.Set( j );
                }
            }
        }
    }
    // Double bond have been assigned, set the remaining aromatic bonds to single
    //std::cout << "Set not assigned single bonds\n"; 
    FOR_BONDS_OF_MOL( bond, *this ) {
        //std::cout << "bond " << bond->GetBeginAtomIdx() << " " << bond->GetEndAtomIdx() << " ";   
        if ( bond->GetOrder() == 5 && bond->IsInRing() ) {
            // just make CO2/SO3/PO3 happy
            bond->SetKSingle();
            //bond->SetType("SB.AR");
            bond->SetOrder( 1 );
            //std::cout << "single\n";
        }
        //else
        //  std::cout << "double\n";
    }

    return;
}

///////////////////////////////////////////////////////////////////////////////////////
//! \brief Start kekulizing one or a fused set of aromatic ring(s)

//! The initial electronic state indicates if an atoms must make a double bond or not
//! Kekulizing is attempted recursively for all the atoms bonded to the first atom
//! of the cycle. 
void CCMol::start_kekulize( std::vector <CCAtom*>& cycle,
                            std::vector<int>& electron )
{
	int natom = GetNumAtom();
	int nbond = GetNumBond();
	// init the atom arrays ( natoms+1 elements )
    std::vector<int> initState( natom + 1, -1 );
    std::vector<int> currentState;
	// init the bond arrays ( nbonds+1 elements )
    std::vector<int> binitState( nbond + 1, SINGLE );
    std::vector<int> bcurrentState;
    std::vector<bool> mark( natom + 1, false );
    unsigned int i, Idx;
    CCAtom *atom, *atom2;
    CCBond* bond;

    //set the electron number
    for ( i = 0; i < cycle.size(); ++i ) {
        atom = cycle[i];
        Idx = atom->GetIdx();
        if ( electron[i] == 1 ) {
            initState[Idx] = 1;
        } // make 1 double bond
        else {
            initState[Idx] = 2;
        } // make 2 single bonds
    }

    std::vector<CCBond*>::iterator b;
    CCAtom* nbr;
	atom = cycle[0];
    bool kelulize_successful = false;
    for ( nbr = atom->BeginNbrAtom( b ); nbr; nbr = atom->NextNbrAtom( b ) ) {
        if ( initState[nbr->GetIdx()] == -1 ) //neighbor atom not in the cycle, try next one
        {
            continue;
        } 

		currentState = initState;
		bcurrentState = binitState;

		for ( i = 1; i <= natom; ++i ) {
			mark[i] = false;
		}

        //std::cout << "Expand kekulize\n";
        expand_kekulize( atom, nbr, currentState, initState, bcurrentState,
                         binitState, mark ) ; 
        //Control that all the electron have been given to the cycle(s)
        kelulize_successful = true;
        for ( i = 0; i < cycle.size(); ++i ) {
            atom2 = cycle[i];
            Idx = atom2->GetIdx();
            //cout << "\t" << currentState[Idx];
            if ( currentState[Idx] == 1 ) {
                kelulize_successful = false;
				break;
            }
        }
        //cout << endl;
        if ( kelulize_successful ) {
            break;
        }
    }
    if ( !kelulize_successful ) {
        stringstream errorMsg;
        errorMsg << "Kekulize Error for molecule " << GetName() << endl;
        gErrorLog.Throw( __FUNCTION__, errorMsg.str(), ccWarning );
    }

    // Set the double bonds
    // std::cout << "Set double bonds\n";
    for ( i = 1; i <= nbond; ++i ) {
        bond = GetBond( i );
        // std::cout << "bond " << bond->GetBeginAtomIdx() << " " << bond->GetEndAtomIdx() << " ";
        if ( bond->GetOrder() == 5 && bcurrentState[i] == DOUBLE ) {
            if ( ( bond->GetBeginAtom() )->IsSulfur()
              && bond->GetEndAtom()->IsSulfur() ) {
                // no double bonds between aromatic sulfur atoms -- PR#1504089
                continue;
            }

            bond->SetKDouble();
            bond->SetOrder( 2 );
            //std::cout << bond->GetIdx() << "double\n";
        }
    }

    return;
}

/** 
* @brief Recursively assign single and double bonds according to the
*        electronical state of the atoms of the current bond.
* @param[in] atom1 first atom.
* @param[in] atom2 second atom.
* @param[out] currentState current kekulize state for atoms.
* @param[in] initState initial kekulize state for atoms.
* @param[out] bcurrentState current kekulize state for bonds.
* @param[in] binitState initial kekulize state for bonds.
* @param[out] mark mark whether atoms have been set.
* @return true if atom1 and atom2 can be kekulized in a ring system.
*/
bool CCMol::expand_kekulize( CCAtom* atom1, CCAtom* atom2,
                            std::vector<int>& currentState,
                            std::vector<int>& initState,
                            std::vector<int>& bcurrentState,
                            std::vector<int>& binitState,
                            std::vector<bool>& mark )
{
    bool done;
    int Idx1 = atom1-> GetIdx(), Idx2 = atom2->GetIdx();
    CCBond* bond;
    std::vector<CCBond*>::iterator i;
    CCAtom* nbr;
    int natom;

    mark[Idx1] = true;
    bond = atom1->GetBond( atom2 );
    int bIdx = bond->GetIdx();      // open babel starts from 0

	if ( Idx1 == 55 ) {
		bool test = false;
	}

	// currentState[i] = 0 --> atom i has been kekulized in previous step
	//                         and can only form single bond.
	// currentState[i] = 1 --> atom i can form single/double bond.
	// currentState[i] = 2 --> atom i contribute 2 electrons and can only
	//                         form single bond.
    //cout << "assign bond state for atoms " 
	//     << Idx1 << " and " << Idx2 << endl;
    if ( currentState[Idx1] == 1 && currentState[Idx2] == 1 ) {
        // reset currentState for atom1 and atom2 to avoid 
		// them make double bond.
        currentState[Idx1] = 0;
        currentState[Idx2] = 0;
        // set bond to double
        //std::cout << "bond " << Idx1 << " " << Idx2 << " double\n";
        bcurrentState[bIdx] = DOUBLE;
    }
    else if ( ( currentState[Idx1] == 0 && currentState[Idx2] == 1 )
		   //|| ( currentState[Idx1] == 0 && currentState[Idx2] == 2 )
           || ( currentState[Idx1] == 2 && currentState[Idx2] == 1 )
           || ( currentState[Idx1] == 2 && currentState[Idx2] == 2 ) ){
        //std::cout << "bond " << Idx1 << " " << Idx2 << " single\n";
        // leave bond to single
    }
    else if ( ( currentState[Idx1] == 1 && currentState[Idx2] == 0 )
           || ( currentState[Idx1] == 1 && currentState[Idx2] == 2 ) ) {
        mark[Idx1] = false;
        //std::cout << "bond " << Idx1 << " " << Idx2 << " error\n";
        return false; // error
    }
    else if ( ( currentState[Idx1] == 0 && currentState[Idx2] == 0 )
           || ( currentState[Idx1] == 2 && currentState[Idx2] == 0 ) ) {
        //std::cout << "bond " << Idx1 << " " << Idx2 << " done\n";
        mark[Idx2] = true;
        return true; //done
    }
    else if ( currentState[Idx1] == 0 && currentState[Idx2] == 2 ) {
        currentState[Idx2] = 0;
        //std::cout << "bond " << Idx1 << " " << Idx2 << " single\n";
        // leave bond to single
    }
    else {
        stringstream errorMsg;

        errorMsg << "unexpected state:" << "atom " << Idx1 << " "
                 << currentState[Idx1] << " atom " << Idx2 << " "
                 << currentState[Idx2] << endl;
        gErrorLog.Throw( __FUNCTION__, errorMsg.str(), ccDebug );
        return( false );
    }

	//for( int i = 1; i < currentState.size(); ++i ) {
	//	cout << i << " -- " << currentState[i] << endl;
	//}

	// Backup the atom and the bond states before expanding again
    vector<int> previousState = currentState;
    vector<int> bpreviousState = bcurrentState;
	// note: OpenBabel does not backup this state which causes some
	//       probelm when kekulizing certain hybrid ring.
	vector<bool> previousMark = mark;

    bool expand_failed = false;
    // for each neighbor of atom2 not already kekulized
    for ( nbr = atom2->BeginNbrAtom( i );
          nbr;
          nbr = atom2->NextNbrAtom( i ) ) {
        natom = nbr->GetIdx();

		 //neighbor atom not in the cycle, try next one
        if ( initState[natom] == -1 ) {
            continue;
        }

        if ( !mark[natom] ) {
            done = expand_kekulize( atom2, nbr, currentState, initState,
                                    bcurrentState, binitState, mark );
            if ( !done )  // kekulize failed
            {
                expand_failed = true;
            }
            else {
                expand_failed = false;
            }
        }
    }
    if ( expand_failed ) {
        // no good solution found 
        //cout << "return_false: no good solution" << endl
        //     << "reset state of " << Idx1 << " and " 
		//     << Idx2 << " from "  << currentState[Idx1]
        //     << " " << currentState[Idx2] << " to ";

        // retrieve the states that might have been changed
		// during kekulize expansion
        currentState = previousState;
        bcurrentState = bpreviousState;
		mark = previousMark;

        // reset the bond and the atom states
        if ( bcurrentState[bIdx] == DOUBLE ) {
            currentState[Idx1] = initState[Idx1];
        }

        currentState[Idx2] = initState[Idx2];
        bcurrentState[bIdx] = binitState[bIdx]; // should be always single

        //cout << currentState[Idx1] << " " << currentState[Idx2] << endl;
        return false;
    }
    // atom2 cannot make any bond, should not have 1 electron 
    if ( currentState[Idx2] == 1 ) {
        //cout << "return true but " << Idx2 << " state = 1\n";
        mark[Idx1] = false;
        return false;
    }
    else {
        // if we found a good solution, then the state of Idx2 may have 
		// shifted from 1 to 0 during the kekulization. If it is the case,
		// we should check if there is a remaining unmarked neighbor because
		// it is possible that kekulizing from this neigbor failed just 
		// because Idx2 was equal to 1

        if ( previousState[Idx2] == 1 ) {
            // Since now Idx2 is equal to 0 because it kekulized well the
			// kekulizing of the failed neigbor could be successfull
            // If there is still an unmarked neigbor try to kekulize it again
            expand_failed = false;
            //cout << "retry kekulizing from " << Idx2 << endl;

            for ( nbr = atom2->BeginNbrAtom( i );
                  nbr;
                  nbr = atom2->NextNbrAtom( i ) ) {
                natom = nbr->GetIdx();

				//neighbor atom not in the cycle, try next one
                if ( initState[natom] == -1 ) {
                    continue;
                }

                if ( !mark[natom] ) {
                    //cout << "atom " << natom 
					//     << " not marked, expand again" << endl;
                    done = expand_kekulize( atom2, nbr, currentState,
                                            initState, bcurrentState,
                                            binitState, mark );
                    if ( !done )  // kekulize failed
                    {
                        expand_failed = true;
                    }
                    else {
                        expand_failed = false;
                    }
                }
            }

            // if we cannot kekulize the remaining neigbor again then
			// we have to return false. we do not have to reset the 
			// states because the kekulize will fail anyway.
            if ( expand_failed ) {
                //cout << "rekekulize failed" << endl;
                return false;
            }
            else {
                //cout << "rekekulized successfull" << endl;
                return true;
            }
        }

        //cout << "return_true: good solution" << endl;
        return true;
    }
}

//! Give the priority to give two electrons instead of 1
int CCMol::getorden( CCAtom* atom )
{
    if ( atom->IsSulfur() ) {
        return 1;
    }
    if ( atom->IsOxygen() ) {
        return 2;
    }
    if ( atom->GetAtomicNum() == 34 || atom->GetAtomicNum() == 52 ) {
        return 3;
    }
	if ( atom->IsAmideNitrogen() ) {
		return 4;
	}
    if ( atom->IsNitrogen()
      && atom->GetFormalCharge() == 0
      && atom->GetValence() == 3 ) {
		FOR_NBORS_OF_ATOM( nbr, atom ) {
			if ( nbr->IsTerOxygen() ) {
				return 10;
			}
		}
        return 5;
    }
    if ( atom->IsNitrogen() && atom->GetFormalCharge() == -1 ) {
        return 6;
    }
    if ( atom->IsNitrogen()
      && atom->GetFormalCharge() == 0
      && atom->IsInRingSize( 5 ) ) {
        return 7;
    }
    if ( atom->IsNitrogen() && atom->GetFormalCharge() == 0 ) {
        return 8;
    }
    if ( atom->IsCarbon() && atom->GetFormalCharge() == -1 ) {
        return 9;
    }
    //if ( atom->IsCarbon() ) return 9;

    return ( 100 ); //no atom found
}

//! 
/** 
* @brief Recursively find the aromatic atoms with an aromatic bond to
*        the current atom.
* @param[in] atom current atom.
* @param[out] avisit aromatic atoms index bit vector.
*/
void CCMol::expandcycle( CCAtom* atom, CCBitVec& avisit )
{
    CCAtom* nbr;
    std::vector<CCBond*>::iterator i;
    int natom;
    //for each neighbour atom test if it is in the aromatic ring
    for ( nbr = atom->BeginNbrAtom( i ); nbr; nbr = atom->NextNbrAtom( i ) ) {
        natom = nbr->GetIdx();
        if ( !avisit[natom] && ( *i )->GetOrder() == 5 && ( *i )->IsInRing() ) {
            avisit.Set( natom );
            expandcycle( nbr, avisit );
        }
    }
}
